/*
 * ia32_files.c
 *
 * wrapper functions for syscalls that deal with files. 
 * common to both internal and external wrappers.
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Gordon Jin <gordon.jin@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>
#include "user32.h"
#include "ia32ss_common.h"
#include "ia32ss_list.h"
#include "user32_mem_regions.h"
#if NOKERNELALTROOT
#include "user32_pathwalk.h"
#endif

inline int HIDDEN iovec_conv (struct iovec32 * iov32, struct iovec * iov);


long HIDDEN
sys32_fcntl (int fd, int cmd, __u32 arg)
{
	struct flock f;
	int ret;

	switch (cmd) {
	case F_GETLK:
	case F_SETLK:
	case F_SETLKW:
	{
		struct flock32 fl32;
		if (g_safe_memcpy (&fl32, (struct flock32 *)I2P(arg), sizeof (fl32)) < 0) {
			return -EFAULT;
		}
			
		f.l_type = fl32.l_type;
		f.l_whence = fl32.l_whence;
		f.l_start = fl32.l_start;
		f.l_len = fl32.l_len;
		f.l_pid = fl32.l_pid;
			
		ret = LONG_SYSCALL3 (fcntl, fd, cmd, &f);

		if (ret < 0) {
			return ret;
		}
		if (cmd == F_GETLK) {
			fl32.l_type = f.l_type;
			fl32.l_whence = f.l_whence;
			fl32.l_start = f.l_start;
			fl32.l_len = f.l_len;
			fl32.l_pid = f.l_pid;
			
			if (g_safe_memcpy ((struct flock32 *)I2P(arg), 
					   &fl32, sizeof (fl32)) < 0) {
				return -EFAULT;
			}
		}
		return ret;			
	}
	default:
		return LONG_SYSCALL3 (fcntl, fd, cmd, arg);
	}
}

long HIDDEN
sys32_fcntl64 (int fd, int cmd, __u32 arg)
{
	int ret;
	__u32 new_cmd = cmd;

	switch (cmd) {
	case F_GETLK64:
		new_cmd = F_GETLK;
		break;
	case F_SETLK64:
		new_cmd = F_SETLK;
		break;
	case F_SETLKW64:
		new_cmd = F_SETLKW;
		break;
	}
	ret = sys32_fcntl(fd, new_cmd, arg);
	return ret;
}

long HIDDEN
sys32_truncate64 (__u32 path, __u32 low, __u32 high)
{
#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	return pws_truncate (I2P(path), ((unsigned long)high << 32) | low);
#else
	return SYSCALL (truncate, (I2P(path), ((unsigned long)high << 32) | low));
#endif
}

long HIDDEN
sys32_ftruncate64 (int fd, __u32 low, __u32 high)
{
	return SYSCALL (ftruncate, (fd, ((unsigned long)high << 32) | low));
}

static inline int
stat64_conv (struct stat64 *buf64, struct stat *buf)
{
	struct stat64 buf_tmp;

	memset (&buf_tmp, 0, sizeof(buf_tmp));

	buf_tmp.st_dev = buf->st_dev;
	buf_tmp.__st_ino = buf->st_ino;
	buf_tmp.st_ino_lo = buf->st_ino;
	buf_tmp.st_ino_hi = buf->st_ino >> 32;
	buf_tmp.st_mode = buf->st_mode;
	buf_tmp.st_nlink = buf->st_nlink;
	buf_tmp.st_uid = buf->st_uid;
	buf_tmp.st_gid = buf->st_gid;
	buf_tmp.st_rdev = buf->st_rdev;
	buf_tmp.st_size_lo = buf->st_size;
	buf_tmp.st_size_hi = buf->st_size >> 32;
	buf_tmp.st_atim = buf->st_atim.tv_sec;
	buf_tmp.st_atim_nsec = buf->st_atim.tv_nsec;
	buf_tmp.st_mtim = buf->st_mtim.tv_sec;
	buf_tmp.st_mtim_nsec = buf->st_mtim.tv_nsec;
	buf_tmp.st_ctim = buf->st_ctim.tv_sec;
	buf_tmp.st_ctim_nsec = buf->st_ctim.tv_nsec;
	buf_tmp.st_blksize = buf->st_blksize;
	buf_tmp.st_blocks = buf->st_blocks;

	if (g_safe_memcpy (buf64, &buf_tmp, sizeof (buf_tmp)) < 0) {
		return -EFAULT;
	}
	return 0;
}

long HIDDEN
sys32_stat64 (__u32 path, __u32 buf)
{
	struct stat s;
	int ret;

#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	ret = pws_stat (I2P(path), &s);
#else
	ret = SYSCALL (stat, (I2P(path), &s));
#endif
	if (ret < 0) {
		return ret;
	}
	if (stat64_conv(I2P(buf), &s)) {
		ret = -EFAULT;
	}
	return ret;
}

long HIDDEN
sys32_lstat64 (__u32 filename, __u32 statbuf)
{
	struct stat s;
	int ret;

#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	ret = pws_lstat (I2P(filename), &s);
#else
	ret = SYSCALL (lstat, (I2P(filename), &s));
#endif
	if (ret < 0) {
		return ret;
	}
	if (stat64_conv(I2P(statbuf), &s)){
		ret = -EFAULT;
	}
	return ret;
}

long HIDDEN
sys32_fstat64 (int fd, __u32 statbuf)
{
	struct stat s;
	int ret;

	ret = SYSCALL (fstat, (fd, &s));
	if (ret < 0) {
		return ret;
	}
	if (stat64_conv(I2P(statbuf), &s)){
		ret = -EFAULT;
	}
	return ret;
}


static inline int
stat_conv (struct stat32 *buf32, struct stat *buf)
{
	struct stat32 buf_tmp;

	memset (&buf_tmp, 0, sizeof (buf_tmp));

	buf_tmp.st_dev = buf->st_dev;
	buf_tmp.st_ino = buf->st_ino;
	buf_tmp.st_mode = buf->st_mode;
	buf_tmp.st_nlink = buf->st_nlink;
	buf_tmp.st_uid = buf->st_uid;
	buf_tmp.st_gid = buf->st_gid;
	buf_tmp.st_rdev = buf->st_rdev;
	buf_tmp.st_size = buf->st_size;
	buf_tmp.st_atim = buf->st_atime;
	buf_tmp.st_mtim = buf->st_mtime;
	buf_tmp.st_ctim = buf->st_ctime;
	buf_tmp.st_blksize = buf->st_blksize;
	buf_tmp.st_blocks = buf->st_blocks;

	if (g_safe_memcpy (buf32, &buf_tmp, sizeof (buf_tmp)) < 0) {
		return -EFAULT;
	}
	return 0;

}

long HIDDEN
sys32_newstat (__u32 filename, __u32 statbuf)
{
	struct stat s;
	int ret;

#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	ret = pws_stat (I2P(filename), &s);
#else
	ret = SYSCALL (stat, (I2P(filename), &s));
#endif
	if (ret < 0) {
		return ret;
	}
	ret = stat_conv(I2P(statbuf), &s);
	return ret;
}


long HIDDEN
sys32_newlstat (__u32 filename, __u32 statbuf)
{
	struct stat s;
	int ret;

#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	ret = pws_lstat (I2P(filename), &s);
#else
	ret = SYSCALL (lstat, (I2P(filename), &s));
#endif
	if (ret < 0) {
		return ret;
	}
	ret = stat_conv(I2P(statbuf), &s);
	return ret;
}


long HIDDEN
sys32_newfstat (int fd, __u32 statbuf)
{
	struct stat s;
	int ret;

	ret = SYSCALL (fstat, (fd, &s));
	if (ret < 0) {
		return ret;
	}
	ret = stat_conv(I2P(statbuf), &s);
	return ret;
}


static inline int
statfs_conv (struct statfs32 *buf32, struct k_statfs *buf)
{
	struct statfs32 buf_tmp;

	buf_tmp.f_type = buf->f_type;
	buf_tmp.f_bsize = buf->f_bsize;
	buf_tmp.f_blocks = buf->f_blocks;
	buf_tmp.f_bfree = buf->f_bfree;
	buf_tmp.f_bavail = buf->f_bavail;
	buf_tmp.f_files = buf->f_files;
	buf_tmp.f_ffree = buf->f_ffree;
	buf_tmp.f_namelen = buf->f_namelen;
	buf_tmp.f_fsid.val[0] = buf->f_fsid.val[0];
	buf_tmp.f_fsid.val[1] = buf->f_fsid.val[1];
	memcpy (&buf_tmp.f_spare, &buf->f_spare, sizeof (buf_tmp.f_spare));

	if (g_safe_memcpy (buf32, &buf_tmp, sizeof (buf_tmp)) < 0) {
		return -EFAULT;
	}
	return 0;
}

long HIDDEN
sys32_statfs (__u32 path, __u32 buf)
{
	struct k_statfs s;
	int ret;

#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	ret = pws_statfs (I2P(path), &s);
#else
	ret = SYSCALL (statfs, (I2P(path), &s));
#endif
	if (ret < 0) {
		return ret;
	}
	ret = statfs_conv(I2P(buf), &s);
	return ret;
}


long HIDDEN
sys32_fstatfs (int fd, __u32 buf)
{
	struct k_statfs s;
	int ret;

	ret = SYSCALL (fstatfs, (fd, &s));
	if (ret < 0) {
		return ret;
	}
	ret = statfs_conv(I2P(buf), &s);
	return ret;
}


long HIDDEN
sys32_lseek (int fd, int offset, int whence)
{
	return SYSCALL (lseek, (fd, offset, whence));
}

long HIDDEN
sys32_pread (int fd, __u32 buf, __u32 count, __u32 low, __u32 high)
{
	return SYSCALL (pread, (fd, I2P(buf), count, ((unsigned long) high << 32) | low));
}

long HIDDEN
sys32_pwrite (int fd, __u32 buf, __u32 count, __u32 low, __u32 high)
{
	return SYSCALL (pwrite, (fd, I2P(buf), count, ((unsigned long) high << 32) | low));
}


inline int HIDDEN
iovec_conv(struct iovec32 *iov32, struct iovec *iov)
{
	struct iovec32 iov_tmp;

	if (g_safe_memcpy (&iov_tmp, iov32, sizeof (iov_tmp)) < 0) {
		return -EFAULT;
	}
	iov->iov_base = I2P(iov_tmp.iov_base);
	iov->iov_len = iov_tmp.iov_len;
	return 0;
}

long HIDDEN
sys32_readv (int fd, __u32 vector, __u32 count)
{
	struct iovec *iov;
	int ret;


	if ((int)count < 0) {
		return -EINVAL;
	}

	if (count > UIO_MAXIOV) {
		return -EINVAL;
	}
	
	iov = (struct iovec *)malloc(count * sizeof(struct iovec));
	if (iov || count == 0) {
		__u32 i;
		for (i = 0; i < count; i++) {
			ret = iovec_conv(&((struct iovec32 *)I2P(vector))[i], &iov[i]);
			if (ret < 0) {
				free(iov);
				return ret;
			} 
		}
		ret = LONG_SYSCALL3 (readv, fd, iov, count);
		free(iov);
		return ret;
	}
	return -ENOMEM;
}

long HIDDEN
sys32_writev (int fd, __u32 vector, __u32 count)
{
	struct iovec *iov;
	int ret;
	
	if ((int)count < 0) {
		return -EINVAL;
	}
	
	if (count > UIO_MAXIOV) {
		return -EINVAL;
	}
	
	iov = (struct iovec*)malloc(count * sizeof(struct iovec));
	if (iov || count == 0) {
		__u32 i;
		for (i = 0; i < count; i++) {
			ret = iovec_conv(&((struct iovec32 *)I2P(vector))[i], &iov[i]);
			if (ret < 0) {
				free(iov);
				return ret;
			} 
		}
		ret = LONG_SYSCALL3 (writev, fd, iov, count);
		free(iov);
		return ret;
	}
	return -ENOMEM;
}



/* map between aio_ctx32_t and aio_ctx_t */
struct ctxmap_t {
	aio_ctx32_t 	 ctx32_id;  /* 32 bit ctx id */
	unsigned         size;
	aio_ctx_t	 ctx_id;    /* 64 bit ctx id */
	struct ctxmap_t  *next;
};

static struct ctxmap_t *m_ctxmap_list = NULL;


long HIDDEN
sys32_io_setup (__u32 nr_events, __u32 ctx32p) 
{
	int          ret;
	aio_ctx32_t   ctx32_id;
	aio_ctx_t     ctx_id = 0;
	unsigned      size;
	unsigned long new_addr;

	struct ctxmap_t *ctxmap = NULL;

	if (g_safe_memcpy (&ctx32_id, (aio_ctx32_t *)I2P(ctx32p), sizeof(ctx32_id)) < 0) {
		return -EFAULT;
	}

	/* *ctx32p must be initialized with 0 */ 
	if (ctx32_id != 0) {
		return -EINVAL;
	}

	/* calculate the size of memory allocated by kernel, depends on 
	   kernel implementation!!! */
	size = sizeof(struct aio_ring) + sizeof(struct io_event) * nr_events;
	size = ((size + PAGE_SIZE - 1) >> PAGE_SHIFT) * PAGE_SIZE;


	/* try to find a free IA-32 address area and remap it with kernel 
	   returned native address area. Need to be synchronized with other
	   IA-32 memory operations */
	mutex_lock (g_memory_sync);

	new_addr = get_ia32_free_memory (0, size, 0, 1);
	if ((long)new_addr < 0) {
		/* cannot find an unmapped area in 32-bit address space */
		mutex_unlock (g_memory_sync);
		return (int)new_addr;
	}

	ret = SYSCALL (io_setup, (nr_events, &ctx_id));
	if (ret != 0) {
		mutex_unlock (g_memory_sync);
		return ret;
	}
	assert (ctx_id != 0);
	
	ret = SYSCALL (mremap, (ctx_id, size, size, MREMAP_FIXED | MREMAP_MAYMOVE, 
				new_addr));
	assert (ret > 0);

	ctx32_id = (aio_ctx32_t)new_addr;
	if (g_safe_memcpy ((aio_ctx32_t*)I2P(ctx32p), &ctx32_id, sizeof (ctx32_id)) < 0) {
		/* clean up if failed: recover orignial map and release 
		   kernel aio ctx */
		ret = SYSCALL (mremap, (new_addr, size, size, 
					MREMAP_FIXED | MREMAP_MAYMOVE, ctx_id));
		assert (ret >= 0);

		ret = SYSCALL (io_destroy, (ctx_id));
		assert (ret == 0);

		mutex_unlock (g_memory_sync);
		return -EFAULT;
	}

	ret = insert_region (new_addr, size, PROT_WRITE | PROT_READ, 0);
	assert (ret == 0);
	mutex_unlock (g_memory_sync);

	/* add a new map */ 
	mutex_lock (g_ctxmap_sync);
	ctxmap = malloc (sizeof(struct ctxmap_t));
	if (ctxmap == NULL) {
		ret = -ENOMEM;
	}
	ctxmap->next = m_ctxmap_list;
	ctxmap->ctx32_id = ctx32_id;
	ctxmap->size = size;
	ctxmap->ctx_id = ctx_id;
	m_ctxmap_list = ctxmap;
	mutex_unlock (g_ctxmap_sync);

	return ret;
}

long HIDDEN
sys32_io_destroy (__u32 ctx32_id) 
{
	int        ret;
	struct ctxmap_t *p = NULL, *q = NULL;

	/* go over the list to find a map */
	mutex_lock (g_ctxmap_sync);
	for (p = q = m_ctxmap_list; p != NULL; p = p->next) {
		if (p->ctx32_id == (aio_ctx32_t)ctx32_id)
			break;
		q = p;
	}
	
	/* remove the map */
	if (p == NULL) {
		mutex_unlock (g_ctxmap_sync);
		return -EINVAL;
	} else {
		assert (q != NULL);
		q->next = p->next;
	}
	mutex_unlock (g_ctxmap_sync);


	/* original ctx_id should be recoverd for io_destory */
	ret = SYSCALL (mremap, (p->ctx32_id, p->size, p->size, 
				MREMAP_FIXED | MREMAP_MAYMOVE, p->ctx_id));

	assert (ret >= 0);

	mutex_lock (g_memory_sync);
	ret = remove_region (p->ctx32_id, p->size);
	assert (ret == 0);
	mutex_unlock (g_memory_sync);

	ret = SYSCALL (io_destroy, (p->ctx_id));
	free (p);
	return ret;
}

long HIDDEN
sys32_io_getevents (__u32 ctx32_id, int min_nr, int nr, __u32 events,
                    __u32 timeout) 
{

	struct timespec   *pt = (struct timespec *)I2P(timeout);
	struct timespec   t;
	struct timespec32 t32;
	struct ctxmap_t   *p = NULL;
	int   ret;
	
	if ((struct timespec32 *)I2P(timeout) != NULL) {
	 	if (g_safe_memcpy (&t32, (struct timespec32 *)I2P(timeout),
				   sizeof(t32)) < 0) {
			ret = -EFAULT;
			goto out;
		}
	
		t.tv_sec = t32.tv_sec;
		t.tv_nsec = t32.tv_nsec;
		pt = &t;
	}


	mutex_lock (g_ctxmap_sync);
	for (p = m_ctxmap_list; p != NULL; p = p->next) {
		if (p->ctx32_id == (aio_ctx32_t)ctx32_id)
			break;
	}
	mutex_unlock (g_ctxmap_sync);

	if (p == NULL) {
		ret = -EINVAL;
		goto out;
	}

	ret = LONG_SYSCALL5 (io_getevents, p->ctx_id, min_nr, nr, events, pt);

out:
	return ret;
}


long HIDDEN 
sys32_io_submit (__u32 ctx32_id, int nr, __u32 iocbpp)
{
	struct ctxmap_t *p;
	int    ret;
	int    i;
	int    addr32;
	void   **iocbpp64 = NULL;
	unsigned *q;

	
	if (nr < 0) {
		ret =  -EINVAL;
		goto out;
	}
	
	iocbpp64 = malloc (nr * sizeof (void *));

	if (iocbpp64 == NULL) {
		ret =  -ENOMEM;
		goto out;
	}

	q = (unsigned *)I2P (iocbpp);
	for (i = 0; i < nr; i++) {
		if (g_safe_memcpy (&addr32, q, sizeof (addr32)) < 0) {
			ret = -EFAULT;
			free (iocbpp64);
			goto out;
		}
		q++;
		iocbpp64[i] = I2P (addr32);
	}
	

	mutex_lock (g_ctxmap_sync);
	for (p = m_ctxmap_list; p != NULL; p = p->next) {
		if (p->ctx32_id == (aio_ctx32_t)ctx32_id)
			break;
	}
	mutex_unlock (g_ctxmap_sync);
	if (p == NULL) {
		ret = -EINVAL;
		goto out;
	}

	ret = SYSCALL (io_submit, (p->ctx_id, nr, iocbpp64));
out:
	return ret;
}


long HIDDEN 
sys32_io_cancel (__u32 ctx32_id, __u32 iocb, __u32 result)
{
	int ret;
	struct ctxmap_t *p;

	mutex_lock (g_ctxmap_sync);
	for (p = m_ctxmap_list; p != NULL; p = p->next) {
		if (p->ctx32_id == (aio_ctx32_t)ctx32_id)
			break;
	}
	mutex_unlock (g_ctxmap_sync);
	if (p == NULL) {
		ret = -EINVAL;
		goto out;
	}

	ret = SYSCALL (io_cancel, (p->ctx_id, (void *)I2P(iocb), (void *)I2P(result)));
out:
	return ret;
}

struct sel_arg_struct {
	__u32 n;
	__u32 readfds;
	__u32 writefds;
	__u32 exceptfds;
	__u32 timeout;
};

long HIDDEN
sys32_old_select (__u32 arg)
{
	struct sel_arg_struct a;

	if (g_safe_memcpy(&a, (struct sel_arg_struct *)I2P(arg), sizeof(a))) {
		return -EFAULT;
	}
	return sys32_select(a.n, a.readfds, a.writefds, a.exceptfds, a.timeout);
}

long HIDDEN
sys32_fstatfs64(__u32 fd, __u32 sz, __u32 buf)
{
	struct statfs64 sf64;
	struct statfs64_32 sf32;
	int retval;
	int i;

	if (sz != sizeof (struct statfs64_32)) {
		return -EINVAL;
	}
	
	if (IS_VALID_PTR(I2P(buf))) {
		if (g_safe_memcpy(&sf32, I2P(buf), sizeof(struct statfs64_32))) {
			return -EFAULT;
		}
		sf64.f_type = sf32.f_type;
		sf64.f_bsize = sf32.f_bsize;
		sf64.f_blocks = sf32.f_blocks;
		sf64.f_bfree = sf32.f_bfree;
		sf64.f_bavail = sf32.f_bavail;
		sf64.f_files = sf32.f_files;
		sf64.f_ffree = sf32.f_ffree;
		sf64.f_fsid.val[0] = sf32.f_fsid.val[0];
		sf64.f_fsid.val[1] = sf32.f_fsid.val[1];
		sf64.f_namelen = sf32.f_namelen;
		sf64.f_frsize = sf32.f_frsize;
		for (i=0; i<5; i++)
			sf64.f_spare[i] = sf32.f_spare[i];
	}
	
	retval = SYSCALL (fstatfs64, (fd, sizeof(struct statfs64),
					IS_VALID_PTR(I2P(buf))? &sf64: NULL));

	if (retval == 0) {
		sf32.f_type = sf64.f_type;
		sf32.f_bsize = sf64.f_bsize;
		sf32.f_blocks = sf64.f_blocks;
		sf32.f_bfree = sf64.f_bfree;
		sf32.f_bavail = sf64.f_bavail;
		sf32.f_files = sf64.f_files;
		sf32.f_ffree = sf64.f_ffree;
		sf32.f_fsid.val[0] = sf64.f_fsid.val[0];
		sf32.f_fsid.val[1] = sf64.f_fsid.val[1];
		sf32.f_namelen = sf64.f_namelen;
		sf32.f_frsize = sf64.f_frsize;
		for (i=0; i<5; i++)
			sf32.f_spare[i] = sf64.f_spare[i];
		
		if (g_safe_memcpy(I2P(buf), &sf32, sizeof(struct statfs64_32))) {
			return -EFAULT;
		}
	}
	
	return retval;
}

