/*
 * ia32_ioctl.c
 *
 * wrapper functions for syscalls that deal with ioctl. 
 * common to both internal and external wrappers.
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Kevin Tian <kevin.tian@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "user32.h"
#include "ia32_ioctl.h"

static long vfat_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct dirent32 d32[2];
	struct dirent d64[2];
	int ret;
	
	ret = LONG_SYSCALL3 (ioctl, fd, cmd, (unsigned long)&d64);
	
	if (!ret) {
		int i;		
		for (i = 0; i < 2; i++) {
			d32[i].d_ino = (__u32)d64[i].d_ino;
			d32[i].d_off = (__u32)d64[i].d_off;
			d32[i].d_reclen = d64[i].d_reclen;
			memcpy (d32[i].d_name, d64[i].d_name, d64[i].d_reclen);
		}
		
		if (g_safe_memcpy(I2P(arg), &d32, sizeof(d32)) < 0) {
			return -EFAULT;
		}
	}
	
	return ret;
}

static long ifconf_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ifconf32 ifc32;
	struct ifconf ifc64;
	struct ifreq32 *ifr32p;
	struct ifreq *ifr64p;
	int v, ret;	
	
	if (g_safe_memcpy(&ifc32, I2P(arg), sizeof(ifc32)) < 0)
		return -EFAULT;
	
	v = IS_VALID_PTR(I2P(ifc32.ifcbuf));
	if (v) {
		ifc64.ifc_len = (ifc32.ifc_len / sizeof (struct ifreq32)) * sizeof (struct ifreq);
		ifc64.ifc_buf = malloc (ifc64.ifc_len);
		if (!ifc64.ifc_buf) {
			return -ENOMEM;
		}
	} else {
		ifc64.ifc_len = 0;
		ifc64.ifc_buf = NULL;
	}
	
	ret = LONG_SYSCALL3 (ioctl, fd, cmd, (unsigned long)&ifc64);
	
	if (!ret) {
		if (v) {
			ifr32p = (struct ifreq32*)I2P(ifc32.ifcbuf);
			ifr64p = ifc64.ifc_req;
			
			for (; (char*)ifr64p < ifc64.ifc_buf + ifc64.ifc_len; ifr32p++, ifr64p++) {
				if (g_safe_memcpy(ifr32p, ifr64p, sizeof(struct ifreq32)) < 0) {
					ret = -EFAULT;
					goto out;
				}
			}
		}
		
		ifc32.ifc_len = (ifc64.ifc_len / sizeof(struct ifreq)) * sizeof(struct ifreq32);
		g_safe_memcpy(I2P(arg), &ifc32, sizeof(ifc32));
	}
	
out:
	if (v) {
		free (ifc64.ifc_buf);
	}
	return ret;
}

static long hdio_getgeo32(int fd, unsigned int cmd, unsigned long arg)
{
	int ret;
	struct hd_geometry32 geo32;
	struct hd_geometry   geo64;
	
	ret = LONG_SYSCALL3 (ioctl, fd, cmd, (unsigned long)&geo64);

	if (ret == 0) {
		geo32.heads = geo64.heads;
		geo32.sectors = geo64.sectors;
		geo32.cylinders = geo64.cylinders;

		ASSERT_HIGH_ZERO(geo64.start);
		geo32.start = geo64.start;

		if (g_safe_memcpy(I2P(arg), &geo32, sizeof(geo32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long rawgetbind32 (int fd, unsigned cmd, unsigned long arg)
{
	int ret;
	__u32 rq32[5];
	union {
		struct raw_config_request rq;
		__u32 int_field[sizeof(struct raw_config_request)/sizeof(__u32)];
	} cnvrt;

	assert (sizeof (struct raw_config_request32) == sizeof(int)+2*sizeof(long));

	if (g_safe_memcpy (&rq32, I2P(arg), sizeof (rq32)) < 0) {
		return -EFAULT;
	}
	memset (&cnvrt.int_field, 0, sizeof (cnvrt.int_field));
	cnvrt.rq.raw_minor = rq32[0];

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, (unsigned long)&cnvrt.rq);

	if ( ! IS_SYS_ERROR (ret) ) {

		/* avoid misalignments */
		rq32[1] = cnvrt.int_field[2];
		rq32[2] = cnvrt.int_field[3];
		rq32[3] = cnvrt.int_field[4];
		rq32[4] = cnvrt.int_field[5];

		if (g_safe_memcpy (I2P(arg), &rq32, sizeof (rq32)) < 0) {
			return -EFAULT;
		}
	}
	return ret;
}
static long rawsetbind32 (int fd, unsigned cmd, unsigned long arg)
{
	int ret;
	__u32 rq32[5];
	union {
		struct raw_config_request rq;
		__u32 int_field[sizeof(struct raw_config_request)];
	} cnvrt;

	assert (sizeof (struct raw_config_request32) == sizeof(int)+2*sizeof(long));

	if (g_safe_memcpy (&rq32, I2P(arg), sizeof (rq32)) < 0) {
		return -EFAULT;
	}
	/* avoid misalignments */
	cnvrt.int_field[0] = rq32[0];
	cnvrt.int_field[1] = 0;
	cnvrt.int_field[2] = rq32[1];
	cnvrt.int_field[3] = rq32[2];
	cnvrt.int_field[4] = rq32[3];
	cnvrt.int_field[5] = rq32[4];

	DBPRINT (0, "raw_minor=%d block_major=%ld block_minor=%ld\n",
			cnvrt.rq.raw_minor, cnvrt.rq.block_major, 
			cnvrt.rq.block_minor);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, (unsigned long)&cnvrt.rq);

	return ret;
}
static long arg32_to_arg64_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	unsigned int arg32;
	unsigned long arg64;
	long ret;

	if (g_safe_memcpy (&arg32, I2P(arg), sizeof (arg32)) < 0) {
                return -EFAULT;
        }

	arg64 = arg32;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &arg64);

	if (ret == 0) {
		arg32 = arg64;
		if (g_safe_memcpy(I2P(arg), &arg32, sizeof (arg32)) < 0) {
			return -EFAULT;
		}
	}
	return ret;
}

static long arg32_to_arg64_writeback_only(int fd, unsigned int cmd, unsigned long arg)
{
	unsigned int arg32;
	unsigned long arg64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &arg64);

	if (ret == 0) {
		arg32 = arg64;
		if (g_safe_memcpy(I2P(arg), &arg32, sizeof (arg32)) < 0) {
			return -EFAULT;
		}
	}
	return ret;
}

static long socket_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ifreq   ifreq64;
	long   ret;

	if (g_safe_memcpy (&ifreq64, (struct ifreq32*)arg, sizeof (struct ifreq32)) < 0) {
		return -EFAULT;
	}

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ifreq64);
	if (ret == 0) {
		if (g_safe_memcpy((struct ifreq32*)arg, &ifreq64, sizeof (struct ifreq32)) < 0) {
			return -EFAULT;
		}

	}

	return ret;
}

static long siocgifmap_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ifreq   ifreq64;
	long   ret;

	if (g_safe_memcpy (&ifreq64, (struct ifreq32*)arg, sizeof (struct ifreq32)) < 0) {
		return -EFAULT;
	}

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ifreq64);
	if (ret == 0) {
		struct ifreq32   ifreq32;

		memcpy (&ifreq32, &ifreq64, sizeof (ifreq64.ifr_ifrn));
		ifreq32.ifr_ifru.ifru_map.mem_start = ifreq64.ifr_ifru.ifru_map.mem_start;
		ifreq32.ifr_ifru.ifru_map.mem_end = ifreq64.ifr_ifru.ifru_map.mem_end;
		ifreq32.ifr_ifru.ifru_map.base_addr = ifreq64.ifr_ifru.ifru_map.base_addr;
		ifreq32.ifr_ifru.ifru_map.irq = ifreq64.ifr_ifru.ifru_map.irq;
		ifreq32.ifr_ifru.ifru_map.dma = ifreq64.ifr_ifru.ifru_map.dma;
		ifreq32.ifr_ifru.ifru_map.port = ifreq64.ifr_ifru.ifru_map.port;

		if (g_safe_memcpy((struct ifreq32*)arg, &ifreq32, sizeof (struct ifreq32)) < 0) {
                        return -EFAULT;
		}
	}

	return ret;
}

static long siocsifmap_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ifreq   ifreq64;
	struct ifreq32 ifreq32;
	long   ret;

	if (g_safe_memcpy (&ifreq32, (struct ifreq32*)arg, sizeof (struct ifreq32)) < 0) {
		return -EFAULT;
	}

	memcpy (&ifreq64, &ifreq32, sizeof (ifreq32.ifr_ifrn));
	ifreq64.ifr_ifru.ifru_map.mem_start = ifreq32.ifr_ifru.ifru_map.mem_start;
	ifreq64.ifr_ifru.ifru_map.mem_end = ifreq32.ifr_ifru.ifru_map.mem_end;
	ifreq64.ifr_ifru.ifru_map.base_addr = ifreq32.ifr_ifru.ifru_map.base_addr;
	ifreq64.ifr_ifru.ifru_map.irq = ifreq32.ifr_ifru.ifru_map.irq;
	ifreq64.ifr_ifru.ifru_map.dma = ifreq32.ifr_ifru.ifru_map.dma;
	ifreq64.ifr_ifru.ifru_map.port = ifreq32.ifr_ifru.ifru_map.port;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ifreq64);

	return ret;
}

static long socket_putdata_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ifreq   ifreq64;
	struct ifreq32 ifreq32;
	long   ret;

	if (g_safe_memcpy (&ifreq32, (struct ifreq32*)arg, sizeof (struct ifreq32)) < 0) {
		return -EFAULT;
	}

	memcpy (&ifreq64.ifr_ifrn, &ifreq32.ifr_ifrn, sizeof (ifreq32.ifr_ifrn));
	ifreq64.ifr_ifru.ifru_data = (char *)UL(ifreq32.ifr_ifru.ifru_data);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ifreq64);

	return ret;
}

static long socket_read_arg32_only_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ifreq   ifreq64;
	long   ret;

	if (g_safe_memcpy (&ifreq64, (struct ifreq32*)arg, sizeof (struct ifreq32)) < 0) {
		return -EFAULT;
	}

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ifreq64);

	return ret;
}

static long blkpg_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct blkpg_ioctl_arg32   arg32;
	struct blkpg_ioctl_arg     arg64;
	struct blkpg_partition32   par32;
	struct blkpg_partition     par64;
	long   ret;

	if (g_safe_memcpy (&arg32, (struct blkpg_ioctl_arg32*)arg, sizeof (struct blkpg_ioctl_arg32)) < 0) {
		return -EFAULT;
	}

	if ((arg32.op != BLKPG_ADD_PARTITION) && (arg32.op != BLKPG_DEL_PARTITION)) {
		return -EINVAL;
	}

	arg64.op = arg32.op;
	arg64.flags = arg32.flags;

	arg64.datalen = sizeof (struct blkpg_partition);

	if (g_safe_memcpy (&par32, (struct blkpg_partition32*)I2P(arg32.data), sizeof (struct blkpg_partition32)) < 0) {
		return -EFAULT;
	}
	par64.start = par32.start;
	par64.length = par32.length;
	par64.pno = par32.pno;
	memcpy (par64.devname, par32.devname, BLKPG_DEVNAMELTH + BLKPG_VOLNAMELTH);

	arg64.data = (void *)&par64;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &arg64);

	return ret;
}

static long socket_rtentry_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct rtentry    rte64;
	struct rtentry32  rte32;
	long   ret;

	if (g_safe_memcpy (&rte32, (struct rtentry32*)arg, sizeof (struct rtentry32)) < 0) {
		return -EFAULT;
	}

	memcpy (&rte64.rt_dst, &rte32.rt_dst, sizeof (rte32.rt_dst));
	memcpy (&rte64.rt_gateway, &rte32.rt_gateway, sizeof (rte32.rt_gateway));
	memcpy (&rte64.rt_genmask, &rte32.rt_genmask, sizeof (rte32.rt_genmask));
	rte64.rt_flags = rte32.rt_flags;
	rte64.rt_metric = rte32.rt_metric;
	rte64.rt_dev = (char *)UL(rte32.rt_dev);
	rte64.rt_mtu = rte32.rt_mtu;
	rte64.rt_window = rte32.rt_window;
	rte64.rt_irtt = rte32.rt_irtt;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &rte64);

	return ret;
}

static long siocgstamp_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct timeval    tv64;
	struct timeval32  tv32;
	long   ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &tv64);

	if (ret == 0) {
		tv32.tv_sec = tv64.tv_sec;
		tv32.tv_usec = tv64.tv_usec;

		if (g_safe_memcpy((struct timeval32*)arg, &tv32, sizeof (struct timeval32)) < 0) {
                        return -EFAULT;
		}
	}

	return ret;
}

static long loopset_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct loop_info    li64;
	struct loop_info32  li32;
	long   ret;

	if (g_safe_memcpy (&li32, (struct loop_info32*)arg, sizeof (struct loop_info32)) < 0) {
		return -EFAULT;
	}

	li64.lo_number = li32.lo_number;
	li64.lo_device = li32.lo_device;
	li64.lo_inode = li32.lo_inode;
	li64.lo_rdevice = li32.lo_rdevice;
	li64.lo_offset = li32.lo_offset;
	li64.lo_encrypt_type = li32.lo_encrypt_type;
	li64.lo_encrypt_key_size = li32.lo_encrypt_key_size;
	li64.lo_flags = li32.lo_flags;
	memcpy (li64.lo_name, li32.lo_name, LO_NAME_SIZE);
	memcpy (li64.lo_encrypt_key, li32.lo_encrypt_key, LO_KEY_SIZE);
	li64.lo_init[0] = li32.lo_init[0];
	li64.lo_init[1] = li32.lo_init[1];
	memcpy (li64.reserved, li32.reserved, sizeof (char) * 4);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &li64);

	return ret;
}

static long loopget_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct loop_info    li64;
	struct loop_info32  li32;
	long   ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &li64);

	li32.lo_number = li64.lo_number;
	li32.lo_device = li64.lo_device;
	li32.lo_inode = li64.lo_inode;
	li32.lo_rdevice = li64.lo_rdevice;
	li32.lo_offset = li64.lo_offset;
	li32.lo_encrypt_type = li64.lo_encrypt_type;
	li32.lo_encrypt_key_size = li64.lo_encrypt_key_size;
	li32.lo_flags = li64.lo_flags;
	memcpy (li32.lo_name, li64.lo_name, LO_NAME_SIZE);
	memcpy (li32.lo_encrypt_key, li64.lo_encrypt_key, LO_KEY_SIZE);
	li32.lo_init[0] = li64.lo_init[0];
	li32.lo_init[1] = li64.lo_init[1];
	memcpy (li32.reserved, li64.reserved, sizeof (char) * 4);

	if (g_safe_memcpy ((struct loop_info32*)arg, &li32, sizeof (struct loop_info32)) < 0) {
		return -EFAULT;
	}

	return ret;
}

static long mtdoob_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct mtd_oob_buf    ob64;
	struct mtd_oob_buf32  ob32;
	long   ret;

	if (IS_VALID_PTR ((struct mtd_oob_buf32*)arg)) {
        	if (g_safe_memcpy (&ob32, (struct mtd_oob_buf32*)arg, sizeof (struct mtd_oob_buf32)) < 0) {
        	        return -EFAULT;
	        }

		ob64.start = ob32.start;
		ob64.ength = ob32.ength;
		ob64.ptr = (unsigned char *)UL(ob32.ptr);
	}

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ob64);

	if (ret == 0) {
		if (IS_VALID_PTR ((struct mtd_oob_buf32*)arg)) {
			ob32.start = ob64.start;
			ob32.ength = ob64.ength;
			ob32.ptr = (__u32)UL(ob64.ptr);

			if (g_safe_memcpy ((struct mtd_oob_buf32*)arg, &ob32, sizeof (struct mtd_oob_buf32)) < 0) {
				return -EFAULT;
			}

		}
	}

	return ret;
}

static long tiocgserial_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct serial_struct32  ss32;
	struct serial_struct    ss64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ss64);

	if (ret == 0) {
		memcpy (&ss32, &ss64, sizeof (__s32) * 10);
		ss32.closing_wait = ss64.closing_wait;
		ss32.closing_wait2 = ss64.closing_wait2;
		ss32.iomem_base = IS_HIGH_ZERO(UL(ss64.iomem_base))?
					(__u32)UL(ss64.iomem_base) : (__u32)NULL;
		ss32.iomem_reg_shift = ss64.iomem_reg_shift;
		ss32.port_high = ss64.port_high;
		ss32.reserved[0] = ss64.reserved[0];

		if (g_safe_memcpy(I2P(arg), &ss32, sizeof (struct serial_struct32)) < 0) {
			return -EFAULT;
		}
	}
	return ret;
}

static long tiocsserial_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct serial_struct32  ss32;
	struct serial_struct    ss64;
	long ret;

	if (g_safe_memcpy (&ss32, I2P(arg), sizeof (struct serial_struct32)) < 0) {
                return -EFAULT;
        }

	memcpy (&ss64, &ss32, sizeof (__s32) * 10);
	ss64.closing_wait = ss32.closing_wait;
	ss64.closing_wait2 = ss32.closing_wait2;
	ss64.iomem_base = (unsigned char *)UL(ss32.iomem_base);
	ss64.iomem_reg_shift = ss32.iomem_reg_shift;
	ss64.port_high = ss32.port_high;
	ss64.reserved[0] = ss32.reserved[0];

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ss64);

	return ret;
}

static long pppiocgidle_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ppp_idle32  pi32;
	struct ppp_idle    pi64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &pi64);

	if (ret == 0) {
		pi32.xmit_idle = pi64.xmit_idle;
		pi32.recv_idle = pi64.recv_idle;
		
		if (g_safe_memcpy(I2P(arg), &pi32, sizeof (struct ppp_idle32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long pppfilter_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct sock_fprog32  sf32;
	struct sock_fprog    sf64;
	long ret;

	if (g_safe_memcpy (&sf32, I2P(arg), sizeof (struct sock_fprog32)) < 0) {
                return -EFAULT;
        }

	sf64.len = sf32.len;
	sf64.filter = (struct sock_filter*)UL(sf32.filter);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &sf64);

	return ret;
}

static long pppiocscompress_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct ppp_option_data32  od32;
	struct ppp_option_data    od64;
	long ret;

	if (g_safe_memcpy (&od32, I2P(arg), sizeof (struct ppp_option_data32)) < 0) {
                return -EFAULT;
        }

	od64.ptr = (__u8 *)UL(od32.ptr);
	od64.length = od32.length;
	od64.transmit = od32.transmit;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &od64);

	return ret;
}

static long fbgfscreeninfo_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct fb_fix_screeninfo32  fs32;
	struct fb_fix_screeninfo    fs64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &fs64);

	if (ret == 0) {
		memcpy (&fs32, &fs64, sizeof (char) * 16);
		fs32.smem_start = (__u32)fs64.smem_start;
		fs32.smem_len = fs64.smem_len;
		fs32.type = fs64.type;	
		fs32.type_aux = fs64.type_aux;	
		fs32.visual = fs64.visual;
		fs32.xpanstep = fs64.xpanstep;	
		fs32.ypanstep = fs64.ypanstep;	
		fs32.ywrapstep = fs64.ywrapstep;	
		fs32.line_length = fs64.line_length;	
		fs32.mmio_start = (__u32)fs64.mmio_start;	
		fs32.mmio_len = fs64.mmio_len;	
		fs32.accel = fs64.accel;	
		memcpy (&fs32.reserved, &fs64.reserved, sizeof (__u16) * 3);
		
		if (g_safe_memcpy(I2P(arg), &fs32, sizeof (struct fb_fix_screeninfo32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long fbiogetcmap_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct fb_cmap32  fc32;
	struct fb_cmap    fc64;
	long ret;

	if (IS_VALID_PTR ((struct fb_cmap32 *)arg)) {
		if (g_safe_memcpy(&fc32, I2P(arg), sizeof (struct fb_cmap32)) < 0) {
			return -EFAULT;
		}

		fc64.start = fc32.start;
		fc64.len = fc32.len;
		fc64.red = (__u16 *)UL(fc32.red);
		fc64.green = (__u16 *)UL(fc32.green);
		fc64.blue = (__u16 *)UL(fc32.blue);
		fc64.transp = (__u16 *)UL(fc32.transp);
	}

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &fc64);

	if (ret == 0) {
		fc32.start = fc64.start;
		fc32.len = fc64.len;
		fc32.red = (__u32)UL(fc64.red);
		fc32.green = (__u32)UL(fc64.green);
		fc32.blue = (__u32)UL(fc64.blue);
		fc32.transp = (__u32)UL(fc64.transp);
		
		if (g_safe_memcpy(I2P(arg), &fc32, sizeof (struct fb_cmap32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long fbioputcmap_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct fb_cmap32  fc32;
	struct fb_cmap    fc64;
	long ret;

	if (g_safe_memcpy(&fc32, I2P(arg), sizeof (struct fb_cmap32)) < 0) {
		return -EFAULT;
	}

	fc64.start = fc32.start;
	fc64.len = fc32.len;
	fc64.red = (__u16 *)UL(fc32.red);
	fc64.green = (__u16 *)UL(fc32.green);
	fc64.blue = (__u16 *)UL(fc32.blue);
	fc64.transp = (__u16 *)UL(fc32.transp);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &fc64);

	return ret;
}

static long giounimap_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct unimapdesc32  um32;
	struct unimapdesc    um64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &um64);

	if (ret == 0) {
		um32.entry_ct = um64.entry_ct;
		um32.entries = (__u32)UL(um64.entries);
		
		if (g_safe_memcpy(I2P(arg), &um32, sizeof (struct unimapdesc32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long piounimap_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct unimapdesc32  um32;
	struct unimapdesc    um64;
	long ret;

	if (g_safe_memcpy(&um32, I2P(arg), sizeof (struct unimapdesc32)) < 0) {
		return -EFAULT;
	}

	um64.entry_ct = um32.entry_ct;
	um64.entries = (struct unipair *)UL(um32.entries);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &um64);

	return ret;
}

static long giofontx_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct consolefontdesc32  cf32;
	struct consolefontdesc    cf64;
	long ret;

	cf64.chardata = (char *) UL (((struct consolefontdesc32 *)arg)->chardata);
	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &cf64);

	if (ret == 0) {
		cf32.charcount = cf64.charcount;
		cf32.charheight = cf64.charheight;
		cf32.chardata = (__u32)UL(cf64.chardata);
		
		if (g_safe_memcpy(I2P(arg), &cf32, sizeof (struct consolefontdesc32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long piofontx_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct consolefontdesc32  cf32;
	struct consolefontdesc    cf64;
	long ret;

	if (g_safe_memcpy(&cf32, I2P(arg), sizeof (struct consolefontdesc32)) < 0) {
		return -EFAULT;
	}
	
	cf64.charcount = cf32.charcount;
	cf64.charheight = cf32.charheight;
	cf64.chardata = (char *)UL(cf32.chardata);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &cf64);

	return ret;
}

static long kdfonttop_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct console_font_op32  fo32;
	struct console_font_op    fo64;
	long ret;

	if (g_safe_memcpy(&fo32, I2P(arg), sizeof (struct console_font_op32)) < 0) {
		return -EFAULT;
	}
	
	fo64.op = fo32.op;
	fo64.flags = fo32.flags;
	fo64.width = fo32.width;
	fo64.height = fo32.height;
	fo64.charcount = fo32.charcount;
	fo64.data = (unsigned char *)UL(fo32.data);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &fo64);

	if (ret == 0) {
		fo32.op = fo64.op;
		fo32.flags = fo64.flags;
		fo32.width = fo64.width;
		fo32.height = fo64.height;
		fo32.charcount = fo64.charcount;
		fo32.data = (__u32)UL(fo64.data);

		if (g_safe_memcpy(I2P(arg), &fo32, sizeof (struct console_font_op32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long mtiocget_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct  mtget32  mg32;
	struct  mtget    mg64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &mg64);

	if (ret == 0) {
		mg32.mt_type = (__s32)mg64.mt_type;
		mg32.mt_resid = (__s32)mg64.mt_resid;
		mg32.mt_dsreg = (__s32)mg64.mt_dsreg;
		mg32.mt_gstat = (__s32)mg64.mt_gstat;
		mg32.mt_erreg = (__s32)mg64.mt_erreg;
		mg32.mt_fileno = mg64.mt_fileno;
		mg32.mt_blkno = mg64.mt_blkno;
		
		if (g_safe_memcpy(I2P(arg), &mg32, sizeof (struct  mtget32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long mtiocpos_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct  mtpos32  mp32;
	struct  mtpos    mp64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &mp64);

	if (ret == 0) {
		mp32.mt_blkno = (__s32)mp64.mt_blkno;
		
		if (g_safe_memcpy(I2P(arg), &mp32, sizeof (struct  mtpos32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long mtiocgetconf_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct mtconfiginfo32  ci32;
	struct mtconfiginfo    ci64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ci64);

	if (ret == 0) {
		ci32.mt_type = ci64.mt_type;
		ci32.ifc_type = ci64.ifc_type;
		ci32.irqnr = ci64.irqnr;
		ci32.dmanr = ci64.dmanr;
		ci32.port = ci64.port;
		ci32.debug = ci64.debug;
		memcpy ((unsigned char*)(&ci32.debug) + sizeof (__u32), 
			(unsigned char*)(&ci64.debug) + sizeof (unsigned long), 
			sizeof (__u32));
		memcpy (ci32.reserved, ci64.reserved, sizeof (char) * 10);

		if (g_safe_memcpy(I2P(arg), &ci32, sizeof (struct mtconfiginfo32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long mtiocsetconf_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct mtconfiginfo32  ci32;
	struct mtconfiginfo    ci64;
	long ret;

	if (g_safe_memcpy(&ci32, I2P(arg), sizeof (struct mtconfiginfo32)) < 0) {
		return -EFAULT;
	}

	ci64.mt_type = ci32.mt_type;
	ci64.ifc_type = ci32.ifc_type;
	ci64.irqnr = ci32.irqnr;
	ci64.dmanr = ci32.dmanr;
	ci64.port = ci32.port;
	ci64.debug = ci32.debug;
	memcpy ((unsigned char*)(&ci64.debug) + sizeof (unsigned long), 
		(unsigned char*)(&ci32.debug) + sizeof (__u32), 
		sizeof (__u32));
	memcpy (ci64.reserved, ci32.reserved, sizeof (char) * 10);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ci64);

	return ret;
}

static long cdromreadaudio_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct cdrom_read_audio32  ra32;
	struct cdrom_read_audio    ra64;
	long ret;

	if (g_safe_memcpy(&ra32, I2P(arg), sizeof (struct cdrom_read_audio32)) < 0) {
		return -EFAULT;
	}

	memcpy (&ra64.addr, &ra32.addr, sizeof (ra32.addr));
	ra64.addr_format = ra32.addr_format;
	ra64.nframes = ra32.nframes;
	ra64.buf = (__u8 *)UL(ra32.buf);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ra64);

	return ret;
}

static long cdromsendpacket_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct cdrom_generic_command32  gc32;
	struct cdrom_generic_command    gc64;
	long ret;

	if (g_safe_memcpy(&gc32, I2P(arg), sizeof (struct cdrom_generic_command32)) < 0) {
		return -EFAULT;
	}

	memcpy (gc64.cmd, gc32.cmd, CDROM_PACKET_SIZE);
	gc64.buffer = (unsigned char *)UL(gc32.buffer);
	gc64.buflen = gc32.buflen;
	gc64.stat = gc32.stat;
	gc64.sense = (void *)UL(gc32.sense);
	gc64.data_direction = gc32.data_direction;
	gc64.quiet = gc32.quiet;
	gc64.timeout = gc32.timeout;
	gc64.reserved[0] = (void *)UL(gc32.reserved[0]);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &gc64);

	return ret;
}

static long sgio_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	sg_io_hdr_t32  si32;
	sg_io_hdr_t    si64;
	sg_iovec_t32 * p32 = NULL;
	sg_iovec_t   * p64 = NULL;
	int need_free_iovec32 = 0;
	int need_free_iovec = 0;
	long ret;

	if (g_safe_memcpy(&si32, I2P(arg), sizeof (sg_io_hdr_t32)) < 0) {
		return -EFAULT;
	}

	si64.interface_id = si32.interface_id;
	si64.dxfer_direction = si32.dxfer_direction;
	si64.cmd_len = si32.cmd_len;
	si64.mx_sb_len = si32.mx_sb_len;
	si64.iovec_count = si32.iovec_count;
	si64.dxfer_len = si32.dxfer_len;
	si64.cmdp = (unsigned char *)UL(si32.cmdp);
	si64.sbp = (unsigned char *)UL(si32.sbp);
	si64.timeout = si32.timeout;
	si64.flags = si32.flags;
	si64.pack_id = si32.pack_id;
	si64.usr_ptr = (void *)UL(si32.usr_ptr);
	
	if (si64.iovec_count == 0) {
		si64.dxferp = (void *)UL(si32.dxferp); 
	}
	else {
		int i;

		p32 = (sg_iovec_t32 *)malloc (sizeof (sg_iovec_t32) * si64.iovec_count);
		if (p32 == NULL) {
			return -ENOMEM;
		}
		need_free_iovec32 = 1;

		if (g_safe_memcpy(p32, (sg_iovec_t32 *)UL(si32.dxferp), 
				sizeof (sg_iovec_t32) * si64.iovec_count) < 0) {
        	        ret = -EFAULT;
			goto out;
	        }

		p64 = (sg_iovec_t *)malloc (sizeof (sg_iovec_t) * si64.iovec_count);
		if (p64 == NULL) {
			ret =  -ENOMEM;
			goto out;
		}
		need_free_iovec = 1;
		
		for (i = 0; i < si64.iovec_count; i++) {
			p64[i].iov_base = (void *)UL(p32[i].iov_base);
			p64[i].iov_len = (size_t)UL(p32[i].iov_len);
		}
		si64.dxferp = p64;

		free (p32);
		need_free_iovec32 = 0;
	}

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &si64);

	if (ret == 0) {
		si32.pack_id = si64.pack_id;
		si32.usr_ptr = IS_HIGH_ZERO(UL(si64.usr_ptr))?
				(__u32)UL(si64.usr_ptr) : (__u32)NULL;
		si32.status = si64.status;
		si32.masked_status = si64.masked_status;
		si32.msg_status = si64.msg_status;
		si32.sb_len_wr = si64.sb_len_wr;
		si32.host_status = si64.host_status;
		si32.driver_status = si64.driver_status;
		si32.resid = si64.resid;
		si32.duration = si64.duration;
		si32.info = si64.info;

		if (g_safe_memcpy(I2P(arg), &si32, sizeof (sg_io_hdr_t32)) < 0) {
			ret = -EFAULT;
			goto out;
		}
	}

out:
	if (need_free_iovec32) {
		free (p32);
	}
	if (need_free_iovec) {
		free (p64);
	}
	return ret;
}

static long usbdevfs_control_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct usbdevfs_ctrltransfer32  uc32;
	struct usbdevfs_ctrltransfer    uc64;
	long ret;

	if (g_safe_memcpy(&uc32, I2P(arg), sizeof (struct usbdevfs_ctrltransfer32)) < 0) {
		return -EFAULT;
	}

	uc64.requesttype = uc32.requesttype;
	uc64.request = uc32.request;
	uc64.value = uc32.value;
	uc64.index = uc32.index;
	uc64.length = uc32.length;
	uc64.timeout = uc32.timeout;
	uc64.data = (void *)UL(uc32.data);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &uc64);

	if (ret == 0) {
		uc32.requesttype = uc64.requesttype;
		uc32.request = uc64.request;
		uc32.value = uc64.value;
		uc32.index = uc64.index;
		uc32.length = uc64.length;
		uc32.timeout = uc64.timeout;
		uc32.data = (__u32)UL(uc64.data);

		if (g_safe_memcpy(I2P(arg), &uc32, sizeof (struct usbdevfs_ctrltransfer32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long usbdevfs_bulk_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct usbdevfs_bulktransfer32  ub32;
	struct usbdevfs_bulktransfer    ub64;
	long ret;

	if (g_safe_memcpy(&ub32, I2P(arg), sizeof (struct usbdevfs_bulktransfer32)) < 0) {
		return -EFAULT;
	}

	ub64.ep = ub32.ep;
	ub64.len = ub32.len;
	ub64.timeout = ub32.timeout;
	ub64.data = (void *)UL(ub32.data);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ub64);

	if (ret == 0) {
		ub32.ep = ub64.ep;
		ub32.len = ub64.len;
		ub32.timeout = ub64.timeout;
		ub32.data = (__u32)UL(ub64.data);

		if (g_safe_memcpy(I2P(arg), &ub32, sizeof (struct usbdevfs_bulktransfer32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long usbdevfs_discsignal_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct usbdevfs_disconnectsignal32  ud32;
	struct usbdevfs_disconnectsignal    ud64;
	long ret;

	if (g_safe_memcpy(&ud32, I2P(arg), sizeof (struct usbdevfs_disconnectsignal32)) < 0) {
		return -EFAULT;
	}

	ud64.signr = ud32.signr;
	ud64.context = (void *)UL(ud32.context);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ud64);

	return ret;
}

#ifndef UNUSED_IOCTL
static long arg32_to_arg64_read_only(int fd, unsigned int cmd, unsigned long arg)
{
	unsigned int arg32;
	unsigned long arg64;
	long ret;

	if (g_safe_memcpy (&arg32, I2P(arg), sizeof (arg32)) < 0) {
                return -EFAULT;
        }

	arg64 = arg32;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &arg64);

	return ret;
}

static long vidiocgtuner_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct video_tuner32  vt32;
	struct video_tuner    vt64;
	long ret;

	if (g_safe_memcpy(&vt32, I2P(arg), sizeof (struct video_tuner32)) < 0) {
		return -EFAULT;
	}

	vt64.tuner = vt32.tuner;
	memcpy (vt64.name, vt32.name, 32);
	vt64.rangelow = vt32.rangelow;
	vt64.rangehigh = vt32.rangehigh;
	vt64.flags = vt32.flags;
	vt64.mode = vt32.mode;
	vt64.signal = vt32.signal;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &vt64);

	if (ret == 0) {
		vt32.tuner = vt64.tuner;
		memcpy (vt32.name, vt64.name, 32);
		vt32.rangelow = (__u32)vt64.rangelow;
		vt32.rangehigh = (__u32)vt64.rangehigh;
		vt32.flags = vt64.flags;
		vt32.mode = vt64.mode;
		vt32.signal = vt64.signal;

		if (g_safe_memcpy(I2P(arg), &vt32, sizeof (struct video_tuner32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long vidiocstuner_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct video_tuner32  vt32;
	struct video_tuner    vt64;
	long ret;

	if (g_safe_memcpy(&vt32, I2P(arg), sizeof (struct video_tuner32)) < 0) {
		return -EFAULT;
	}

	vt64.tuner = vt32.tuner;
	memcpy (vt64.name, vt32.name, 32);
	vt64.rangelow = vt32.rangelow;
	vt64.rangehigh = vt32.rangehigh;
	vt64.flags = vt32.flags;
	vt64.mode = vt32.mode;
	vt64.signal = vt32.signal;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &vt64);

	return ret;
}

static long vidiocgwin_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct video_window32  vw32;
	struct video_window    vw64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &vw64);

	if (ret == 0) {
		vw32.x = vw64.x;
		vw32.y = vw64.y;
		vw32.width = vw64.width;
		vw32.height = vw64.height;
		vw32.chromakey = vw64.chromakey;
		vw32.flags = vw64.flags;
		vw32.clips = (__u32)UL(vw64.clips);
		vw32.clipcount = vw64.clipcount;

		if (g_safe_memcpy(I2P(arg), &vw32, sizeof (struct video_window32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long vidiocswin_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct video_window32  vw32;
	struct video_window    vw64;
	struct video_clip32 *  p32 = NULL;
	struct video_clip *    p64 = NULL;
	int need_free_clip32 = 0;
	int need_free_clip = 0;
	long ret;

	if (g_safe_memcpy(&vw32, I2P(arg), sizeof (struct video_window32)) < 0) {
		return -EFAULT;
	}

	vw64.x = vw32.x;
	vw64.y = vw32.y;
	vw64.width = vw32.width;
	vw64.height = vw32.height;
	vw64.chromakey = vw32.chromakey;
	vw64.flags = vw32.flags;
	vw64.clipcount = vw32.clipcount;
	
	if (vw64.clipcount > 0) {
		int i;

		p32 = (struct video_clip32 *)malloc (sizeof (struct video_clip32) * vw64.clipcount);
		if (p32 == NULL) {
			return -ENOMEM;
		}
		need_free_clip32 = 1;

		if (g_safe_memcpy(p32, (struct video_clip32 *)UL(vw32.clips), 
				sizeof (struct video_clip32) * vw64.clipcount) < 0) {
			ret = -EFAULT;
			goto out;
		}

		p64 = (struct video_clip *)malloc (sizeof (struct video_clip) * vw64.clipcount);
		if (p64 == NULL) {
                        ret = -ENOMEM;
			goto out;
                }
                need_free_clip = 1;

		for (i = 0 ; i < vw64.clipcount ; i++) {
			memcpy (&p64[i], &p32[i], sizeof (__s32) * 4);
			/* next is unused */
			p64[i].next = NULL;
		}
		vw64.clips = p64;
		
		free (p32);
		need_free_clip32 = 0;
	}
	else {
		vw64.clips = (struct  video_clip *)UL(vw32.clips);
	}

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &vw64);

out:
	if (need_free_clip32) {
		free (p32);
	}
	if (need_free_clip) {
		free (p64);
	}
	return ret;
}

static long vidiocgfbuf_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct video_buffer32  vb32;
	struct video_buffer    vb64;
	long ret;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &vb64);

	if (ret == 0) {
		vb32.base = (__u32)UL(vb64.base);
		vb32.height = vb64.height;
		vb32.width = vb64.width;
		vb32.depth = vb64.depth;
		vb32.bytesperline = vb64.bytesperline;

		if (g_safe_memcpy(I2P(arg), &vb32, sizeof (struct video_buffer32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long vidiocsfbuf_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct video_buffer32  vb32;
	struct video_buffer    vb64;
	long ret;

	if (g_safe_memcpy(&vb32, I2P(arg), sizeof (struct video_buffer32)) < 0) {
		return -EFAULT;
	}

	vb64.base = (void *)UL(vb32.base);
	vb64.height = vb32.height;
	vb64.width = vb32.width;
	vb64.depth = vb32.depth;
	vb64.bytesperline = vb32.bytesperline;

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &vb64);

	return ret;
}

static long atmif_sioc_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct atmif_sioc32   as32;
	struct atmif_sioc     as64;
	long ret;

	if (g_safe_memcpy (&as32, I2P(arg), sizeof (struct atmif_sioc32)) < 0) {
                return -EFAULT;
        }

	as64.number = as32.number;
	as64.length = as32.length;
	as64.arg = (void *)UL(as32.arg);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &as64);

	if (ret == 0) {
		as32.number = as64.number;
		as32.length = as64.length;
		as32.arg = (__u32)UL(as64.arg);

		if (g_safe_memcpy(I2P(arg), &as32, sizeof (struct atmif_sioc32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

static long atm_iobuf_ioctl32(int fd, unsigned int cmd, unsigned long arg)
{
	struct atm_iobuf32   ai32;
	struct atm_iobuf     ai64;
	long ret;

	if (g_safe_memcpy (&ai32, I2P(arg), sizeof (struct atm_iobuf32)) < 0) {
                return -EFAULT;
        }

	ai64.length = ai32.length;
	ai64.buffer = (void *)UL(ai32.buffer);

	ret = LONG_SYSCALL3 (ioctl, fd, cmd, &ai64);

	if (ret == 0) {
		ai32.length = ai64.length;
		ai32.buffer = (__u32)UL(ai64.buffer);

		if (g_safe_memcpy(I2P(arg), &ai32, sizeof (struct atm_iobuf32)) < 0) {
			return -EFAULT;
		}
	}

	return ret;
}

#endif




long HIDDEN sys32_ioctl (int fd, unsigned int cmd, unsigned long arg)
{
	int ret;
	unsigned int hash_value;
	long t;

	DBPRINT (0, "fd=%d cmd=%u arg=%lx\n", fd, cmd, arg);

	hash_value = ioctl_hash_value(cmd);

	t = hash_value;
	while ((t != -1) && (ioctl_handler_table[t].cmd != cmd)) {
		t = ioctl_handler_table[t].next;
	}
	
	if (t != -1) {
		/* the cmd is found in handler-table */
		if (ioctl_handler_table[t].handler == NULL) {
			/* no special handler needed */
			ret = LONG_SYSCALL3 (ioctl, fd, ioctl_handler_table[t].new_cmd, arg);
		}
		else {
			ret = ioctl_handler_table[t].handler(fd, ioctl_handler_table[t].new_cmd, arg);
		}
	} else {
		ret = LONG_SYSCALL3 (ioctl, fd, cmd, arg);
	}
	
	return ret;
}
