/*
 * ia32_memory.c
 *
 * wrapper functions to memory-related syscalls that are common for
 * both external and internal wrappers
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Arun Sharma <arun.sharma@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "user32.h"
#include "user32_mem_regions.h"

long HIDDEN
sys32_msync (unsigned int start, unsigned int len, int flags)
{
	unsigned long addr;
	int page_exists;
	long retval;

	DBPRINT (0, "start=%x len=%x flags=%x\n", start, len, flags);
	if (IA32_PAGE_OFF(start))
		return -EINVAL;
	addr = PAGE_START(start);
	/* check explicitly for argument that fall inside page 0 - 
	 * this page is mapped for native 64 processes
	 */
	if (addr == 0) {
 	 	get_page_prot_e(addr, &page_exists);
		if (!page_exists)
			return -ENOMEM;
	}
	retval = SYSCALL (msync, (addr, len + (start - addr), flags));
	return retval;
}

long HIDDEN
sys32_mlock (unsigned int addr, unsigned int len)
{
	int page_exists;
	long retval;

	/* check explicitly for argument that fall inside page 0 - 
	 * this page is mapped for native 64 processes
	 */
	if (addr == 0) {
 	 	get_page_prot_e(addr, &page_exists);
		if (!page_exists)
			return -ENOMEM;
	}
	retval = SYSCALL (mlock, (addr, len));
	return retval;
}

long HIDDEN
sys32_munlock (unsigned int addr, unsigned int len)
{
	int page_exists;
	long retval;

	/* check explicitly for argument that fall inside page 0 - 
	 * this page is mapped for native 64 processes
	 */
	if (addr == 0) {
 	 	get_page_prot_e(addr, &page_exists);
		if (!page_exists)
			return -ENOMEM;
	}
	retval = SYSCALL (munlock, (addr, len));
	return retval;
}

/* Emulating a smaller page size than the system page size requires setting
 * the permissions more liberally than on a native system. Some faults which
 * are expected on a native system may never happen on the emulated
 * system. In practice, this is not a problem.
 */
static inline long 
subpage_manip_mprotect32 (unsigned long addr, int prot)
{
	int old_prot = 0;

	old_prot = get_page_prot(addr);
	return  SYSCALL (mprotect, (addr, PAGE_SIZE, prot | old_prot));
}

/*
 * check if the address is out the region range
 */
static inline long
is_out_of_region (unsigned long addr, unsigned long end)
{
	void * region, * next;
	long retval = 0;
	
	retval = find_region (addr, &region);
	if (retval != 0) {
		return retval;
	}
	
	if (REGION_ENDADDR(region) < end) {
		next = NEXT_REGION (region);
		if ((next == 0) || (REGION_STARTADDR (next) > REGION_ENDADDR (region))) {
			/* [addr, end] is not totally mmapped */
			retval = -1;
		}
	}

	return retval;
}

/* check the mem region of mprotect, try to reduce the 4K
 * page size impact.
 */
static inline long 
mprotect32_check_region (unsigned long *p_addr, unsigned long *p_end, int prot)
{
	void *region, *prev, *next;
	unsigned long pstart, pend;
	int need_subpage_manip = 0;
	long retval = 0;
	
	retval = find_region (*p_addr, &region);
	if (retval != 0) {
		return -ENOMEM;
	}
	
	pstart = PAGE_START(*p_addr);
	if (*p_addr > pstart) {
		need_subpage_manip = 0;
		if (*p_addr == REGION_STARTADDR (region)) {
			/* need more judgements */
			prev = PREV_REGION (region);
		
			if (prev != 0) {
				if (pstart < PAGE_ALIGN(REGION_ENDADDR (prev))) {
					/* pstart and PAGE_ALIGN(region_end) is in one page */
					need_subpage_manip = 1;
				}
			}
		} else {
			need_subpage_manip = 1;
		}
		
		if (need_subpage_manip == 1) {
			retval = subpage_manip_mprotect32 (pstart, prot);
			if (retval != 0) {
				return retval;
			}
				
			*p_addr = PAGE_ALIGN(*p_addr);
		} else {
			*p_addr = pstart;
		}
	}
	
	pend = PAGE_START(*p_end);
	if (pend < *p_end) {
		need_subpage_manip = 1;

		if (*p_end == REGION_ENDADDR (region)) {
			next = NEXT_REGION (region);
		
			if (next != 0) {
				if (PAGE_START(REGION_STARTADDR (next)) > PAGE_ALIGN(*p_end)) {
					need_subpage_manip = 0;
				}
			}
		}

		if (need_subpage_manip) {
			retval = subpage_manip_mprotect32 (pend, prot);
			if (retval != 0) {
				return retval;
			}
		
			*p_end = pend;
		}
	}

	return retval;
}

/*
 * resetup the memory permission in [start, end], return the real length
 * of the memory which is impacted by mprotect.
 * in this function, we repeat calling mprotect() for every native page
 * in the range until we get a failure.
 */
static inline unsigned long
resetup_region_prot (unsigned long start, unsigned long end, int prot)
{
	unsigned long addr = start;
	unsigned long len = 0;
	long retval;

	while ((addr < end) && (end - addr >= PAGE_SIZE)) {
		retval = SYSCALL (mprotect, (addr, PAGE_SIZE, prot));
		if (IS_SYS_ERROR (retval)) {
			/* mprotect failed on this native page, no more work needed */
			addr -= PAGE_SIZE;
			break;
		}
		
		len += PAGE_SIZE;
		addr += PAGE_SIZE;
	}

	/* the last page may be missed by above loop */
	if ((addr < end) && (end - addr < PAGE_SIZE)) {
		retval = SYSCALL (mprotect, (addr, end - addr, prot));
		if (!IS_SYS_ERROR (retval)) {
			len += end - addr;
		}
	}

	return len;
}

long HIDDEN
mprotect32_none4k (unsigned int start, unsigned int len, int prot)
{
	unsigned long addr = start;
	unsigned long end = addr + len;
	unsigned long pstart;
	long retval;
	int  page_exists;

	prot = convert_prot64_to_prot32(prot);

	DBPRINT (0, "start=%x len=%x prot=%x\n", start, len, prot);

	pstart = PAGE_START(start);

	/* filter out illegal requests */
	retval = -EINVAL;
	if (pstart == 0) {
 	 	get_page_prot_e (pstart, &page_exists);
		if (!page_exists) {
			return -ENOMEM;
		}
	}

	/* addr is not 4k page aligned */
	if (IA32_PAGE_OFF(addr)) {
		goto out;
	}

	end = IA32_PAGE_ALIGN(end);
	if (end < addr) {
		goto out;
	}

	mutex_lock (g_memory_sync);
	
	retval = mprotect32_check_region (&addr, &end, prot);
	if (retval != 0) {
		goto out_rel_sem;
	}

	if (end <= addr) 
		goto out_set_region;

	retval = SYSCALL (mprotect, (addr, end - addr, prot));
	if (IS_SYS_ERROR (retval)) {
		if (retval != -EINVAL) {
			/* 
			 * POSIX: When mprotect() fails for reasons other
			 *        than [EINVAL], the protections on some of
			 *        the pages in the range [addr, addr + len) 
			 *        may have been changed.
			 * So we need to check the real region permission.
			 * As POSIX does not point out which rule mprotect() will
			 * follow in this kind of situation, below resetup_region_prot ()
			 * is just imitating what kernel acts now, and may be 
			 * changed in the future.
			 */
			len = resetup_region_prot (addr, end, prot);
			set_region_prot (addr, IA32_PAGE_ALIGN(len), prot);
		}
		
		goto out_rel_sem;
	}

out_set_region:
	if (is_out_of_region (start, start + len) != 0) {
		retval = -ENOMEM;
	} else {
		retval = set_region_prot (start, IA32_PAGE_ALIGN(len), prot);
	}

out_rel_sem:
	mutex_unlock (g_memory_sync);
out:
	DBPRINT (0, "retval=%ld\n", retval);
	return retval;
}

