/* 
 * ia32_networking.c
 *
 * wrapper functions for syscalls related with networking.
 * common for both internal and external wrappers.
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>

#include "user32.h"
#include "ia32_networking.h"

#define SOCKETCALL_ARGUMENTS_SIZE(i) (i * sizeof(__u32))

static const unsigned char socketcall_arguments_size[18]={
	0,                            /* dummy */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_socket(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_bind(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_connect(2) */
	SOCKETCALL_ARGUMENTS_SIZE(2),  /* sys_listen(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_accept(2) */
    	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_getsockname(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_getpeername(2) */
	SOCKETCALL_ARGUMENTS_SIZE(4),  /* sys_socketpair(2) */
	SOCKETCALL_ARGUMENTS_SIZE(4),  /* sys_send(2) */
	SOCKETCALL_ARGUMENTS_SIZE(4),  /* sys_recv(2) */
	SOCKETCALL_ARGUMENTS_SIZE(6),  /* sys_sendto(2)	*/
    	SOCKETCALL_ARGUMENTS_SIZE(6),  /* sys_recvfrom(2) */
	SOCKETCALL_ARGUMENTS_SIZE(2),  /* sys_shutdown(2) */
	SOCKETCALL_ARGUMENTS_SIZE(5),  /* sys_setsockopt(2) */
	SOCKETCALL_ARGUMENTS_SIZE(5),  /* sys_getsockopt(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3),  /* sys_sendmsg(2) */
	SOCKETCALL_ARGUMENTS_SIZE(3)}; /* sys_recvmsg(2) */


long HIDDEN
sys32_socketcall (int call, __u32 args)
{
	int retval;
	__u32 syscall_parameters[6];

	if (g_safe_memcpy(syscall_parameters, I2P(args), 
			  socketcall_arguments_size[call]) < 0) {
		return -EFAULT;
	}

	switch(call) {
	case SYS_SOCKET:
		retval = SYSCALL(socket,
			         (syscall_parameters[0], 
			          syscall_parameters[1], 
			          syscall_parameters[2]));
		break;
	case SYS_BIND:
		retval = SYSCALL(bind,
			         (syscall_parameters[0], 
			          (struct sockaddr *)UL(syscall_parameters[1]), 
			          syscall_parameters[2]));
		break;
	case SYS_CONNECT:
		retval = SYSCALL(connect,
				 (syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
				  syscall_parameters[2]));
		break;
	case SYS_LISTEN:
		retval = LONG_SYSCALL2(listen,
				       syscall_parameters[0], 
				       syscall_parameters[1]);
		break;
	case SYS_ACCEPT:
		retval = SYSCALL(accept,
			         (syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
			          (int *)UL(syscall_parameters[2])));
		break;
	case SYS_GETSOCKNAME:
		retval = SYSCALL(getsockname,
				 (syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
				  (int *)UL(syscall_parameters[2])));
			break;
	case SYS_GETPEERNAME:
		retval = SYSCALL(getpeername,
				 (syscall_parameters[0], 
				  (struct sockaddr *)UL(syscall_parameters[1]), 
				  (int *)UL(syscall_parameters[2])));
		break;
	case SYS_SOCKETPAIR:
		retval = SYSCALL(socketpair,
				 (syscall_parameters[0], 
				  syscall_parameters[1], 
				  syscall_parameters[2], 
				  (int *)UL(syscall_parameters[3])));
		break;
	case SYS_SEND:
		retval = LONG_SYSCALL4(send, 
				       syscall_parameters[0], 
				       (void *)UL(syscall_parameters[1]), 
				       syscall_parameters[2], 
				       syscall_parameters[3]);
		break;
	case SYS_SENDTO:
		retval = LONG_SYSCALL6(sendto, 
				       syscall_parameters[0], 
				       (void *)UL(syscall_parameters[1]), 
				       syscall_parameters[2], 
				       syscall_parameters[3], 
				       (struct sockaddr*)UL(syscall_parameters[4]), 
				       syscall_parameters[5]);
		break;
	case SYS_RECV:
		retval = LONG_SYSCALL4(recv, 
				       syscall_parameters[0], 
				       (void *)UL(syscall_parameters[1]), 
				       syscall_parameters[2], 
				       syscall_parameters[3]);
		break;
	case SYS_RECVFROM:
		retval = LONG_SYSCALL6(recvfrom, 
				       syscall_parameters[0], 
				       (void*)UL(syscall_parameters[1]), 
				       syscall_parameters[2], 
				       syscall_parameters[3], 
				       (struct sockaddr*)UL(syscall_parameters[4]), 
				       (void*)UL(syscall_parameters[5]));
		break;
	case SYS_SHUTDOWN:
		retval = SYSCALL(shutdown,
				 (syscall_parameters[0],
				  syscall_parameters[1]));
		break;
	case SYS_SETSOCKOPT:
		retval = SYSCALL(setsockopt,
				 (syscall_parameters[0], 
				  syscall_parameters[1], 
				  syscall_parameters[2], 
				  (char *)UL(syscall_parameters[3]), 
				  syscall_parameters[4]));
		break;
	case SYS_GETSOCKOPT:
		retval = SYSCALL(getsockopt,
				 (syscall_parameters[0], 
				  syscall_parameters[1], 
				  syscall_parameters[2], 
				  (void*)UL(syscall_parameters[3]), 
				  (void*)UL(syscall_parameters[4])));
			break;
	case SYS_SENDMSG:
		retval = sys32_sendmsg(syscall_parameters[0], 
				       (__u32) UL(syscall_parameters[1]), 
				       syscall_parameters[2]);
		break;
	case SYS_RECVMSG:
		retval = sys32_recvmsg(syscall_parameters[0], 
			   	       (__u32) UL(syscall_parameters[1]), 
				       syscall_parameters[2]);
		break;
	default:
		retval = -EINVAL;
		break;
	}

	return retval;
}


static int
get_epoll_event (struct epoll_event32 *epev, struct epoll_event *epev64)
{
	int err = 0;
	struct epoll_event32 epev32;

	if (g_safe_memcpy (&epev32, epev, sizeof(epev32)) < 0) {
		return -EFAULT;
	}
	epev64->events = epev32.events;
	memcpy(&epev64->data, &epev32.data_high, sizeof(epev64->data));
	
	return err;
}

static int
put_epoll_event (struct epoll_event32 *epev, struct epoll_event *epev64)
{
	int err = 0;
	struct epoll_event32 epev32;

	epev32.events = epev64->events;
	memcpy(&epev32.data_high, &epev64->data, sizeof(epev64->data));
	if (g_safe_memcpy (epev, &epev32, sizeof(epev)) < 0) {
		err = -EFAULT;
	}
	
	return err;
}

long HIDDEN
sys32_epoll_ctl(int epfd, int op, int fd, __u32 event)
{
	int retval;
	struct epoll_event epev64;

	if (get_epoll_event((struct epoll_event32 *)I2P(event), &epev64) < 0) {
		return -EFAULT;
	}

	retval = SYSCALL (epoll_ctl, (epfd, op, fd, &epev64));
	
	return retval;
}

long HIDDEN
sys32_epoll_wait(int epfd, __u32 events, int maxevents, int timeout)
{
	int retval;
	int i;
	struct epoll_event *pepev64;

	if (maxevents <= 0) {
		return -EINVAL;
	}
	pepev64 = malloc(maxevents * sizeof(struct epoll_event));

	if (pepev64 == NULL) {
		return -ENOMEM;
	}
	retval = LONG_SYSCALL4 (epoll_wait, epfd, pepev64, maxevents, timeout);
	if (retval < 0) {
		free(pepev64);
		return retval;
	}
	for (i = 0; i < retval; i++) {
		if (put_epoll_event(&(((struct epoll_event32 *)I2P(events))[i]), 
				    &pepev64[i]) < 0) {
			return -EFAULT;
		}				
	}	
	free(pepev64);
	return retval;
}

