/*
 *  ia32_process.c
 *
 * wrapper functions for syscalls related with process.
 * common for both internal and external wrappers.
 *  
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>
#include "user32.h"

static int
put_rusage (struct rusage32 *ru, struct rusage *r)
{
	int err = 0;
	struct rusage32 ru32;

	ru32.ru_utime.tv_sec = r->ru_utime.tv_sec;
	ru32.ru_utime.tv_usec = r->ru_utime.tv_usec;
	ru32.ru_stime.tv_sec = r->ru_stime.tv_sec;
	ru32.ru_stime.tv_usec = r->ru_stime.tv_usec;
	ru32.ru_maxrss = r->ru_maxrss;
	ru32.ru_ixrss = r->ru_ixrss;
	ru32.ru_idrss = r->ru_idrss;
	ru32.ru_isrss = r->ru_isrss;
	ru32.ru_minflt = r->ru_minflt;
	ru32.ru_majflt = r->ru_majflt;
	ru32.ru_nswap = r->ru_nswap;
	ru32.ru_inblock = r->ru_inblock;
	ru32.ru_oublock = r->ru_oublock;
	ru32.ru_msgsnd = r->ru_msgsnd;
	ru32.ru_msgrcv = r->ru_msgrcv;
	ru32.ru_nsignals = r->ru_nsignals;
	ru32.ru_nvcsw = r->ru_nvcsw;
	ru32.ru_nivcsw = r->ru_nivcsw;
	if (g_safe_memcpy (ru, &ru32, sizeof (ru32)) < 0) {
		err = -EFAULT;
	}
	return err;
}


long HIDDEN
sys32_wait4 (int pid, int *status, int options, __u32 rusage)
{
	int    retval;
	struct rusage ru64;
	
	retval = LONG_SYSCALL4 (wait4, pid, status, options, 
				I2P(rusage)? &ru64 : NULL);
	
	// if rusage is not null, we should fill it with accounting information
	if (I2P(rusage) && retval > 0) { 
		if (put_rusage((struct rusage32 *)I2P(rusage), &ru64) < 0) {
			retval = -EFAULT;
		}
	}
	
	return retval;
}


long HIDDEN
sys32_waitpid (int pid, int *status, int options)
{
	// same functionality as wait4 with ru argument as null
	return sys32_wait4(pid, status, options, 0); 
}


long HIDDEN
sys32_getrusage (int who, __u32 rusage)
{
	int retval;
	struct rusage ru64;

	retval = SYSCALL (getrusage, (who, &ru64));

	if (retval == 0) {
		if (put_rusage((struct rusage32 *)I2P(rusage), &ru64) < 0) {
			retval = -EFAULT;
		}
	}
	return retval;
}


static int
put_timespec (struct timespec32 *tp, struct timespec *tp64)
{
	int err = 0;
	struct timespec32 tp32;

	tp32.tv_sec = tp64->tv_sec;
	tp32.tv_nsec = tp64->tv_nsec;
	if (g_safe_memcpy (tp, &tp32, sizeof (tp32)) < 0) {
		err = -EFAULT;
	}
	return err;
}


long HIDDEN
sys32_sched_rr_get_interval (pid_t pid, __u32 tp)
{
	int retval;
	struct timespec tp64;
	
	retval = SYSCALL (sched_rr_get_interval, (pid, &tp64));
	if (retval == 0) {
		if (put_timespec((struct timespec32 *)I2P(tp), &tp64) < 0) {
			retval = -EFAULT;
		}
	}
		
	return retval;
}

struct rlimit32 {
	unsigned rlim_cur;
	unsigned rlim_max;
};

#define OLD_GETRLIMIT 1
#define NEW_GETRLIMIT 2
static int
put_rlimit (struct rlimit32 *rlim, struct rlimit *rlim64, int version)
{
	int err = 0;
	struct rlimit32 rlim32;

	assert((version == OLD_GETRLIMIT) || (version == NEW_GETRLIMIT));
	if (version == OLD_GETRLIMIT) {
		rlim32.rlim_cur = (rlim64->rlim_cur > LEGACY_IA32_RLIM_INFINITY) ? 
			LEGACY_IA32_RLIM_INFINITY : rlim64->rlim_cur;
		rlim32.rlim_max = (rlim64->rlim_max > LEGACY_IA32_RLIM_INFINITY) ? 
			LEGACY_IA32_RLIM_INFINITY : rlim64->rlim_max;
	}
	else { // version == NEW_GETRLIMIT
		rlim32.rlim_cur = (rlim64->rlim_cur >= IA32_RLIM_INFINITY) ? 
			IA32_RLIM_INFINITY : rlim64->rlim_cur;
		rlim32.rlim_max = (rlim64->rlim_max >= IA32_RLIM_INFINITY) ? 
			IA32_RLIM_INFINITY : rlim64->rlim_max;
	}
	if (g_safe_memcpy (rlim, &rlim32, sizeof (rlim32)) < 0) {
		err = -EFAULT;
	}
	return err;
}

long HIDDEN
sys32_getrlimit (unsigned int resource, __u32 rlim)
{
	int retval;
	struct rlimit rlim64;

	retval = SYSCALL (getrlimit, (resource, &rlim64));

	if (!IS_SYS_ERROR (retval)) { //syscall64 succeeded
		if (put_rlimit ((struct rlimit32 *)I2P(rlim), 
				&rlim64, NEW_GETRLIMIT) < 0) {
			retval = -EFAULT;
		}	
	}
	return retval;
}


long HIDDEN
sys32_old_getrlimit (unsigned int resource, __u32 rlim)
{
	int retval;
	struct rlimit rlim64;

	retval = SYSCALL (getrlimit, (resource, &rlim64));

	if (!IS_SYS_ERROR (retval)) { //syscall64 succeeded
		if (put_rlimit ((struct rlimit32 *)I2P(rlim), 
				&rlim64, OLD_GETRLIMIT) < 0) {
			retval = -EFAULT;
		}	
	}
	return retval;
}


long HIDDEN
sys32_setrlimit (unsigned int resource, __u32 rlim)
{
	int retval;
	struct rlimit rlim64;
	struct rlimit32 rlim32;

	if (resource >= RLIM_NLIMITS) {
		return -EINVAL;
	}

	if (g_safe_memcpy (&rlim32, I2P(rlim), sizeof(rlim32)) < 0) {
		return -EFAULT;
	}

	DBPRINT (0, "resource=%u rlim_cur=%u<%x> rlim_max=%u<%x>\n",
		 resource, rlim32.rlim_cur, rlim32.rlim_cur,
		 rlim32.rlim_max, rlim32.rlim_max);

	rlim64.rlim_cur = (rlim32.rlim_cur == IA32_RLIM_INFINITY) ? 
		RLIM_INFINITY : rlim32.rlim_cur;
	rlim64.rlim_max = (rlim32.rlim_max == IA32_RLIM_INFINITY) ? 
		RLIM_INFINITY : rlim32.rlim_max;
	retval = SYSCALL (setrlimit, (resource, &rlim64));	


	return retval;
}


long HIDDEN
sys32_getresuid16 (__u32 ruid, __u32 euid, __u32 suid)
{
	int retval;
	uid_t ruid64, euid64, suid64;
	__u16 ruid32, euid32, suid32;

	retval = SYSCALL (getresuid, (&ruid64, &euid64, &suid64));
	
	if (retval == 0) {
		ruid32 = ruid64;
		euid32 = euid64;
		suid32 = suid64;
		if ((g_safe_memcpy(I2P(ruid), &ruid32, sizeof(ruid)) < 0) || 
		    (g_safe_memcpy(I2P(euid), &euid32, sizeof(euid)) < 0) || 
		    (g_safe_memcpy(I2P(suid), &suid32, sizeof(suid)) < 0)) {
			return -EFAULT;  
		}	
	}
	return retval;	
}


long HIDDEN
sys32_getresgid16 (__u32 rgid, __u32 egid, __u32 sgid)
{
	int retval;
	gid_t rgid64, egid64, sgid64;
	__u16 rgid32, egid32, sgid32;

	retval = SYSCALL (getresgid, (&rgid64, &egid64, &sgid64));
	
	if (retval == 0) {
		rgid32 = rgid64;
		egid32 = egid64;
		sgid32 = sgid64;
		if ((g_safe_memcpy(I2P(rgid), &rgid32, sizeof(rgid)) < 0) || 
		    (g_safe_memcpy(I2P(egid), &egid32, sizeof(egid)) < 0) || 
		    (g_safe_memcpy(I2P(sgid), &sgid32, sizeof(sgid)) < 0)) {
			return -EFAULT;  
		}	
	}
	return retval;	
}


long HIDDEN
sys32_nanosleep (__u32 req, __u32 rem)
{
	int retval;
	struct timespec req64;
	struct timespec rem64;
	struct timespec32 ts32;
	
	if (g_safe_memcpy (&ts32, I2P(req), sizeof(ts32)) < 0) {
		return -EFAULT;
	}
	req64.tv_sec = ts32.tv_sec;
	req64.tv_nsec = ts32.tv_nsec;

	retval = LONG_SYSCALL2 (nanosleep, &req64, VALID_PTR (I2P(rem), &rem64));

	if (retval == -EINTR) { // syscall64 was interrupted
		if (I2P(rem)) { // we need to set the remaining time
			if (put_timespec ((struct timespec32 *)I2P(rem), &rem64) < 0) {
				return -EFAULT;
			}	
		}
	}
	return retval;
}


#define LEGAL_GROUP_NUMBER(group_number) \
	(((group_number > 0) && (group_number <= NGROUPS)) ? 1 : 0)

long HIDDEN
sys32_getgroups16 (int gidsetsize, __u32 grouplist)
{
	int retval;
	gid_t gl64[NGROUPS];
	__u32 gl32[NGROUPS];
	int i;

	retval = SYSCALL (getgroups, (gidsetsize, gl64));

	if ((gidsetsize > 0) && LEGAL_GROUP_NUMBER(retval)) {
		for (i = 0; i < retval; i++) {
			gl32[i] = gl64[i];
		}	
		if (g_safe_memcpy (I2P(grouplist), gl32, 
				   gidsetsize * sizeof(__u32)) < 0) {
			return -EFAULT;
		}
	}
	
	return retval;
}


long HIDDEN
sys32_setgroups16 (int gidsetsize, __u32 grouplist)
{
	int retval;
	gid_t gl64[NGROUPS];
	__u32 gl32[NGROUPS];
	int i;

	if (g_safe_memcpy (gl32, I2P(grouplist), gidsetsize * sizeof(__u32)) < 0) {
		return -EFAULT;
	}
	for (i = 0; i < gidsetsize; i++) {
		gl64[i] = gl32[i];
	}

	retval = SYSCALL (setgroups, (gidsetsize, gl64));

	return retval;
}


static int
get_timespec (struct timespec32 *ts, struct timespec *ts64)
{
	int err = 0;
	struct timespec32 ts32;

	if (g_safe_memcpy (&ts32, ts, sizeof (ts32)) < 0) {
		err = -EFAULT;
	}
	ts64->tv_sec = ts32.tv_sec;
	ts64->tv_nsec = ts32.tv_nsec;
	
	return err;
}

long HIDDEN
sys32_futex (__u32 futex, int op, int val, __u32 timeout, __u32 futex2)
{
	int retval;
	struct timespec ts64;
	struct timespec *pts64 = (struct timespec *)(long)timeout;

	if ((op == FUTEX_WAIT) && (I2P(timeout) != NULL)) {
		if (get_timespec ((struct timespec32 *)I2P(timeout), &ts64) < 0) {
			return -EFAULT;
		}
		pts64 = &ts64;	
	}

	retval = LONG_SYSCALL5 (futex, I2P(futex), op, val, pts64, I2P(futex2));
	
	return retval;
}


/* the following 2 functions do not need any casting, */
/* but they do not exist on 64 bit platform, and thus */
/* need to be handled in the wrapper                  */

long HIDDEN
sys32_getpgrp ()
{
	return SYSCALL (getpgid, (0));
}

long HIDDEN
sys32_pause (void)
{
	int retval;
	k_sigset_t set;

	retval = SYSCALL (rt_sigprocmask, (SIG_BLOCK, NULL, &set, sizeof(set)));
	assert (retval == 0);

	return LONG_SYSCALL2 (rt_sigsuspend, &set, sizeof(set));
}

