/* 
 * ia32_time.c
 *
 * wrapper functions for syscalls related with time.
 * common for both internal and external wrappers.
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Gordon Jin <gordon.jin@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>

#include "ia32ss_common.h"
#include "user32.h"
#if NOKERNELALTROOT
#include "user32_pathwalk.h"
#endif

struct itimerval32
{
	struct timeval32 it_interval;
	struct timeval32 it_value;
};


static inline int
lp64_itimerval (struct itimerval *it, struct itimerval32 *it32)
{
	struct itimerval32 it32_tmp;

	if (g_safe_memcpy (&it32_tmp, it32, sizeof(it32_tmp)) < 0) {
		return -EFAULT;
	}

	it->it_interval.tv_sec = it32_tmp.it_interval.tv_sec ;
	it->it_interval.tv_usec = it32_tmp.it_interval.tv_usec ;
	it->it_value.tv_sec = it32_tmp.it_value.tv_sec ;
	it->it_value.tv_usec = it32_tmp.it_value.tv_usec;

	return 0;
}

static inline int
ilp32_itimerval (struct itimerval32 *it32, struct itimerval *it)
{
	struct itimerval32 it32_tmp;

	it32_tmp.it_interval.tv_sec = it->it_interval.tv_sec;
	it32_tmp.it_interval.tv_usec = it->it_interval.tv_usec;
	it32_tmp.it_value.tv_sec = it->it_value.tv_sec;
	it32_tmp.it_value.tv_usec = it->it_value.tv_usec;

	if (g_safe_memcpy (it32, &it32_tmp, sizeof (it32_tmp)) < 0) {
		return -EFAULT;
	}
	return 0;
}


long HIDDEN
sys32_getitimer (int which, __u32 it)
{
	struct itimerval it_tmp;
	int result;

	result = SYSCALL (getitimer, (which, &it_tmp));
	if (result) {
		return result;
	}
	result = ilp32_itimerval((struct itimerval32 *)I2P(it), &it_tmp);
	return result;
}


long HIDDEN
sys32_setitimer (int which, __u32 value, __u32 ovalue)
{
	struct itimerval kvalue, kovalue;
	struct itimerval * kvaluep;
	int result;

	if (value) {
		DBPRINT (0, "interval=%d:%d\n", 
							((struct itimerval32*)I2P(value))->it_value.tv_sec, ((struct itimerval32*)I2P(value))->it_value.tv_usec);
		if (lp64_itimerval(&kvalue, (struct itimerval32*)I2P(value))) {
			return -EFAULT;
		}
		kvaluep = &kvalue;
	} else {
		kvaluep = NULL;
	}

	result = SYSCALL (setitimer, (which, kvaluep, VALID_PTR(I2P(ovalue),&kovalue)));

	if ( ! IS_SYS_ERROR (result)) {
		if (IS_VALID_PTR(I2P(ovalue))) {
			result = ilp32_itimerval((struct itimerval32 *)I2P(ovalue), &kovalue);
		}
	}

	return result;
}

long HIDDEN
sys32_alarm (__u32 seconds)
{
	struct itimerval old, new;
	int retval;

	DBPRINT (0, "seconds=%u\n", seconds);
	new.it_interval.tv_sec = 0;
	new.it_interval.tv_usec = 0;
	new.it_value.tv_sec = seconds;
	new.it_value.tv_usec = 0;
	retval = SYSCALL (setitimer, (ITIMER_REAL, &new, &old));
	DBPRINT (0, "retval=%d\n", retval);
	if (retval < 0) {
		return 0;
	}
	retval = old.it_value.tv_sec;
	if (old.it_value.tv_usec)
		++retval;
	DBPRINT (0, "retval=%u\n", retval);
	return retval;
}


struct utimbuf32 {
	int	atime;
	int	mtime;
};


long HIDDEN
sys32_utime (__u32 filename, __u32 buf)
{
	struct timeval timevals[2];
	struct timeval *tvp;
	struct utimbuf32 buf32_tmp;
	int ret;

	if (buf) {
		if (g_safe_memcpy (&buf32_tmp, I2P(buf), sizeof (buf32_tmp)) < 0) {
			return -EFAULT;
		}
		timevals[0].tv_sec = buf32_tmp.atime;
		timevals[1].tv_sec = buf32_tmp.mtime;
		timevals[0].tv_usec = 0;
		timevals[1].tv_usec = 0;
		tvp = timevals;
	} else {
		tvp = NULL;
	}
#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	ret = pws_utimes ((char *)I2P(filename), tvp);
#else
	ret = SYSCALL (utimes, ((char *)I2P(filename), tvp));
#endif
	return ret;
}


long HIDDEN
sys32_gettimeofday (__u32 tv, __u32 tz)
{
	struct timeval tv_tmp;
	int ret = 0;

	if (tv) {
		ret = SYSCALL (gettimeofday, (&tv_tmp, (struct timezone *)I2P(tz)));
		if (ret == 0) {
			if (ILP32_TIMEVAL(*(struct timeval32 *)((I2P(tv))), tv_tmp) < 0) {
				ret = -EFAULT;
			}
		}
	}
	return ret;
}

long HIDDEN
sys32_settimeofday (__u32 tv, __u32 tz)
{
	struct timeval tv_tmp;

	if (IS_VALID_PTR(I2P(tv))) {
		if (LP64_TIMEVAL(tv_tmp, *(struct timeval32 *)((I2P(tv)))) < 0) {
			return -EFAULT;
		}
	}

	return SYSCALL (settimeofday, (VALID_PTR((struct timeval32 *)I2P(tv), &tv_tmp), (struct timezone *)I2P(tz)));
}


long HIDDEN
sys32_time (__u32 t)
{
	struct timeval tv_tmp;
	__s32	tv_sec_tmp;
	int ret;

	ret = SYSCALL(gettimeofday, (&tv_tmp, NULL));
	if (ret < 0) {
		return ret;
	}
	
	if (t) {
		tv_sec_tmp = (__s32)(tv_tmp.tv_sec);
		if (g_safe_memcpy (I2P(t), &tv_sec_tmp, sizeof(__s32)) < 0) {
			return -EFAULT;
		}
	}

	ret = tv_tmp.tv_sec;
	return ret;
}


struct tms32 {
	int tms_utime;
	int tms_stime;
	int tms_cutime;
	int tms_cstime;
};


long HIDDEN
sys32_times (__u32 buf)
{
	struct tms t;
	struct tms32 tm32;
	int ret;

	ret = SYSCALL (times, (VALID_PTR((struct tms32 *)I2P(buf), &t)));
	if (IS_SYS_ERROR (ret)) {
		return ret;
	}

	if (IS_VALID_PTR(I2P(buf))) {
		tm32.tms_utime = IA32_TICK(t.tms_utime);
		tm32.tms_stime = IA32_TICK(t.tms_stime);
		tm32.tms_cutime = IA32_TICK(t.tms_cutime);
		tm32.tms_cstime = IA32_TICK(t.tms_cstime);

		if (g_safe_memcpy (I2P(buf), &tm32, sizeof(tm32)) < 0) {
			return -EFAULT;
		}
	}

	return IA32_TICK(ret);
}

struct timex32 {
	__u32 modes;
	__s32 offset, freq, maxerror, esterror;
	__s32 status, constant, precision, tolerance;
	struct timeval32 time;
	__s32 tick;
	__s32 ppsfreq, jitter, shift, stabil;
	__s32 jitcnt, calcnt, errcnt, stbcnt;
	__s32  :32; __s32  :32; __s32  :32; __s32  :32;
	__s32  :32; __s32  :32; __s32  :32; __s32  :32;
	__s32  :32; __s32  :32; __s32  :32; __s32  :32;
};

long HIDDEN
sys32_adjtimex(__u32 buf)
{
	struct timex tx;
	struct timex32 tx32;
	int ret;

	if (g_safe_memcpy(&tx32, I2P(buf), sizeof(tx32)) < 0)
		return -EFAULT;

	tx.modes = tx32.modes;
	tx.offset = tx32.offset;
	tx.freq = tx32.freq;
	tx.maxerror = tx32.maxerror;
	tx.esterror = tx32.esterror;
	tx.status = tx32.status;
	tx.constant = tx32.constant;
	tx.precision = tx32.precision;
	tx.tolerance = tx32.tolerance;
	tx.time.tv_sec = tx32.time.tv_sec;
	tx.time.tv_usec = tx32.time.tv_usec;
	tx.tick = tx32.tick;
	tx.ppsfreq = tx32.ppsfreq;
	tx.jitter = tx32.jitter;
	tx.shift = tx32.shift;
	tx.stabil = tx32.stabil;
	tx.jitcnt = tx32.jitcnt;
	tx.calcnt = tx32.calcnt;
	tx.errcnt = tx32.errcnt;
	tx.stbcnt = tx32.stbcnt;

	ret = SYSCALL (adjtimex, (&tx));
	if (ret < 0)
		return ret;

	tx32.modes = tx.modes;
	tx32.offset = tx.offset;
	tx32.freq = tx.freq;
	tx32.maxerror = tx.maxerror;
	tx32.esterror = tx.esterror;
	tx32.status = tx.status;
	tx32.constant = tx.constant;
	tx32.precision = tx.precision;
	tx32.tolerance = tx.tolerance;
	tx32.time.tv_sec = tx.time.tv_sec;
	tx32.time.tv_usec = tx.time.tv_usec;
	tx32.tick = tx.tick;
	tx32.ppsfreq = tx.ppsfreq;
	tx32.jitter = tx.jitter;
	tx32.shift = tx.shift;
	tx32.stabil = tx.stabil;
	tx32.jitcnt = tx.jitcnt;
	tx32.calcnt = tx.calcnt;
	tx32.errcnt = tx.errcnt;
	tx32.stbcnt = tx.stbcnt;

	if (g_safe_memcpy (I2P(buf), &tx32, sizeof(tx32)) < 0) {
		ret = -EFAULT;
	}

	return ret;
}

long HIDDEN
sys32_stime(__u32 t)
{
	struct timeval tv;
	int t32;

	if (IS_VALID_PTR(I2P(t))) {
		if (g_safe_memcpy(&t32, I2P(t), sizeof(t32)) < 0)
			return -EFAULT;	
		tv.tv_sec = t32;
		tv.tv_usec = 0;
	}
	return SYSCALL (settimeofday, (t ? &tv : NULL,  NULL));
}

long HIDDEN
sys32_timer_create(__s32 clockid, __u32 evp, __u32 timerid)
{
	struct sigevent se64;
	struct sigevent32 se32;
	int retval;

	if (IS_VALID_PTR(I2P(evp))) {
		if (g_safe_memcpy(&se32, (struct sigevent32 *)I2P(evp),
			sizeof(struct sigevent32)) < 0) {
			return -EFAULT;
		}
		se64.sigev_value.sival_int = se32.sigev_value.sival_int;
		se64.sigev_value.sival_ptr = I2P(se32.sigev_value.sival_ptr);
		se64.sigev_signo = se32.sigev_signo;
		se64.sigev_notify = se32.sigev_notify;
		memcpy (se64._sigev_un._pad, se32._sigev_un._pad, sizeof(se64._sigev_un._pad));
	}

	retval = SYSCALL (timer_create, (clockid, &se64, I2P(timerid)));

	return retval;
}

long HIDDEN
sys32_timer_settime(__s32 timerid, __s32 flags, __u32 value, __u32 ovalue)
{
	struct itimerspec is64;
	struct itimerspec ois64;
	struct itimerspec32 is32;
	int retval;

	if (IS_VALID_PTR(I2P(value))) {
		if (g_safe_memcpy(&is32, I2P(value), sizeof(struct itimerspec32)) < 0) {
			return -EFAULT;
		}
		is64.it_interval.tv_sec = is32.it_interval.tv_sec;
		is64.it_interval.tv_nsec = is32.it_interval.tv_nsec;
		is64.it_value.tv_sec = is32.it_value.tv_sec;
		is64.it_value.tv_nsec = is32.it_value.tv_nsec;
	}
	else {
		return -EINVAL;
	}

	retval = SYSCALL (timer_settime, (timerid, flags, &is64, IS_VALID_PTR(I2P(ovalue))? &ois64: NULL));

	if ((retval == 0) && IS_VALID_PTR(I2P(ovalue))) {
		is32.it_interval.tv_sec = ois64.it_interval.tv_sec;
		is32.it_interval.tv_nsec = ois64.it_interval.tv_nsec;
		is32.it_value.tv_sec = ois64.it_value.tv_sec;
		is32.it_value.tv_nsec = ois64.it_value.tv_nsec;

		if (g_safe_memcpy(I2P(ovalue), &is32, sizeof(struct itimerspec32)) < 0) {
			return -EFAULT;
		}
	}

	return retval;
}

long HIDDEN
sys32_timer_gettime(__s32 timerid, __u32 value)
{
	struct itimerspec is64;
	struct itimerspec32 is32;
	int retval;

	retval = SYSCALL (timer_gettime, (timerid, &is64));

	if (retval == 0) {
		is32.it_interval.tv_sec = is64.it_interval.tv_sec;
		is32.it_interval.tv_nsec = is64.it_interval.tv_nsec;
		is32.it_value.tv_sec = is64.it_value.tv_sec;
		is32.it_value.tv_nsec = is64.it_value.tv_nsec;

		if (g_safe_memcpy(I2P(value), &is32, sizeof(struct itimerspec32)) < 0) {
			return -EFAULT;
		}
	}

	return retval;
}

long HIDDEN
sys32_clock_settime(__s32 clk_id, __u32 p)
{
	int retval;
	struct timespec ts64;
	struct timespec32 ts32;

	if (g_safe_memcpy (&ts32, I2P(p), sizeof(struct timespec32)) < 0) {
		return -EFAULT;
	}
	
	ts64.tv_sec = ts32.tv_sec;
	ts64.tv_nsec = ts32.tv_nsec;

	retval = sys_clock_settime (clk_id, &ts64);
	return retval;
}

long HIDDEN
sys32_clock_gettime(__s32 clk_id, __u32 p)
{
	int retval;
	struct timespec ts64;
	struct timespec32 ts32;

	retval = sys_clock_gettime (clk_id, &ts64);
	if (retval == 0) {
		ts32.tv_sec = ts64.tv_sec;
		ts32.tv_nsec = ts64.tv_nsec;

		if (g_safe_memcpy (I2P(p), &ts32, sizeof(struct timespec32)) < 0) {
			return -EFAULT;
		}
	}
	
	return retval;
}

long HIDDEN
sys32_clock_getres(__s32 clk_id, __u32 p)
{
	int retval;
	struct timespec ts64;
	struct timespec32 ts32;

	retval = sys_clock_getres (clk_id, &ts64);
	if ((retval == 0) && IS_VALID_PTR(I2P(p))) {
		ts32.tv_sec = ts64.tv_sec;
		ts32.tv_nsec = ts64.tv_nsec;

		if (g_safe_memcpy (I2P(p), &ts32, sizeof(struct timespec32)) < 0) {
			return -EFAULT;
		}
	}
	
	return retval;
}

long HIDDEN
sys32_clock_nanosleep(__s32 clk_id, __s32 flags, __u32 qp, __u32 mp)
{
	int retval;
	struct timespec tq64, tm64;
	struct timespec32 ts32;

	if (g_safe_memcpy (&ts32, I2P(qp), sizeof(struct timespec32)) < 0) {
		return -EFAULT;
	}
	tq64.tv_sec = ts32.tv_sec;
	tq64.tv_nsec = ts32.tv_nsec;

	retval = sys_clock_nanosleep (clk_id, flags, &tq64, &tm64);
	if ((retval == -ERESTART_RESTARTBLOCK) && IS_VALID_PTR(I2P(mp))) {
		ts32.tv_sec = tm64.tv_sec;	
		ts32.tv_nsec = tm64.tv_nsec;

		if (g_safe_memcpy (I2P(mp), &ts32, sizeof(struct timespec32)) < 0) {
			return -EFAULT;
		}
	}
	
	return retval;
}

long HIDDEN
sys32_utimes(__u32 filename, __u32 tvp)
{
	struct timeval tv64[2];
	struct timeval32 tv32;
	int retval;
	void * utp;

	if (IS_VALID_PTR(I2P(tvp))) {
		if (g_safe_memcpy (&tv32, &((struct timeval32 *)I2P(tvp))[0],
					sizeof(struct timeval32)) < 0) {
			return -EFAULT;
		}
		tv64[0].tv_sec = tv32.tv_sec;
		tv64[0].tv_usec = tv32.tv_usec;

		if (g_safe_memcpy (&tv32, &((struct timeval32 *)I2P(tvp))[1],
					sizeof(struct timeval32)) < 0) {
			return -EFAULT;
		}
		tv64[1].tv_sec = tv32.tv_sec;
		tv64[1].tv_usec = tv32.tv_usec;
	}
	
	utp = IS_VALID_PTR(I2P(tvp))? &tv64 : NULL;
#if NOKERNELALTROOT
	/* use the pathwalk sycall wrapper in user32_pathwalk.c */
	retval = pws_utimes ((char*)I2P(filename), utp);
#else
	retval = SYSCALL (utimes, ((char*)I2P(filename), utp));
#endif
	
	return retval;
}

