/*
 * ia32ss_list.c
 *
 * implementation of the list interface ( described in ia32ss_list.h )
 * using a sorted doubly linked list
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>
#include "ia32ss_list.h"
# include "user32.h"

struct list_t {
	/* the 'key' and 'data' fields are generic */
	long key;
	unsigned long data;
	/* the following fields depend on implementation */
	struct list_t * next, * prev;
};

#if DEBUG
static void print_list (struct list_t ** list) UNUSED;
#endif


long HIDDEN list_init (struct list_t ** list)
{
	*list = NULL;
	return 0;
}

long HIDDEN list_destroy (struct list_t ** list)
{
	*list = NULL;
	return 0;
}

list_handle HIDDEN list_insert (struct list_t ** list, long key, unsigned long data)
{
	struct list_t * new = NULL;

	/* allocate memory and init the new element */
	new = (struct list_t*) malloc (sizeof (*new));
	if (new == NULL) {
		DBPRINT (TRUE, "malloc failed\n");
		return (list_handle)-ENOMEM;
	}
	new->prev = new->next = NULL;
	new->key = key;
	new->data = data;

	/* insert the new element */
	if (*list == NULL) {
		*list = new;
	} else {
		struct list_t * next = *list;
	   	struct list_t * prev = NULL;

		/* find the insertion point */
		while ((next != NULL) && (next->key < key)) {
			prev = next;
			next = next->next;
		}
		/* see if the key already exists */
		if (prev && prev->key == key) {
			return (list_handle)-EINVAL;
		}

		/* insert */
		new->next = next;
		new->prev = prev;
		if (next != NULL) {
			next->prev = new;
		}
		if (prev == NULL) {
			*list = new;
		} else {
			prev->next = new;
		}
	}

	return (list_handle)new;
}

long HIDDEN list_remove (struct list_t ** list, list_handle handle)
{
	struct list_t * tbr = (struct list_t*)handle;

	/* disconnect from the list */
	if (tbr->prev != NULL) {
		tbr->prev->next = tbr->next;
	} else {
		*list = tbr->next;
	}

	if (tbr->next != NULL) {
		tbr->next->prev = tbr->prev;
	}

	/* free the memory */
	free (tbr);

	return 0;
}
list_handle HIDDEN list_retrieve (struct list_t ** list, long key)
{
	struct list_t * p = *list;
	list_handle ret = 0;

	while (p != NULL) {
		if (p->key == key) {
			ret = (list_handle)p;
		}
		if (p->key > key) {
			/* not found */
			break;
		}
		p = p->next;
	}
	return ret;
}

list_handle HIDDEN list_get_first (struct list_t ** list)
{
	return (list_handle)*list;
}
list_handle HIDDEN get_element_next (list_handle handle)
{
	struct list_t * h = (struct list_t *) handle;
	return h->next;
}
list_handle HIDDEN get_element_prev (list_handle handle)
{
	struct list_t * h = (struct list_t *) handle;
	return h->prev;
}

long HIDDEN get_element_key (list_handle handle)
{
	struct list_t * h = (struct list_t *) handle;
	return h->key;
}

unsigned long HIDDEN get_element_data (list_handle handle)
{
	struct list_t * h = (struct list_t *) handle;
	return h->data;
}

#if DEBUG

static void print_list (struct list_t ** list)
{
	struct list_t * elem = * list;
	char buffer[16*1024];
	char * pb = buffer;

	while (elem != NULL) {
		sprintf (pb, " [key=%lx(%lu)-data=%lx(%lu)]", 
				elem->key, elem->key, elem->data, elem->data);
		pb += strlen (pb);
		elem = elem->next;
	}

	DBPRINT (TRUE, "%s\n", buffer);
}

#endif
