/*
 * user32_fs.c
 *
 * wrapper functions for syscalls that deal with file systems.
 * used only by external wrapper
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>
#include "user32.h"
#include "user32_ia32_syscall_prototypes.h"
#if NOKERNELALTROOT
#include "user32_pathwalk.h"
#endif

/* 
 * flag to indicate if d_type is written into dirent structure 
 *	-1: uninitializeed
 *	0: d_type is not written into dirent
 *	1: d_type is written into dirent
 */
static int m_is_d_type_filled = -1;

#define DIRENT32_FIX_SIZE offsetof (struct dirent32, d_name)
#define DIRENT_FIX_SIZE offsetof (struct dirent, d_name)
#define LAST_CHAR_IN_DIRENT(p)	(char*)((char *)(p) + (p)->d_reclen - 1)

inline void dirent64_to_dirent32 (struct dirent32 * p32, struct dirent * p64);
void check_d_type_policy (void);

/*
 * check the kernel policy if d_type is written into dirent structure
 */
void check_d_type_policy (void)
{
	struct dirent d64;
	struct dirent * p;
	long retval;
	int fd;

	m_is_d_type_filled = 0;
		
	fd = SYSCALL (open, (".", O_RDONLY, 0));
	if (fd < 0) {
		DBPRINT (1, "open failed!\n");
		goto out;
	}

	/* the entry "." is a good candidate, catch it */
	do {
		/* fill dirent all 0xFF */
		memset (&d64, 0xFF, sizeof (d64));
		retval = SYSCALL (getdents, (fd, &d64, sizeof (struct dirent)));
		if (retval <= 0) {
			DBPRINT (1, "getdents failed!\n");
			goto out_close;
		}

		p=&d64;
		while (((char *)p < (char *)&d64 + retval)
				&& (p->d_ino != 0)
				&& (p->d_reclen > 0)) {
			if ((p->d_name[0] == '.')
				&& (p->d_name[1] == '\0')) {
				/* check if d_type is written into dirent */
				if (*LAST_CHAR_IN_DIRENT (p) != (char)0xFF) {
					m_is_d_type_filled = 1;
				}
				
				goto out_close;
			}

			p = (struct dirent *)(void *)((char *)p + p->d_reclen);
		}
		
	} while (1);
	
out_close:	
	retval = SYSCALL (close, (fd));
	if (retval < 0) {
		DBPRINT (1, "close failed!\n");
	}
	
out:
	return;
}

/*
 * copy struct dirent32 to struct dirent
 */
inline void dirent64_to_dirent32 (struct dirent32 * p32, struct dirent * p64)
{
	unsigned int name_len;
	
	p32->d_ino = p64->d_ino;
	p32->d_off = p64->d_off;

	/* d_name ends with '\0', so name_len should be added by 1 */
	name_len = strnlen (p64->d_name, sizeof (p64->d_name)) + 1;
	assert (name_len <= sizeof (p64->d_name));
	strncpy (p32->d_name, p64->d_name, name_len);

	/* check if d_type is written into dirent */
	if (m_is_d_type_filled == -1) {
		check_d_type_policy ();
	}
	
	/* d_reclen in struct dirent32 is 4 byte aligned */
	p32->d_reclen = ROUNDUP (DIRENT32_FIX_SIZE + name_len, 4);
	if (m_is_d_type_filled == 1) {
		/* 
		 * Since kernel 2.6.4, d_type is stored, we need to cosmeticize it.
		 *	and this may be changed in the future.
		 */
		p32->d_reclen = ROUNDUP (DIRENT32_FIX_SIZE + name_len + 1, 4);
		/* make sure we are dealing with the right situation */
		if (p64->d_reclen == ROUNDUP (DIRENT_FIX_SIZE + name_len + 1, 8)) {
			*(LAST_CHAR_IN_DIRENT (p32)) = *(LAST_CHAR_IN_DIRENT (p64));
		}
	}
	
	return;
}

long HIDDEN sys32_getdents (int fd, __u32 dirent, unsigned int count)
{
	struct dirent d64;
	struct dirent * p64;
	struct dirent32 d32;
	struct dirent32 * p32;
	off_t orig_offset;
	long UNUSED r;
	long retval, new_ret;

	DBPRINT (0, "fd:%d, dirent:%p, count:%d\n", fd, I2P(dirent), count);
	
	if (count < DIRENT32_FIX_SIZE) {
		return -EINVAL;
	}

	/* back up the origibal offset */
	orig_offset = SYSCALL (lseek, (fd, 0, SEEK_CUR));
	if (IS_SYS_ERROR (orig_offset)) {
		return (long)orig_offset;
	}
	
	/* 
	 * since number of dirents can be know only after ia64 system call returns.
	 *  it's impossible for us to allocate enough memory for ia64 system call.
	 *  here we get a dentry one time.
	 */
	new_ret = 0;
	p32 = (struct dirent32 *)I2P(dirent);

	retval = SYSCALL (getdents, (fd, &d64, sizeof (struct dirent)));
	if (retval < 0) {
		new_ret = retval;
		goto out;
	}
	p64 = &d64;
	
	/* 
	 * when we encounter the end of the file, no need to do more work
	 */
	if (retval == 0) {
		if ((p64->d_ino != 0) && (p64->d_reclen > 0)) {
			/* copy info to the user given structure */
			dirent64_to_dirent32 (&d32, p64);

			/* count is too small */
			if (d32.d_reclen > count) {
				return -EINVAL;
			}
			
			if (g_safe_memcpy (p32, &d32, d32.d_reclen) < 0) {
				return -EFAULT;
			}
		}

		return 0;
	}

	while (1) {
		while (((char *)p64 < (char *)&d64 + retval)
				&& (p64->d_ino != 0)
				&& (p64->d_reclen > 0)) {
			/* copy info to the user given structure */
			dirent64_to_dirent32 (&d32, p64);
		
			if (new_ret + d32.d_reclen > count) {
				/* 
				 * the ia32 memory is full-filled, 
				 * we need to resetup the offset
				 */
				r = SYSCALL (lseek, (fd, orig_offset, SEEK_SET));
				DBPRINT (IS_SYS_ERROR (r), "lseek failed:%ld\n", r);
				
				goto out;
			}
			
			if (g_safe_memcpy (p32, &d32, d32.d_reclen) < 0) {
				new_ret = -EFAULT;
				goto out;
			}
			
			DBPRINT (0, "d32:d_ino:%d, d_off:%d, d_reclen:%d, d_name:%s\n",
					d32.d_ino,
					d32.d_off,
					d32.d_reclen,
					d32.d_name);

			p32 = (struct dirent32 *)(void *)((char *)p32 + d32.d_reclen);
			p64 = (struct dirent *)(void *)((char *)p64 + p64->d_reclen);
			new_ret += d32.d_reclen;
			orig_offset = d32.d_off;
		}

		if (new_ret >= count) {
			break;
		}

		p64 = &d64;
		retval = SYSCALL (getdents, (fd, &d64, sizeof (struct dirent)));
		if (retval < 0) {
			new_ret = retval;
			break;
		}
		if (retval == 0) {
			/* the end! */
			break;
		}
	}
	
out:
	return new_ret;
}

long HIDDEN sys32_readdir (int fd, __u32 dirent, 
                           unsigned int count UNUSED)
{
	struct dirent32 d32;
	struct dirent d64;
	off_t orig_offset;
	long UNUSED r;
	long retval;
	
	orig_offset = SYSCALL (lseek, (fd, 0, SEEK_CUR));
	if (IS_SYS_ERROR (orig_offset)) {
		return (long)orig_offset;
	}
	
	retval = SYSCALL (getdents, (fd, &d64, sizeof (struct dirent)));
	if (retval < 0) {
		goto out;
	}

	/* readdir always return 1 or 0 when success */
	if (retval != 0)
		retval = 1;

	/* copy info to the user given structure */
	dirent64_to_dirent32 (&d32, &d64);

	if (g_safe_memcpy (I2P(dirent), &d32, d32.d_reclen) < 0) {
		retval = -EFAULT;
		goto out_rewind;
	}

	goto out;

out_rewind:
	r = SYSCALL (lseek, (fd, orig_offset, SEEK_SET));
	DBPRINT (IS_SYS_ERROR (r), "lseek failed:%ld\n", r);
	
out:
	return retval;
}

long HIDDEN sys32_oldumount (__u32 name)
{
#if NOKERNELALTROOT
	/* use the pathwalk syscall wrapper in user32_pathwalk.c */
	return pws_umount ((char*)I2P(name), 0);
#else
	return SYSCALL (umount, ((char*)I2P(name), 0));
#endif
}


