/*
 * user32_files.c
 *
 * wrapper functions for syscalls that deal with ipc. 
 * for internal wrapper.
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Gordon Jin <gordon.jin@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>
#include "user32.h"

#include "ia32ss_list.h"
#include "user32_mem_regions.h"
#include "ia32toia64.h"

long HIDDEN remove_shm_region (unsigned long addr);

/* we need to know the size of the shm at the time of attachement.
 * the only way for us to get it is to store it upon the call to shmget.
 * the list shm_list store the size of the shm and is accessed by the shm id.
 * 
 * since we will not need this data after shmat returns, we remove the
 * elements as soon as the attachment succeeds
 */
static struct list_t * m_addr_list = NULL;


long HIDDEN 
sys32_pipe (int fd)
{
	long retval, err;
	int  kfd[2];

	/* libc's pipe is not good enough */
	kfd[0] = SYSCALL (pipe, ());
	__asm__ __volatile__ ("mov %0 = r10;;\n" : "=r"(err)); 
	__asm__ __volatile__ ("mov %0 = r9;;\n" : "=r"(kfd[1])); 

	if (err) {
		retval = kfd[0];
	} else {
		retval = (g_safe_memcpy (I2P(fd), kfd, sizeof(kfd)) < 0) ? -EFAULT : 0;
	}

	return retval;
}
	
int HIDDEN 
shmat32 (int shmid, int shmflag, __u32 ret_ptr, void *shmaddr)
{
	unsigned long raddr;
	unsigned long size;
	int  err;
	int  memory_locked = 0;
	list_handle   handle;


	if (NOT_VALID_PTR (shmaddr) && (shmflag & SHM_REMAP)) {
		err = -EINVAL;
		goto out;
	}

	if ((UL(shmaddr) & (PAGE_SIZE - 1)) && !(shmflag & SHM_RND)) {
		err = -EINVAL;
		goto out;
	}

	/* get the size of the shm */
	{
		struct shmid64_ds s64;
		int  ret;

		ret = SYSCALL (shmctl, (shmid, IPC_STAT, (void *) &s64));
		assert ((ret != -EFAULT) && (ret != -EOVERFLOW));
		
		if (IS_SYS_ERROR (ret)) {
			if ((ret == -EIDRM) || (ret == -EPERM)) {
				ret = -EINVAL;
			}
			err = ret;
			goto out;
		}
		size = s64.shm_segsz;
	}

	mutex_lock (g_memory_sync);

	/* if shmaddr==0 then we need to find a valid IA-32 address
	 * for the attachment, otherwise the kernel will return
	 * with an address above 4G
	 */
	if (shmaddr == NULL) {
		memory_locked = 1;
		shmaddr = (void*)get_ia32_free_memory (0, size + SHMLBA, MAP_SHARED, 1);
		/* align shmaddr on a SHMLBA boundry */
#define SHMLBA_MASK (~(SHMLBA - 1))
		shmaddr = (void*)(((unsigned long)shmaddr - 1) & SHMLBA_MASK) + SHMLBA;
	}

	DBPRINT (FALSE, "shmid=%d shmaddr=%p shmflag=%d\n",
		 shmid, shmaddr, shmflag);

	/* call the system */
	raddr = SYSCALL (shmat, (shmid, shmaddr, shmflag));
	__asm__ __volatile__ ("mov %0=r10;;\n":"=r"(err));

	DBPRINT (FALSE, "err=%d<%x> raddr=%ld<%lx>\n",
		 err, err, raddr, raddr);

	/* check the return value */
	if (err != 0) {
		err = (int)raddr;
		goto out_up;
	}

	ASSERT_HIGH_ZERO (raddr);

	/* keep the region size for shmdt */
	handle = list_insert (&m_addr_list, raddr, size);
	if ((long)handle < 0) {
		err = (int)(long)handle;
		goto out_up;
	}

	/* update the memory regions */
	err = (int)insert_region (raddr, size, 
				  PROT_READ | PROT_WRITE | PROT_EXEC, MAP_SHARED);
	if (err < 0) {
		list_remove (&m_addr_list, handle);
		goto out_up;
	}
	
	if (g_safe_memcpy (I2P(ret_ptr), &raddr, sizeof(__u32)) < 0) {
		list_remove (&m_addr_list, handle);
		remove_region (raddr, size);
		err = -EFAULT;
	}

out_up:
	mutex_unlock (g_memory_sync);

out:
	return err;
}

/*****************************************************************************/
/* remove a shared memory from the region list
 * input:	addr - the address of the shm
 * output:	0 on success, negative value on failure
 * remarks:	the shm should be inserted into the addr-list by the function
 * 		that emulates the IA-32-shmat syscall.
 * 		this function should be called by the function that emulates
 * 		the IA-32-shmdt syscall
 */
/*****************************************************************************/
long HIDDEN remove_shm_region (unsigned long addr)
{
	list_handle h;
	long ret = 0;
	unsigned long size;

	/* find the removed memory in the list */
	h = list_retrieve (&m_addr_list, addr);
	if (h == 0) {
		goto out;
	}

	/* get the region size */
	ret = get_element_data (h);
	if (ret == 0) {
		goto out;
	}
	size = ret;

	/* remove the memory from the list of memory-regions */
	ret = remove_region (addr, size);

	/* remove from the list */
	list_remove (&m_addr_list, h);
out:
	return ret;
}


