/*
 * user32_mem_regions.c
 *
 * function to handle the list of memory regions that were 
 * mapped by the application
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "user32.h"
#include "user32_mem_regions.h"
#include "ia32toia64.h"

void HIDDEN print_region (list_handle region);

/* in the region list -
 *   key == region base addr
 *   data == region attrib
 */
struct list_t * m_regions = NULL;

static long remove_overlapping_region (list_handle old_region, 
                                       unsigned long start, 
                                       unsigned long len);
static list_handle merge_two_regions (list_handle region1, list_handle region2);
static long do_remove_region (unsigned long start, unsigned long len);
static long do_insert_region (unsigned long start, unsigned long len, unsigned prot, unsigned flags);
static long do_is_inside_region (unsigned long start, unsigned long len);

/*****************************************************************************/
/* check whether the given memory chunk overlapps with one
 * of the regions in the list
 */
/*****************************************************************************/
long HIDDEN is_inside_region (unsigned long start, unsigned long len)
{
	long ret = 0;
	ret = do_is_inside_region (start, len);
	return ret;
}

static long do_is_inside_region (unsigned long start, unsigned long len)
{
	list_handle region;
	unsigned long end;
	unsigned long region_start, region_len, region_end;
	region_attrib * ra;
	long ret = 0;

	end = start + len;
	region = list_get_first (&m_regions);
	if (region == 0) {
		/* list empty */
		goto out;
	}
	region_start = get_element_key (region);

	while (region_start < end) {
		ra = (void*)get_element_data (region);
		region_len = ra->len;
		region_end = region_start + region_len;

		if (region_end > start) {
			/* overlapping */
			ret = 1;
			break;
		}
		region = get_element_next (region);
		if (region == 0) {
			/* end of list */
			break;
		}
		region_start = get_element_key (region);
	}
out:
	return ret;
}

/*****************************************************************************/
/* find the region containing the addr 
 * out:
 *     p_region contains the found region's pointer
 * returns 0 on success and negative on failure */
/*****************************************************************************/
long HIDDEN find_region (unsigned long addr, void **p_region)
{
	list_handle region;
	region_attrib *ra;
	unsigned long region_end;
	unsigned long region_len;
	unsigned long region_start;
	
	region = list_get_first (&m_regions);
	
	while (1) {
		if (region == 0) {
			return -1;
		}

		region_start = get_element_key (region);
		ra = (void *)get_element_data (region);
		region_len = ra->len;
		region_end = region_start + region_len;

		if (region_start > addr) {
			return -1;
		} else if (region_end >= addr) {
			/* here, region_start <= addr */
			if (p_region != NULL) {
				*p_region = region;
			}
			break;
		}
		
		region = get_element_next (region);
	}

	return 0;
}

/*****************************************************************************/
/* remove a region from the list 
 * returns 0 on success and a negative value on failure */
/*****************************************************************************/
long HIDDEN remove_region (unsigned long start, unsigned long len)
{
	long retval;
	retval = do_remove_region (start, len);
	return retval;
}
static long do_remove_region (unsigned long start, unsigned long len)
{
	list_handle next = 0;
	unsigned long end;
	unsigned long next_len, next_start, next_end;
	long ret;

	end = start + len;
	len = IA32_PAGE_ALIGN (len);
	DBPRINT (0, "start=%lx len=%lx\n", start, len);

	next = list_get_first (&m_regions);
	if (next == 0) {
		/* if the list is empty than no extra work is needed */
		DBPRINT (0, "try to remove a region from an empty list\n");
		ret = 0;
		goto out;
	}

	next_start = get_element_key (next);
	while (next_start < end) {
		list_handle tmp = next;
		region_attrib * ra;

		ra = (void*)get_element_data (next);
		next_len = ra->len;
		next_end = next_start + next_len;

		/* get the next element before calling remove_overlapping_region
		 * because the current element  might be removed from the list
		 */
		next = get_element_next (next);

		if (next_end > start) {
			ret = remove_overlapping_region (tmp, start, len);
			if (ret < 0) {
				goto out;
			}
		}
		if (next == 0) {
			break;
		}
		next_start = get_element_key (next);
	}

	ret = 0;

out:
	DBCODE (0, print_regions ());
	return ret;
}

/*****************************************************************************/
/* insert a new region to the list 
 * returns 0 on success and a negative value on failure */
/*****************************************************************************/
long HIDDEN insert_region (unsigned long start, unsigned long len, 
                           unsigned prot, unsigned flags)
{
	long retval;
	retval = do_insert_region (start, len, prot, flags);
	return retval;
}

static long do_insert_region (unsigned long start, unsigned long len, 
                              unsigned prot, unsigned flags)
{
	list_handle next_region, prev_region, new_region, tmp_region;
	region_attrib * ra = NULL;
	long ret;

	len = IA32_PAGE_ALIGN (len);
	DBPRINT (0, "start=%lx len=%lx prot=%x\n", start, len, prot);
	/* remove previously mapped region */
	ret = do_remove_region (start, len);
	if (ret < 0) {
		goto out;
	}

	/* create a new region */
	ra = malloc (sizeof (*ra));
	if (ra == NULL) {
		ret = -ENOMEM;
		goto out;
	}

	ra->len = len;
	ra->prot = prot;
	ra->flags = flags;
	new_region = list_insert (&m_regions, start, (unsigned long)ra);
	if ((long)new_region < 0) {
		free (ra);
		ret = (long)new_region;
		goto out;
	}

	/* get the regions on both sides of start */
	next_region = get_element_next (new_region);
	prev_region = get_element_prev (new_region);

	/* try to merge with the existing regions */
	if (prev_region != 0) {

		tmp_region = merge_two_regions (prev_region, new_region);
		if ((long)tmp_region < 0) {
			ret = (long)tmp_region;
			goto out;
		} else if (tmp_region != 0) {
			new_region = tmp_region;
		}

	}

	if (next_region != 0) {

		tmp_region = merge_two_regions (next_region, new_region);
		if ((long)tmp_region < 0) {
			ret = (long)tmp_region;
			goto out;
		} else if (tmp_region != 0) {
			new_region = tmp_region;
		}

	}

	DBCODE (0, print_regions ());

out:
	return ret;
}

/*****************************************************************************/
/* get the permissions of the region where the given address resides.
 * input:  addr -
 *         prot - will hold the permission at the given address
 *         start - will hold the region's base address
 *         len - will hold the region's size
 *         flags - will hold the region's flags 
 * output: 1 if the address is within a mapped memory area, 0 otherwise
 */
/*****************************************************************************/
long HIDDEN get_region_prot (unsigned long addr, unsigned *prot, 
                             unsigned long *start, unsigned long *len, 
			     unsigned *flags)
{
	unsigned p = 0, fl = 0;
	unsigned long s = 0, l = 0;
	int found = 0;
	list_handle region;

	DBPRINT (0, "addr=%lx\n", addr);

	region = list_get_first (&m_regions);

	while (1) {

		region_attrib * ra;
		unsigned long region_end;
		unsigned long region_len;
		unsigned long region_start = 0;


		if (region == 0) {
			/* no more regions to check */
			break;
		}

		
		region_start = get_element_key (region);
		if (region_start > addr) {
			/* the list is guranteed to be sorted, so there is no point
			 * in checking further
			 */
			break;
		}

		/* get the region's size */
		ra = (void*)get_element_data (region);
		region_len = ra->len;
		region_end = region_start + region_len;

		if (region_end > addr) {
			found = 1;
			p = ra->prot;
			s = region_start;
			l = region_len;
			fl = ra->flags;
			break;
		}
		region = get_element_next (region);
	}

	if (!found) {
		p = 0;
		s = IA32_PAGE_ALIGN (addr);
		l = IA32_PAGE_SIZE;
		fl = 0;
	}

	DBPRINT (0, "addr=%lx found=%d perm=%x start=%lx len=%lx flags=%x\n", 
		 addr, found, p, s, l, fl);

	if (prot != NULL) {
		*prot = p;
	}
	if (start != NULL) {
		*start = s;
	}
	if (len != NULL) {
		*len = l;
	}
	if (flags != NULL) {
		*flags = fl;
	}

	return found;
}

/*****************************************************************************/
/* change the permissions of the given memory area
 * input:  start - base address of the area
 *         len - size of area
 *         prot - new permissions for the area
 * output: 0 on success, negative error value on error
 */
/*****************************************************************************/
long HIDDEN set_region_prot (unsigned long start, unsigned long len, unsigned prot)
{
	long ret;
	unsigned flags;

	DBPRINT (0, "start=%lx len=%lx prot=%x\n", start, len, prot);
	/* mprotect fails for non mapped memory, so we can simple remove the region
	 * and re-insert it with the new permissions
	 */
	ret = get_region_prot (start, NULL, NULL, NULL, &flags);
	if (ret < 0) {
		goto out;
	}
	ret = do_remove_region (start, len);
	if (ret < 0) {
		goto out;
	}
	ret = do_insert_region (start, len, prot, flags); 
out:
	return ret;
}

/*****************************************************************************/
/* try to merge two regions 
 * input:  reg0 and reg1 are handles of the regions to be merged.
 * output: in case of error return a negative value.
 *         in case the regions cannot be merged return 0. 
 *         otherwise return a handle to a region that is the result of the
 *         merge operation of the two regions.
 * comments :the given regions can be destroyed, that is - removed from the
 *           list and get their attribute freed.
 */
/*****************************************************************************/
static list_handle merge_two_regions (list_handle reg0, list_handle reg1)
{
	region_attrib * ra0, * ra1;
	unsigned long start0, start1, end0, end1;
	unsigned long new_len = 0, new_start = 0;
	list_handle new_reg = (list_handle)0;
	unsigned prot;
	unsigned flags;
	long ret = 0;

	/* finish with trivial cases right here */
	assert (reg0 != 0);
	assert (reg1 != 0);

	/* get info of the regions */

	start0 = get_element_key (reg0);
	start1 = get_element_key (reg1);

	ra0 = (region_attrib *)get_element_data (reg0);
	ra1 = (region_attrib *)get_element_data (reg1);

	end0 = start0 + ra0->len;
	end1 = start1 + ra1->len;

	/* see whether the regions can be merged */

	if ((end0 < start1) || (end1 < start0)) {
		/* cannot merge */
		goto out;
	}

	if (ra1->prot != ra0->prot) {
		/* cannot merge */
		goto out;
	}
	prot = ra0->prot;

	if (ra1->flags != ra0->flags) {
		/* cannot merge */
		goto out;
	}
	flags = ra0->flags;
	
	/* the regions can be merged - there are few possible scenarios
	 * 1) the new region contains the base region:
	 * 	the base region should be removed
	 * 	and the new region should be returned as the result
	 * 	of the merge operation
	 * 2) the base region contains the new one:
	 * 	the new region has to be removed and the base
	 * 	region should be returned as the result of the merge
	 * 	operation
	 * 3) the regions overlap ( or touch ):
	 * 	both regions should be removed and a new region
	 * 	should be created with the appropriate settings
	 */ 
	if (start0 >= start1) {
		if (end0 <= end1) {
			/* reg1 contains reg0 */
			new_reg = reg1;
		} else { /* end0 > end1 */
			new_start = start1;
			new_len = end0 - new_start;
		}
	} else { /* start1 > start0 */

		if  (end1 <= end0) {
			/* reg0 contains reg1 */
			new_reg = reg0;
		} else { /* end1 > end0 */
			new_start = start0;
			new_len = end1 - new_start;
		}
	}

	if (new_reg != reg0) {
		/* remove region 0 */
		free (ra0);
		ret = list_remove (&m_regions, reg0);
		if (ret < 0) {
			new_reg = (list_handle)ret;
		}
	}
	if (ret < 0) {
		goto out;
	}

	if (new_reg != reg1) {
		/* remove region 1 */
		free (ra1);
		ret = list_remove (&m_regions, reg1);
		if (ret < 0) {
			new_reg = (list_handle)ret;
		}
	}
	if (ret < 0) {
		goto out;
	}

	if (new_reg == 0) {
        	region_attrib * ra = malloc (sizeof (*ra)); 
        	if (ra == NULL) { 
            		new_reg = (list_handle)-ENOMEM;
        	} else {
	        	ra->len = new_len;
	        	ra->prot = prot; 
			ra->flags = flags;
	        	new_reg = list_insert (&m_regions, new_start, (unsigned long)ra);
			if ((long)new_reg < 0) {
				free (ra);
			}
		}
	}

out:
	return new_reg;
}

/*****************************************************************************/
/* remove a memory region from an existing region
 * input:   old_region - existing region
 *          start - base address of the region to be removed
 *          len - size of region to be removed
 * output:  0 on success, negative value on error
 */
/*****************************************************************************/
static long remove_overlapping_region (list_handle old_region, 
                                       unsigned long start, unsigned long len)
{
	unsigned long old_start, old_len, old_end, end;
	unsigned long new_start[2] = {0};
	unsigned long new_len[2] = {0};
	region_attrib * old_ra;
	long j;
	long ret;
	region_attrib * new_ra[2] = {0};

	old_start = get_element_key (old_region);
	old_ra = (void*)get_element_data (old_region);
	old_len = old_ra->len;
	old_end = old_start + old_len;
	end = start + len;

	assert (old_end > start);
	assert (end > old_start);

	/* the old region is removed in any case */
	ret = list_remove (&m_regions, old_region);
	if (ret < 0) {
		return ret;
	}

	if (old_start >= start) {
		if (old_end <= end) {
			/* the existing old_region is inside the tbr region 
			 * we have nothing more to do
			 */
		} else { /* old_end > end */
			/* shrink the old region */
			new_start[0] = end;
			new_len[0] = old_end - new_start[0];
		}
	} else { /* old_start < start */
		/* we will need at least one new region */
		new_start[0] = old_start;
		new_len[0] = start - old_start;
		if (old_end > end) {
			/* the tbr region is inside the old region. the
			 * old region has to split
			 */
			new_start[1] = end;
			new_len[1] = old_end - end;
		}
	}

	/* create the new regions */
	for (j = 0; j < 2; j ++) {
		if (new_len[j] != 0) {
			new_ra[j] = malloc (sizeof (region_attrib));
			if (new_ra[j] == NULL) {
				ret = -ENOMEM;
				if (j == 1) {
					free (new_ra[0]);
				}
				goto out;
			}
			new_ra[j]->len = new_len[j];
			new_ra[j]->prot = old_ra->prot;
			new_ra[j]->flags = old_ra->flags;
			ret = (long)list_insert (&m_regions, new_start[j], 
						(unsigned long)new_ra[j]);
			if (ret < 0) {
				long i;
				for (i = 0; i <= j; i ++) {
					free (new_ra[i]);
				}
				goto out;
			}
		}
	}

out:
	free (old_ra);
	return ret;
}


/*****************************************************************************/
/* look for an unmapped area in the process' IA-32 memory space
 * input:  start - the address from which to start the search
 *         len - size of requested area
 * output: -ENOMEM if no free memory is found, address of area on success
 *         the returned address is aligned on system page-size for most
 *         cases, and on SHMLBA if MAP_SHARED is asserted in 'flags'.
 */
/*****************************************************************************/
unsigned long HIDDEN get_ia32_free_memory (unsigned long start, 
                                           unsigned long len, 
                                           unsigned flags,
					   int is_ia32_user_space)
{
#define SHMLBA_MINUS_ONE (SHMLBA - 1)
	int found = 0;
	unsigned long unmapped_area = 0;
	unsigned shared;
	list_handle region;
	long retval;
	unsigned long prev_region_end = 0ul;
	unsigned long max_valid_addr;

	DBPRINT (0, "start=%lx len=%lx flags=%x\n", start, len, flags);

	shared = flags & MAP_SHARED;

	if (start == 0) { 
		start = IA32_TASK_UNMAPPED_BASE;
	}


	/* if this area is going to be mapped than make sure the
	 * whole page is free. otherwise, if the end of the last
	 * page is allocated, its contents wil be erased once the
	 * page is mapped again */
	len = PAGE_ALIGN (len);


	if (shared) {
		/* shared memory must be aligned on SHMLBA */
		start = (start + SHMLBA_MINUS_ONE) & ~SHMLBA_MINUS_ONE;
	}

	/* go over the list of mapped memory regions and look for a free slot */
	region = list_get_first (&m_regions);

	while (1) {

		region_attrib * ra;
		unsigned long region_end;
		unsigned long region_len;
		unsigned long region_start = 0;
		unsigned long slot_size = 0;


		if (region == 0) {
			/* no more regions to check */
			break;
		}

		
		region_start = get_element_key (region);
		assert (region_start >= prev_region_end);

		/* get the region's size */
		ra = (void*)get_element_data (region);
		region_len = ra->len;
		region_end = region_start + region_len;



		/* don't get areas lower than 'start' */
		if (region_start >= start) {

			/* the memory needs to be aligned at the native page size */
			unmapped_area = PAGE_ALIGN (MAX(start, prev_region_end));

			if (shared) {
				/* shared memory must be aligned at SHMLBA */
				unmapped_area = (unmapped_area + SHMLBA_MINUS_ONE) & 
						~SHMLBA_MINUS_ONE;
			}
			
			/* the free slot stretches from 'unmapped_area' to 'region_start' .
			 * unmapped_area might be bigger than region_start because it
			 * is aligned on a page boundry, while region_start not */
			if (region_start >= unmapped_area) {
				slot_size = region_start - unmapped_area;
				if (slot_size >= len) {
					found = 1; /* free area was found */
					break;
				}
			}
		}
	
		prev_region_end = region_end;
		region = get_element_next (region);
	}

	DBPRINT (0, "found=%d\n", found);

	/* there is no free slot between mapped areas - try the end of the list */
	if (!found) {

		/* the memory needs to be aligned at the native page size */
		unmapped_area = PAGE_ALIGN (MAX(start, prev_region_end));

		if (shared) {
			/* shared memory must be aligned at SHMLBA */
			unmapped_area = (unmapped_area + SHMLBA_MINUS_ONE) & 
					~SHMLBA_MINUS_ONE;
		}
	} 

	/* unmapped_area is the address we want to return */
	assert (unmapped_area >= start);

	/* if the first available page is too high then return error */
	max_valid_addr = is_ia32_user_space? IA32_MAX_USER_ADDR : IA32_MAX_KERNEL_ADDR;
	retval = (unmapped_area + len > max_valid_addr) ?
			-ENOMEM : 
			(long)unmapped_area;
	DBPRINT (0, "retval=%lx\n", (unsigned long)retval);
	return (unsigned long)retval;
}

#if DEBUG
void HIDDEN print_region (list_handle region)
{
	unsigned long start, len, end;
	unsigned prot;
	unsigned flags;
	region_attrib * ra = (void*)get_element_data (region);

	start = get_element_key (region);
	len = ra->len;
	prot = ra->prot;
	flags = ra->flags;
	end = start + len;
	fprintf (g_log_file, "%lx:%lx-%x-%x ", start, end, prot, flags);
}

void HIDDEN print_regions ()
{
	list_handle h;

	h = list_get_first (&m_regions);
	while (h != 0) {
		print_region (h);
		h = get_element_next (h);
	}
	fprintf (g_log_file, "\n");
}
#endif

