/* 
 * user32_mem_utils.c
 *
 *  utility functions related to memory
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "user32.h"
#include "user32_mem_regions.h"
#include "ia32toia64.h"

#define MREMAP_MAYMOVE  1
#define MREMAP_FIXED    2

#define HEAP_FLAGS MAP_ANON | MAP_PRIVATE

unsigned long HIDDEN g_page_size;
unsigned int HIDDEN g_ldt[IA32_LDT_ENTRIES];

long HIDDEN (*g_safe_memcpy)(void*,void*,unsigned long);
long iti_set_brk (unsigned end_code, unsigned brk_value);
long HIDDEN get_file_contents (int fd, unsigned long offset, char * buf, unsigned long size);

static unsigned m_end_code;
static unsigned m_brk;

extern long HIDDEN sys32_munmap (__u32 start, __u32 len);
extern long do_mmap32 (unsigned long addr, unsigned long len, unsigned int prot, 
                       unsigned int flags, unsigned int fd, unsigned long offset);
/*****************************************************************************/
/*
 * initialize the memory module
 */
/*****************************************************************************/
long HIDDEN init_mem_module (int page_size, 
                             long (*p_safe_memcpy)(void *, void *, unsigned long))
{
	assert (page_size != 0);
	g_page_size = page_size;
	g_safe_memcpy = p_safe_memcpy;
	memset (g_ldt, 0, sizeof (g_ldt));

	return 0;
}

long iti_set_brk (unsigned end_code, unsigned brk_value)
{
	m_end_code = end_code;
	m_brk = brk_value;

	return 0;
}

/*****************************************************************************/
/*
 * allocate one page in ia32's kernel space, that is: 3G - 4G.
 * this func should always be called by btlib, should never be called by 
 * ia32 binaries.
 */
/*****************************************************************************/
long iti_alloc_kernel_page (unsigned int prot)
{
	unsigned int flags = MAP_SHARED | MAP_ANON;
	long start;
	long retval = -1;

	/* this func is called by btlib, and does NOT convert prot */	
	if (prot & ~(PROT_READ | PROT_WRITE | PROT_EXEC)) {
		return -1;
	}
	
	mutex_lock (g_memory_sync);
	
	start = get_ia32_free_memory (IA32_MIN_KERNEL_ADDR, g_page_size, prot, 0);
	if (start < 0) {
		DBPRINT (TRUE, "getting free area failed\n");
		goto out;
	}
	
	flags |= MAP_FIXED;
	retval = SYSCALL (mmap, (start, g_page_size, prot, flags, 0, 0));
	if (retval < 0) {
		DBPRINT (TRUE, "mmap failed\n");
		retval = -1;
		goto out;
	} else {
		/* update the memory regions */
		assert (start == retval);
		assert (start < (long)(IA32_MAX_KERNEL_ADDR - g_page_size));
		assert (start >= (long)IA32_MIN_KERNEL_ADDR);

		retval = insert_region (start, g_page_size, prot, MAP_SHARED);
		if (retval < 0) {
			SYSCALL (munmap, (start, g_page_size)); /* don't need return value */
			DBPRINT (TRUE, "insert_region failed\n");
			retval = -1;
			goto out;
		}

		retval = start;
	}

out:
	mutex_unlock (g_memory_sync);
	
	return retval;
}

/*****************************************************************************/
/*
 * IA-32 mremap has to be managed in userland because the kernel may move the
 * mapping to a region above 4G.
 */
/*****************************************************************************/
long HIDDEN sys32_mremap (unsigned addr, unsigned old_len, unsigned new_len, 
                          unsigned flags, unsigned new_addr)
{
	long ret = 0;
	unsigned long mremap_retval;
	unsigned prot;
	unsigned mem_regions_flags;

	mutex_lock (g_memory_sync);

	/* 
	 * in the following cases we can call the 64 bit syscall directly -
	 * 1)	MREMAP_FIXED is asserted 
	 * 2)	MREMAP_MAYMOVE is not asserted.
	 * 3)	new_len < old_len
	 *
	 * in all these cases we can be sure that the address of the newly
	 * mapped area is inside the memory space of IA-32 processes.
	 *
	 * in all other cases we have to find a free memory area and call
	 * mremap with the new address with the MREMAP_FIXED flag asserted,
	 * otherwise, we will get memory above 4G.
	 */
	if (!(flags & MREMAP_FIXED) && 
	    (flags & MREMAP_MAYMOVE) &&
	    (old_len <= new_len)) {

		ret = get_ia32_free_memory (0, new_len, 0, 1);
		if (ret < 0) {
			goto out;
		}

		assert ((unsigned long)ret < IA32_MAX_USER_ADDR);
		new_addr = (unsigned) ret;
		flags |= MREMAP_FIXED;
	}

	/* filter out illegal requests */
	if (IA32_PAGE_OFF (addr)) {
		/* return -EINVAL when addr is not page aligned */
		ret = -EINVAL;
		goto out;
	}

	if (find_region (addr, NULL) != 0) {
		/* return -EFAULT when addr is not mapped */
		ret = -EFAULT;
		goto out;
	}
	
	if (flags & MREMAP_FIXED) {
		if (new_addr + new_len > IA32_MAX_USER_ADDR) {
			ret = -EINVAL;
		}
	}
	if (ret < 0) {
		goto out;
	}

	if (!(flags & MREMAP_MAYMOVE)) {
		if (addr + new_len > IA32_MAX_USER_ADDR) {
			ret = -EINVAL;
		}
	}
	if (ret < 0) {
		goto out;
	}


	/* do the syscall */
	mremap_retval = SYSCALL (mremap, (addr, old_len, new_len, flags, new_addr));

	if ((long)mremap_retval < 0) {
		ret = (long)mremap_retval;
		goto out;
	}
	assert (mremap_retval < IA32_MAX_USER_ADDR);

	/* update the region list */
	ret = get_region_prot (addr, &prot, NULL, NULL, &mem_regions_flags);
	if ( ! ret ) {
		prot = 0;
		mem_regions_flags = 0;
	}

	ret = remove_region (addr, old_len);
	if (ret < 0) {
		goto out;
	}
	
	ret = insert_region (mremap_retval, new_len, prot, mem_regions_flags);
	if (ret < 0) {
		goto out;
	}

	ret = (long)mremap_retval;

out:
	mutex_unlock (g_memory_sync);
	return ret;
}

/*****************************************************************************/
/*
 * IA-32 brk has to be managed in userland because the kernel does not allow
 * to put the heap below the text
 */
/*****************************************************************************/
long HIDDEN sys32_brk (unsigned req_brk)
{
	unsigned long newbrk, oldbrk;
	struct rlimit rlim;
	long retval;
	unsigned long addr;
	unsigned long len;
	unsigned int  prot;

	DBPRINT (0, "req_brk=%x m_brk=%x\n", req_brk, m_brk);
	
	if (req_brk < m_end_code){
		goto out;
	}


	/* newbrk holds the address of the beginning of the 32 bit page
	 * pass the requsted req_brk.
	 * oldbrk holds the address of the beginning of the 32 bit page
	 * pass the current req_brk.
	 */
	newbrk = IA32_PAGE_ALIGN (req_brk);
	oldbrk = IA32_PAGE_ALIGN (m_brk);
	DBPRINT (0, "oldbrk=%lx newbrk=%lx\n", oldbrk, newbrk);

	if (oldbrk == newbrk){
		/* the new req_brk and the current req_brk fall in the same page
		 * no need to map more memory 
		 */

		goto set_brk;
	}

	if (req_brk < m_brk) {
		assert (newbrk < oldbrk);

		/* Always allow shrinking req_brk */
		retval = sys32_munmap (newbrk, (unsigned)(oldbrk - newbrk));
		if (retval < 0) {
			goto out;
		} else {
			goto set_brk;
		}
	}
	
	/* if we got here that means we need to allocate more
	 * memory for the heap
	 */
	assert (newbrk > oldbrk);

	/* see if the request does not exceed process limits */
	retval = SYSCALL (getrlimit, (RLIMIT_DATA, &rlim));
	if (retval < 0) {
		DBPRINT (TRUE, "getrlimit failed: %ld\n", retval);
		goto out;
	}
	if (req_brk - m_end_code > rlim.rlim_cur) {
		goto out;
	}
	
	/* allocate more memory for the heap */
	addr = oldbrk;
	len = newbrk - oldbrk;

	/* request the memory as FIXED. it means that we need to check
	 * whether is is already in use, which requires a lock on the 
	 * memory sub-system */
	mutex_lock (g_memory_sync);
	if (is_inside_region (addr, len)) {
		DBPRINT (0, "memory already allocated\n");
		mutex_unlock (g_memory_sync);
		goto out;
	}

	prot = RX_PERSON(thread_personality)? PROT_EXEC : 0;
	prot |=	PROT_READ | PROT_WRITE;
	
	retval = do_mmap32 (addr, len, prot, HEAP_FLAGS | MAP_FIXED, -1, 0);
	mutex_unlock (g_memory_sync);

	DBPRINT (0, "addr=%lx retval=%lx\n", addr, retval);
	
	if (retval < 0) {
		goto out;
	}

set_brk:
	{
		/* clear newly allocated memory */
		unsigned long s = IA32_PAGE_ALIGN (m_brk);
		unsigned long e = MIN (req_brk, PAGE_ALIGN (s));
		if (e > s) {
			memset ((void*)s, 0, e - s);
		}
	}
	m_brk = req_brk;

out:
	DBPRINT (0, "m_brk=%x\n", m_brk);
	
	return m_brk;
}

/*****************************************************************************/
/* reads a file into memory
 * input:  fd - file descriptor
 *         offset - where to start the read
 *         buf - memory into which the file is to be read
 *         size - size of buffer
 * output:  error indication ( 0 on success 
 */
/*****************************************************************************/
long HIDDEN get_file_contents (int fd, unsigned long offset, char *buf, 
                               unsigned long size)
{
	unsigned long cur_pos;
	long ret; 
	char * p = NULL;
	unsigned long left = size;
														

	DBPRINT (0, "fd=%d offset=%lx buf=%p size=%lx\n", fd, offset, buf, size);
	/* keep the current file location */
	cur_pos = SYSCALL (lseek, (fd, 0, SEEK_CUR));

	/* move to the beginning of the 'mapped' area */
	ret = SYSCALL (lseek, (fd, offset, SEEK_SET));
	if (ret < 0) {
		goto out;
	}

	/* read the file contents */
	ret = 0;
	p = buf;
	while (left > 0) {
		ret = SYSCALL (read, (fd, p, left));
		if (ret < 0) {
			goto out;
		}
		if (ret == 0) {
			break;
		}
		left = left - ret;
		p += ret;
	}
	memset (buf + (size - left), 0, left);

	/* return to the saved file location */
	ret = SYSCALL (lseek, (fd, cur_pos, SEEK_SET));
	if (ret < 0) {
		goto out;
	}
out:
	return ret;
}

/*****************************************************************************/
/* get the access permission of the native page that contains the given 
 * address
 * input:  addr -
 *         is_exist - output param. will hold 1 if the page exists and 0 
 *                    otherwise
 * output: numeric representation of the permissions ( 0 in case the page
 *         does not exist
 */
/*****************************************************************************/
int HIDDEN get_page_prot_e (unsigned long addr, int *is_exist)
{
	int exist = 0;
	unsigned perm = 0;
	unsigned long p;

	/* linux-64 bit maps the first page as read-only */
	if (addr < g_page_size) {
		goto out;
	}
	/* the memory-regions list holds the IA-32 pages - the native pages 
	 * are the conjuction of these pages, both existance and permissions
	 */
	p = PAGE_START(addr);
	while (PAGE_START(p) == PAGE_START(addr)) { 
		unsigned long start, len;
		unsigned prot;
		exist |= (int) get_region_prot (p, &prot, &start, &len, NULL);
		perm |= prot;
		p = start + len;
	}
out:
	DBPRINT (0, "addr=%lx exist=%d perm=%x\n", addr, exist, perm);
	if (is_exist != NULL) {
		*is_exist = exist;
	}
	return perm;
}

