/* 
 * user32_memory.c
 *
 * wrappers for IA-32 memory related syscalls.
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>

#include "user32.h"
#include "user32_mem_regions.h"
#include "ia32toia64.h"

long HIDDEN sys32_munmap (__u32 start, __u32 len);
long do_mmap32 (unsigned long addr, unsigned long len, unsigned int prot, unsigned int flags, unsigned int fd, unsigned long offset);
static long mmap32_none4k (unsigned long start, unsigned long len, int prot, int flags, int fd, unsigned long off);
static long subpage_manip_mmap32 (unsigned long start, unsigned long end, int prot, int flags, int fd, unsigned long off);
static long subpage_manip_munmap32 (unsigned long start, unsigned long end);
extern long HIDDEN get_file_contents (int fd, unsigned long offset, char *buf, unsigned long size);

/*****************************************************************************/
/*
 * wrapper for mmap syscall 
 */
/*****************************************************************************/
long HIDDEN sys32_mmap (void *arg)
{
	long retval;
	struct legacy_mmap_arg_struct map;

	if (g_safe_memcpy (&map, arg, sizeof(map))) {
		return -EFAULT;
	}

	retval = do_mmap32 (map.addr, map.len, map.prot, map.flags, 
			 map.fd, map.offset);
	return retval;
}

/*****************************************************************************/
/*
 * wrapper for mmap2 syscall 
 */
/*****************************************************************************/
long HIDDEN sys32_mmap2 (unsigned int addr, unsigned int len, unsigned int prot, 
                         unsigned int flags, unsigned int fd, unsigned int pgoff)
{
	long retval;

	retval = do_mmap32 (addr, len, prot, flags, fd,
			(unsigned long)pgoff << IA32_PAGE_SHIFT);
	return retval;
}

/*****************************************************************************/
/*
 * wrapper for munmap syscall 
 */
/*****************************************************************************/
long HIDDEN sys32_munmap (__u32 start, __u32 len)
{
	unsigned long end = start + len;
	unsigned long subp_start, subp_end;
	long ret = 0;

	DBPRINT (0, "start=%x len=%x\n", start, len);

	if ((start & ~IA32_PAGE_MASK) || 
	    (start > IA32_MAX_USER_ADDR) || 
	    ((start + len) > IA32_MAX_USER_ADDR)) {
		return -EINVAL;
	}

	mutex_lock (g_memory_sync);
	if (g_page_size == IA32_PAGE_SIZE) {
		ret = SYSCALL (munmap, (start, end - start));
	} else {
		if (start >= end) {
			ret = -EINVAL;
			goto out;
		}
		if (PAGE_START(start) == PAGE_START(end)) {
			/* the whole region resides inside a single page */
			ret = subpage_manip_munmap32 (start, end);
		} else {

			if (start != PAGE_ALIGN (start)) {
				/* the beginning of the region is a sub-page */
				subp_start = start;
				start = PAGE_ALIGN (start);
				ret = subpage_manip_munmap32 (subp_start, start);
				if (ret < 0) {
					goto out;
				}
			}

			if (end != PAGE_START (end)) {
				/* the end of the region is a sub-page */
				subp_end = end;
				end = PAGE_START (end);
				ret = subpage_manip_munmap32 (end, subp_end);
				if (ret < 0) {
					goto out;
				}
			}

			if (start != end) {
				ret = SYSCALL (munmap, (start, end - start));
				if (IS_SYS_ERROR (ret)) {
					goto out;
				}
				/* update the region list */
				ret = remove_region (start, end - start);
				if (ret < 0) {
					goto out;
				}
			}
		}
	}
out:
	mutex_unlock (g_memory_sync);
	DBPRINT (0, "ret=%ld\n", ret);
	return ret;
}

/*****************************************************************************/
/*
 * wrapper for mprotect syscall 
 */
/*****************************************************************************/
long HIDDEN sys32_mprotect (unsigned int start, unsigned int len, int prot)
{
	long retval;


	if (g_page_size == IA32_PAGE_SIZE) {
		/* not update memory map, to be fixed */
		retval = SYSCALL (mprotect, (start, len, prot));
	} else {
		/* mprotect32_none4k locks the memory for itself */
		retval = mprotect32_none4k (start, len, prot);
	}


	return retval;
}

/*****************************************************************************/
/*
 * do the actual work of the mmap syscalls.
 * input and output are same as those of mmap
 */
/*****************************************************************************/
long do_mmap32 (unsigned long addr, unsigned long len, unsigned int prot, 
                unsigned int flags, unsigned int fd, unsigned long offset)
{
	long retval;
	long new_addr;


	len = IA32_PAGE_ALIGN(len);
	if (len == 0) {
		return addr;
	}

	if ((len > IA32_MAX_USER_ADDR) || (addr > (IA32_MAX_USER_ADDR - len))) {
		/* POSIX, when MAP_FIXED is specified, return ENOMEM */
		if (flags & MAP_FIXED){
			return -ENOMEM;
		} else {
			return -EINVAL;
		}
	}

	if (IA32_PAGE_OFF(offset)) {
		return -EINVAL;
	}
	
	prot = convert_prot64_to_prot32 (prot);

	if (g_page_size == IA32_PAGE_SIZE) {
		/* do not work for 4K page size, to be fixed */
		new_addr = SYSCALL (mmap, (addr, len, prot, flags, fd, offset));
	} else { 
		/* page size != 4K */
		mutex_lock (g_memory_sync);
		new_addr = mmap32_none4k (addr, len, prot, flags, fd, offset);
		mutex_unlock (g_memory_sync);
	}
	if (new_addr < 0) {
		retval = new_addr;
		goto out;
	}

	/* check the returned address */
	if ((unsigned long)new_addr >= IA32_MAX_USER_ADDR) {
		SYSCALL (munmap, (new_addr, len));
		retval = -ENOMEM;
	} else {
		retval = new_addr;
	}
out:
	DBPRINT (0, "retval=%lx\n", retval);
	return retval;
}

/******************************************************************************/
/*
 * do the mapping in systems where page size is other than 4K
 * input and output are same as those of mmap
 */
/******************************************************************************/
static long mmap32_none4k (unsigned long start, unsigned long len, 
			int prot, int flags, int fd, unsigned long offset)
{
	unsigned long ret;
	unsigned long ia32_start = 0ul;	/* return value in the 32 bit level */
	unsigned long orig_len = len;	/* keep the length requested by the user */
	unsigned long end = start + len;
	unsigned long body_start;
	unsigned long body_len; 
	unsigned long body_offset;
	unsigned long mem_regions_flags = flags & (MAP_SHARED | MAP_GROWSDOWN);
	
	DBPRINT (1, "start=%lx len=%lx prot=%x flags=%x fd=%d offset=%lx\n",
		 start, len, prot, flags, fd, offset);

	if (flags & MAP_GROWSDOWN) {
		flags &= ~MAP_GROWSDOWN;
	}

	if ((flags & MAP_FIXED) && (start != IA32_PAGE_START (start))) {
		ret = -EINVAL;
		goto out;
	}
	start = IA32_PAGE_ALIGN (start);

	/* in some cases we map the head and tail of the requested area 
	 * separately by using the subpage_manip_mmap32 function, so we are left
	 * with a well aligned 'body' that can be mapped by the system 
	 * directly */ 
	body_start = start;
       	body_len = len; 
	body_offset = offset;

	assert (start + len <= IA32_MAX_USER_ADDR);

	if (flags & MAP_FIXED) {

		ia32_start = start;

		/* if the requested address is not aligned on system page-size then
		 * map the beginning of the area separately */
		if (PAGE_OFF (start)) {
			unsigned long sub_page_start, sub_page_len;
			
			sub_page_start = start;
			body_start = PAGE_ALIGN (start);

			sub_page_len = MIN(end, body_start) - sub_page_start;
			len -= sub_page_len;
			
			ret = subpage_manip_mmap32 (
					sub_page_start, 
					sub_page_start + sub_page_len, 
					prot, flags, fd, offset);

			if (IS_SYS_ERROR (ret)) {
				DBPRINT (TRUE, "mapping head failed:%ld\n", ret);
				goto out;
			}
			if (! (flags & MAP_ANON)) {
				offset += sub_page_len;
				body_offset = offset;
			}
			body_len = len;
		}

		if ((len > 0) && PAGE_OFF (end)) {
			/* if the end is not aligned on a page size then 
			 * map the end of the area separately */
			unsigned long sub_page_start, sub_page_len;
			unsigned long sub_page_offset = 0;

			sub_page_start = PAGE_START (end);
			sub_page_len = end - sub_page_start;
			len -= sub_page_len;
			end = sub_page_start;
			if ((flags & MAP_ANON) == 0) {
				sub_page_offset = offset + len;
			}

			ret = subpage_manip_mmap32 (
					sub_page_start, 
					sub_page_start + sub_page_len, 
					prot, flags, fd, sub_page_offset);
			if (IS_SYS_ERROR (ret)) {
				DBPRINT (TRUE, "mapping tail failed:%ld\n", ret);
				goto out;
			}
			body_len = len;
		}
		ret = ia32_start;

	} else if (!(flags & MAP_SHARED) || (flags & MAP_ANON)) {
		/* if it's not shared mapping then we don't care about the 
		 * offset. just try to get a free address as close as possible
		 * to the requested 'start' address */
		unsigned long tmp_start, tmp_len;

		/* if the requested address is not aligned on system page-size
		 * then start looking for a free space at a lower address. this
		 * way we have a better chance to satisfy the user request
		 */
		tmp_start = PAGE_START (start);

		/* look for a bigger chunk of memory that requested because we
		 * try to return to the user the original start address. the
		 * end of the area is actually the original end, and tmp_start
		 * is at lower address than original start. */
		tmp_len = end - tmp_start;
		assert (tmp_len >= len);
		ret = get_ia32_free_memory (tmp_start, tmp_len, flags, 1);

		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "get_ia32_free_memory failed:%ld\n", ret);
			goto out;
		}
		if (ret == tmp_start)  {
			/* if the page where the user request resides is free then
			 * we can satisfy the user's request. the body start point
			 * and length should be updated, though. */
			ia32_start = start;
			start = body_start = ret;
			body_len = tmp_len;
		} else {
			/* if we cannot satisfy the user request then we go back to
			 * original start value. the requested length is 
			 * smaller, so we have a better chance to find a free
			 * slot at a point closer to the requested one */
			if (tmp_len != len) {
				/* there is no point in searching for af free
				   area again if tmp_len==orig_len */
				ret = get_ia32_free_memory (start, len, flags, 1);
				if (IS_SYS_ERROR (ret)) {
					DBPRINT (TRUE, "get_ia32_free_memory failed:%ld\n", ret);
					goto out;
				}
			}
			start = ia32_start = body_start = ret;
			assert (body_len == len);
		} 

	} else if (PAGE_OFF (offset)) {
		/* if offset is not aligned on a system page-size then
		 * start the mapping at an earlier point, so that the 
		 * area that needs to be shared will indeed be.
		 * note that the shared part of the file will be larger
		 * than what the user intended.
		 */
		unsigned long tmp_start;
		unsigned long tmp_len;
		if (PAGE_START(start) != 0) {
			tmp_start = start - PAGE_OFF (offset);
			/* also, we start the search for a free area at a start
			 * of the page in the hope to get an address as close
			 * as possible to the requested one */
			tmp_start = PAGE_START (tmp_start);
			/* look for a bigger chunk of memory that requested because we
			 * try to return to the user the original start address. the
			 * end of the area is actually the original end, and tmp_start
			 * is at lower address than original start. */
			tmp_len = end - tmp_start;
		} else {
			tmp_start = start;
			/* make sure we ask for more memory than what was requested
			 * by the user */
			tmp_len = PAGE_ALIGN(len+1);
		}

		ret = get_ia32_free_memory (tmp_start, tmp_len, flags, 1);
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "getting free area failed:%ld\n", ret);
			goto out;
		}
		
		ia32_start = ret + PAGE_OFF (offset);
		start = body_start = ret;
		body_len = len + PAGE_OFF (offset);
		body_offset = PAGE_START(offset);
	} else { 
		/* MAP_SHARED && PAGE_OFF(offset)==0 */
		/* we just need to make sure the start point is aligned
		 * on a system page-size */
		ret = get_ia32_free_memory (start, len, flags, 1);
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "getting free area failed:%ld\n", ret);
			goto out;
		}
		start = ia32_start = body_start = ret;
		assert (body_offset == offset);
		assert (body_len == len);
	}


	/* are we done ? */
	if (len == 0) {
		goto out;
	}


	/* when going to the system MAP_FIXED must be asserted so
	 * that we don't get addresses above 4G */
	flags |= MAP_FIXED;

	DBPRINT (0, "start=%lx len=%lx prot=%x flags=%x fd=%d offset=%lx\n",
		 body_start, body_len, prot, flags, fd, body_offset);
	if ((flags & MAP_ANON) || PAGE_OFF (body_offset) == 0) {
		/* anonymous mappings and mappings which offset is a multiple of
		 * the system page-size can be handled by the system */
		ret = SYSCALL (mmap, (body_start, body_len, prot, flags, fd, body_offset));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "mmap(0) failed:%ld\n", ret);
			goto out;
		}
	} else {
		/* mapping a file with an offset which is not multiple of a page-size
		 * cannot be done by the system. the best we can do here is to
		 * make anonymous mapping and read the file contents to memory.
		 * this mapping cannot be shared of-course */
		flags |= MAP_ANON;
		ret = SYSCALL (mmap, (body_start, body_len, prot | PROT_WRITE, flags, -1, 0));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (IS_SYS_ERROR (ret), "mmap(1) failed:%ld\n", ret);
			goto out;
		}
		
		assert (ret == body_start);

		ret = get_file_contents (fd, body_offset, (void*)ret, body_len);
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "reading file failed:%ld\n", ret);
			goto unmap;
		}

		if (! (prot & PROT_WRITE)) {
			ret = SYSCALL (mprotect, (body_start, body_len, prot));
			if (IS_SYS_ERROR (ret)) {
				DBPRINT (TRUE, "mprotect failed: body_start=%lx body_len=%lx "
					       "prot=%x ret=%ld \n", 
					 body_start, body_len, prot, ret);
				goto unmap;
			}
		}
	}

out:

	if (!IS_SYS_ERROR (ret)) {
		/* update the memory regions */
		ret = insert_region (ia32_start, IA32_PAGE_ALIGN(orig_len), prot, mem_regions_flags);
		if (IS_SYS_ERROR(ret)) {
			DBPRINT (TRUE, "insert_region failed:%ld\n", ret);
			goto unmap;
		}
	}


	if (IS_SYS_ERROR (ret)) {
		ia32_start = ret;
	}

	DBPRINT (0, "retval=%lx\n", ia32_start);

	return ia32_start;
unmap:
	SYSCALL (munmap, (start, len)); /* don't need return value */
	goto out;
}

/*****************************************************************************/
/* map a subpage
 * input:  same as that of sys32_mmap
 * output: 0 on success, error code on failure
 */
/*****************************************************************************/
static long subpage_manip_mmap32 (unsigned long start, unsigned long end, int prot,
                                  int flags, int fd, unsigned long off)
{
	void *temp_page = NULL;
	long r, ret = 0, sub_page;
	int old_prot, page_exists;
	int temp_prot = prot | PROT_WRITE;
	int tflags;
	void * p = (void*)PAGE_START(start);

	DBPRINT (0, "fd=%d start=%lx end=%lx prot=%x flags=%x off=%lx\n", 
		 fd, start, end, prot, flags, off);

	assert (start == IA32_PAGE_ALIGN (start));

 	old_prot = get_page_prot_e(start, &page_exists);

	if (page_exists) {
		/* keep the contents of the page */
		temp_page = (void*)
			SYSCALL (mmap, (0, PAGE_SIZE, PROT_READ | PROT_WRITE, 
					MAP_PRIVATE | MAP_ANON, -1, 0));

		if (IS_SYS_ERROR (temp_page)) {
			return -ENOMEM;
		}

		g_safe_memcpy (temp_page, p, PAGE_SIZE);
	}

	tflags = flags | MAP_FIXED | MAP_ANONYMOUS;
	/* map the page */
	sub_page = SYSCALL (mmap, ((unsigned long)p, PAGE_SIZE, temp_prot, tflags, -1, 0));


	if (IS_SYS_ERROR(sub_page)) {
		DBPRINT (TRUE, "mmap failed p=%p size=%lx prot=%x flags=%x retval=%ld\n", 
			 p, PAGE_SIZE, temp_prot, tflags, sub_page);
		ret = sub_page;
		goto free_dummy_page;
	}

	if (page_exists) {
		/* copy back the old page contents.*/
		if (PAGE_OFF(start)) {
			g_safe_memcpy (p, temp_page, PAGE_OFF(start));
		}
		if (PAGE_OFF(end)) {
			g_safe_memcpy((void *) end, temp_page + PAGE_OFF(end),
				      PAGE_SIZE - PAGE_OFF(end));
		}
		/* clear anonymous mappings */
		if (flags & MAP_ANON) {
			memset ((void*)start, 0, IA32_PAGE_ALIGN (end - start));
		}
		
	}
	/* copy the file contents into the memory */
	if (!(flags & MAP_ANONYMOUS)) {
		ret = get_file_contents (fd, off, (void*)start, end - start);
		if (IS_SYS_ERROR(ret)) {
			DBPRINT (TRUE, "get_file_contents failed:%ld\n", ret);
			goto free_sub_page;
		}
	}

	/* set the permission for the page */
	if (temp_prot != (prot | old_prot)) {
		ret = SYSCALL (mprotect, ((unsigned long)p, PAGE_SIZE, prot | old_prot));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "mprotect failed:%ld\n", ret);
			goto free_sub_page;
		}
	}

	DBPRINT (0, "ret=%ld\n", ret);
	return ret;

free_sub_page:

	if (!page_exists) {
		r = SYSCALL (munmap, (sub_page, g_page_size));
		DBPRINT (r < 0, "munmap failed: %ld\n", r);
	}

free_dummy_page:

	r = SYSCALL (munmap, ((unsigned long)temp_page, PAGE_SIZE));
	DBPRINT (r < 0, "munmap failed: %ld\n", r);

	DBPRINT (0, "ret=%ld\n", ret);
	return ret;
}

/*****************************************************************************/
/* remove a subpage from the list
 * input:    start - start IA-32 region
 *           end - end IA-32 region
 * output:   error status - 0 on success, negative value on failure
 * remarks:  the function updates the bitmap of the page that holds
 *           the IA-32 region. if the bitmap becomes 0 then the page
 *           is unmapped.
 */
/*****************************************************************************/
static long subpage_manip_munmap32 (unsigned long start, unsigned long end)
{
	long page_needed;
	unsigned long len;
	long ret;

	end = IA32_PAGE_ALIGN (end);

	if (start == end) {
		return 0;
	}
	if (PAGE_START (start) != PAGE_START (end - 1)) {
		return -EINVAL;
	}
	ret = remove_region (start, end - start);
	if (ret != 0) {
		return -EINVAL;
	}

	len = PAGE_ALIGN (end) - PAGE_START (start);
	page_needed = is_inside_region (PAGE_START (start), len);
	if (!page_needed) {
		ret = SYSCALL (munmap, (PAGE_START (start), g_page_size));
	}

	return ret;
}


