/* 
 * user32_networking.c
 *
 * wrapper functions for sys_sendmsg and sys_recvmsg.
 * used by the external wrapper only
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/


#include <ia32x_utils.h>
#include "ia32ss_common.h"
#include "user32.h"
#include "ia32_networking.h"

static int msghdr32_to_msghdr(struct msghdr * msg, struct msghdr32 * msg32);
static int cmsghdr32_to_cmsghdr(struct msghdr * msg);
static int cmsg32_to_cmsg(struct msghdr * msg, struct msghdr32 * msg32);
extern inline int iovec_conv(struct iovec32 * iov32, struct iovec * iov);
/*****************************************************************************/
/* from /usr/include/bits/socket.h */
/*****************************************************************************/
inline struct cmsghdr *
__cmsg_nxthdr (struct msghdr *__mhdr, struct cmsghdr *__cmsg)
{
  if ((size_t) __cmsg->cmsg_len < sizeof (struct cmsghdr))
    /* The kernel header does this so there may be a reason.  */
    return 0;

  __cmsg = (struct cmsghdr *) ((unsigned long) __cmsg
			       + CMSG_ALIGN (__cmsg->cmsg_len));
  if ((unsigned char *) (__cmsg + 1) > ((unsigned char *) __mhdr->msg_control
					+ __mhdr->msg_controllen)
      || ((unsigned char *) __cmsg + CMSG_ALIGN (__cmsg->cmsg_len)
	  > ((unsigned char *) __mhdr->msg_control + __mhdr->msg_controllen)))
    /* No more entries.  */
    return 0;
  return __cmsg;
}

static inline struct cmsghdr32 * cmsg32_nxthdr (struct msghdr * mhdr, 
						struct cmsghdr32 * cmsg32, 
						int cmsg_len)
{
	cmsg32 = (struct cmsghdr32 *) UL((unsigned char *) cmsg32 +
		CMSG32_ALIGN (cmsg_len));
	if ((unsigned char *) (cmsg32 + 1) > ((unsigned char *) mhdr->msg_control
			+ mhdr->msg_controllen)
		|| ((unsigned char *) cmsg32 + CMSG32_ALIGN (cmsg_len) 
			> ((unsigned char *) mhdr->msg_control + mhdr->msg_controllen))) {
		/* No more entries.  */
		return NULL;
	}
	return cmsg32;
}

/* 
 * simulate sendmsg system call, due to 32-bit and 64-bit issue, 
 * we must handle: 
 *	1. io vector 
 *	2. control message header 
 */
long HIDDEN 
sys32_sendmsg (int fd, __u32 msg32, unsigned flags)
{
	struct msghdr msg;
	struct iovec * iov = NULL;
	int retval, ret;
	int free_msg_iov = 0;
	unsigned int i;

	DBPRINT (0, "fd=%d msg32=%p flags=%x\n", fd, I2P(msg32), flags);

	retval = msghdr32_to_msghdr (&msg, (struct msghdr32 *)UL(msg32));
	if (retval < 0) {
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_iov) && (msg.msg_iovlen > 0)) {
		iov = (struct iovec *) malloc (sizeof(struct iovec) * msg.msg_iovlen);
		if (iov == NULL) {
			retval = -ENOMEM;
			goto out;
		}
		free_msg_iov= 1;
		
		for (i = 0; i < msg.msg_iovlen; i++) {
			ret = iovec_conv((struct iovec32 *)&msg.msg_iov[i], &iov[i]);
			if (ret < 0) {
				retval = -EFAULT;
				goto out;
			} 
		}
		msg.msg_iov = iov;
	}

	if (msg.msg_controllen > INT_MAX) {
		retval = -ENOBUFS;
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_control)) {
		if (msg.msg_controllen > 0) {
			retval = cmsghdr32_to_cmsghdr (&msg);
			if (retval < 0)
				goto out;
		}
	}

	flags &= 0x7FFFFFFF; /* clear MSG_CMSG_COMPAT bit before calling the system call */
	msg.msg_flags = flags;
	retval = LONG_SYSCALL3 (sendmsg, fd, &msg, flags);

	if (msg.msg_controllen != 0) {
		free(msg.msg_control);
	}

out:
	if (free_msg_iov) {
		free (iov);
	}
	return retval;
}

/* 
 * simulate recvmsg system call, due to 32-bit and 64-bit issue, 
 * we must handle: 
 *	1. io vector 
 *	2. control message header 
 */
long HIDDEN 
sys32_recvmsg (int fd, __u32 msg32, unsigned int flags)
{
	struct msghdr msg;
	struct iovec * iov = NULL;
	int retval, ret;
	int free_msg_iov = 0, free_msg_control = 0;
	unsigned int cmsg_num, i;

	retval = msghdr32_to_msghdr (&msg, (struct msghdr32 *)UL(msg32));
	if (retval < 0) {
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_iov) && (msg.msg_iovlen > 0)) {
		iov = (struct iovec *) malloc (sizeof(struct iovec) * msg.msg_iovlen);
		if (iov == NULL) {
			retval = -ENOMEM;
			goto out;
		}
		free_msg_iov = 1;
		
		for (i = 0; i < msg.msg_iovlen; i++) {
			ret = iovec_conv((struct iovec32 *)&msg.msg_iov[i], &iov[i]);
			if (ret < 0) {
				retval = -EFAULT;
				goto out;
			} 
		}
		msg.msg_iov = iov;
	}

	cmsg_num = msg.msg_controllen / (sizeof(struct cmsghdr32) + sizeof(int));
	msg.msg_controllen += cmsg_num * 8;

	if (IS_VALID_PTR(msg.msg_control) && (msg.msg_controllen > 0)) {
		msg.msg_control = malloc(msg.msg_controllen);
		if (msg.msg_control == NULL) {
			retval = -ENOMEM;
			goto out;
		}
		free_msg_control = 1;
	}
	
	retval = LONG_SYSCALL3 (recvmsg, fd, &msg, flags);
	if (IS_SYS_ERROR (retval)) {
		goto out;
	}

	if (IS_VALID_PTR(msg.msg_control)) {
		ret = cmsg32_to_cmsg(&msg, (struct msghdr32 *)UL(msg32));
		if (ret < 0)
			retval = ret;
	}

	/* free the localy allocated memory */
out:
	if (free_msg_iov) {
		free (msg.msg_iov);
	}
	if (free_msg_control) {
		free (msg.msg_control);
	}

	return retval;
}

/*****************************************************************************/
/* convert msghdr32 to msghdr
 * input: msg - destination, 64-bit msghdr 
 *		msg32 - source, 32-bit msghdr
 * output: 0 on success, negative error code otherwise
 */					
/*****************************************************************************/
static int msghdr32_to_msghdr (struct msghdr * msg, struct msghdr32 * msg32)
{
	struct msghdr32 lm32;

	if (g_safe_memcpy (&lm32, msg32, sizeof(lm32)) < 0) {
		return -EFAULT;
	}

	if (lm32.msg_iovlen > UIO_MAXIOV) {
		return -EMSGSIZE;
	}

	msg->msg_name		= I2P(lm32.msg_name);
	msg->msg_namelen	= lm32.msg_namelen;
	msg->msg_iovlen		= lm32.msg_iovlen;
	msg->msg_flags		= lm32.msg_flags;
	msg->msg_iov		= I2P(lm32.msg_iov);	
	msg->msg_control	= I2P(lm32.msg_control);
	msg->msg_controllen	= lm32.msg_controllen;

	return 0;
}

/*****************************************************************************/
/* 
 * convert 32-bit style cmsghdr to 64-bit style cmsghdr. memory to hold 64-bit
 * cmsghdr information is allocted in this function and must be freed by the 
 * caller.
 * input: msg->msg_control - 32-bit cmsghdr 
 * output: 0 on success, negative error code otherwise
 */					
/*****************************************************************************/
static int cmsghdr32_to_cmsghdr(struct msghdr * msg)
{
	struct cmsghdr * cmsg, * cmsg_bak = NULL;
	struct cmsghdr32 * cmsg32;
	struct cmsghdr32_list * cmsghdr32_list_head = NULL, * cmsghdr32_list_tail = NULL, * p;
	unsigned long	cmsglen = 0;
	unsigned long cmsg32len = 0;
	unsigned long	cmsg_now_offset = UL(CMSG32_FIRSTHDR (msg));
	long retval = 0;
	int	free_cmsg = 0;

	/* record cmsg32 */
	for (cmsg32 = CMSG32_FIRSTHDR (msg);
		cmsg32 != NULL;
		cmsg32 = cmsg32_nxthdr(msg, cmsg32, cmsg32len)) {
		
		struct cmsghdr32 cmsg32_tmp;
		if (g_safe_memcpy (&cmsg32_tmp, cmsg32, sizeof (cmsg32_tmp)) < 0) {
			retval = -EFAULT;
			goto out;
		}
		cmsg32len = CMSG32_LEN(cmsg32_tmp);

		p = (struct cmsghdr32_list *)malloc(sizeof(struct cmsghdr32_list));
		if (p == NULL) {
			retval = -ENOBUFS;
			goto out;
		}
		
		if(cmsghdr32_list_head == NULL) {
			cmsghdr32_list_head = cmsghdr32_list_tail = p;
		}
		else {
			cmsghdr32_list_tail->next = p;
			cmsghdr32_list_tail = p;
		}
		
		p->item = cmsg32;
		p->next = NULL;		
	}

	/* get cmsglen */
	for (cmsg32 = CMSG32_FIRSTHDR (msg);
		cmsg32 != NULL;
		cmsglen += cmsg32len + CMSG_DIFF) {

		struct cmsghdr32 cmsg32_tmp;
		if (g_safe_memcpy (&cmsg32_tmp, cmsg32, sizeof (cmsg32_tmp)) < 0) {
			retval = -EFAULT;
			goto out;
		}
		
		cmsg32len = CMSG32_LEN(cmsg32_tmp);
		cmsg_now_offset += cmsg32len;
		if ((CMSG32_ALIGN(cmsg32len) < CMSG32_ALIGN(sizeof(struct cmsghdr32)))
			|| (cmsg32 > (struct cmsghdr32 *)cmsg_now_offset)) {
			retval = -EINVAL;
			goto out;
		}
		
		cmsg32 = cmsg32_nxthdr(msg, cmsg32, cmsg32len);
	}

	/* generate cmsg! */
	cmsg = (struct cmsghdr *)malloc(cmsglen);
	if (cmsg == NULL) {
		retval = -ENOBUFS;
		goto out;
	}
	free_cmsg = 1;

	/* construct cmsg */
	p = cmsghdr32_list_head;
	cmsg_bak = cmsg;
	while (p) {
		assert(IS_VALID_PTR(p->item));
		cmsg->cmsg_len = p->item->cmsg_len + CMSG_DIFF;
		cmsg->cmsg_level = p->item->cmsg_level;
		cmsg->cmsg_type = p->item->cmsg_type;
		if (g_safe_memcpy (CMSG_DATA(cmsg), CMSG32_DATA(p->item), 
			(p->item->cmsg_len  - CMSG32_ALIGN(sizeof(struct cmsghdr32)))) < 0) {
			retval = -EFAULT;
			goto out;
		}
		
		cmsg = (struct cmsghdr *)UL(((unsigned char *)cmsg + CMSG_ALIGN(cmsg->cmsg_len)));
		p = p->next;
	}

	msg->msg_controllen = cmsglen;
	msg->msg_control = cmsg_bak;
	free_cmsg = 0;

out:
	if (free_cmsg) {
		free (cmsg_bak);
	}
	while (cmsghdr32_list_head) {
		p = cmsghdr32_list_head->next;
		free (cmsghdr32_list_head);
		cmsghdr32_list_head = p;
	}
	return retval;
		
}
/* 
 * convert 64-bit control/ancillary data to 32-bit 
 * input: msg - 64-bit msghdr 
 *		msg32 - 32-bit msghdr
 * output: void
 */ 
static int cmsg32_to_cmsg(struct msghdr * msg, struct msghdr32 * msg32)
{
	struct cmsghdr * cmsg;
	struct cmsghdr32 * cmsg32, cmsg32_tmp;
	unsigned len = 0;
	unsigned int datalen;

	cmsg32 = (struct cmsghdr32 *)UL(msg32->msg_control);
	for (cmsg = CMSG_FIRSTHDR(msg);
		cmsg != NULL; 
		cmsg = CMSG_NXTHDR(msg, cmsg)) {

		cmsg32_tmp.cmsg_level = cmsg->cmsg_level;
		cmsg32_tmp.cmsg_type = cmsg->cmsg_type;
		datalen = cmsg->cmsg_len - CMSG_ALIGN(sizeof(*cmsg));
		cmsg32_tmp.cmsg_len = (CMSG32_ALIGN(sizeof(struct cmsghdr32))) + 
					datalen;

		if (g_safe_memcpy (cmsg32, &cmsg32_tmp, sizeof (cmsg32_tmp)) < 0) {
			return -EFAULT;
		}

		if (g_safe_memcpy (CMSG32_DATA(cmsg32), CMSG_DATA(cmsg), datalen) < 0) {
			return -EFAULT;
		}
		
		len += cmsg32_tmp.cmsg_len;

		cmsg32 = (struct cmsghdr32 *)UL((unsigned char *)cmsg32 + CMSG32_ALIGN(cmsg32->cmsg_len));
	}
	msg32->msg_controllen = len;

	return 0;
}
