/* 
 * user32_pathwalk.c
 *
 * contains wrappers for syscalls that need the altroot mechanism
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/



#if NOKERNELALTROOT
#include "user32.h"
#include "user32_pathwalk.h"

static long get_real_absolute_path (char * path, char * output_buf, unsigned long output_buf_len, char ** real_path);
#define ALTROOT "/emul/ia32-linux"
#define MIN_ALTROOT_BUF_SIZE (unsigned long)(PATH_MAX + sizeof (ALTROOT))
#define SIZEOF_ALTROOT sizeof(ALTROOT)-1

/*****************************************************************************/
/* emulate the altroot mechanism in the kernel. the function
 * returns the 'real' absolute path of the given file, meaning
 * that it adds the emulation directory to the path in case the
 * file exists in the emulation directory.
 * input:	path - file name (memory not trusted)
 *	output_buf - a buffer that will hold the real file name
 *			(memory trusted to be valid)
 *	output_buf_len - size of output_buf
 *	real_path - output variable - will hold the address of 
 *			the real file name
 * output: 0 on success, negative error value on failure */
/*****************************************************************************/
static long get_real_absolute_path (char * path, char * output_buf,
		unsigned long output_buf_len, char ** real_path)
{
	char * orig_name = output_buf + SIZEOF_ALTROOT;
	char * on;
	long safe_stringcopy (char * dest, char * src, unsigned long len);
	long ret;
	unsigned long orig_name_max_len = output_buf_len - sizeof (orig_name);
#if DEBUG
	int error = 0;
#endif

	DBPRINT (1, "path=%p outputbuf=%p output_buf_len=%ld real_path=%p\n",
			path, output_buf, output_buf_len, real_path);
	assert (output_buf_len >= MIN_ALTROOT_BUF_SIZE);

	if (path == NULL) {
		/* optimize away simple cases */
		*real_path = path;
		ret = 0;
		DBCODE (TRUE, error = 1);
		goto out;
	}
	/* copy the file name to a local buffer */
	ret = safe_stringcopy (orig_name, path, orig_name_max_len);
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "copy_string failed\n");
		/* let the kernel deal with it */
		*real_path = path;
		ret = 0;
		DBCODE (TRUE, error = 1);
		goto out;
	}

	assert ((unsigned long)ret <= orig_name_max_len);
	if ((unsigned long)ret == orig_name_max_len) {
		/* no null char in the string given by the user */
		DBPRINT (TRUE, "path too long\n");
		ret = -ENAMETOOLONG;
		goto out;
	}
	*(orig_name + ret) = '\0';
	DBPRINT (1, "orig_name='%s'\n", orig_name);

	if (*orig_name != '/') {
		/* relative path - no need to check emulation directory */
		*real_path = orig_name;
		ret = 0;
		goto out;
	}

	on = orig_name;
	do {
		on ++;
	} while (*on == '/');
	if (*on == '\0') {
		/* '/' must not be looked up in the emulation directory */
		*real_path = orig_name;
		ret = 0;
		goto out;
	}

	memcpy (output_buf, ALTROOT, SIZEOF_ALTROOT);

	struct stat statbuf;
	ret = SYSCALL (stat, (output_buf, &statbuf));
	if (ret == -ENOENT) {
		ret = 0;
		*real_path = orig_name;
	} else if (ret == 0) {
		*real_path = output_buf;
	}
out:
	DBPRINT (1, "real_path=%p<%s> ret=%ld\n", *real_path,
			(IS_SYS_ERROR(ret) || error) ? "nil" : *real_path, ret);
	return ret;
}

#define PATHWALKSYSCALL1(name)						\
long HIDDEN pws_##name (char * param1)					\
{									\
	char realname[MIN_ALTROOT_BUF_SIZE];				\
	char * realnamep;						\
	long ret;							\
									\
	ret = get_real_absolute_path (param1, realname,	\
			sizeof (realname), &realnamep);			\
	if (IS_SYS_ERROR (ret)) {					\
		goto out;						\
	}								\
									\
	ret = sys_##name (realnamep);					\
out:									\
	return ret;							\
}

PATHWALKSYSCALL1(unlink)
PATHWALKSYSCALL1(chdir)
PATHWALKSYSCALL1(rmdir)
PATHWALKSYSCALL1(acct)
PATHWALKSYSCALL1(chroot)
PATHWALKSYSCALL1(swapoff)

#define PATHWALKSYSCALL2(name,type2)					\
long HIDDEN pws_##name (char * param1,type2 param2)			\
{									\
	char realname[MIN_ALTROOT_BUF_SIZE];				\
	char * realnamep;						\
	long ret;							\
									\
	ret = get_real_absolute_path (param1, realname,	\
			sizeof (realname), &realnamep);			\
	if (IS_SYS_ERROR (ret)) {					\
		goto out;						\
	}								\
									\
	ret = sys_##name (realnamep,param2);				\
out:									\
	return ret;							\
}

PATHWALKSYSCALL2(creat,int)
PATHWALKSYSCALL2(chmod,mode_t)
PATHWALKSYSCALL2(umount,int)
PATHWALKSYSCALL2(utimes,void *)
PATHWALKSYSCALL2(access,int)
PATHWALKSYSCALL2(mkdir,int)
PATHWALKSYSCALL2(stat,void *)
PATHWALKSYSCALL2(lstat,void *)
PATHWALKSYSCALL2(truncate,unsigned long)
PATHWALKSYSCALL2(statfs,void *)
PATHWALKSYSCALL2(removexattr,char *)
PATHWALKSYSCALL2(lremovexattr,char *)
PATHWALKSYSCALL2(swapon,int)


#define PATHWALKSYSCALL3(name,type2,type3)				\
long HIDDEN pws_##name (char * param1,type2 param2,type3 param3)	\
{									\
	char realname[MIN_ALTROOT_BUF_SIZE];				\
	char * realnamep;						\
	long ret;							\
									\
	ret = get_real_absolute_path (param1, realname,			\
			sizeof (realname), &realnamep);			\
	if (IS_SYS_ERROR (ret)) {					\
		goto out;						\
	}								\
									\
	ret = sys_##name (realnamep,param2,param3);			\
out:									\
	return ret;							\
}

PATHWALKSYSCALL3(open,int,int)
PATHWALKSYSCALL3(mknod,int,dev_t)
PATHWALKSYSCALL3(chown,uid_t,gid_t)
PATHWALKSYSCALL3(lchown,uid_t,gid_t)
PATHWALKSYSCALL3(readlink,char *,int)
PATHWALKSYSCALL3(listxattr,char *,size_t)
PATHWALKSYSCALL3(llistxattr,char *,size_t)
PATHWALKSYSCALL3(execve,char **,char **)

#define PATHWALKSYSCALL4(name,type2,type3,type4)				\
long HIDDEN pws_##name (char * param1,type2 param2,type3 param3,type4 param4)	\
{										\
	char realname[MIN_ALTROOT_BUF_SIZE];					\
	char * realnamep;							\
	long ret;								\
										\
	ret = get_real_absolute_path (param1, realname,		\
			sizeof (realname), &realnamep);				\
	if (IS_SYS_ERROR (ret)) {						\
		goto out;							\
	}									\
										\
	ret = sys_##name (realnamep,param2,param3,param4);			\
out:										\
	return ret;								\
}
PATHWALKSYSCALL4(getxattr,char *,void *,size_t)
PATHWALKSYSCALL4(lgetxattr,char *,void *,size_t)

#define PATHWALKSYSCALL5(name,type2,type3,type4,type5)						\
long HIDDEN pws_##name (char * param1,type2 param2,type3 param3,type4 param4,type5 param5)	\
{												\
	char realname[MIN_ALTROOT_BUF_SIZE];							\
	char * realnamep;									\
	long ret;										\
												\
	ret = get_real_absolute_path (param1, realname,				\
			sizeof (realname), &realnamep);						\
	if (IS_SYS_ERROR (ret)) {								\
		goto out;									\
	}											\
												\
	ret = sys_##name (realnamep,param2,param3,param4,param5);				\
out:												\
	return ret;										\
}
PATHWALKSYSCALL5(setxattr,char *,void *,size_t,int)
PATHWALKSYSCALL5(lsetxattr,char *,void *,size_t,int)

#define PATHWALKSYSCALL2_2(name)					\
long HIDDEN pws_##name (char * param1,char * param2)			\
{									\
	long ret;							\
	char realname1[MIN_ALTROOT_BUF_SIZE];				\
	char * realname1p;						\
									\
	ret = get_real_absolute_path (param1, realname1,		\
			sizeof (realname1), &realname1p);		\
	if (IS_SYS_ERROR (ret)) {					\
		goto out;						\
	}								\
									\
	char realname2[MIN_ALTROOT_BUF_SIZE];				\
	char * realname2p;						\
									\
	ret = get_real_absolute_path (param2, realname2,		\
			sizeof (realname2), &realname2p);		\
	if (IS_SYS_ERROR (ret)) {					\
		goto out;						\
	}								\
									\
	ret = sys_##name (realname1p,realname2p);			\
out:									\
	return ret;							\
}

PATHWALKSYSCALL2_2(link)
PATHWALKSYSCALL2_2(pivot_root)
PATHWALKSYSCALL2_2(rename)
PATHWALKSYSCALL2_2(symlink)

/*****************************************************************************/
	/* sys_mount cannot use none of the above macros */
/*****************************************************************************/
long HIDDEN pws_mount (char * dev, char * path, char * type, unsigned long flags, void * data)
{
	char realname[MIN_ALTROOT_BUF_SIZE];
	char * realnamep;
	long ret;

	ret = get_real_absolute_path (path, realname,
			sizeof (realname), &realnamep);
	if (IS_SYS_ERROR (ret)) {
		goto out;
	}

	ret = SYSCALL (mount, (dev, realnamep, type, flags, data));
out:
	return ret;
}

#endif
