/* 
 * user32_process.c
 *
 * wrapper functions for process related syscalls.
 * used by the external wrapper only
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include "user32.h"
#if NOKERNELALTROOT
#include "user32_pathwalk.h"
#endif

#ifndef S_IFMT
#define S_IFMT	0170000
#define S_IFREG	0100000
#endif

static int get_nargs (__u32 args);

#ifdef NOKERNELALTROOT
static long check_noexec_mount (struct stat *stp)
{
	int fd = SYSCALL (open, ("/proc/mounts", O_RDONLY, 0));
	size_t alen = 1024;
	ssize_t len;
	long ret = 0;
	char *buf, *rp, *p, *q, *path;
	struct stat st;

	if (fd < 0)
		return 0;
	buf = malloc (alen);
	if (buf == NULL)
		goto out;
	p = rp = buf;
	while ((len = SYSCALL (read, (fd, rp, buf + alen - rp))) > 0) {
		while (1) {
			/* Lack of memchr... */
			for (q = p; q < p + len && *q != '\n'; ++q);
			if (q == p + len)
				break;
			*q = '\0';
			len -= q - p + 1;
			while (p < q && *p != ' ')
				p++;
			while (*p == ' ')
				p++;
			if (p == q)
				goto out;
			path = p;
			while (p < q && *p != ' ')
				p++;
			if (p == q)
				goto out;
			*p++ = '\0';
			if (strstr (p, "noexec")) {
				if (SYSCALL (stat, (path, &st)) >= 0
				    && st.st_dev == stp->st_dev) {
					ret = 1;
					goto out;
				}
			}
			p = q + 1;
		}
		if (p == buf) {
			if (rp + len == buf + alen) {
				char *new_buf = realloc (buf, 2 * alen);
				if (new_buf == NULL)
					goto out;
				rp = new_buf + (rp - buf);
				buf = new_buf;
				p = buf;
				alen *= 2;
				/* Should never happen.  */
				if (alen >= 0x8000000000000000)
					goto out;
			}
			rp += len;
		} else {
			memmove (buf, p, len);
			p = buf;
			rp = buf + len;
		}
	}
out:
	if (buf)
		free (buf);
	SYSCALL (close, (fd));
	return ret;
}

static long script_execve (char *interp, char *arg, char *filename,
			   int fd, char **argv, char **envp)
{
	/* Should use VLA here too.  */
	char name[sizeof "/emul/ia32-linux" + 128];
	char *args[4];
	struct stat st;
	long ret = 1;
	size_t len;

	memcpy (name, "/emul/ia32-linux", sizeof ("/emul/ia32-linux") - 1);
	/* Caller guarantees interp is at most 127 chars long.  */
	len = strlen (interp);
	if (len == 0 || len >= 128)
		goto out;
	memcpy (name + sizeof ("/emul/ia32-linux") - 1, interp, len + 1);
	if (IS_SYS_ERROR (SYSCALL (stat, (name, &st)))) {
		/* ret=1 - let the system handle this case */
		goto out;
	}
	if ((st.st_mode & 0111) == 0) {
		ret = -EACCES;
		goto out;
	}
	if (argv == NULL || *argv == NULL) {
		int i = 0;
		args[i++] = interp;
		if (arg)
			args[i++] = arg;
		args[i++] = filename;
		args[i++] = NULL;
		argv = args;
	} else {
		*argv = filename;
		if (arg)
			*--argv = arg;
		*--argv = interp;
	}
	/* To check executable perm's, we can't use access here, as that
	   checks uid/gid, not fsuid/fsgid.  Well, euid/egid is not
	   fsuid/fsgid either...  */
	if (SYSCALL (fstat, (fd, &st)) < 0)
		goto out;

	ret = -EACCES;
	if ((st.st_mode & 0111) != 0111) {
		uid_t uid = SYSCALL (geteuid, ());
		/* FIXME: this should actually be capable (CAP_DAC_OVERRIDE).  */

		if (uid == 0 && (st.st_mode & 0111)) {
		/* if the file is executable at all than root can exec it */
			/* do nothing */;
		} else if (uid == st.st_uid) {
			/* check owner perm */
			if ((st.st_mode & 0100) == 0) {
				goto out;
			}
		} else if (SYSCALL (getegid, ()) == st.st_gid) {
			/* check main group perm */
			if ((st.st_mode & 010) == 0) {
				goto out;
			}
		} else if ((st.st_mode & 011) != 011) {
			/* check secondary group perm */
			gid_t grps[64], *groups = grps;
			long r;
			if ((st.st_mode & 011) == 0)
				goto out;
			r = SYSCALL (getgroups, (64, grps));
			if (r == -EINVAL) {
				r = SYSCALL (getgroups, (0, NULL));
				if (r < 0)
					goto out;
				groups = malloc (r * sizeof (gid_t));
				if (groups == NULL) {
					ret = -ENOMEM;
					goto out;
				}
				if (SYSCALL (getgroups, (r, groups)) != r) {
					free (groups);
					goto out;
				}
			} else if (r < 0)
				goto out;
			while (r > 0)
				if (groups[--r] == st.st_gid) {
					r = 3;
					break;
				}

			if (groups != grps)
				free (groups);

			if ((st.st_mode & (1 << r)) == 0)
				goto out;
		}
	}
	if (check_noexec_mount (&st))
		goto out;
	SYSCALL (close, (fd));
	return SYSCALL (execve, (name, argv, envp));
out:
	SYSCALL (close, (fd));
	return ret;
}

static long execve_altroot_script (char *filename, char **argv, char **envp)
{
	long ret = 1;
	char buf[128]; /* Kernel's BINPRM_BUF_SIZE */
	size_t len;
	int fd = pws_open (filename, O_RDONLY, 0);

	if (fd < 0)
		return fd;

	len = SYSCALL (pread, (fd, buf, sizeof (buf) - 1, (off_t)0));
	if (len > 2 && buf[0] == '#' && buf[1] == '!') {
		char *p, *interp = buf + 2, *arg = NULL;

		buf[len] = '\0';
		while (*interp == ' ' || *interp == '\t')
			interp++;
		if (*interp == '\0') {
			ret = -ENOEXEC;
			goto out;
		}
		p = interp;
		while (*p && *p != ' ' && *p != '\t' && *p != '\n')
			p++;
		if (*p != '\0' && *p != '\n') {
			*p++ = '\0';
			while (*p == ' ' || *p == '\t')
				p++;
			if (*p != '\0' && *p != '\n') {
				char *argend;

				arg = p++;
				/* Kernel passes at most one argument.
				   If there are multiple words, it
				   strips trailing whitespace and passes
				   everything between start of 1st word
				   and end of last word as first
				   argument.  */
				for (argend = NULL; *p != '\0' && *p != '\n'; p++)
					if (*p != ' ' && *p != '\t')
						argend = NULL;
					else if (argend == NULL)
						argend = p;
				if (argend == NULL)
					argend = p;
				*argend = '\0';
			}
		} else
			*p = '\0';
		return script_execve (interp, arg, filename,
				      fd, argv, envp);
	}
out:
	SYSCALL (close, (fd));
	return ret;
}
#endif

/* convert the 4 bytes pointer arrays to 8 bytes pointer arrays */
long HIDDEN sys32_execve (__u32 filename, __u32 argv, __u32 envp)
{
	char      **ia64_argv_envp, **ia64_argv, **ia64_envp;
	long      ret;
	int       argc, envc;  /* number of list */
	int       i;
	unsigned  addr32;
	unsigned  *p;

	DBPRINT (0, "filename=<%s> argv=%p envp=%p\n", 
		 (I2P(filename) != NULL) ? (char *)I2P(filename) : "(nil)", 
		 I2P(argv), I2P(envp));

	if ((argc = get_nargs (argv)) < 0) {
		return argc;
	}

	if ((envc = get_nargs (envp)) < 0) {
		return envc;
	}
	
	if ((unsigned)argc > (MAX_ARG_PAGES * IA32_PAGE_SIZE) / sizeof (__u32) - 1) {
		DBPRINT (FALSE, "Too many arguments\n");
		return -E2BIG;
	}

	if ((unsigned)envc > (MAX_ARG_PAGES * IA32_PAGE_SIZE) / sizeof (__u32) - 1) {
		DBPRINT (FALSE, "Too many environs\n");
		return -E2BIG;
	}

	/* allocate memory for the argument/enviroment list,
	   leave 2 space for NULL, 2 for altroot script.  */
	ia64_argv_envp = malloc ((argc + envc + 4) * sizeof (void *));
	if (ia64_argv_envp == NULL) {
		DBPRINT (FALSE, "malloc failed\n");
		return -ENOMEM;
	}
	ia64_argv = ia64_argv_envp + 2;
	ia64_envp = ia64_argv + argc + 1;

	/* copy argument/environment list */
	p = I2P (argv);
	if (p != NULL) {
		for (i = 0; i <= argc; i++, p++) {
			if (g_safe_memcpy (&addr32, p, sizeof (addr32)) < 0) {
				ret = -EFAULT;
				goto out;
			}
			ia64_argv[i] = I2P (addr32);
#if DEBUG & 0
			if (ia64_argv[i]) {
				DBPRINT (TRUE, "\targv[%d]='%s'\n", i, ia64_argv[i]);
			}
#endif
		}
	}
	p = I2P (envp);
	if (p != NULL) {
		for (i = 0; i <= envc; i++, p++) {
			if (g_safe_memcpy (&addr32, p, sizeof (addr32)) < 0) {
				ret = -EFAULT;
				goto out;
			}
			ia64_envp[i] = I2P (addr32);
#if DEBUG & 0
			if (ia64_envp[i]) {
				DBPRINT (TRUE, "\tenvp[%d]='%s'\n", i, ia64_envp[i]);
			}
#endif
		}
	}

#if NOKERNELALTROOT
	/* First see if we aren't trying to execute a script.
	   If so and the interpreter is available in altroot, execute
	   using that interpreter instead.  */
	struct stat s;
	ret = pws_stat ((char *)I2P(filename), &s);
	if (ret >= 0 && (s.st_mode & S_IFMT) == S_IFREG && (s.st_mode & 0111)) {
		ret = execve_altroot_script ((char *)I2P(filename),
					     VALID_PTR(I2P(argv), ia64_argv),
					     VALID_PTR(I2P(envp), ia64_envp));
	} else
		ret = 1;
	if (ret == 1)
		/* use the pathwalk sycall wrapper in user32_pathwalk.c */
		ret = pws_execve ((char *)I2P(filename), 
				  VALID_PTR(I2P(argv), ia64_argv),
				  VALID_PTR(I2P(envp), ia64_envp));
#else
	ret = SYSCALL (execve, ((char *)I2P(filename), 
				VALID_PTR(I2P(argv), ia64_argv),
				VALID_PTR(I2P(envp), ia64_envp)));
#endif

out:
	/* if we got here then the syscall failed */
	free (ia64_argv_envp);

	DBPRINT (0, "ret=%ld\n", ret);
	return ret;
}

long HIDDEN sys32_getpriority (int who, int which)
{
	long err;
	long ret;

	/* getpriority returns a value in the range -20 to 20 */
	ret = SYSCALL (getpriority, (who, which));
	__asm__ __volatile__ ("mov %0=r10;;\n" : "=r"(err)); 
	/* in IA-32 - the syscall returns the priority with offset 20 */
	if (err == 0) {
		ret = 20 - ret;
	}

	return ret;
}

long HIDDEN sys32_nice(int increment)
{
	long newprio;
	long oldprio;
	long retval;
	long err;
	int tid = SYSCALL (gettid, ());

	oldprio = SYSCALL (getpriority, (PRIO_PROCESS, tid));
	__asm__ __volatile__ ("mov %0=r10;;\n" : "=r"(err)); 
	if (err < 0) {
		return oldprio;
	}

	newprio = oldprio + increment;
	newprio = MAX (-20, newprio);
	newprio = MIN (19, newprio);

	retval = SYSCALL (setpriority, (PRIO_PROCESS, tid, newprio));
	
	return retval;
}

/* calculate number of arguments/enviroment variables */
static int get_nargs (__u32 args)
{
	int      n = 0;
	unsigned arg;		/* pointer to argument */
	unsigned *p = I2P(args);
	
	if (p == NULL)
		return 0;

	for (;;) {
		if (g_safe_memcpy (&arg, p, sizeof (arg)) < 0) {
			return -EFAULT;	
		}
		if (arg == 0) {
			break;
		}
		n++;
		p++;
	}
	return n;
}

