/* 
 * btlib_api.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "btlib.h"
#include "btlib_sync.h"
#include "btlib_asm.h"
#include "btlib_ptemu.h"

void BTLIB_NYI (void);

/* Override special cases */
#define btlib_api_setjmp 		_setjmp
#define btlib_api_longjmp 		longjmp
#define btlib_api_pseudo_open_file	BTLIB_NYI
#define btlib_api_pseudo_close_file	BTLIB_NYI
#define btlib_api_pseudo_write_file	BTLIB_NYI

/* BTLib_API_Table */
API_TABLE_ELEMENT_TYPE btl_api_table_elements[NO_OF_BTLIB_APIS] = {
	#include "btrans.inc/ia32exec_os_api_inc.c"
};

/* BTLib placeholder table for BTGeneric plable pointers */
PLABEL_PTR_TYPE btl_place_holder_table[NO_OF_BTGENERIC_APIS];

/*
 --------------------------------------------------------------------------------
 * BTLIB APIs
 --------------------------------------------------------------------------------
 */

/*****************************************************************************/
/* 
 * btlib_api_get_thread_id
 *
 * Purpose: return the thread id of the current thread
 */
/*****************************************************************************/
__u64
btlib_api_get_thread_id (void)
{
	DBPRINT (btl_prt_api, "\n");
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	return cur_thread_data->tid;
}

/*****************************************************************************/
void
btlib_api_ia32_reenter (BTGENERIC_IA32_CONTEXT * bc32)
{
	DBPRINT (btl_prt_api, "\n");
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
#ifdef ALL_THREAD_AFFINITY
	if (SHOULD_TRACK_THREAD_SCLBLTY) {
		btlib_record_thread_exit_simulation ();
	}
#endif
	restart_btg (bc32, NULL);

}

/*****************************************************************************/
void
btlib_api_ia32_lcall (BTGENERIC_IA32_CONTEXT * bc32 UNUSED,
                      __u32 returnAddress UNUSED, __u32 targetAddress UNUSED)
{
	DBPRINT (btl_prt_api, "NYI\n");
}

/*****************************************************************************/
/*
 * void btlib_api_ia32_interrupt
 *
 * Purpose: Simulates exceptions initiated by IA32EL
 *
 * Input:  interrupt-code and IA-32 context
 * Output: none
 */
/*****************************************************************************/
void
btlib_api_ia32_interrupt (BTGENERIC_IA32_CONTEXT * bc32,
                          const BT_EXCEPTION_CODE except_code,
                          const __u32 return_address)
{

	assert (bc32 == cur_thread_data->bc32);

	DBPRINT (btl_prt_api,
		 "eip=%x except_code=%d return_address=%x\n",
		 bc32->eip, except_code, return_address);

	DBCODE (0, print_btgcontext32 (bc32));
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
#ifdef ALL_THREAD_AFFINITY
	if (SHOULD_TRACK_THREAD_SCLBLTY) {
		btlib_record_thread_exit_simulation ();
	}
#endif

	/* the normal case is 0x80 - system call */
	if (except_code == IA32_SYSCALL_INT) {
		longjmp (cur_thread_data->btgeneric_run_context, IA32_SYSCALL);
	} else {
		siginfo_t siginfo;
		unsigned int trapno = 0;

		memset (&siginfo, 0, sizeof (siginfo));

		/* map the interrupt code to siginfo codes */
		if (except_code < BT_MAX_INTERRUPT_NUMBER) {
			/* for int n instruction, only "int 3" and "into" are "real" trap, 
			   other software interrupt will generate a GP */
			if (except_code == 3) {
				/* int 3 instruction */
				siginfo.si_signo = SIGTRAP;
				trapno = 3;
			} else if (except_code == 4) {
				/* into */
				siginfo.si_signo = SIGSEGV;
				trapno = 4;
			} else {
				siginfo.si_signo = SIGSEGV;
			}
			if (except_code == 3 || except_code == 4) {
				bc32->eip = return_address;
			}
			siginfo.si_code = SI_KERNEL;
		} else {
			/* 
			 * the exception code is btgeneric specific - 
			 * convert it to linux siginfo
			 */
			assert (except_code > BT_NO_EXCEPT);
			except2signal (bc32, except_code, &siginfo, &trapno);
		}
		DBPRINT (btl_prt_sig, "signo=%d\n", siginfo.si_signo);

		/* 
		 * run the IA-32 sig handler
		 * no need to see if the signal is masked because
		 * it is not possible to mask signals that are the 
		 * result of hw interrupt (which is what this function emulates)
		 */
		assert (cur_thread_data->syscall_id == IA32_SYSCALL_NONE);
#if PTEMU_ENABLED
		if (cur_thread_data->ptrace & PTEMU_PTRACED) {
			long retval;

			/* 
			 * notify the debugger of the signal -
			 * the debugger will look at the tia_siganal field
			 * in the tls, for the signo
			 */
			cur_thread_data->tia_signal.siginfo.si_signo =
			    siginfo.si_signo;

			DBPRINT (btl_prt_sig,
				 "ptrace=%lx signo=%d eip=%x\n",
				 cur_thread_data->ptrace,
				 cur_thread_data->tia_signal.siginfo.si_signo,
				 bc32->eip);
			assert (cur_thread_data->tid == (int) SYSCALL (gettid, ()));
			retval = SYSCALL (tkill, (cur_thread_data->tid, PTEMU_SIG0));
			sys_err_chk (retval, "kill");
		}
#endif
		btlib_run_ia32_handler (&siginfo, NULL, trapno);
	}
	assert (!"should not get here");
}

/*****************************************************************************/
/*
 * btlib_api_ia32_jmp_ia64
 *
 * Purpose:
 *
 * Input:
 * Output:
 *
 */
/*****************************************************************************/
void
btlib_api_ia32_jmp_ia64 (BTGENERIC_IA32_CONTEXT *ia32context UNUSED,
                         __u32 returnAddress UNUSED, __u32 targetAddress UNUSED)
{
	assert (!"should never get here in Linux!");
}

/*****************************************************************************/
/*
 * btlib_api_ia32_sysenter
 *
 * Purpose:
 *
 * Input:
 * Output:
 *
 */
/*****************************************************************************/
void
btlib_api_ia32_sysenter (BTGENERIC_IA32_CONTEXT *ia32context UNUSED)
{
	DBPRINT (btl_prt_api, "NYI\n");
}

/*****************************************************************************/
/*
 * void btlib_api_lock_signals
 *
 * Purpose: Report to BTLib that BTGeneric is in a critical section. 
 *          BTLib should not deliver signals until BtlUnlockSignals is called.
 *
 */
/*****************************************************************************/
void
btlib_api_lock_signals (void)
{
	DBPRINT (btl_prt_api, "\n");
}

/*****************************************************************************/
/*
 * void btlib_api_unlock_signals
 *
 * Purpose: Report to BTLib that BTGeneric is out of a critical section.
 *          Future signals will be handled and not queued.
 *
 */
/*****************************************************************************/

void
btlib_api_unlock_signals (void)
{
	DBPRINT (btl_prt_api, "\n");
}

/*****************************************************************************/
/*
 * void * btlib_api_memory_alloc
 *
 * Purpose: map memory for BTGeneric
 *
 * Input:	start address, size, and protection
 * Output: pointer to the allocated memory
 *
 */
/*****************************************************************************/
void *
btlib_api_memory_alloc (void *start_address, __u32 size, __u64 prot)
{
	void * retval;

	DBPRINT (btl_prt_api | btl_prt_mem,
		 "start_address=%p size=%u prot=%lx\n", start_address, size,
		 prot);

	/*
	 * For allocations (code or data) that originate by IA32EL, we do the
	 * following:
	 * On the first call, we receive the special values INITIAL_CODE_ALLOC or
	 * INITIAL_DATA_ALLOC and replace them by the base code and data areas
	 * (defined in btlib.h). the following calls will pass addresses that are
	 * based on the return values of former memory allocation calls, and will
	 * therefore be passed to mmap unchanged.
	 */
	if (start_address == INITIAL_DATA_ALLOC) {
		start_address = (void *) INITIAL_DATA_ADDRESS;
	} else if (start_address == (void *) INITIAL_CODE_ALLOC) {
		start_address = (void *) INITIAL_CODE_ADDRESS;
	} else if (start_address == 0) {
		start_address = (void *) INITIAL_DEFAULT_ADDRESS;
	}
	DBPRINT (btl_prt_mem, "start_address=%p\n", start_address);

	retval = btlib_mem_alloc ((unsigned long) start_address,
					 (unsigned long) size, (unsigned) prot);

	DBPRINT (btl_prt_mem, "retval=%p\n", retval);

	if (IS_SYS_ERROR (retval)) {
		retval = NULL;
	} else {
		unsigned hi_act_addr = GET_HI_DWORD ((unsigned long)retval);
		unsigned hi_req_addr = GET_HI_DWORD ((unsigned long)start_address);
		int is_within_same_4G = (hi_act_addr == hi_req_addr) ? 1 : 0;

		assert (is_within_same_4G);
		if ( ! is_within_same_4G ) {
			/* this is not perfect, since in kernels that map
			 * memory below given start addr, the value of max_mem
			 * needs to be updated with the unmapping, and this is
			 * not done here. it means that in such kernels, all
			 * memory mapping requests from btg will fail from this
			 * point on */
			btlib_mem_free (retval, (unsigned long)size);
			retval = NULL;
		}
	}
	return retval;
}

/*****************************************************************************/
/*
 * void * btlib_api_memory_free
 *
 * Purpose: map memory for BTGeneric
 *
 * Input:	start address, size
 * Output: pointer to the allocated memory
 *
 */
/*****************************************************************************/

BT_STATUS_CODE
btlib_api_memory_free (void *start_address, __u32 size)
{
	long retval;
	DBPRINT (btl_prt_api | btl_prt_mem,
		 "start_address=%p size=%u\n", start_address, size);
	retval = btlib_mem_free (start_address, (unsigned long) size);
	DBPRINT (btl_prt_mem, "retval=%ld\n", retval);
	retval = (retval == 0) ? BT_STATUS_SUCCESS : BT_STATUS_UNSUCCESSFUL;
	return (BT_STATUS_CODE) retval;
}

/*****************************************************************************/
/*
 * btlib_api_memory_page_size
 *
 * Purpose: return the system's page size
 *
 */
/*****************************************************************************/
__u32
btlib_api_memory_page_size (void)
{
	DBPRINT (btl_prt_api, "\n");
	return IA32_PAGE_SIZE;
}

/*****************************************************************************/
/*
 * btlib_api_memory_change_permissions
 *
 * Purpose: call mprotect for BTGeneric
 *
 * Input: New memory permissions
 * Output: 0 = success
 *         current memory mask on failure
 *
 */
/*****************************************************************************/
__u64
btlib_api_memory_change_permissions (void *start_address, __u32 size,
                                     __u64 prot)
{

	DBPRINT (btl_prt_api,
		 "start_address=%p size=%u prot=%lx\n", start_address, size,
		 prot);
	return btlib_memory_change_permissions ((unsigned long) start_address,
						(unsigned long) size,
						(unsigned) prot);

}

/*****************************************************************************/
/*
 * int btlib_api_memory_query_permissions
 *
 * Purpose: return the permission of the memory at given address.
 *
 * Input:  Address to be queried
 * Output: Permission mask of address
 *
 */
/*****************************************************************************/
__u64
btlib_api_memory_query_permissions (void *address,
                                    void **region_start, __u32 *region_size)
{
	long perm = 0;
	unsigned long size;
	void *start;

	DBPRINT (btl_prt_api, "address=%p\n", address);

	/* get the permissions.
	 *  it is not possible to cast int* to long* because of alignment
	 *  difference, so we use a local variable for the size
	 */
	perm = btlib_memory_query_permissions (address, &start, &size);
	if (perm < 0) {
		assert (!"btlib_memory_query_permissions failed");
		/* treat errors as no-access */
		size = btl_page_size;
		start = (void *) ((unsigned long) address & ~(btl_page_size - 1));
		perm = 0;
	}

	if (region_start != NULL) {
		*region_start = start;
	}

	if (region_size != NULL) {
		*region_size = (__u32) size;
		assert (size <= (unsigned long) (*region_size));
	}
#if PTEMU_ENABLED
	/* TODO:
	 * this is a work around till the flush_instruction_cache_remote is
	 * implemented - make all pages writable, so smc mode is enforced
	 */
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	if (btl_is_ptraced != IS_PTRACED_NO) {
		DBPRINT (btl_prt_mem, "under debug\n");

		/* on IA-32 CPU, READ implicits EXECUTE */
		if ((perm & MEM_READ) || (perm & MEM_EXECUTE)) {
			perm |= MEM_WRITE;
		}
	}
#endif	// PTEMU_ENABLED

	assert (perm >= 0);
	return (__u64) perm;
}

/*****************************************************************************/
/*
 * int btlib_api_suspend_thread
 * Purpose: suspend execution of the given thread
 * Input:  handle - identifies the thread
 *         counter - how many times to try before giving up
 * Output: status code
 *
 */
/*****************************************************************************/

BT_STATUS_CODE
btlib_api_suspend_thread (__u64 handle, __u32 counter)
{
	long retval;

	DBPRINT (btl_prt_api, "handle=%lx counter=%u\n", handle, counter);
	retval = btlib_suspend_thread ((int) handle, counter);
	return (retval == 0) ? BT_STATUS_SUCCESS : BT_STATUS_UNSUCCESSFUL;
}

/*****************************************************************************/
/*
 * int btlib_api_resume_thread
 * Purpose: restart a thread that was suspended by BtlSuspendThread
 * Input:  handle - identifies the thread
 * Output: status code
 */
/*****************************************************************************/
BT_STATUS_CODE
btlib_api_resume_thread (__u64 handle)
{
	long retval;
	DBPRINT (btl_prt_api, "handle=%lx\n", handle);

	retval = btlib_resume_thread ((int) handle);
	return (retval == 0) ? BT_STATUS_SUCCESS : BT_STATUS_UNSUCCESSFUL;
}

/*****************************************************************************/
/*
 * long btlib_api_init_access_lock
 * Purpose: initialize a sync object
 * Input: pointer to a memory where the sync object resides
 * Output: status code
 */
/*****************************************************************************/

BT_STATUS_CODE
btlib_api_init_access_lock (void *lock)
{
	cs_handle ch = (cs_handle) lock;
	DBPRINT (btl_prt_api, "lock=%p\n", lock);
	btlib_init_critical_section (ch);
	return BT_STATUS_SUCCESS;
}

/*****************************************************************************/
/*
 * int btlib_api_lock_access
 * Purpose: aqcuire sync object
 * Input:  lock - pointer to sync object
 *         flag - whether to block
 * Output: status code
 */
/*****************************************************************************/
BT_STATUS_CODE
btlib_api_lock_access (void *lock, __u64 flag)
{
	long retval;
	cs_handle ch = (cs_handle) lock;
	DBPRINT (btl_prt_api, "lock=%p flag=%lu\n", lock, flag);
	
	/* all cs but the top most one, are not allowed to be held when calling
	btgeneric. this is because btgenric may call btlib, which, in turn
	will try to acquire another lock. if this other lock is the top most one,
	then the order in which cs must be taken will be broken */
	DBCODE (TRUE, assert (!btlib_is_inside_lock (cur_thread_data) ||
			       	btlib_is_inside_lock (cur_thread_data) == BTL_LOCK_IND_SUSPENSION));

	retval = btlib_enter_critical_section (ch, (flag == BLOCK) ? 1 : 0);
	return retval ? BT_STATUS_UNSUCCESSFUL : BT_STATUS_SUCCESS;
}

/*****************************************************************************/
/*
 * void btlib_api_unlock_access
 * Purpose: release sync object
 * Input: lock - sync object
 */
/*****************************************************************************/
void
btlib_api_unlock_access (void *lock)
{
	cs_handle ch = (cs_handle) lock;
	DBPRINT (btl_prt_api, "lock=%p\n", lock);
	btlib_leave_critical_section (ch);
}

/*****************************************************************************/
/*
 * void btlib_api_invalidate_access_lock
 * Purpose: free resources kept for the given sync object
 * Input: lock - the sync object
 */
/*****************************************************************************/
void
btlib_api_invalidate_access_lock (void *lock UNUSED)
{
#if DEBUG
	cs_handle ch = (cs_handle) lock;
	DBPRINT (btl_prt_api, "\n");
	assert (!who_holds_critical_section (ch));
#endif
}

/*************************************************************************/
/*   btlib_api_query_static_info
 *
 * Description:
 *
 *   Fill the given information structure (StaticInfoP).
 *   Fill only fields requested by the RequestedFields bit-field input.
 *   The information given by this API is fixed and valid for the entire
 *   process life-time.
 * Arguments:
 *   StaticInfoP - a pointer to buffer containing the BT_StaticInfo.
 *   RequestedFields - which fields are requested to be written.
 *   The function ignores all other fields as well as bits in this input,
 *   That have no match in the API definition.
 * Return Value:
 *   The given RequestedFields when all bits that correspond to fields that
 *   were successfully written, are *cleared*. 
 */
/**************************************************************************/
__u64
btlib_api_query_static_info (BT_STATIC_INFO * static_info,
                             __u64 requested_fields)
{
	DBPRINT (btl_prt_api, "\n");
	if (requested_fields & BT_STATIC_INFO_QUERY_JMP_BUF_SIZE) {
		BT_STATIC_INFO_JMP_BUF_SIZE (*static_info) =
			(__u32) (sizeof (jmp_buf));
		requested_fields &= ~BT_STATIC_INFO_QUERY_JMP_BUF_SIZE;
	}
	if (requested_fields & BT_STATIC_INFO_QUERY_LOGICAL_CPUS_NUMBER) {
		/* !!Temporary - should get from kernel */
		BT_STATIC_INFO_LOGICAL_CPUS_NUMBER (*static_info) = 1;
		requested_fields &= ~BT_STATIC_INFO_QUERY_LOGICAL_CPUS_NUMBER;
	}
	return requested_fields;
}

/*****************************************************************************/
/* 
 * Gets the ready print buffer from DBT_bt_printf() in dbt_substitutes.c
 * 
 */
/*****************************************************************************/
void
btlib_api_debug_print (__u8 * print_buffer UNUSED)
{
#if ! RELEASE
	fprintf (g_log_file, "%s", print_buffer);
#endif
}

/*****************************************************************************/
/*
 * void btlib_api_abort
 *
 * Purpose: abort thread execution
 *
 */
/*****************************************************************************/
void
btlib_api_abort (void)
{
	DBPRINT (btl_prt_api, "!!!!!!!!!!!!!!!\n");
	btlib_abort ();
}

/*****************************************************************************/
/*
 * void BTLib_NYI
 *
 * Purpose: Place Holder For Unimplemeted Apis
 *
 */
/*****************************************************************************/
void
BTLIB_NYI (void)
{
	DBPRINT (btl_prt_api, "\n");
}

/*****************************************************************************/
/* btlib_api_yield_thread_execution
 *
 * Description:
 *      Relinquish the remainder of the current thread's time slice 
 *      to any other thread that is ready to run
 * Arguments: None
 * Return Value: None
 */
/*****************************************************************************/
void
btlib_api_yield_thread_execution (void)
{
	DBPRINT (btl_prt_api, "\n");
	YIELD_THREAD_EXECUTION ();
}

/*****************************************************************************/
/* btlib_api_flush_ia64_instruction_cache
 *
 * Description:
 *       Notify kernel about a modification in 64 bit code made within 
 *       the given region
 * Arguments:
 *       address - IN Pointer to start of the region
 *       size    - IN Size of the region
 * Return Value:
 *       none
 */
/*****************************************************************************/
void
btlib_api_flush_ia64_instruction_cache (void *address, __u32 size)
{
#define CACHE_LINE_SIZE 32
	void *new_address = address;
	__u32 new_size = size;

	DBPRINT (btl_prt_api, "address=%p, size=%d\n", address, size);

	/* address and size must be 16 bytes aligned */
	assert (((__u64) address & 0xF) == 0);
	assert ((size & 0xF) == 0);
	assert (address != NULL);
	assert (size > 0);

	/* for address not aligned on 32 bytes boundary, adjust the address and size */
	if ((((__u64) address) & ((__u64) CACHE_LINE_SIZE - 1)) != 0) {
		new_address =
		    (void *) (((__u64) address) & (~((__u64) CACHE_LINE_SIZE - 1)));
		new_size += CACHE_LINE_SIZE / 2;
	}

	__flush_instruction_cache (new_address, new_size);
}

/*****************************************************************************/
BT_STATUS_CODE
btlib_api_memory_read_remote (BT_HANDLE process_handle,
                              void *addr, void *buffer, __u32 size)
{
	int pid;
	DBPRINT (btl_prt_api, "handle=%p,addr=%p,buffer=%p,size=%u\n",
		 process_handle, addr, buffer, size);
	pid = (int) (unsigned long) process_handle;
	return (btlib_peekdata (pid, addr, buffer, (__u64) size) == 0) ?
		BT_STATUS_SUCCESS : BT_STATUS_UNSUCCESSFUL;
}

/*****************************************************************************/
BT_STATUS_CODE
btlib_api_memory_write_remote (BT_HANDLE process_handle,
                               void *addr, const void *buffer, __u32 size)
{
	int pid;
	DBPRINT (btl_prt_api,
	         "handle=%p,addr=%p,buffer=%p,size=%u\n",
	         process_handle, addr, buffer, size);
	pid = (int) (unsigned long) process_handle;
	return (btlib_pokedata (pid, addr, buffer, (__u64) size) == 0) ?
		BT_STATUS_SUCCESS : BT_STATUS_UNSUCCESSFUL;
}

