/*  
 * btlib_core_file.c
 *  
 * contains the mechanism for support of IA-32 core files 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include "btlib.h"
#include "btlib_core_file.h"
#include "btlib_sync.h"
#include "btlib_ptemu.h"
#include "btlib_thread_data.h"
#include <ia32toia64.h>

extern int g_dont_core_suid_own;
extern btlib_thread_data_t *m_used_thread_list;
extern btrans_binary_map_t g_application_info;
static thread_core_info_t *m_thread_core_info_list = NULL;
int g_core_file_policy = DONT_SET_CORE_BIT; /* default */
extern char **g_envp;
static int m_fildes = -1;
static size_t m_core_file_size = 0;
static size_t m_core_file_limit;

/*****************************************************************************/
/* decide wheather the signal should activate the core dump mechanism        */
/* decision is based on:                                                     */
/*    - the signal is a signal that its default behavoir is to core dump     */
/*    - the thread is not under debug                                        */
/*    - core file size limit in the system permits creation of core files    */
/* input - the signal we want to decide for                                  */
/* output - 1 if core dump mechanism should be activated, 0 if not           */ 
/*****************************************************************************/

int should_core_dump(int signo)
{
	int retval;
	int core_limit;
	struct rlimit rlp;
	DBPRINT(0, "signo=%d\n", signo);
	retval = ((signo == SIGQUIT) ||
		 (signo == SIGILL) ||
		 (signo == SIGTRAP) ||
		 (signo == SIGABRT) ||
		 (signo == SIGFPE) ||
		 (signo == SIGSEGV) ||
		 (signo == SIGBUS) ||
		 (signo == SIGSYS) ||
		 (signo == SIGXCPU) ||
		 (signo == SIGXFSZ)) ? 1 : 0;
	retval = (retval && !(cur_thread_data->ptrace));
        if (retval) {
		core_limit = COREFILE_SYSCALL (getrlimit, (RLIMIT_CORE, &rlp));
		if (rlp.rlim_cur < MIN_COREDUMP) {
			retval = 0;
		} else {
			m_core_file_limit = rlp.rlim_cur;
		}
	}
	if (g_dont_core_suid_own) {
		retval = 0;	
	}
	DBPRINT(0, "retval=%d\n",retval);
	return retval;
}


/*****************************************************************************/
/* fill the elf header                                                       */
/* input - elf header structure to fill                                      */
/*         number of memory regions                                          */ 
/*****************************************************************************/

static inline void init_elf_header(Elf32_Ehdr *hdr, int mem_reg_num)
{
	memset(hdr->e_ident, 0, (size_t)EI_NIDENT);
	hdr->e_ident[EI_MAG0] = ELFMAG0;
	hdr->e_ident[EI_MAG1] = ELFMAG1;
	hdr->e_ident[EI_MAG2] = ELFMAG2;
	hdr->e_ident[EI_MAG3] = ELFMAG3;
	hdr->e_ident[EI_CLASS] = ELFCLASS32;
	hdr->e_ident[EI_DATA] = ELFDATA2LSB;
	hdr->e_ident[EI_VERSION] = EV_CURRENT;
	hdr->e_type = ET_CORE;
	hdr->e_machine = EM_386;
	hdr->e_version = EV_CURRENT;
	hdr->e_entry = 0;
	hdr->e_phoff = sizeof(Elf32_Ehdr);
	hdr->e_shoff = 0;
	hdr->e_flags = 0;
	hdr->e_ehsize = sizeof(Elf32_Ehdr);
	hdr->e_phentsize = sizeof(Elf32_Phdr);
	hdr->e_phnum = mem_reg_num + 1;
	hdr->e_shentsize = 0;
	hdr->e_shnum = 0;
	hdr->e_shstrndx = 0;
	return;
}


/*****************************************************************************/
/* fill the note program header                                              */ 
/* input - elf program header structure to fill                              */
/*         offset of the note section in the core file                       */
/*         size of the note section                                          */
/*****************************************************************************/

static inline void init_note_phdr(Elf32_Phdr *phdr, __u32 offset, __u32 size)
{
	memset(phdr, 0, sizeof(*phdr));
	phdr->p_type = PT_NOTE;
	phdr->p_offset = offset;
	phdr->p_vaddr = 0;
	phdr->p_paddr = 0;
	phdr->p_filesz = size;
	phdr->p_memsz = 0;
	phdr->p_flags = 0;
	phdr->p_align = 0;
	return;
}


/*****************************************************************************/
/* convert from regular protection to elf program header protection          */
/* input - regular protection                                                */
/* output - elf protection                                                   */
/*****************************************************************************/

static unsigned int convert_prot_elf_phdr(unsigned int prot)
{
	int new_prot = 0;
	
	DBPRINT(0, "regular prot=%x\n", prot);
	switch (prot) {
		case PROT_READ:
			new_prot = PF_R;
			break;
		case PROT_WRITE:
			new_prot = PF_W;
			break;
		case PROT_READ | PROT_WRITE:
			new_prot = PF_R | PF_W;
			break;
		case PROT_EXEC:
			new_prot = PF_X;
			break;
		case PROT_READ | PROT_EXEC:
			new_prot = PF_R | PF_X;
			break;
		case PROT_WRITE | PROT_EXEC:
			new_prot = PF_W | PF_X;
			break;
		case PROT_READ | PROT_WRITE | PROT_EXEC:
			new_prot = PF_R | PF_W | PF_X;
			break;
		case PROT_NONE:
			break;
		default:
			assert (!"should never get here");
			break;
	}
	DBPRINT(0, "elf prot=%x\n", new_prot);	

	return new_prot;
}


/*****************************************************************************/
/* calcualte the load program header file size based on its corresponding    */
/* memory region's length and protection                                     */
/* input - length of the memory region related to the program header         */
/*         protection of the memory region related to the program header     */
/*         indication whether the memory region is shared                    */
/* output - length of the memory region if this memory region should be      */
/*          dumped, 0 otherwise                                              */
/*****************************************************************************/

static unsigned long calculate_phdr_filesz(unsigned long len, 
		unsigned int prot, unsigned int shared)
{
	unsigned long ret = 0;

	DBPRINT(0, "len =%ld prot=%d\n", len, prot);
	if (!(prot & PROT_READ)) {
		return 0;
	}
	if (prot & PROT_WRITE) {
		return len;
	}
	if (prot & (PROT_READ | PROT_EXEC)) {
		return 0;
	}
	if (shared & MAP_SHARED) {
		return 0;
	}
	DBPRINT(0, "ret=%ld\n", ret);
	return ret;
}


/*****************************************************************************/
/* fill the load program header                                              */
/* input - elf program header structure to fill                              */
/*         corresponding memory region                                       */
/*         current offset in core file                                       */ 
/*****************************************************************************/

static inline void init_load_phdr(Elf32_Phdr *phdr, struct memory_region *mem_reg, 
		unsigned int *off)
{
	memset(phdr, 0, sizeof(*phdr));
	phdr->p_type = PT_LOAD;
	phdr->p_vaddr = mem_reg->address;
	phdr->p_paddr = 0;
	phdr->p_memsz = mem_reg->len;
	phdr->p_offset = (*off);
	phdr->p_filesz = calculate_phdr_filesz(mem_reg->len, mem_reg->prot, 
			mem_reg->flags);
        if (phdr->p_filesz > 0) {
		mem_reg->dumpable = 1;
	}	
	*off += phdr->p_filesz;
	phdr->p_flags = convert_prot_elf_phdr(mem_reg->prot);
	phdr->p_align = IA32_PAGE_SIZE;
	return;
}


/*****************************************************************************/
/* copy registers                                                            */
/* input - registers structure to fill                                       */
/*         pointer to the thread's 32 bit context                            */
/*****************************************************************************/

static void copy_registers(elf_greg_t *regs, BTGENERIC_IA32_CONTEXT *bc32)
{
	regs[EBX] = bc32->ebx;	
	regs[ECX] = bc32->ecx;
	regs[EDX] = bc32->edx;	
	regs[ESI] = bc32->esi;	
	regs[EDI] = bc32->edi;	
	regs[EBP] = bc32->ebp;	
	regs[EAX] = bc32->eax;	
	regs[DS] = bc32->ds;	
	regs[ES] = bc32->es;	
	regs[FS] = bc32->fs;
	regs[GS] = bc32->gs;	
	regs[ORIG_EAX] = cur_thread_data->syscall_id;
	regs[EIP] = bc32->eip;	
	regs[CS] = bc32->cs;	
	regs[EFL] = bc32->eflags;	
	regs[UESP] = bc32->esp;
	regs[SS] = bc32->ss;	
	return;
}


/*****************************************************************************/
/* fill the prstatus structure                                               */ 
/* input - prstatus structure to fill                                        */
/*         signal information of the signal triggered the core mechanism     */
/*         thread id of the related thread                                   */
/*         pointer to the thread's 32 bit context                            */
/*****************************************************************************/

static void init_prstatus(prstatus_t *prstat, siginfo_t *siginfo, 
		int tid, BTGENERIC_IA32_CONTEXT *bc32)
{
	memset(prstat, 0, sizeof(*prstat));	
	prstat->pr_info.si_signo = siginfo->si_signo;
	prstat->pr_info.si_code = siginfo->si_code;
	prstat->pr_info.si_errno = siginfo->si_errno;
	prstat->pr_cursig = siginfo->si_signo;
	/*prstat->pr_sigpend = 0; */
	/*prstat->pr_sighold = 0; */
	prstat->pr_pid = tid;
	prstat->pr_ppid = COREFILE_SYSCALL (getppid, ());
	prstat->pr_pgrp = COREFILE_SYSCALL (getpgid, (prstat->pr_pid));
	prstat->pr_sid = COREFILE_SYSCALL (getsid, (prstat->pr_pid));
	/*prstat->pr_utime = 0; */
	/*prstat->pr_stime = 0; */
	/*prstat->pr_cutime = 0; */
	/*prstat->pr_cstime = 0; */
	copy_registers(prstat->pr_reg, bc32);
	prstat->pr_fpvalid = 1;
	return;
}


/*****************************************************************************/
/* fill the prpsinfo structure                                               */ 
/* input - prpsinfo structure to fill                                        */
/*****************************************************************************/

static void init_prpsinfo(prpsinfo_t *prinfo)
{
	unsigned int i;
	size_t args_size;
	size_t app_arg_size; 

	memset(prinfo, 0, sizeof(*prinfo));
	prinfo->pr_state = 0;
	prinfo->pr_sname = 'R';
	prinfo->pr_zomb = 0;
	/*prinfo->pr_flag = 0; */	 
	prinfo->pr_uid = COREFILE_SYSCALL (getuid, ());
	prinfo->pr_gid = COREFILE_SYSCALL (getgid, ());
	prinfo->pr_pid = COREFILE_SYSCALL (getpid, ());
	prinfo->pr_ppid = COREFILE_SYSCALL (getppid, ());
	prinfo->pr_pgrp = COREFILE_SYSCALL (getpgid, (prinfo->pr_pid));
	prinfo->pr_sid = COREFILE_SYSCALL (getsid, (prinfo->pr_pid));
	prinfo->pr_nice = COREFILE_SYSCALL (getpriority, (PRIO_PROCESS, prinfo->pr_pid));
	strncpy(prinfo->pr_fname, btl_ia32_bin_name, sizeof(prinfo->pr_fname));
	app_arg_size = g_application_info.arg_end - g_application_info.arg_start;
	args_size = (sizeof(prinfo->pr_psargs) > app_arg_size) ?
		app_arg_size : sizeof(prinfo->pr_psargs);
	memcpy(prinfo->pr_psargs, U2P(g_application_info.arg_start), args_size);
	for (i = 0; i < args_size; i++) {	
		if (prinfo->pr_psargs[i] == '\0') {
			prinfo->pr_psargs[i] = ' '; 
		}
	}
	return;
}


/*****************************************************************************/
/* copy floating point context part from btgeneric to core file structure    */ 
/* input - btgeneric context                                                 */
/*         core floating point structure to fill                             */ 
/*****************************************************************************/

static inline void btg2fsave_core (const BTGENERIC_IA32_CONTEXT * bc32, 
		struct i387_fsave_core_struct *fsave)
{
	fsave->cwd = bc32->fpu_state.control_word;
	fsave->swd = bc32->fpu_state.status_word;
	fsave->twd = bc32->fpu_state.tag_word;
	fsave->fip = bc32->fpu_state.instruction_pointer_offset;
	fsave->fcs = bc32->fpu_state.instruction_pointer_selector;
	fsave->foo = bc32->fpu_state.data_pointer_offset;
	fsave->fos = bc32->fpu_state.data_pointer_selector;
	memcpy (&fsave->st_space, &bc32->fpu_state.stack,
		sizeof (fsave->st_space));
	return;
}


/*****************************************************************************/
/* initialize fpu structure                                                  */ 
/* input - core fsave structure to fill                                      */
/*         pointer to the thread's 32 bit context                            */
/*****************************************************************************/

static inline void init_fpu (struct i387_fsave_core_struct *fpu, 
		BTGENERIC_IA32_CONTEXT *bc32)
{
	btg2fsave_core(bc32,fpu);
	return;
}


/*****************************************************************************/
/* initialize xfpu structure                                                 */ 
/* input - core fxsave structure to fill                                     */
/*         pointer to the thread's 32 bit context                            */
/*****************************************************************************/

static inline void init_xfpu (struct i387_fxsave_struct *xfpu, 
		BTGENERIC_IA32_CONTEXT *bc32)
{
	btg2fxsave(bc32,xfpu);
	return;
}


/*****************************************************************************/
/* fill the note structure                                                   */
/* input - note structure to fill                                            */
/*         note's name                                                       */
/*         note's type                                                       */
/*         note's description                                                */
/*         size of note's description                                        */
/*         size of note section till far (this function increments it)       */
/*****************************************************************************/

static void init_note(note *n, char *name, int type, void *desc, 
		size_t descsz, __u32 *note_size)
{
	n->nhdr.n_namesz = strlen(name) + 1;
	n->nhdr.n_descsz = descsz;
	n->nhdr.n_type = type;
	n->name = name;
	n->desc = desc;
	*note_size += sizeof(n->nhdr) + ROUND(n->nhdr.n_namesz, 4) + 
		ROUND(n->nhdr.n_descsz, 4);
	return;
}


/*****************************************************************************/
/* write a note to the core file                                             */
/* input - file descriptior of the core file                                 */
/*         the note we want to write                                         */
/*****************************************************************************/

static void write_note(int fd, note *n)
{
	off_t cur_offset;
	COREFILE_WRITE (sizeof(n->nhdr), (fd, &(n->nhdr), sizeof(n->nhdr)));
	COREFILE_WRITE ((size_t)n->nhdr.n_namesz, (fd, n->name, (size_t)n->nhdr.n_namesz));
	cur_offset = COREFILE_SYSCALL (lseek, (fd,(off_t)0,SEEK_CUR));
	m_core_file_size = COREFILE_SYSCALL (lseek, (fd,ROUND(cur_offset,4),SEEK_SET));
	COREFILE_WRITE ((size_t)n->nhdr.n_descsz, (fd, n->desc, (size_t)n->nhdr.n_descsz));
	cur_offset = COREFILE_SYSCALL (lseek, (fd,(off_t)0,SEEK_CUR));
	m_core_file_size = COREFILE_SYSCALL (lseek, (fd,ROUND(cur_offset,4),SEEK_SET));
	return;
}


/*****************************************************************************/
/* add thread core information element to thread core info list              */
/* input - thread id                                                         */
/* output - pointer to the thread core info element that was added           */
/*****************************************************************************/

static thread_core_info_t *add_thread_core_info (int tid)
{
	thread_core_info_t *tci = NULL;

	/* allocate memory for the new element */
	tci = (thread_core_info_t *) malloc (sizeof (thread_core_info_t));
	if (tci == NULL) {
		DBPRINT (TRUE, "malloc failed\n");
		btlib_abort ();
	}

	/* init the new element */
	memset (tci, 0, sizeof (*tci));
	tci->tid = tid;
	
	/* add the new element to the list */
	tci->next = m_thread_core_info_list;
	m_thread_core_info_list = tci;
	
	return tci;
}


/*****************************************************************************/
/* remove thread core info element from thread core info list                */
/* input - pointer to the thread core info element we want to remove         */
/*****************************************************************************/

static void remove_thread_core_info (thread_core_info_t *tci)
{
	thread_core_info_t *tci1, *tci2;

	/* find the previous element in the list */
	tci1 = m_thread_core_info_list;
	tci2 = NULL;
	while (tci1->tid != tci->tid) {
		tci2 = tci1;
		assert (tci1 != NULL);
		tci1 = tci1->next;
	}
	/* disconect from the list */
	if (tci2 == NULL) {	/* tci2 now points to the prev element */
		m_thread_core_info_list = tci->next; /* one element list */
	} else {
		tci2->next = tci->next;
	}
	/* free the memory */
	free (tci);
	return;
}


/*****************************************************************************/
/* gather thread core info for a specific thread                             */
/* the thread is canonized at this point, thus we can use its ia-32 context. */
/* we don't resume the thread. it will wait for the core dump mecanism to    */
/* end, and then it will be terminated with the whole process                */
/* input - thread id                                                         */
/*         pointer to the thread's 32 bit context                            */
/*         relevant signal information                                       */
/*         note section size so far (to modify)                              */
/*****************************************************************************/

static void dump_thread_info(int tid, BTGENERIC_IA32_CONTEXT *bc32, 
		siginfo_t *siginfo, __u32 *note_size)
{
	thread_core_info_t *tci;
	long retval;
	
	retval = btlib_suspend_thread(tid, (unsigned)-1);
	DBPRINT (retval, "btlib_suspend_thread failed\n");
	if (retval == 0) {
		DBPRINT(0, "adding %d to thread core list\n",tid);
		tci = add_thread_core_info(tid);	

		/* get thread's needed information */
		init_prstatus(&tci->eprstat, siginfo, tid, bc32);
		init_note(&tci->prstatus_note, "CORE", NT_PRSTATUS, &tci->eprstat, 
				sizeof(tci->eprstat), note_size);
		init_fpu(&tci->fpu, bc32);
		init_note(&tci->fpu_note, "CORE", NT_FPREGSET, &tci->fpu, 
				sizeof(tci->fpu), note_size); 

		init_xfpu(&tci->xfpu, bc32);
		init_note(&tci->xfpu_note, "LINUX", NT_PRXFPREG, &tci->xfpu, 
				sizeof(tci->xfpu), note_size);
	}

	return;
}

/*****************************************************************************/
/* read core file settings from relevant /proc files.                        */
/* input: filename (either /proc/sys/kernel/core_patterns or                 */
/*                 /proc/sys/kernel/core_uses_pid)                           */
/*        buffer to store the settings                                       */
/*        buffer's size                                                      */
/*****************************************************************************/

static void read_corefile_settings (char *filename, char *buf, size_t buf_len)
{
	int fd;
	size_t size;
	
	assert (!strncmp(filename, PROC_CORE_PATTERN, sizeof(PROC_CORE_PATTERN)) || 
		(!strncmp(filename, PROC_CORE_USES_PID, sizeof(PROC_CORE_USES_PID))));
	
	fd = COREFILE_SYSCALL (open, (filename, O_RDONLY, 0));
	if ((size = COREFILE_SYSCALL (read, (fd, buf, buf_len))) >= buf_len) {
		/* content of the file exceeds limit */
		DBPRINT (1, "content of the file exceeds limit\n");
		longjmp(cur_thread_data->btgeneric_run_context, m_fildes);	
	}
	DBPRINT (0, "size=%ld\n", size);
	buf[size - 1] = '\0'; /* end buffer with '\0' */ 	
	return;
}

/*****************************************************************************/
/* set core file name.                                                       */ 
/* core file name is decided based on the policy (policy 1 - create only 32  */
/* bit core file and don't set core bit; policy 2 - create also 64 bit core  */ 
/* file and set core bit) to avoid collision.                                */ 
/* the name is determined by analyzing /proc/sys/kernel/core_pattern and     */
/* /proc/sys/kernel/core_uses_pid                                            */
/*****************************************************************************/

static void set_corefile_name (char *name, size_t name_size)
{
	int pid;
	char core_uses_pid[32];
	char extension_buf[4] = ".32";
	char pid_buf[32];
	
	read_corefile_settings ("/proc/sys/kernel/core_pattern", 
				name, name_size);
	DBPRINT(0, "name=%s\n", name);
	read_corefile_settings ("/proc/sys/kernel/core_uses_pid", 
				core_uses_pid, sizeof(core_uses_pid));
	DBPRINT(0, "core_uses_pid=%s\n", core_uses_pid);

	pid = COREFILE_SYSCALL (getpid, ());
	sprintf (pid_buf, ".%d", pid);
	if (core_uses_pid[0] != '0') {
		if ((strlen (name) + strlen (pid_buf) + 1) >= name_size ) {
			/* length exceeds the buffer size */
			DBPRINT (1, "lenght exceeds the buffer size\n");
			longjmp(cur_thread_data->btgeneric_run_context, m_fildes);	
		}
		strncat (name, pid_buf, strlen (pid_buf));
		if (g_core_file_policy == SET_CORE_BIT) {
			if ((strlen (name) + strlen (extension_buf) + 1) >= name_size ) {
				/* length exceeds the buffer size */
				DBPRINT (1, "lenght exceeds the buffer size\n");
				longjmp(cur_thread_data->btgeneric_run_context, m_fildes);	
			}
			strncat (name, extension_buf, strlen (extension_buf));
		} 
	} else {
		if (g_core_file_policy == SET_CORE_BIT) {
			if ((strlen (name) + strlen (extension_buf) + 1) >= name_size ) {
				/* length exceeds the buffer size */
				DBPRINT (1, "lenght exceeds the buffer size\n");
				longjmp(cur_thread_data->btgeneric_run_context, m_fildes);	
			}
			strncat (name, extension_buf, strlen (extension_buf));
		} 
	}

	return;
}

/*****************************************************************************/
/* main function for creating an elf core file for the 32-bit application    */
/* this function is called when btlib decides it should create a core file   */
/* for the running process. it doea all the work involving the creation      */
/* of the core file                                                          */
/* input - relevant signal information                                       */ 
/*****************************************************************************/

void core_file(siginfo_t *siginfo)
{
	Elf32_Ehdr ehdr;
	Elf32_Phdr note_phdr;
	prstatus_t eprstat;
	prpsinfo_t eprinfo;
	struct i387_fsave_core_struct fpu;
	struct i387_fxsave_struct xfpu;
	__u32 note_off;
	__u32 note_sz = 0;
	note prstatus_note;
	note prpsinfo_note;
	note fpu_note;
	note xfpu_note;
	off_t cur_offset;
	char core_file_name[PATH_MAX];
	int number_of_program_headers;
	struct memory_region *mem_regions_buffer;	
	int i;	
	unsigned int cur_off;
	btlib_thread_data_t *susp_thread;
	thread_core_info_t *tci = NULL;
	static unsigned not_first = 0;

	/* allow only one thread in a time pass this point.
	the first thread trigerred the core file mechanism will do all the 
	work, and any other thread (if any) that will arrive afterwards, will
	do nothing. */
	
	btlib_acquire_lock (BTL_LOCK_IND_SUSPENSION);
	
	DBPRINT(0,"entered core_file with thread %d\n",cur_thread_data->tid);
	if (not_first) {
		btlib_release_lock (BTL_LOCK_IND_SUSPENSION);
		return;
	}
	not_first++;
	assert(not_first == 1); /* if false, more than one thread entered,
	or that the same thread entered twice, and this indicates that the
	lock mechanism is not safe and deadlock can occur */

	/* determine policy regarding IA32EL_DUMP_64_CORE_FILE_64 environment 
	variable */ 
	size_t core_len = strnlen (IA32EL_DUMP_64_CORE_FILE, (unsigned long)PATH_MAX);

	if (g_envp != (char **) NULL) {
		/* Go through the environment strings until we find 
		"IA32EL_DUMP_64_CORE_FILE=" */
		while (*g_envp) {
			if (strncmp (*g_envp, IA32EL_DUMP_64_CORE_FILE, 
					core_len) == 0) {
				break;
			} else {
				g_envp++;
			}
		}

		/* If *g_envp is not NULL, then we found IA32EL_DUMP_64_CORE_FILE */
		if (*g_envp) {
			if (strncmp((*g_envp + core_len), "0", (size_t)PATH_MAX)) {
				g_core_file_policy = SET_CORE_BIT;
			}
		}
		DBPRINT(0, "core file policy is set to %s\n", g_core_file_policy? 
					"DONT_SET_CORE_BIT" : "SET_CORE_BIT");
	}

	/* create the core file */
	set_corefile_name (core_file_name, sizeof(core_file_name));
	m_fildes = COREFILE_SYSCALL (open, (core_file_name, O_RDWR|O_CREAT, 00600));
	DBPRINT(0, "core file %s was created\n", core_file_name);

	/* get status of all other threads */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	susp_thread = m_used_thread_list;
	btlib_release_lock (BTL_LOCK_IND_THREAD);
	while (susp_thread != NULL) {
		if (susp_thread->tid == cur_thread_data->tid) {
			/* don't suspend current thread... */
			DBPRINT(0, "thread %d should not be suspended\n",
					susp_thread->tid);
		} else {
			DBPRINT(0, "get status of thread %d\n",susp_thread->tid);
			dump_thread_info(susp_thread->tid, susp_thread->bc32,
					siginfo, &note_sz);
		}
		btlib_acquire_lock (BTL_LOCK_IND_THREAD);
		susp_thread = susp_thread->next;
		btlib_release_lock (BTL_LOCK_IND_THREAD);
	}

	/* get information from the wrapper about all mapped memory regions */	
	mem_regions_buffer = calloc((size_t)MAX_MEMORY_REGIONS, 
			sizeof(struct memory_region));
	if (mem_regions_buffer == NULL) {
		/* can't allocate memory for mem_regions_buffer */ 
		longjmp(cur_thread_data->btgeneric_run_context, m_fildes);
	}
	number_of_program_headers = iti_enumerate_memory_regions(mem_regions_buffer, 
			MAX_MEMORY_REGIONS);
	if (number_of_program_headers == -1) {
		/* number of memory regions exceeds maximum allowed */
		longjmp(cur_thread_data->btgeneric_run_context, m_fildes);
	}

	/* write the elf header */
	init_elf_header(&ehdr, number_of_program_headers);
	COREFILE_WRITE (sizeof(ehdr), (m_fildes, &ehdr, sizeof(ehdr)));

	/* initialize the notes */
	init_prstatus(&eprstat, siginfo, COREFILE_SYSCALL(gettid, ()), cur_thread_data->bc32);
	init_note(&prstatus_note, "CORE", NT_PRSTATUS, &eprstat, 
			sizeof(eprstat), &note_sz);
	init_prpsinfo(&eprinfo);	
	init_note(&prpsinfo_note, "CORE", NT_PRPSINFO, &eprinfo, 
			sizeof(eprinfo), &note_sz);
	init_fpu(&fpu, cur_thread_data->bc32);
	init_note(&fpu_note, "CORE", NT_FPREGSET, &fpu, sizeof(fpu), &note_sz); 
	init_xfpu(&xfpu, cur_thread_data->bc32);
	init_note(&xfpu_note, "LINUX", NT_PRXFPREG, &xfpu, sizeof(xfpu), &note_sz);
	
	/* write the note program header */
	note_off = sizeof(Elf32_Ehdr) + (number_of_program_headers + 1) * 
		sizeof(Elf32_Phdr);
	init_note_phdr(&note_phdr, note_off, note_sz);
	COREFILE_WRITE (sizeof(note_phdr), (m_fildes, &note_phdr, sizeof(note_phdr)));

	/* write the load program headers */
	cur_off = ROUND(note_sz, IA32_PAGE_SIZE);
	for (i = 0; i < number_of_program_headers; i++) { 
		init_load_phdr(&note_phdr, &mem_regions_buffer[i], &cur_off);
		COREFILE_WRITE (sizeof(note_phdr), (m_fildes, &note_phdr, sizeof(note_phdr)));
	}
	
	/* write the notes (for current thread) */
	DBPRINT(0, "write status of current thread %d\n", cur_thread_data->tid);
	write_note(m_fildes, &prstatus_note);
	write_note(m_fildes, &prpsinfo_note);
	write_note(m_fildes, &fpu_note);
	write_note(m_fildes, &xfpu_note);
	/* write thread status notes */
	tci = m_thread_core_info_list; 
	while (tci != NULL) {
		DBPRINT(0, "write status of thread %d\n",tci->tid);
		write_note(m_fildes, &tci->prstatus_note);
		write_note(m_fildes, &tci->fpu_note);
		write_note(m_fildes, &tci->xfpu_note);
		tci = tci->next;
	}
	
	/* write dumpable memory regions */
	cur_offset = COREFILE_SYSCALL (lseek, (m_fildes,(off_t)0,SEEK_CUR));
	m_core_file_size = COREFILE_SYSCALL (lseek, (m_fildes,(off_t)ROUND(cur_offset,IA32_PAGE_SIZE),SEEK_SET));
	for (i = 0; i < number_of_program_headers; i++) {
		if (mem_regions_buffer[i].dumpable == 1) {
			COREFILE_WRITE (mem_regions_buffer[i].len, (m_fildes, 
					     (void *)mem_regions_buffer[i].address, 
					      mem_regions_buffer[i].len));
		}
	}
	
	/* cleanup */	 
	if (mem_regions_buffer) {
		free (mem_regions_buffer);
	}
	tci = m_thread_core_info_list; 
	while (tci != NULL) {
		remove_thread_core_info(tci);
		tci = m_thread_core_info_list;
	}
	
	COREFILE_SYSCALL (close, (m_fildes));
	
	btlib_release_lock (BTL_LOCK_IND_SUSPENSION);
	return;
}

