/* 
 * btlib_critical_section.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "btlib.h"
#include "btlib_sync.h"
#include "btlib_asm.h"

#define SPIN_LOCK_COUNT 2000

typedef struct critical_section_t {
	mutex_opaque mutex;
	struct critical_section_t *next;
} critical_section;

#define CSP(x) (critical_section*)(x)
static critical_section *first_element_in_cs_list = NULL;

void
btlib_init_cs_module ()
{
	assert (sizeof (critical_section) <= ACCESS_LOCK_OBJECT_SIZE);
}

/**************************************************************************/
/* is tid holding one of the cs objects */
/**************************************************************************/
long
is_holding_critical_section (btlib_thread_data_t * tls)
{
	critical_section *cs = first_element_in_cs_list;
	int found = 0;

	btlib_acquire_lock (BTL_LOCK_IND_CS);

	while (cs != NULL) {
		if (get_mutex_owner(&cs->mutex) == tls->__r13) {
			found = 1;
			break;
		}
		cs = cs->next;
	}
	btlib_release_lock (BTL_LOCK_IND_CS);
	DBPRINT (btl_prt_thread | btl_prt_cs,
	       "tls->tid=%d cs=%p\n", tls->tid, found ? cs : 0);
	return found;
}

/**************************************************************************/
/* free a cs object
 * input: address of cs object to be removed
 * output: none
 */
/**************************************************************************/
void
btlib_free_critical_section (cs_handle ch)
{
	critical_section *cs = CSP (ch);
	critical_section *head = first_element_in_cs_list;
	critical_section *tail = NULL;
#if DEBUG
	int found = 0;
#endif

	DBPRINT (btl_prt_cs, "cs=%p\n", cs);

	if (cs == NULL) {
		return;
	}

	/* remove from the first_element_in_cs_list */

	btlib_acquire_lock (BTL_LOCK_IND_CS);

	while (head != NULL) {
		if (head == cs) {
			DBCODE (TRUE, found = 1);
			break;
		}
		tail = head;
		head = head->next;
	}
	assert (found == 1);

	if (tail == NULL) {
		first_element_in_cs_list = cs->next;
	} else {
		tail->next = head->next;
	}

	btlib_release_lock (BTL_LOCK_IND_CS);

	mutex_destroy (&cs->mutex);
	free (cs);
}

/**************************************************************************/
/* create a cs object
 * input: none
 * output: address of a new NON-INITIALIZED cs object
 */
/**************************************************************************/
cs_handle
btlib_create_critical_section ()
{
	critical_section *cs = NULL;

	cs = (critical_section *) malloc (sizeof (critical_section));

	if (cs == NULL) {
		DBPRINT (TRUE, "malloc failed\n");
		btlib_abort ();
	}

	DBPRINT (btl_prt_cs, "cs=%p\n", cs);
	return (cs_handle) cs;
}

/**************************************************************************/
/* init critical section
 * input:   cs - the critical section
 * output:  none
 * remarks: the memory space for the critical section object is
 *          allocated by btgeneric. the size allocated for each 
 *          object is given by the ACCESS_LOCK_OBJECT_SIZE constant
 *          defined in btgeneric.h.
 */
/**************************************************************************/
long
btlib_init_critical_section (cs_handle ch)
{
	critical_section *cs = CSP (ch);
	long retval;

	DBPRINT (btl_prt_cs, "%p\n", cs);

	retval = mutex_init (&cs->mutex);
	if (retval < 0) {
		return retval;
	}
	cs->next = NULL;

	/* insert the cs to the first_element_in_cs_list */

	btlib_acquire_lock (BTL_LOCK_IND_CS);

	if (first_element_in_cs_list != NULL) {
		cs->next = first_element_in_cs_list;
	}
	first_element_in_cs_list = cs;

	btlib_release_lock (BTL_LOCK_IND_CS);
	return retval;
}

/**************************************************************************/
/* enter the given critical section
 * input:   cs - the critical section object
 *          block - whether or not to block the thread if the critial section
 *          is busy
 * output:  0 if the cs is entered, -1 otherwise
 * remarks: 1 -	a thread does not block itself when re-entering a cs
 *          2 - the function is not reentrant. if called from a signal
 *              hanlder than the handler should block all the signal it serves.
 *          3 - btlib MUST NEVER call btgeneric while inside a critical
 *              section. ( when btlib is inside the cs, that is )
 */
/**************************************************************************/
long
btlib_enter_critical_section (cs_handle ch, int block)
{
	critical_section *cs = CSP (ch);
	long ret;

	DBPRINT (btl_prt_cs, "cs=%p block=%d\n", cs, block);

	if (block) {
		ret = mutex_lock (&cs->mutex);
	} else {
		ret = mutex_trylock (&cs->mutex);
		if (ret == -EBUSY) {
			return -1;
		}
	}
	sys_err_chk (ret, "enter_cs failed");

	return 0;
}

/**************************************************************************/
/* leave critical section
 * input:	cs - the critical section
 * output: 	none
 */
/**************************************************************************/
void
btlib_leave_critical_section (cs_handle ch)
{
	critical_section *cs = CSP (ch);
	long ret;

	DBPRINT (btl_prt_cs, "cur_thread_data=%p cs=%p\n", cur_thread_data, cs);

	ret = mutex_unlock (&cs->mutex);
	sys_err_chk (ret, "leave_cs failed");
}

#if DEBUG
/**************************************************************************/
/* returns the tid of the thread that holds the cs */
/**************************************************************************/
unsigned long
who_holds_critical_section (cs_handle ch)
{
	unsigned long owner;
	critical_section *cs = CSP (ch);
	owner = get_mutex_owner (&cs->mutex);
	DBPRINT (btl_prt_cs, "owner=%lx\n", owner);
	return owner;
}
#endif				//DEBUG
