/* 
 * btlib_memory.c
 *
 * 
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include <ia32toia64.h>

#include "btlib.h"
#include "btlib_sync.h"

/* highest address mapped for btgeneric so far in the three memory
 * sections
 */
static unsigned long m_max_code = 0;
static unsigned long m_max_data = 0;
static unsigned long m_max_default = 0;
static unsigned long m_max_btlib = 0;

/* function prototypes */
static unsigned long get_unmapped_area (unsigned long start, unsigned long len);
static long set_btrans_permissions (unsigned long prot);
static long mmap32 (unsigned int addr, unsigned int len, unsigned int prot, unsigned int flags, unsigned int fd, unsigned int pgoff);
static unsigned set_linux_permissions (unsigned prot);
static long munmap32 (unsigned addr, unsigned len);

/*****************************************************************************/
/* initialize the memory module. must be called before any
 * exported function is called
 */
/*****************************************************************************/
void
btlib_init_memory_module ()
{
	DBPRINT (btl_prt_mem, "\n");

}

/*****************************************************************************/
/* convert from linux permissions to BTGeneric permissions */
/*****************************************************************************/
static long
set_btrans_permissions (unsigned long prot)
{
	assert ((MEM_READ == PROT_READ) &&
		(MEM_WRITE == PROT_WRITE) && (MEM_EXECUTE == PROT_EXEC));
	return prot;
}

/*****************************************************************************/
/* convert from BTGeneric permissions to linux permissions */
/*****************************************************************************/
static unsigned
set_linux_permissions (unsigned prot)
{
	assert ((MEM_READ == PROT_READ) &&
		(MEM_WRITE == PROT_WRITE) && (MEM_EXECUTE == PROT_EXEC));
	return prot;
}

/*****************************************************************************/
/* get the permissions of the memory at a given address 
 * returns the permssions of the page where the given address falls, or 
 * a negative error code on failure.
 */
/*****************************************************************************/
long
btlib_memory_query_permissions (void *address, void **region_start,
                                unsigned long *region_size)
{
	unsigned perm = 0;
	long ret;

	/* first try to query the wrapper lib */
	ret = iti_get_memory_access_permissions (address, &perm, region_start,
						 region_size);
	if (ret == 0) {

		DBPRINT (btl_prt_mem, "address not mapped\n");

		/* the address was not found in the wrapper lib */

		/* linux maps the first page as ro */
		if (PAGESTART ((unsigned long) address, btl_page_size) != 0ul) {
			perm = PROT_NONE;
		}

		if (region_start != NULL) {
			*region_start = (void *) PAGESTART ((unsigned long) address, 
							    btl_page_size);
		}
		if (region_size != NULL) {
			*region_size = btl_page_size;
		}
	}

	DBPRINT (0, "perm=%x\n", perm);

	return perm;
}

/*****************************************************************************/
/* do the memory mapping for IA-32 addresses */
/*****************************************************************************/
static long
mmap32 (unsigned int addr, unsigned int len, unsigned int prot, 
		unsigned int flags, unsigned int fd, unsigned int pgoff)
{
	unsigned long retval;

	/* be careful with the following debug-print -
	 * we cannot be sure that mmap_args is a valid pointer 
	 */
	DBPRINT (1, "addr=%x len=%x prot=%x flags=%x fd=%d pgoff=%x\n",
		 addr, len, prot, flags, fd, pgoff);

	/* do the mapping */
	retval = ILP32_WRAPPER6 (IA32_SYSCALL_MMAP2, addr, len, prot,
			flags, fd, pgoff);

	DBPRINT (btl_prt_mem, "map32 retval=%lx\n", retval);

	if (!IS_SYS_ERROR (retval)) {


		/* make sure we don't return a high address to the application */
		if (retval >= 0x100000000) {
			DBPRINT (btl_prt_mem, "mmap32:un-mapping\n");
			retval = SYSCALL (munmap, (retval, (unsigned long)len));
			sys_err_chk (retval, "munmap");
			return (unsigned long)-ENOMEM;
		} else {
			/* if the mapped region is executable then notify btgeneric */
			if ((prot & PROT_EXEC) == PROT_EXEC) {
				DBPRINT (btl_prt_mem, "report_load %lx %u\n", retval, len);
				BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (
					(void *) retval, len, BT_FLUSH_ALLOC);
			}
		}

#if defined(SHMEM_PROC_AFFINITY) || defined(SHMEM_ORDER_SOLUTION)
		if (flags & MAP_SHARED) {
#if SHMEM_PROC_AFFINITY
			apply_shared_memory_affinity_solution ();
#elif SHMEM_ORDER_SOLUTION
			apply_shared_memory_ordering_solution ();
#endif
		}
#endif
	}


	return retval;
}

/*****************************************************************************/
/* do the memory un-mapping for IA-32 addresses */
/*****************************************************************************/
static long
munmap32 (unsigned addr, unsigned len)
{
	long retval;

	DBPRINT (btl_prt_mem, "addr=%x len=%d\n", addr, len);
	/* notify btgeneric */
	BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (U2P (addr), len, BT_FLUSH_FREE);

	/* do the unmapping */
	retval = ILP32_WRAPPER2 (IA32_SYSCALL_MUNMAP, (unsigned long) addr,
				 (unsigned long) len);
	DBPRINT (btl_prt_mem, "retval=%ld\n", retval);
	return retval;
}

/*****************************************************************************/
/* do the memory re-mapping for IA-32 addresses */
/*****************************************************************************/
long
btlib_ia32_mremap (const BTGENERIC_IA32_CONTEXT *bc32)
{
	void *retval;
	unsigned addr, old_len, new_len;
	unsigned flags, new_addr;
	int perm;

	DBPRINT (btl_prt_mem, "\n");

	addr = IA32_SYSCALL_PARAM1 (bc32);
	old_len = IA32_SYSCALL_PARAM2 (bc32);
	new_len = IA32_SYSCALL_PARAM3 (bc32);
	flags = IA32_SYSCALL_PARAM4 (bc32);
	new_addr = IA32_SYSCALL_PARAM5 (bc32);

	perm = btlib_memory_query_permissions (U2P (addr), NULL, NULL);
	if (perm < 0) {
		retval = (void *) (long) perm;
		goto out;
	}
	if (perm & MEM_EXECUTE) {
		BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (U2P (addr), old_len,
							BT_FLUSH_FREE);
	}

	DBPRINT (btl_prt_mem, "addr=%x old_len=%d perm=%x\n", addr, old_len, perm);
	DBPRINT (btl_prt_mem, "new_len=%d flags=%x new_addr=%x\n", new_len, flags, 
		 new_addr);

	/* do the remapping */
	retval = (void *) ILP32_WRAPPER5 (IA32_SYSCALL_MREMAP, 
					  (unsigned long) addr,
					  (unsigned long) old_len,
					  (unsigned long) new_len,
					  (unsigned long) flags,
					  (unsigned long) new_addr);

	DBPRINT (btl_prt_mem, "retval=%p\n", retval);

	if (perm & PROT_EXEC) {
		if (IS_SYS_ERROR ((long) retval)) {
			/* undo the report_unload */
			BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (U2P (addr),
								old_len,
								BT_FLUSH_ALLOC);
		} else {
			assert ((uintptr_t) retval < 0x100000000ul);
			BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE ((void *) retval,
								new_len,
								BT_FLUSH_ALLOC);
		}
	}
      out:
	return (long) retval;
}

/*****************************************************************************/
/* implement the IA-32 mmap2 syscall */
/*****************************************************************************/
long
btlib_ia32_mmap2 (const BTGENERIC_IA32_CONTEXT * bc32)
{
	long retval = 0;

	DBPRINT (btl_prt_mem, "\n");

	/* put the arguments in a struct fitting to mmap */
	retval = mmap32 (IA32_SYSCALL_PARAM1 (bc32),
			IA32_SYSCALL_PARAM2 (bc32),
			IA32_SYSCALL_PARAM3 (bc32),
			IA32_SYSCALL_PARAM4 (bc32),
			IA32_SYSCALL_PARAM5 (bc32),
			IA32_SYSCALL_PARAM6 (bc32));

	/* do the mapping */
	DBPRINT (0, "mmap32=%lx\n", retval);
	return retval;
}

/*****************************************************************************/
/* implement the IA-32 mmap syscall */
/*****************************************************************************/
long
btlib_ia32_mmap (const BTGENERIC_IA32_CONTEXT * bc32)
{
	legacy_mmap_args_t args;
	long retval = 0;

	retval = safe_memcpy (&args, U2P(IA32_SYSCALL_PARAM1(bc32)), sizeof (args));
	if (retval == 0/*success*/) {
		unsigned pgoff;
		if (IA32_PAGE_OFF(args.offset)) {
			return -EINVAL;
		}
		pgoff = args.offset >> IA32_PAGE_SHIFT;
	
		/* do the mapping */
		retval = mmap32 (args.addr, args.len, args.prot, args.flags,
				args.fd, pgoff);
	}
	return retval;
}

/*****************************************************************************/
/* implement the IA-32 munmap syscall */
/*****************************************************************************/
long
btlib_ia32_munmap (const BTGENERIC_IA32_CONTEXT *bc32)
{
	long retval;

	DBPRINT (btl_prt_mem, "\n");
	retval = munmap32 (IA32_SYSCALL_PARAM1 (bc32), IA32_SYSCALL_PARAM2 (bc32));
	DBPRINT (btl_prt_mem, "retval=%ld\n", retval);
	return retval;
}

/*****************************************************************************/
/* implement the IA-32 mprotect syscall */
/*****************************************************************************/
long
btlib_ia32_mprotect (BTGENERIC_IA32_CONTEXT *bc32)
{
	long perm = 0;
	int no_of_pages;
	long retval;
	unsigned long size, addr, prot;

	addr = (unsigned long) IA32_SYSCALL_PARAM1 (bc32);
	size = (unsigned long) IA32_SYSCALL_PARAM2 (bc32);
	prot = (unsigned long) IA32_SYSCALL_PARAM3 (bc32);
	DBPRINT (btl_prt_mem, "addr=%lx size=%lu prot=%lx\n", addr, size, prot);

	perm = set_btrans_permissions (prot);
	no_of_pages = (size - 1) / btl_page_size + 1;

	/* notify ia32exec.bin of the change */
	BTGENERIC_NOTIFY_CHANGE_PERMISSION_REQUEST (addr, no_of_pages, perm);

	/* do the syscall */
	retval = ILP32_WRAPPER3 (IA32_SYSCALL_MPROTECT, addr, size, prot);
	return retval;
}

/*****************************************************************************/
/* allocate memory for IA-32 Execution Layer 
 * all anonymous mappings above 4G must go through this function - no direct 
 * calls to mmap are allowed.
 */
/*****************************************************************************/
void *
btlib_mem_alloc (unsigned long start, size_t length, unsigned prot)
{
	long retval;
	int linux_prot;
	void *start_addr;
	int flags = MAP_PRIVATE | MAP_ANONYMOUS;
	static long kernel_alloc_below_hint = 0;
	int locked = 0;

	linux_prot = set_linux_permissions (prot);

	while (1) {
		/* 
		 * most linux kernels map memory above the requested address. most but
		 * not all. 
		 * we must gurantee this behavior since btgeneric depends on it. so -
		 * once we find that the current kernel does allocate memory 
		 * below the requested start-address we need to find an  unmpapped
		 * area ourselves.
		 */
		if (kernel_alloc_below_hint && (start != 0)) {
			btlib_acquire_lock (BTL_LOCK_IND_INTERNAL_MEM);
			locked = 1;
			start_addr = (void *)get_unmapped_area (start, length);
			flags |= MAP_FIXED;
		} else {
			start_addr = (void *)start;
		}

		retval = (long) SYSCALL (mmap, ((unsigned long)start_addr, length, 
					 	linux_prot, flags, -1, 0ul));

		DBPRINT (btl_prt_mem, "length=%lu retval=%lx\n", length, retval);

		if (IS_SYS_ERROR (retval)) {
			goto out;
		}
		assert ((retval >> 32) != 0);

		if ((unsigned long) retval < start) {
			assert (kernel_alloc_below_hint == 0);
			DBPRINT (btl_prt_mem, "kernel allocates below hint\n");
			kernel_alloc_below_hint = 1;
			retval = SYSCALL (munmap, ((unsigned long)retval, length));
			if (IS_SYS_ERROR (retval)) {
				goto out;
			}
		} else {
			break;
		}
	}

	/* store the max mapped memory in the section */
	switch (start & BTGENERIC_MEMORY_MASK) {
	case INITIAL_CODE_ADDRESS:
		m_max_code = retval + length;
		break;
	case INITIAL_DATA_ADDRESS:
		m_max_data = retval + length;
		break;
	case INITIAL_DEFAULT_ADDRESS:
		m_max_default = retval + length;
		break;
	case INITIAL_BTLIB_ADDRESS:
		m_max_btlib = retval + length;
		break;
	default:
		assert (!"unknown memory section");
	}

      out:
	if (locked) {
		btlib_release_lock (BTL_LOCK_IND_INTERNAL_MEM);
	}

	return (void *) retval;
}

/*****************************************************************************/
/* free a mapped memory area */
/*****************************************************************************/
long
btlib_mem_free (void *start, size_t length)
{
	long retval;
	DBPRINT (btl_prt_mem, "start=%p length=%lu\n", start, length);
	retval = SYSCALL (munmap, ((unsigned long)start, length));
	return retval;
}

/*****************************************************************************/
/* change permissions for ia32exec.bin needs */
/*****************************************************************************/
long
btlib_memory_change_permissions (unsigned long start, unsigned long size,
                                 unsigned prot)
{
	unsigned linux_prot = 0;
	long retval;

	linux_prot = set_linux_permissions (prot);
	retval = SYSCALL (mprotect, (start, (size_t) size, linux_prot));
	/* TODO: need to return the currrent permissions on failure ? */
	return retval;
}

/*****************************************************************************/
/* look for an unmapped area in the process mapping memory space
 * input:   start - the address from which to start the search
 *          len - size of requested area
 * output:  -ENOMEM if no free memory is found, address of area on success
 * remarks: as for now, a very simplistic algorithm will do - keep the highest
 *          address requested by btgeneric in the three different sections.
 *          next request will get the memory at the end of the mapped area
 *          in the section.
 */
/*****************************************************************************/
static unsigned long
get_unmapped_area (unsigned long start, unsigned long len UNUSED)
{
	unsigned long addr = 0;
	switch (start & BTGENERIC_MEMORY_MASK) {
	case INITIAL_CODE_ADDRESS:
		addr = PAGEALIGN (m_max_code, btl_page_size);
		break;
	case INITIAL_DATA_ADDRESS:
		addr = PAGEALIGN (m_max_data, btl_page_size);
		break;
	case INITIAL_DEFAULT_ADDRESS:
		addr = PAGEALIGN (m_max_default, btl_page_size);
		break;
	case INITIAL_BTLIB_ADDRESS:
		addr = PAGEALIGN (m_max_btlib, btl_page_size);
		break;
	default:
		assert (!"unknown memory section");
	}

	return addr;
}

