/*  
 * btlib_misc.c
 *  
 * this file contains the code for emulation of the IA-32 ptrace syscall. 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include "btlib.h"

/*****************************************************************************/
/* kill the whold app */
/*****************************************************************************/
void
btlib_abort_all ()
{
	btlib_thread_data_t *p;

	DBPRINT (TRUE, "\n");

	/* print rolling trace buffers */
	DBCODE (btl_btg_initialized, BTGENERIC_EXCEPTION_DEBUG_PRINT ());
	DBCODE (TRUE, btlib_print_rolling_trace_array ());

	p = cur_thread_data->next;
	while (p != NULL) {
		KILL_THREAD_SIGNAL (p->tid, SIGKILL);
		p = p->next;
	}
	p = cur_thread_data->prev;
	while (p != NULL) {
		KILL_THREAD_SIGNAL (p->tid, SIGKILL);
		p = p->prev;
	}
	KILL_THREAD_SIGNAL (cur_thread_data->tid, SIGKILL);
}

/*****************************************************************************/
/* kill the current thread */
/*****************************************************************************/
void
btlib_abort ()
{
	/* print rolling trace buffers */
	DBCODE (TRUE, BTGENERIC_EXCEPTION_DEBUG_PRINT ());
	DBCODE (TRUE, btlib_print_rolling_trace_array ());
	/* kill all the processes in the process-group */
	abort ();
}

#if ALL_THREAD_AFFINITY
#define CPUINFO_FILE "/proc/cpuinfo"
#define CPUINFO_STR_LENGTH 512ul
#define DEFAULT_MHZ 1400
unsigned long
btlib_get_cpu_mhz ()
{
	int cpuinfo_file;
	char cpuinfo_str[CPUINFO_STR_LENGTH];
	int  retval;
	unsigned long mhz;

	/* read from /proc/cpuinfo file, to get mhz of processor */
	cpuinfo_file = SYSCALL (open, (CPUINFO_FILE, O_RDONLY, 0));
	if (IS_SYS_ERROR (cpuinfo_file)) {
		DBPRINT (btl_prt_thread, "Error: open of %s failed: %d\n", CPUINFO_FILE, cpuinfo_file);
		return DEFAULT_MHZ;
	} else {
		mhz = 0;
		while (0 == mhz) {
			retval = SYSCALL (read, (cpuinfo_file, cpuinfo_str, CPUINFO_STR_LENGTH));
			if (IS_SYS_ERROR (retval)) {
				DBPRINT (btl_prt_thread,
					"Error: failed to read from %s: %d\n", 
					CPUINFO_FILE, retval);
				mhz = DEFAULT_MHZ;
			} else if (-1==retval) {
				mhz = DEFAULT_MHZ;
			} else {
				char * p = strstr (cpuinfo_str, "cpu MHz"), *start_mhz_str;
				if (p) {
					p = strchr ((p+(strlen ("cpu MHz"))), ':');
					if (p) {
						p = strchr (p+1, ' ');
						if (p) {
							start_mhz_str = p+1;
							p = strchr (p+1, '.');
							if (p) {
								*p = '\0';
								mhz = atol (start_mhz_str);
								DBPRINT (btl_prt_thread, "mhz - %u\n",  (unsigned int)mhz);
							}
						}
					}
				}
			}
		}
		SYSCALL (close, (cpuinfo_file));
	}

	return mhz;
}
#endif

#if DEBUG

#if USE_ROLLING_TRACE
#define ROLLING_TRACE_STR_NUM 1000
#define ROLLING_TRACE_STR_LEN 2048
static char rolling_trace_array[ROLLING_TRACE_STR_NUM][ROLLING_TRACE_STR_LEN];
static long rolling_trace_wrap = 0;
static long rolling_trace_ptr = 0;
#endif	// USE_ROLLING_TRACE

/*****************************************************************************/
/* add an entry to the rolling trace array */
/*****************************************************************************/
void
btlib_insert_rolling_trace_str (char *UNUSED str)
{
	/* workaround to gcc warning */
	char *UNUSED dummy = str;

#if USE_ROLLING_TRACE
	assert (strlen (str) < ROLLING_TRACE_STR_LEN);
	strncpy (rolling_trace_array[rolling_trace_ptr], str, ROLLING_TRACE_STR_LEN - 1);

	rolling_trace_ptr = (rolling_trace_ptr + 1) % ROLLING_TRACE_STR_NUM;
	if (rolling_trace_ptr == 0) {
		rolling_trace_wrap = 1;
	}
#endif	// USE_ROLLING_TRACE
}

/*****************************************************************************/
/* print the contents of the rolling trace array */
/*****************************************************************************/
void
btlib_print_rolling_trace_array ()
{
#if USE_ROLLING_TRACE
	int j;

	if ((rolling_trace_ptr == 0) && (rolling_trace_wrap == 0)) {
		/* no records to print */
		return;
	}

	j = rolling_trace_wrap ? rolling_trace_ptr : 0;
	do {
		char buf[ROLLING_TRACE_STR_LEN + 64];
		fprintf (g_log_file, "%s", rolling_trace_array[j]);
		assert (!IS_SYS_ERROR (retval));
		j = (j + 1) % ROLLING_TRACE_STR_NUM;
	} while (j != rolling_trace_ptr);
#endif				// USE_ROLLING_TRACE
}

/* NOTE: it's just a very simple unwinding and can only handle cases  
   where ebp is frame pointer. It can be only called when IA-32 state
   is consistent */
void 
unwind_ia32stack (void) 
{
	unsigned ebp;
	unsigned ret_addr;

	ebp = cur_thread_data->bc32->ebp;

	DBPRINT (TRUE, "\nStart unwinding IA-32 stack!\n");
	while (ebp != 0) {
		ret_addr = *(unsigned *)(unsigned long)(ebp + 4);
		DBPRINT (TRUE, "ret_addr = %x\n", ret_addr);
		ebp = *(unsigned *)(unsigned long)(ebp);
	}
}

#define MAP_LEN 1024 * 80 
void print_maps (void)
{
	long ret;
	int  fd;
	static char maps[MAP_LEN];

	fd  = SYSCALL (open, ("/proc/self/maps", 0, 0));
	if (!IS_SYS_ERROR (fd)) {
		ret = SYSCALL (read, (fd, maps, sizeof(maps)));

		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "read failed: %ld\n", ret);
		} else {
			maps[(ret < MAP_LEN? ret : MAP_LEN) - 1] = 0;
			DBPRINT (TRUE, "\n%s\n", maps);
		}
		SYSCALL (close, (fd));
	} else {
		DBPRINT (TRUE, "open failed: %d\n", fd);
	}
}

/*****************************************************************************/
/*****************************************************************************/
void
print_thread_data (btlib_thread_data_t * blst)
{

	fprintf (g_log_file, "(%d) *** print_thread_data: blst=%p\n",
		 SYSCALL (gettid, ()), blst);
	fprintf (g_log_file, "******** next=%p\n", blst->next);
	fprintf (g_log_file, "******** stackblock=%p\n", blst->stackblock);
	fprintf (g_log_file, "******** sighandlers=%p\n", blst->sighandlers);
	fprintf (g_log_file, "******** pending_que=%p\n", blst->pending_que);
	fprintf (g_log_file, "******** ptrace=%lx\n", blst->ptrace);
	fprintf (g_log_file, "******** debugee_list=%p\n",
		 blst->debugee_list);
	fprintf (g_log_file, "******** tia_signal.si_signo=%d\n",
		 blst->tia_signal.siginfo.si_signo);
	fprintf (g_log_file, "******** tia_signal.si_errno=%d\n",
		 blst->tia_signal.siginfo.si_errno);
	fprintf (g_log_file, "******** tia_signal.si_code=%d\n",
		 blst->tia_signal.siginfo.si_code);
	fprintf (g_log_file, "******** alt_stack.ss_sp=%u\n",
		 blst->alt_stack.ss_sp);
	fprintf (g_log_file, "******** alt_stack.ss_size=%u\n",
		 blst->alt_stack.ss_size);
	fprintf (g_log_file, "******** alt_stack.ss_flags=%u\n",
		 blst->alt_stack.ss_flags);
	fprintf (g_log_file, "******** syscall_id=%d\n", blst->syscall_id);
	fprintf (g_log_file, "******** tid=%d\n", blst->tid);
	fprintf (g_log_file, "******** is_exited=%d\n", blst->is_exited);
	fprintf (g_log_file, "******** self_canon_semkey=%d\n",
		 blst->self_canon_semkey);
	fprintf (g_log_file, "******** locks=%d\n", blst->locks);
	fprintf (g_log_file, "******** longjmp_status=%d\n",
		 blst->longjmp_status);
	/*
	   for (j = 0; j < (sizeof(struct jmp_buffer)/sizeof(uintptr_t)); j ++) {
	   unsigned long * p = ((unsigned long*)&cur_thread_data->btgeneric_run_context) + j;
	   fprintf (g_log_file,"******** jmp_buffer[%d] %p=%lx\n", j, p,*p);
	   }
	 */
}
#endif	// DEBUG

