/*  
 * btlib_pending_que.c
 *  
 * 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include "btlib.h"
#include "btlib_sig_utils.h"
#include "btlib_sync.h"
#include "btlib_stats.h"
#include "btlib_pending_que.h"

/* shared signal que */
signal_pending_que *btl_shared_que = NULL;

static void remove_que_element (signal_pending_que *que, unsigned indx, unsigned prev);
static long enlarge_que (signal_pending_que *pque);
#if DEBUG
static void print_que (signal_pending_que *que);
#endif

/****************************************************************************** 
 * allocate memory for the pending que of the given thread
 *****************************************************************************/
void
btlib_que_init (signal_pending_que **ppque)
{
	unsigned long j;
	unsigned long que_size = btl_page_size / sizeof (sig_pending);

	/* create the que */

	*ppque = (signal_pending_que *) malloc (sizeof (signal_pending_que));
	if (*ppque == NULL) {
		assert (!"cannot allocate memory for pending que");
		btlib_abort ();
	}
	memset (*ppque, 0, sizeof (signal_pending_que));

	(*ppque)->que = (sig_pending *)btlib_mem_alloc (INITIAL_BTLIB_ADDRESS,
						        btl_page_size,
						        PROT_READ | PROT_WRITE);

	sys_err_chk ((long) (*ppque)->que, "mmap failed");

	/* initialize the list of pending signals */
	for (j = 0; j < que_size - 1; j++) {
		(*ppque)->que[j].next = j + 1;
	}
	(*ppque)->que[j].next = INVALID_ENTRY;
	(*ppque)->first_pending = (*ppque)->last_pending = INVALID_ENTRY;
	(*ppque)->size = btl_page_size;
}



/****************************************************************************** 
   return signal number of last pending signal
 *****************************************************************************/
int 
btlib_que_lastsig (struct signal_pending_que_t *pque)
{
	int signo;

	signo = pque->que[pque->last_pending].info.si_signo;
	return signo;
}

long
btlib_que_is_real_empty (signal_pending_que *pque)
{
	sigset32_t ss = pque->pending_set;
	return isemptyset32 (&ss, WORDS_IN_MASK) ? 1 : 0;
}

/******************************************************************************/
/* return 0 if the que is surely not empty 
 * return 1 if the que is possibly emtpy */
/******************************************************************************/
long
btlib_que_is_empty (signal_pending_que *pque)
{
	sigset32_t ss = pque->pending_set;

	/* not include all signals which is actually  blocked */
	sig_andnot_mask (&ss, &cur_thread_data->sigmask, WORDS_IN_MASK);

	return isemptyset32 (&ss, WORDS_IN_MASK) ? 1 : 0;

}

/******************************************************************************/
/* copy the pending mask to the given argument */
/******************************************************************************/
void
btlib_que_sigpending (signal_pending_que *pque, int size, sigset32_t *set)
{
	assert (pque != NULL);

	memset (set, 0, (unsigned long)size * sizeof(unsigned int));
	memcpy (set, &pque->pending_set, size * sizeof(int));
}


/****************************************************************************** 
 * enque a signal
 * input  : que - pointer to the que
 * 	    signo - the signal id
 *	    siginfo - pointer to the signal info
 * returns: void
 * remarks: the function inserts the signal to the pending list ordered by 
 *          the time stamp, and updates the pending_sigs field. RT signals 
 * 	    are inserted even if there is already same signal pending. 
 * 	    non-RT signals are inserted only once for each type.
 *          the function is NOT reentrant
 * note: this function MUST be called when all signals are blocked
 *****************************************************************************/
void
btlib_que_enque (signal_pending_que *pque, siginfo_t *siginfo, unsigned long time_stamp)
{
	unsigned indx;
	long retval;
	int signo;
	unsigned next, prev;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (pque != NULL);
	assert (siginfo != NULL);

	signo = siginfo->si_signo;

	DBPRINT (0, "signo=%d time=%lu\n", signo, time_stamp);
	DBCODE (0, print_que (pque));

	if ((signo >= SIG_RT_MIN) || (!pque->pending_sigs[signo])) {
		/* enlarge the que if needed */
		if (pque->first_free == INVALID_ENTRY) {
			assert (pque->total_pending ==
				pque->size / sizeof (sig_pending));

			retval = enlarge_que (pque);
			if (retval < 0) {
				assert (!"cannot enlarge que");
				btlib_abort ();
			}
		}
		assert (pque->first_free != INVALID_ENTRY);

		/* initialize a new que element */
		indx = pque->first_free;
		pque->first_free = pque->que[indx].next;

		memcpy (&pque->que[indx].info, siginfo, sizeof (siginfo_t));
		pque->que[indx].time_stamp = time_stamp;
		pque->que[indx].next = INVALID_ENTRY;

		/* insert the new element to the que sorted by the time stamp */

		/* find the point of insertion */
		next = pque->first_pending;
		prev = INVALID_ENTRY;

		while (next != INVALID_ENTRY) {
			if (pque->que[next].time_stamp < time_stamp) {
				/* sig older than the new one. push further */
				prev = next;
				next = pque->que[next].next;
			} else {
				/* sig newer than the new one. insert here */
				assert (pque->que[next].time_stamp > time_stamp); 
				break;
			}
		}
		DBPRINT (0, "next=%x prev=%x\n", next, prev);

		/* insert */
		pque->que[indx].next = next;
		if (prev != INVALID_ENTRY) {
			/* not the first one */
			assert (next != pque->first_pending);
			assert ((prev != pque->last_pending) ||
				(pque->que[prev].next == INVALID_ENTRY));
			pque->que[prev].next = indx;

		} else {
			assert (next == pque->first_pending);
			pque->first_pending = indx;
		}
		if (next == INVALID_ENTRY) {
			/* last in que */
			assert (prev == pque->last_pending);
			pque->last_pending = indx;
		}

		/* increment the counter */

		/* if this assertion fails then the type of pending_sigs 
		 * needs be larger than 'char' */
		assert (pque->pending_sigs[signo] < 255);

		pque->pending_sigs[signo]++;
		pque->total_pending++;
		sigaddset32 (&pque->pending_set, signo);

		STATCODE (btl_stats, 
			  btlib_update_stat (BTL_STAT_MAX_PENDING, MAX_VAL,
					     (unsigned long)pque->total_pending));

	}

	DBPRINT (0, "finished\n");
}

/****************************************************************************
 * free the thread's pending que
 ****************************************************************************/
void
btlib_que_free (signal_pending_que **ppque)
{
	long retval;
	retval = SYSCALL (munmap, ((unsigned long)(*ppque)->que, (*ppque)->size));
	DBPRINT (IS_SYS_ERROR (retval), "munmap failed\n");
	free (*ppque);
	*ppque = NULL;
}

/******************************************************************************
 * remove the given signal from the que
 * input  : que - pointer to the que
 *          signo - the signal to be removed
 * output : void
 * remarks:	
 *****************************************************************************/
void
btlib_que_remove_signo (signal_pending_que *pque, int signo)
{
	unsigned prev, indx;

	DBPRINT (0, "signo=%d\n", signo);
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (pque != NULL);

	BLOCK_ALL_SIGNALS ();

	prev = INVALID_ENTRY;
	indx = pque->first_pending;
	while ((pque->pending_sigs[signo] > 0) && (indx != INVALID_ENTRY)) {
		if (pque->que[indx].info.si_signo == signo) {
			remove_que_element (pque, indx, prev);
		}
		prev = indx;
		indx = pque->que[indx].next;
	}

	RESTORE_SIGNALS ();

}

/******************************************************************************
 * remove a single element from the que
 * input  : que - pointer to the que
 *          indx - the index of the element to be removed
 *          prev - the index of the element that prceeds the element
 *                 to be removed
 * output : void
 * remarks:	
 *****************************************************************************/
static void
remove_que_element (signal_pending_que *pque, unsigned indx, unsigned prev)
{
	int signo;

	DBPRINT (0, "remove_que_element indx=%d prev=%d next=%d\n",
	         indx, prev, pque->que[indx].next);
	signo = pque->que[indx].info.si_signo;
	/* remove the element from the pending list */
	if (indx == pque->last_pending) {
		pque->last_pending = prev;
	}
	if (indx == pque->first_pending) {
		assert (prev == INVALID_ENTRY);
		pque->first_pending = pque->que[indx].next;
	} else {
		assert (prev != INVALID_ENTRY);
		pque->que[prev].next = pque->que[indx].next;
	}
	assert (pque->pending_sigs[signo] > 0);
	pque->pending_sigs[signo]--;
	if (pque->pending_sigs[signo] == 0) {
		sigdelset32 (&pque->pending_set, signo);
	}
	assert (pque->total_pending > 0);
	pque->total_pending--;

	/* add it to the free list */
	pque->que[indx].next = pque->first_free;
	pque->first_free = indx;
}

/******************************************************************************
 * resize the given pending que 
 * input  : pque - pointer to the que
 * output : 0 on success, negative error code on failure
 * remarks: the que is always enlarged by a factor of 2
 *****************************************************************************/
static long
enlarge_que (signal_pending_que *pque)
{
	unsigned long new_size;
	sig_pending *que = pque->que;
	sig_pending *new_que = NULL;
	unsigned j;
	long ret = 0;
	unsigned long elnum, new_elnum;

	assert (pque->first_free == INVALID_ENTRY);
	DBPRINT (btl_prt_mem | btl_prt_sig, "pque->size=%lx\n", pque->size);

	/* reallocate */
	new_size = pque->size + btl_page_size;

	/* to be fixed, what about mremap return an address below 4G? */
	new_que = (sig_pending *) 
		SYSCALL (mremap, ((unsigned long)que, pque->size, new_size, 
			           MREMAP_MAYMOVE, 0ul));
	if (IS_SYS_ERROR (new_que)) {
		ret = -ENOMEM;
		goto out;
	}
	assert (((unsigned long)new_que >> 32) != 0);

	/* que is not valid from this point on */
	elnum = pque->size / sizeof (sig_pending);
	new_elnum = new_size / sizeof (sig_pending);

	/* recalc the free list */
	for (j = elnum; j < new_elnum - 1; j++) {
		new_que[j].next = j + 1;
	}
	assert (j == (new_elnum - 1));
	new_que[j].next = INVALID_ENTRY;
	pque->first_free = elnum;

	/* update the pending que */
	pque->que = new_que;
	pque->size = new_size;

      out:
	return ret;
}

/******************************************************************************
 * returns the first pending signal that is not blocked by the current thread
 * input  : si - will hold the info of the returned signals
 * output : the signo of the returned signal, or 0 of none was found
 * remarks: as a side effect, this function also moves blocked, non-private 
 *          signals that are found in the private que to the shared que
******************************************************************************/
int
btlib_que_deque (siginfo_t *si)
{
	int found = 0;

	if (!btlib_que_is_empty (cur_thread_data->pending_que) ||
	    !btlib_que_is_empty (btl_shared_que)) {
		/* if private/shared queue has pending signal */
		unsigned private = 0, shared = 0;

		signal_pending_que *pque;
		unsigned next, shr_indx, shr_prev, prv_indx, prv_prev;

		BLOCK_ALL_SIGNALS ();

		btlib_acquire_lock (BTL_LOCK_IND_SHARED_QUE);

		/* look in the shared que for the first non-blocked signal */
		pque = btl_shared_que;
		shr_prev = INVALID_ENTRY;
		next = INVALID_ENTRY;
		shr_indx = pque->first_pending;
		while (shr_indx != INVALID_ENTRY) {
			int signo = pque->que[shr_indx].info.si_signo;
			next = pque->que[shr_indx].next;

			/* private blocked signals stay in queue */
			if (sigismember32 (&cur_thread_data->sigmask, signo)) {
				shr_prev = shr_indx;
				shr_indx = next;
			} else {
				/* found */
			shared = 1;
				break;
			}
		}

		/* look in the private que for the first non-blocked signal */
		pque = cur_thread_data->pending_que;
		prv_prev = INVALID_ENTRY;
		next = INVALID_ENTRY;
		prv_indx = pque->first_pending;
		while (prv_indx != INVALID_ENTRY) {
			int signo = pque->que[prv_indx].info.si_signo;
			next = pque->que[prv_indx].next;

			/* private blocked signals stay in queue */
			if (sigismember32 (&cur_thread_data->sigmask, signo)) {
				if (!IS_PRIVATE_SIG (&pque->que[prv_indx].info)) {
					/* move blocked non-private signals from 
					 * private to shared que */
					btlib_que_enque (btl_shared_que,
							 &pque->que[prv_indx].
							 info,
							 pque->que[prv_indx].
							 time_stamp);
					remove_que_element (pque, prv_indx,
							    prv_prev);
				} else {
					prv_prev = prv_indx;
				}
				prv_indx = next;
			} else {
				/* found */
			private = 1;
				break;
			}
		}


		/* if signals were found in both que's then choose the older one */
		if ((shared == 1) && (private == 1)) {
			assert (btl_shared_que->que[shr_indx].time_stamp !=
				cur_thread_data->pending_que->que[prv_indx].time_stamp);
			if (btl_shared_que->que[shr_indx].time_stamp <
			    cur_thread_data->pending_que->que[prv_indx].time_stamp) {
				private = 0;
			} else {
				shared = 0;
			}
		}

		/* deque the signal and return its info */
		if ((shared == 0) && (private == 0)) {
			/* no signal found */
			found = 0;
		} else if (private == 1) {
			/* return the signal from the private que */
			memcpy (si, &cur_thread_data->pending_que->que[prv_indx].info, 
				sizeof(*si));
			found = si->si_signo;
			remove_que_element (cur_thread_data->pending_que,
					    prv_indx, prv_prev);
		} else if (shared == 1) {
			/* return the signal from the shared que */
			memcpy (si, &btl_shared_que->que[shr_indx].info, sizeof(*si));
			found = si->si_signo;
			remove_que_element (btl_shared_que, shr_indx, shr_prev);
		} else {
			assert (!"both private and shared are non zero");
		}

		btlib_release_lock (BTL_LOCK_IND_SHARED_QUE);

		RESTORE_SIGNALS ();
	}

	return found;
}

#if DEBUG
static void
print_que (signal_pending_que *pque)
{
	unsigned j, i;
	char str[4096];
	size_t len = 4096;
	char *p = str;

	*p = 0;
	i = 0;
	j = pque->first_pending;
	while (j != INVALID_ENTRY) {
		char buf[256];
		sprintf (buf, "%d=[entry=%d time=%lu signo=%d next=%x] ",
			 i, j, pque->que[j].time_stamp,
			 pque->que[j].info.si_signo, pque->que[j].next);
		strncpy (p, buf, len - 1);
		p += strlen (buf);
		len -= strlen (buf);
		j = pque->que[j].next;
		i++;
	}
	fprintf (g_log_file, "(%d) pending_que:%s\n", cur_thread_data->tid, str);
}
#endif

