/*  
 * btlib_ptemu.c
 *  
 * this file contains the code for emulation of the IA-32 sys_ptrace syscall. 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include <ia32x_utils.h>
#include <ia32toia64.h>

#include "btlib_segment.h"
#include "btlib_ptemu.h"
#include "btlib_signal.h"
#include "btlib_sync.h"
#include "btlib_syscall.h"

/* global flag that is set by the debugger */
unsigned long btl_is_ptraced = IS_PTRACED_NO;

/* TODO: this is an UGLY workaround for problems with kernel 2.4
 * (see function btlib_stop_debugee) */
static inline int is_kernel_2_6 (void);
static int m_is_kernel_2_6 = 0;

/*****************************************************************************/
/* get the address of the debugee's blst/glst
 * input - debugee
 *         data : output param. the blst address will be written to that 
 *                address
 */
/*****************************************************************************/
#define _blst blst
#define _glst glst
#define GET_DEBUGEE_TD_FIELD(field) 							\
	inline void * get_debugee_##field##_address(debugee_t * debugee)		\
	{										\
		unsigned long ldata; 							\
		unsigned long addr;							\
		long retval;								\
											\
		/* get the pointer to the debugee's thread descriptor */		\
		retval = ia64_peekuser (debugee->pid, (unsigned long)PT_R13, &ldata);	\
		if (retval < 0) {							\
			return (void*)retval;						\
		}									\
		DBPRINT (0, "r13=%lx\n", ldata)	;					\
		if (ldata != 0) {							\
			addr = ldata - sizeof (btlib_thread_desc);			\
			/* get the debugee's address of the ##field */			\
			addr += OFFSETOF (_##field, btlib_thread_desc);			\
			DBPRINT (0, "addr=%lx\n", addr)	;				\
			retval = ia64_peekdata (debugee->pid, addr, &ldata); 		\
			if (retval < 0) {						\
				return (void*)retval;					\
			}								\
			DBPRINT (0, #field "-address=%lx\n", ldata);			\
		} else {								\
			ldata = (unsigned long)-EIO;					\
		}									\
		return (void*)ldata;							\
	}
GET_DEBUGEE_TD_FIELD (blst)
GET_DEBUGEE_TD_FIELD (glst)

/*****************************************************************************/
/* get fields in the debugee's tls 
 * input - debugee
 *         field_name : name of the field
 *         data : output param. the field's value will be written to that 
 *                address
 */
/*****************************************************************************/
#define get_debugee_blst_field(debugee, field_name) \
	({	\
		long _offset = OFFSETOF(field_name, btlib_thread_data_t); \
		unsigned long _addr = debugee->blst + _offset; \
		unsigned long _ldata; \
		long _retval = ia64_peekdata (debugee->pid, _addr, &_ldata); \
		sys_err_chk (_retval, "get_debugee_blst_field: ptrace peek failed");	\
		DBPRINT (0, "%s=%lx\n", #field_name, _ldata);	\
		_ldata;	\
	})
/*****************************************************************************/
/* set fields in the debugee's tls
 * input - debugee
 *         field_name : name of the field
 *         data : will be written into the field
 */
/*****************************************************************************/
#define set_debugee_blst_field(debugee, field_name, data) \
	{	\
		long _retval;	\
		unsigned long _ldata = 0ul; \
		long _offset = OFFSETOF(field_name, btlib_thread_data_t); \
		long _size = sizeof (data);	\
		unsigned long _addr = debugee->blst + _offset; \
		if (_size != 8) {	\
			_retval = ia64_peekdata (debugee->pid, _addr, &_ldata); \
			DBPRINT (0, "<%s> peekdata=%lx\n", #field_name, _ldata);	\
			sys_err_chk (_retval, "set_debugee_blst_field: ptrace peek failed");	\
		}	\
		assert ((_size == sizeof(long)) || (_size == sizeof(int)));	\
		if (_size == sizeof (long)) {	\
			_ldata = data;	\
		} else {	\
			SET_LO_DWORD (&_ldata, (unsigned)data);	\
		}	\
		DBPRINT(0, "<%s> _ldata=%lx\n", #field_name, _ldata);	\
		_retval = SYSCALL (ptrace, (PTRACE_POKEDATA, debugee->pid, _addr, _ldata)); \
		sys_err_chk (_retval, "set_debugee_blst_field: ptrace poke failed");	\
	}
#define TF_FLAG 0x100
/*****************************************************************************/
/* wait for debuggee */
/*****************************************************************************/
static inline int
wait_for_debugee (debugee_t * debugee, int * statusp)
{
	int pid;
	do {
		/* don't let signals interrupt us here */
		DBPRINT (btl_prt_thread | btl_prt_pt,
				"before-wait: pid=%d options=%x\n",
				debugee->pid, __WALL);
		pid = SYSCALL (wait4, (debugee->pid, statusp, __WALL, NULL));
		DBPRINT (btl_prt_thread | btl_prt_pt, "after-wait: ret=%d status=%d\n",
				pid, *statusp);
	} while (pid == -EINTR);
	sys_err_chk ((long) pid, "sys_waitpid");
	return pid;
}
/*****************************************************************************/

/* access functions to the cont_signo part of the ptrace field */
/*****************************************************************************/
static inline int
get_cont_signo (debugee_t * debugee)
{
	unsigned long dpf;
	int cont_signo;
	dpf = get_debugee_blst_field (debugee, ptrace);
	cont_signo = GET_LO_DWORD (dpf);
	return cont_signo;
}
static inline void
set_cont_signo (debugee_t * debugee, int cont_signo)
{
	unsigned long dpf;
	DBPRINT (btl_prt_pt, "debuggee=%d cont_signo=%d\n", 
			debugee->pid, cont_signo);
	dpf = get_debugee_blst_field (debugee, ptrace);
	SET_LO_DWORD (&dpf, (unsigned) cont_signo);
	set_debugee_blst_field (debugee, ptrace, dpf);
}

static long continue_debugee (debugee_t * debugee, int cont_signo, int detach);
static unsigned *get_reg_addr (unsigned, BTGENERIC_IA32_CONTEXT *);
static unsigned get_context_flags (unsigned);
static debugee_t *find_debugee (int);
static debugee_t *add_debugee (int);
static void remove_debugee (debugee_t *);
static void set_ptrace_field (debugee_t * debugee, unsigned long request, int set);
static void set_trap_flag (debugee_t * debugee, int set);
static unsigned get_debugee_handler (debugee_t * debugee, int signo);
static int debugee_handles_signal (debugee_t * debugee, int signo);
static long wait_for_debugee_to_stop_itself (debugee_t * debugee, int * statusp);
static long handle_newly_attached_debugee (debugee_t * debugee, int * statusp);
static int is_ia32_process (int pid);
static int is_debugee_ready (debugee_t * debugee);
static debugee_t *handle_new_debugee (int child_pid, int *statusp);
static long sync_self_canonization (debugee_t * debugee);
static void ptctx2btgctx_64 (struct pt_all_user_regs *ptctx, BTGENERIC_IA64_CONTEXT * btg_ctx);
static void btgctx2ptctx_64 (BTGENERIC_IA64_CONTEXT * btg_ctx, struct pt_all_user_regs *ptctx);
static long is_inside_long_syscall (debugee_t * debugee, struct pt_all_user_regs *ptctx);

static long ptemu_traceme (int pid, unsigned addr, unsigned data);
static long ptemu_peekdata (int pid, unsigned addr, unsigned data);
static long ptemu_peekusr (int pid, unsigned addr, unsigned data);
static long ptemu_pokedata (int pid, unsigned addr, unsigned data);
static long ptemu_pokeusr (int pid, unsigned addr, unsigned data);
static long ptemu_cont (int pid, unsigned addr, unsigned data);
static long ptemu_kill (int pid, unsigned addr, unsigned data);
static long ptemu_singlestep (int pid, unsigned addr, unsigned data);
static long ptemu_getregs (int pid, unsigned addr, unsigned data);
static long ptemu_setregs (int pid, unsigned addr, unsigned data);
static long ptemu_getfpregs (int pid, unsigned addr, unsigned data);
static long ptemu_setfpregs (int pid, unsigned addr, unsigned data);
static long ptemu_attach (int pid, unsigned addr, unsigned data);
static long ptemu_detach (int pid, unsigned addr, unsigned data);
static long ptemu_getfpxregs (int pid, unsigned addr, unsigned data);
static long ptemu_setfpxregs (int pid, unsigned addr, unsigned data);
static long ptemu_syscall (int pid, unsigned addr, unsigned data);
static long ptemu_setoptions (int pid, unsigned addr, unsigned data);
static long ptemu_geteventmsg (int pid, unsigned addr, unsigned data);
static long ptemu_getthreadarea (int pid, unsigned addr, unsigned data);
static long ptemu_setthreadarea (int pid, unsigned addr, unsigned data);

/* 
 * array of helper functions that provide the functionality of the ptrace
 * requests
 */
static long (*ptemu_funcs[]) (int, unsigned, unsigned) = {
	ptemu_traceme,		/* 0 */
	    ptemu_peekdata,	/* 1 */
	    ptemu_peekdata,	/* 2 */
	    ptemu_peekusr,	/* 3 */
	    ptemu_pokedata,	/* 4 */
	    ptemu_pokedata,	/* 5 */
	    ptemu_pokeusr,	/* 6 */
	    ptemu_cont,		/* 7 */
	    ptemu_kill,		/* 8 */
	    ptemu_singlestep,	/* 9 */
	    NULL, 
	    NULL, 
	    ptemu_getregs,	/* 12 */
	    ptemu_setregs,	/* 13 */
	    ptemu_getfpregs,	/* 14 */
	    ptemu_setfpregs,	/* 15 */
	    ptemu_attach,	/* 16 */
	    ptemu_detach,	/* 17 */
	    ptemu_getfpxregs,	/* 18 */
	    ptemu_setfpxregs,	/* 19 */
	    NULL, 
	    ptemu_setoptions,	/* 21 */
	    NULL, 
	    NULL, 
	    ptemu_syscall,	/* 24 */
	    ptemu_getthreadarea,	/* 25 */
	    ptemu_setthreadarea,	/* 26 */
};
static long (*ptemu_funcsX[]) (int, unsigned, unsigned) = {
	ptemu_setoptions,		/* 0x4200 */
	ptemu_geteventmsg,		/* 0x4201 */
	NULL/*ptemu_getsiginfo*/,		/* 0x4202 */
	NULL/*ptemu_setsiginfo*/,		/* 0x4203 */
};

#if DEBUG
static char *pt_req_str[] = {
	"PTRACE_TRACEME",	/* 0 */
	"PTRACE_PEEKTEXT",	/* 1 */
	"PTRACE_PEEKDATA",	/* 2 */
	"PTRACE_PEEKUSR",	/* 3 */
	"PTRACE_POKETEXT",	/* 4 */
	"PTRACE_POKEDATA",	/* 5 */
	"PTRACE_POKEUSR",	/* 6 */
	"PTRACE_CONT",		/* 7 */
	"PTRACE_KILL",		/* 8 */
	"PTRACE_SINGLESTEP",	/* 9 */
	"",
	"",
	"PTRACE_GETREGS",	/* 12 */
	"PTRACE_SETREGS",	/* 13 */
	"PTRACE_GETFPREGS",	/* 14 */
	"PTRACE_SETFPREGS",	/* 15 */
	"PTRACE_ATTACH",	/* 16 */
	"PTRACE_DETACH",	/* 17 */
	"PTRACE_GETFPXREGS",	/* 18 */
	"PTRACE_SETFPXREGS",	/* 19 */
	"",
	"PTRACE_OLD_SETOPTIONS",	/* 21 */
	"",
	"",
	"PTRACE_SYSCALL",	/* 24 */
	"PTRACE_GET_THREAD_AREA",	/* 25 */
	"PTRACE_SET_THREAD_AREA"	/* 26 */
};
static char *ptX_req_str[] = {
	"PTRACE_SETOPTIONS",	/* 0x4200 */
	"PTRACE_GETEVENTMSG",	/* 0x4201 */
	"PTRACE_GETSIGINFO",	/* 0x4202 */
	"PTRACE_SETSIGINFO",	/* 0x4203 */
};
#endif				// DEBUG

/*****************************************************************************/
/* emulation of IA-32 ptrace syscall. 
 * this function only delegate the job to helper functions
 */
/*****************************************************************************/
long
btlib_ptrace (BTGENERIC_IA32_CONTEXT * bc32)
{
	int request, pid;
	unsigned addr, data;
	long retval;
	long (*fn) (int, unsigned, unsigned) = NULL;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	/* get the parameters for the syscall */
	request = IA32_SYSCALL_PARAM1 (bc32);
	pid = IA32_SYSCALL_PARAM2 (bc32);
	addr = IA32_SYSCALL_PARAM3 (bc32);
	data = IA32_SYSCALL_PARAM4 (bc32);

	DBPRINT (btl_prt_pt, "request=%d(%s) pid=%d addr=%x<%d> data=%x<%d>\n",
	       request, 
	       ((request >= 0) && (request <= 26)) ? pt_req_str[request] : 
	       ((request >= 0x4200) && (request <= 0x4203)) ? ptX_req_str[request - 0x4200] :
	       "(nill)",
	       pid, addr, addr, data, data);

	if ((request >= 0 /*PTRACE_TRACEME*/) && 
	    (request <= 26 /*PTRACE_SET_THREAD_AREA*/)) {
		/* call the helper function for the request */
		fn = ptemu_funcs[request];
	}
	if (fn == NULL) {
		if ((request >= 0x4200 /*PTRACE_SETOPTIONS*/) && 
		    (request <= 0x4203 /*PTRACE_SETSIGINFO*/)) {
			/* call the helper function for the request */
			fn = ptemu_funcsX[request - 0x4200];
		}
	}

	if (fn == NULL) {
		DBPRINT (TRUE, "ptrace command %x not implemented\n", request);
		retval = -EIO;
	} else {
		retval = fn (pid, addr, data);
	}


	DBPRINT (0, "retval=%ld\n", retval);
	return retval;
}

/*****************************************************************************/
/* process the wait syscall for the debugger
 * inputs: child_pid - the pid returned by the wait syscall in ia32_wait
 * 	   statusp - the address of the status param supplied by the
 *                   translated process
 * output: if the wait syscall has meaningfull information for the 
 *         translated process then return FINISH_WAIT. otherwise - the 
 *         information returned by wait syscall should be discarded and the 
 *         function returns DISCARD_WAIT
 */
/*****************************************************************************/
long
btlib_ptemu_wait (int child_pid, int *statusp)
{
	debugee_t *debugee;
	long retval;
	int ia32_state_canonized = 0;
	int signo;
	unsigned long dpf;
#if DEBUG
	int orig_status = *statusp;
#endif

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	DBPRINT (btl_prt_pt, "child_pid=%d status=%d \n", child_pid, *statusp);

	/* 
	 * if the child terminated then the wait syscall 
	 * should return at the debugger 
	 */
	if (!__WIFSTOPPED (*statusp)) {
		DBPRINT (btl_prt_pt, "child term - returning\n");
		/* remove the child from the debugee list */
		debugee = find_debugee (child_pid);
		if (debugee != NULL) {
			remove_debugee (debugee);
			debugee = NULL;
		}
		return FINISH_WAIT;
	}

	debugee = find_debugee (child_pid);
	/* 
	 * if the child is not in the debugee list than it can be a child
	 * who issued TRACEME request. 
	 */
	if (debugee == NULL) {
		DBPRINT (btl_prt_pt, "debugee not exist\n");
		debugee = handle_new_debugee (child_pid, statusp);
	} else {	/* debugee-already-exists */

		/* the debugee can issue execve and transform to an 64 bit process */
		if (!is_ia32_process (debugee->pid)) {
			DBPRINT (btl_prt_pt, "not IA-32 process\n");
			remove_debugee (debugee);
			return DISCARD_WAIT;
		}

		/* 
		 * it can issue execve and transform to a brand new IA-32 process
		 * in which case - it is not ready now
		 */
		if (!is_debugee_ready (debugee)) {
			DBPRINT (0, "not ready\n");
			remove_debugee (debugee);
			debugee = handle_new_debugee (child_pid, statusp);

		}

	}

	/* 
	 * if handle_new_debugee returned NULL, then the child is not 
	 * a debuggee and the wait syscall should return at the caller
	 */
	if (debugee == NULL) {
		DBPRINT (btl_prt_pt, "child is not a debuggee\n");
		return FINISH_WAIT;
	}

	DBPRINT (btl_prt_pt, "debugee->flags=%lx\n", debugee->flags);

	/* 
	 * if the debugee is marked as newly-attached ( it happens
	 * when the debugger attaches to a running process )
	 * it means that it needs to be stopped in a canonized IA-32 state
	 */
	if (debugee->flags & PTEMU_NEWLY_ATTACHED) {

		DBPRINT (btl_prt_pt, "PTEMU_NEWLY_ATTACHED\n");

		retval = handle_newly_attached_debugee (debugee, statusp);
		if (IS_SYS_ERROR (retval)) {
			DBPRINT (btl_prt_pt, "child could not be stopped\n");
			remove_debugee (debugee);
			return FINISH_WAIT;
		}

		debugee->blst = (unsigned long) get_debugee_blst_address (debugee);
		sys_err_chk (debugee->blst, "get_debugee_blst_address");

		/* handle_newly_attached_debugee do the canonization */
		ia32_state_canonized = 1;
	}



	/* if we got here then the debugee is stopped due to a signal */
	/**************************************************************/

	signo = __WSTOPSIG (*statusp);
	assert ((signo > 0) && (signo <= _NSIG));
	DBPRINT (btl_prt_pt, "signo=%d\n", signo);
	assert (debugee->stop_signo == 0);

	int new_debugee = debugee->flags & PTEMU_NEW ? 1 : 0;
	if (new_debugee) {
		debugee->flags &= ~PTEMU_NEW;
		/* a new debuggee was created. handle_new_debugee already
		 * took care of canonization. there is no point going thru
		 * the rest of the procedure */
		goto out;
	}
	if (*statusp >> 16) {
		int status;
		/* a message has arrived saying the the debuggee just
		 * created a new thread. the parent is at the end of the 
		 * clone syscall, treat it as if it's inside a long syscall
		 * (see btlib_stop_debugee) */
		DBPRINT (btl_prt_pt, "message arrived\n");
		assert (orig_status == *statusp);
		set_ptrace_field (debugee, PTEMU_IN_CONT_PROCESS, TRUE);
		set_ptrace_field (debugee, PTEMU_SUSPENSION_REQ, TRUE);
		//set_ptrace_field (debugee, PTEMU_CONT, TRUE);
		/* continue debuggee's execution */
		retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 
					(unsigned long)PTEMU_SIG0));
		sys_err_chk (retval, "ptrace cont failed (1)");
		retval = wait_for_debugee_to_stop_itself (debugee, &status);
		if (IS_SYS_ERROR (retval)) {
			*statusp = status;
			return -ESRCH;
		}
		goto out;
	}

	/* 
	 * the debugee sends itself a ptemu signal once it knows
	 * whether the signal should be discarded or delivered.
	 * it can be in the following cases - 
	 * 1) btgeneric called the btlib_ia32_raise_interrupt api func
	 * 2) when the debugee is stopped due to a signal
	 *      in cases when the signal is handled by the debugee - 
	 *      btlib of the debugger lets btlib of the debugee 
	 *      continue execution until it finds whether the signal
	 *      is meaningful in the IA-32 level. once it made this decision,
	 *      btlib of the debugee sends itself the ptemu-signal (see
	 *      next block of code)
	 */
	if (signo != PTEMU_SIG0) {
		if (debugee_handles_signal (debugee, signo)) {
			/* 
			 * if the debugee has a handler for this signal then
			 * we need to let it continue and see if the signal is
			 * meaningfull on the IA-32 level
			 */
			retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul,
				    (unsigned long) signo));
			sys_err_chk (retval, "ptrace(2)");
			retval = wait_for_debugee_to_stop_itself (debugee, statusp);
			if (IS_SYS_ERROR (retval)) {
				DBPRINT (TRUE, "wait_for_debugee_to_stop failed=%ld\n", retval);
				return FINISH_WAIT;
			}
			/* 
			 * check the ptrace field in the debugee's tls
			 * if the PTEMU_DISCARD bit is cleared  
			 * then the signal should be delivered
			 */
			dpf = get_debugee_blst_field (debugee, ptrace);
			signo = (dpf & PTEMU_DISCARD) ? 0 : PTEMU_SIG0;
			DBPRINT (btl_prt_pt, "PTEMU_DISCARD=%d signo=%d\n",
			         (dpf & PTEMU_DISCARD) ? 1 : 0, signo);
			debugee->flags |= PTEMU_SIG_PROCESSED;
		}
	}
	if (signo == PTEMU_SIG0) {
		/* 
		 * when btlib of the debugee stops itself with the ptemu
		 * signal it means that -
		 * if the signal is not discarded then the IA-32 state is
		 * already canonized
		 */
		ia32_state_canonized = 1;
		signo = get_debugee_blst_field (debugee, tia_signal.siginfo.si_signo);
		/* 
		 * if the PTEMU_CANCEL bit is set then this is a cancellation
		 * signal sent by btlib when no signal handler is registered
		 * by the translated process. cancellation signal is sent
		 * after the signal was processed by btlib's signal handler
		 * so this is the second time the debugee gets the same signal.
		 * we need to discard the wait and let the debugee die with
		 * the signal
		 */
		if (signo & PTEMU_CANCEL) {
			signo &= ~PTEMU_CANCEL;
			DBPRINT (btl_prt_pt, "(2) cancellation signo=%d\n",
				 signo);
			retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 
						   (unsigned long) signo));
			return DISCARD_WAIT;
		}
		DBPRINT (btl_prt_pt, "(2) signo=%d\n", signo);
	}

	if (signo == 0) {
		/* 
		 * the signal is discarded by btlib of the debugee. so -
		 * the results of the wait syscall need to be discarded
		 * by btlib of the debugger
		 */
		DBPRINT (btl_prt_pt, "signal is discarded\n");

		/* let the debugee continue */
		retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 0ul));
		sys_err_chk (retval, "ptrace(3)");
		/* discard the result of the wait syscall */
		return DISCARD_WAIT;
	}

	/* 
	 * if we got here then the signal should be delivered to the
	 * debugee, that is - the wait syscall need to return at the debugger 
	 */
	/*********************************************************************/

	DBPRINT (btl_prt_pt, "ia32_state_canonized=%d\n", ia32_state_canonized);
	if (ia32_state_canonized == 0) {
		/* if the state is not canonized yet then do it now */
		assert (signo != PTEMU_SIG0);
		assert ((debugee->flags & PTEMU_SIG_PROCESSED) == 0);

		/* 
		 * we must not detach from the debugee, so pass 1 for
		 * max_count.
		 * if max_count is exceeded ( retval==1 ) then the debugee
		 * has to be restarted and stopped again in the hope that
		 * it would be possible to canonize its state next time
		 * it is stopped. since btlib_stop_debugee uses PTRACE_DETACH 
		 * and ATTACH to start and stop the debugee, it is not possible
		 * to do it here since we may lose signals that arrive at the
		 * debugee when we detached from it. so, we will use PTRACE_CONT
		 * in order to resume the debugee's execution.
		 * problem is that while we let the debugee
		 * continue, it may get a signal, which is different than the
		 * one that brought us here at the first place.
		 * so - we stop the debugee by sending it the same signal that
		 * brought us here. this signal will stop the debugee and 
		 * bring us back here. 
		 * after sending the signal we discard this
		 * wait. next time the debugger calls wait4, everything will
		 * start from the beginning. if wait4 returns with a different
		 * signal then no-problem - the original signal will not get
		 * lost because we've just resent it.
		 */
		retval = btlib_stop_debugee (debugee, 1, 0, statusp);
		if (retval < 0) {
			/* debuggee died */
			return FINISH_WAIT;
		} else if (retval == 1) {
			/* debuggee could not be stopped in a consistent
			 * 32 bit state */
			if (debugee_handles_signal (debugee, signo)) {
				/* let the debuggee continue with the signal that stopped it. 
				 * the debuggee cannot be canonized only in cases where
				 * where it was stopped by async signal. 
				 * we let the debuggee continue with the signal that stopped
				 * it - btlib will enque it and will try to canonize the 32 bit
				 * state localy. since the state cannot be canonized
				 * btlib of the debuggee will ask btgeneric to yield execution
				 * as soon as it finishes its internal works that prevent it
				 * from being canonized. so we can be sure that the debuggee
				 * will soon get to btlib_reenter and raise the signal that
				 * stopped it. */ 
				DBPRINT (btl_prt_pt, "continue debuggee with signal\n");
				retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 
							(unsigned long)signo));
				sys_err_chk (retval, "ptrace (4)");

				/* repeat the wait4 */
				return DISCARD_WAIT;
			} else {
				DBPRINT (btl_prt_pt, "resend signal\n");
				/* let the debugee continue */
				retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 0ul));
				sys_err_chk (retval, "ptrace_cont");
				/* give it a chance to work */
				YIELD_THREAD_EXECUTION ();
				/* resend the signal. if the debuggee does not handle the signal
				 * then this signal will either be ignored or kill the debuggee.
				 * in both cases the debuggee won't care much about the info, so
				 * we can use tkill to resend the signal */
				retval = SYSCALL (tkill, (debugee->pid, signo));
				sys_err_chk (retval, "tkill");
			}
			/* repeat the wait4 */
			return DISCARD_WAIT;
		}
		assert (retval == 0);
	}

	/* 
	 * clear the information bits from signo
	 * TODO: support TRACESYSGOOD 
	 */
       *statusp = __W_STOPCODE (signo & 0x7Fu);
out:
	/* put the signo in the status field and return at the debugger */
	debugee->stop_signo = signo;
	DBPRINT (btl_prt_pt, "*statusp=%d\n", *statusp);
	return FINISH_WAIT;
}

/*****************************************************************************/
/* signal handler for the first ptemu signal */
/* this signal handler is never executed as a result of signal, only as 
 * a result of PTRACE_CONT of the debugger. therefore, we do not have to
 * canonize-suspend-context before restating btgeneric.
 */
/*****************************************************************************/
void
btlib_ptemu0_handler (int signo UNUSED, siginfo_t * siginfo UNUSED, 
                      struct sigcontext *ctx UNUSED)
{
	int is_detached, is_last_thread, is_newly_attached, cont_signo, action = 0;
	int is_cont;
	unsigned long *plabel;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (signo == PTEMU_SIG0);
	DBPRINT (btl_prt_pt, "btl_is_ptraced=%ld cur_thread_data->ptrace=%lx\n", 
		 btl_is_ptraced, cur_thread_data->ptrace);

	/* if the thread is not being debugged then proceed with regular sig
	 * handling */
	if (cur_thread_data->ptrace == 0ul) {
		return btlib_signal_handler (signo, siginfo, ctx);
	}
	
	is_cont = (cur_thread_data->ptrace & PTEMU_CONT) ? 1 : 0;
	is_detached = (cur_thread_data->ptrace & PTEMU_DETACH) ? 1 : 0;
	is_last_thread = (cur_thread_data->ptrace & PTEMU_DETACH_LAST_THREAD) ? 1 : 0;
	cur_thread_data->ptrace &= ~ (PTEMU_DETACH_LAST_THREAD | PTEMU_DETACH | PTEMU_CONT);
	/* a thread is marked as last-thread must be detached */
	assert ( ! is_last_thread || is_detached );
	/* if a thread is detached its ptrace field should be cleared at this point */
	assert ( ! is_detached || (cur_thread_data->ptrace == 0));
	DBPRINT (btl_prt_pt && is_detached, "PTEMU_DETACH\n");
	DBPRINT (btl_prt_pt && is_last_thread, "PTEMU_DETACH_LAST_THREAD\n");
	DBPRINT (btl_prt_pt && is_cont, "PTEMU_CONT\n");

	/* newly attached debuggees need to flush all instruction cache */
	is_newly_attached = (cur_thread_data->ptrace & PTEMU_NEWLY_ATTACHED) ? 1 : 0;

	if (is_newly_attached) {
		/* remove the NEWLY_ATTACHED indication */
		cur_thread_data->ptrace = PTEMU_PTRACED;
		if (btl_is_ptraced != IS_PTRACED_YES) {
			EMULATION_POLICY ep = {
				memory_order_policy: POLICY_UNCHANGED,
				/* don't optimized away changes in eflag */			
				strict_state_policy: STATE_POLICY_PRESERVE_EFLAGS,
			};
	
			DBPRINT (btl_prt_pt, "PTEMU_NEWLY_ATTACHED\n");
			assert (btl_is_ptraced == IS_PTRACED_YES_NOT_READY);

			/* don't optimize away changes in eflags */
			BTGENERIC_SET_EMULATION_POLICY(&ep);
			assert (ep.strict_state_policy == STATE_POLICY_PRESERVE_EFLAGS);
			/* change memory ordering policy */
			apply_shared_memory_ordering_solution ();
			/* flushing instruction cache must be done after policy setting 
			   because policy setting doesn't flush current translations */
			BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (0, ~0u , BT_FLUSH_FORCE);
			btl_is_ptraced = IS_PTRACED_YES;
		}
		/* stop self with SIGTRAP */
		cur_thread_data->longjmp_status = TRAP;
		/* set new ip to do_longjmp */
		plabel = (unsigned long *) do_longjmp;
		ctx->sc_ip = PLABEL_FUNC (plabel);
		ctx->sc_gr[1] = PLABEL_GP (plabel);
	} else {
		
		/* clear the restart signo from the ptrace field */
		cont_signo = GET_LO_DWORD (cur_thread_data->ptrace);
		SET_LO_DWORD (&cur_thread_data->ptrace, 0);

		if (cont_signo != 0) {
			btlib_prepare_cont_signo (cont_signo);
			action = SIGNAL;
		} else {
			/* 
			 * reenter btgeneric.
			 * we need to take care of the first time the 
			 * debugee stopped after ptrace_traceme - the thread
			 * data struct of the main thread is not ready yet
			 * and we must not use the jmpbuf
			 */
			action = is_cont ? CONTINUE : 
				(cur_thread_data->tid == cur_thread_data->is_run_context_ready) ?
				 RUN : CONTINUE;
			/* restart syscalls that need it */
			if ((action == RUN) && 
					syscall_need_restart (cur_thread_data->syscall_id)) {
				DBPRINT (btl_prt_pt, "restart syscall_id=%d\n",
						cur_thread_data->syscall_id);
				cur_thread_data->bc32->eax = cur_thread_data->syscall_id;
				cur_thread_data->bc32->eip -= 2;
				cur_thread_data->syscall_id = IA32_SYSCALL_NONE;
			}
		}
	}

	if (is_last_thread) {
		btl_is_ptraced = IS_PTRACED_NO;
	}

	DBPRINT (btl_prt_pt, "action=%d\n", action);
	if (action != CONTINUE) {
		cur_thread_data->longjmp_status = action;
		/* if we restart btg then clear the syscall indication */
		if (action == RUN) {
			if (cur_thread_data->syscall_id != IA32_SYSCALL_NONE) {
				cur_thread_data->bc32->eax = (unsigned)-EINTR;
				cur_thread_data->syscall_id = IA32_SYSCALL_NONE;
			}
		}
		/* set new ip to do_longjmp */
		plabel = (unsigned long *) do_longjmp;
		ctx->sc_ip = PLABEL_FUNC (plabel);
		ctx->sc_gr[1] = PLABEL_GP (plabel);
                /* if the debuggee does not continue at the 64-bit level then
                 * we need to clear the long-syscall-flag, in case it was stopped
                 * while inside a long syscall
                 */
                *cur_thread_data->long_syscall_flag = NOT_IN_SYSCALL;

	}

	return;
}

/* helper functions for ptrace requests */
/*****************************************************************************/
static long
ptemu_traceme (int pid UNUSED, unsigned addr UNUSED, unsigned data UNUSED)
{
	long retval = 0;

	DBPRINT (btl_prt_pt, "\n");

	/* call the 64 bit ptrace with the traceme request */
	retval = SYSCALL (ptrace, (PTRACE_TRACEME, 0, 0ul, 0ul));
	if (retval == 0) {
		EMULATION_POLICY policy;

		btl_is_ptraced = IS_PTRACED_YES_NOT_READY;
		cur_thread_data->ptrace = PTEMU_PTRACED;

		/* change memory ordering policy */
		apply_shared_memory_ordering_solution ();
		/* disable writable code optimization */
		policy.memory_order_policy = POLICY_UNCHANGED;
		policy.strict_state_policy = STATE_POLICY_PRESERVE_EFLAGS;
		BTGENERIC_SET_EMULATION_POLICY (&policy);
		assert (policy.strict_state_policy == STATE_POLICY_PRESERVE_EFLAGS);
		/* flush translations */
		BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE (0, ~0u, BT_FLUSH_FORCE);
	}
	return retval;
}

/*****************************************************************************/
static long
ptemu_peekdata (int pid, unsigned addr, unsigned data)
{
	long retval;
	unsigned long p = addr;
	int is_hi = 0;
	union {
		unsigned long l;
		struct {
			unsigned lo;
			unsigned hi;
		} s;
	} ldata;

	DBPRINT (btl_prt_pt, "\n");

try_peek:

	retval = ia64_peekdata (pid, p, &ldata.l);
	if (retval < 0) {
		/* 
		 * handle a case where the user tries to read the last 4 
		 * bytes in a page and the next page is not readable 
		 */
		if (is_hi == 0) {
			int is_on_page_boundry =
				(PAGESTART (addr, btl_page_size) !=
				 PAGESTART (addr + 4, btl_page_size)) ? 1 : 0;
			if ((retval == -EIO) && is_on_page_boundry) {
				p -= 4;
				is_hi = 1;
				goto try_peek;
			}
		}
	}

	if (retval == 0) {
		DBPRINT (btl_prt_pt, "ldata=%lx is_hi=%d\n", ldata.l, is_hi);
		retval = safe_memcpy (U2P (data), is_hi ? &ldata.s.hi : &ldata.s.lo,
				      sizeof (unsigned));

		if (retval < 0) {
			retval = -EIO;
		}
	}

	DBPRINT (btl_prt_pt, "retval=%ld \n", retval);
	return retval;
}

/* 
 * BTLibLinux implementation of ptrace with PTRACE_GET_THREAD_AREA 
 *
 * parameters: 
 *     addr - TLS/GDT index
 *     data - 32-bit pointer, address of user_desc
 * return value:  
 *     0               - success
 *     negative number - error number
 */
static long
ptemu_getthreadarea (int pid, unsigned addr, unsigned data)
{
	debugee_t *debuggee;
	desc_struct desc;
	user_desc u_info;
	unsigned long temp;

	debuggee = find_debugee (pid);	/* find debuggee according to pid */
	if (debuggee == NULL) {
		return -ESRCH;
	}

	if (((int) addr > GDT_USER_ENT_LAST)
	    || ((int) addr > GDT_USER_ENT_FIRST)) {
		return -EINVAL;
	}

	/* get thread GDT/TLS information - in btlib_thread_data_t.tls_array */
	temp = get_debugee_blst_field (debuggee, tls_array[(int) addr - GDT_USER_ENT_FIRST]);
	memcpy (&desc, &temp, sizeof (temp));
	get_desc (&u_info, &desc);

	if (safe_memcpy (U2P (data), &u_info, sizeof (u_info)) < 0)
		return -EFAULT;

	return 0;
}

/* 
 * BTLibLinux implementation of ptrace with PTRACE_SET_THREAD_AREA 
 *
 * parameters: 
 *     addr - TLS/GDT index
 *     data - 32-bit pointer, address of user_desc
 * return value:  
 *     0               - success
 *     negative number - error number
 */
static long
ptemu_setthreadarea (int pid, unsigned addr, unsigned data)
{
	debugee_t *debuggee;
	desc_struct desc;
	user_desc u_info;
	unsigned long temp;

	debuggee = find_debugee (pid);	/* find debuggee according to pid */
	if (debuggee == NULL) {
		return -ESRCH;
	}

	if (((int) addr > GDT_USER_ENT_LAST)
	    || ((int) addr > GDT_USER_ENT_FIRST)) {
		return -EINVAL;
	}

	if (safe_memcpy (&u_info, U2P (data), sizeof (u_info)) < 0)
		return -EFAULT;

	if (is_udesc_empty (&u_info)) {
		memset (&desc, 0, sizeof (desc));
	} else {
		set_desc (&desc, &u_info);
		memcpy (&temp, &desc, sizeof (temp));
	}

	/* set thread GDT/TLS information - in btlib_thread_data_t.tls_array */
	set_debugee_blst_field (debuggee,
				tls_array[(int) addr - GDT_USER_ENT_FIRST],
				temp);

	return 0;
}

/*****************************************************************************/
static long
ptemu_peekusr (int pid, unsigned addr, unsigned data)
{
	int *reg_addr;
	unsigned reg_val;
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	void *remote_glst;
	debugee_t *debugee;
	long retval = 0;
	unsigned *up = U2P (data);

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* if the addr param is not valid then return an error */
	if ((addr & 3) || ((int) addr < 0)) {
		return -EIO;
	}

	/* orig_eax is kept in remote_tls */
	if (addr >> 2 == ORIG_EAX) {
		int syscall_id;
		syscall_id = get_debugee_blst_field (debugee, syscall_id);
		reg_val = syscall_id;
		DBPRINT (0, "addr=ORIG_EAX val=%d\n", syscall_id);
		goto write_reg_val;
	}

	/* 
	 * get the address of the field in bc32 that corresponds 
	 * to the user's request
	 */
	reg_addr = get_reg_addr (addr, &bc32);
	DBPRINT (0, "reg_addr=%p\n", reg_addr);
	/* if the address is not of any register then return the value 0 */
	if (reg_addr == NULL) {
		return -EIO;	/* the syscall returns with success */
	}

	/* no more excuses - we need to get the remote context */

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	memset (&bc32, 0, sizeof (bc32));
	bc32.composition_bits = get_context_flags (addr);
	btg_retval = BTGENERIC_IA32_CONTEXT_GET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);

	reg_val = *reg_addr;
	/* verify the access perm to the given address */
write_reg_val:
	/* get the register's value and return */
	if (safe_memcpy (up, &reg_val, sizeof (*up)) < 0) {
		retval = -EIO;
		goto out;
	}
	DBPRINT (btl_prt_pt, "val=%x<%d>\n", *up, *up);
out:
	return retval;
}

/*****************************************************************************/
static long
ptemu_pokedata (int pid, unsigned addr, unsigned data)
{
	long retval;
	unsigned long p = addr;
	int is_hi = 0;
	union {
		unsigned long l;
		struct {
			unsigned lo;
			unsigned hi;
		} s;
	} ldata;

	DBPRINT (btl_prt_pt, "\n");

try_peek:

	/* read 8 bytes from the debuggee */
	retval = ia64_peekdata (pid, p, &ldata.l);
	if (retval < 0) {
		/* handle a case where the user tries to read the last 4 
		 * bytes in a page and the next page is not readable */
		if (is_hi == 0) {
			int is_on_page_boundry =
			    (PAGESTART (addr, btl_page_size) !=
			     PAGESTART (addr + 4, btl_page_size)) ? 1 : 0;
			if ((retval == -EIO) && is_on_page_boundry) {
				p -= 4;
				is_hi = 1;
				goto try_peek;
			}
		}
	}
	DBPRINT (btl_prt_pt, "ia64_peekdata=%ld ldata=%lx\n", retval, ldata.l);

	if (retval < 0) {
		retval = -EIO;
		goto out;
	}

	/* insert the 4 bytes requested by the user */
	if (is_hi) {
		ldata.s.hi = data;
	} else {
		ldata.s.lo = data;
	}

	DBPRINT (btl_prt_pt, "ldata=%lx\n", ldata.l);

	/* write the resulted 8 bytes back to the debuggee */
	retval = SYSCALL (ptrace, (PTRACE_POKEDATA, pid, p, ldata.l));

	DBPRINT (btl_prt_pt, "poke_data=%ld\n", retval);

	if (retval < 0) {
		goto out;
	}

/* remote-flush is not implemented yet in btgeneric. but - except
 * of attaching to running processes and detaching from them -
 * we dont need to flush instruction cache because we work
 * in SMC mode when debugged ( see BtlMemoryQueryPermissions )
 *
		
	if (retval == 0) {
		BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE_REMOTE (
			(void*)(long)pid, 
			(void*)(uintptr_t)addr, 
			sizeof(int), 
			BT_FLUSH_FORCE);
	}
*/
out:
	return retval;
}

/*****************************************************************************/
static long
ptemu_pokeusr (int pid, unsigned addr, unsigned data)
{
	unsigned *reg_addr;
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	void *remote_glst;
	debugee_t *debugee;

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* if the addr param is not valid then return an error */
	if ((addr & 3) || ((int) addr < 0)) {
		return -EIO;
	}
	/* orig_eax is kept in remote_tls */
	if (addr >> 2 == ORIG_EAX) {
		/* TODO: check behavior of native programs */
		assert (!"attempt to change orig_eax");
	}
	/* 
	 * get the address of the field in bc32 that corresponds 
	 * to the user's request
	 */
	reg_addr = get_reg_addr (addr, &bc32);
	DBPRINT (0, "reg_addr=%p\n", reg_addr);
	/* if the address is not of any register then return the value 0 */
	if (reg_addr == NULL) {
		return -EIO;	/* the syscall returns with success */
	}

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	/* get the remote context */
	memset (&bc32, 0, sizeof (bc32));
	bc32.composition_bits = get_context_flags (addr);
	btg_retval = BTGENERIC_IA32_CONTEXT_GET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);

	/* set the register's value */
	*reg_addr = data;
	/* set the remote context with the new value */
	btg_retval = BTGENERIC_IA32_CONTEXT_SET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);

	return 0;
}

/*****************************************************************************/
static long
ptemu_cont (int pid, unsigned addr UNUSED, unsigned data)
{
	long retval;
	int cont_signo;
	debugee_t *debugee;

	DBPRINT (btl_prt_pt, "\n");

	/* the data argument is the signal with which the debugee has to continue */
	cont_signo = (int) data;
	if ((cont_signo < 0) || (cont_signo > _NSIG)
		|| (cont_signo == PTEMU_SIG0)) {
		return -EIO;
	}

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* clear the syscall request from the debugee's tls */
	set_ptrace_field (debugee, PTEMU_SYSCALL, FALSE);

	/* clear the TF flag from the debugee's eflags */
	set_trap_flag (debugee, FALSE);

	/* continue debugee execution */
	retval = continue_debugee (debugee, cont_signo, FALSE);
	return retval;
}

/*****************************************************************************/
static long
ptemu_kill (int pid, unsigned addr UNUSED, unsigned data UNUSED)
{
	long retval;
	debugee_t *debugee;

	DBPRINT (btl_prt_pt, "pid=%d\n", pid);

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	retval = SYSCALL (ptrace, (PTRACE_KILL, pid, 0ul, 0ul));
	if (retval == 0) {
		/* remove from the debugee list */
		remove_debugee (debugee);
	}
	return (int) retval;
}

/*****************************************************************************/
static long
ptemu_singlestep (int pid, unsigned addr UNUSED, unsigned data)
{
	debugee_t *debugee;
	long retval;
	int cont_signo;

	DBPRINT (btl_prt_pt, "\n");

	/* the data argument is the signal with which the debugee has to continue */
	cont_signo = (int) data;
	if ((cont_signo < 0) || (cont_signo > _NSIG)
		|| (cont_signo == PTEMU_SIG0)) {
		return -EIO;
	}

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* clear the syscall request from the debugee's tls */
	set_ptrace_field (debugee, PTEMU_SYSCALL, FALSE);

	/* set the TF flag at the debugee's eflags */
	set_trap_flag (debugee, TRUE);

	/* continue debugee execution */
	retval = continue_debugee (debugee, cont_signo, FALSE);
	return retval;
}

/*****************************************************************************/
static long
ptemu_attach (int pid UNUSED, unsigned addr UNUSED, unsigned data UNUSED)
{
	long retval;
	debugee_t *debugee;

	DBPRINT (btl_prt_pt, "\n");

	/* attach to pid at the 64 bit level */
	retval = SYSCALL (ptrace, (PTRACE_ATTACH, pid, 0ul, 0ul));
	if (IS_SYS_ERROR (retval)) {
		DBPRINT (btl_prt_pt, "ptrace failed %ld\n", retval);
		goto out;
	}
	/* add an element to the debugee list */
	debugee = add_debugee (pid);
	DBPRINT (btl_prt_pt, "debugee=%p\n", debugee);
	/* mark it as not-consistnet at the IA-32 level */
	debugee->flags = PTEMU_NEWLY_ATTACHED;
out:
	return retval;
}

/*****************************************************************************/
static long
ptemu_detach (int pid, unsigned addr UNUSED, unsigned data)
{
	debugee_t *debugee;
	long retval;
	int cont_signo;

	DBPRINT (btl_prt_pt, "\n");

	/* the data argument is the signal with which the debugee has to continue */
	cont_signo = (int) data;
	if ((cont_signo < 0) || (cont_signo > _NSIG)
		|| (cont_signo == PTEMU_SIG0)) {
		return -EIO;
	}


	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

#if (DEBUG && 1)
	{
		unsigned long ptf;
		ptf = get_debugee_blst_field (debugee, ptrace);
		DBPRINT (btl_prt_pt, "ptf=%lx\n", ptf);
	}
#endif
	/* clear the ptrace field of the debuggee, keep only
	 * the DETACH field set */
	set_debugee_blst_field (debugee, ptrace, PTEMU_DETACH);

	/* detach from the debugee */
	retval = continue_debugee (debugee, cont_signo, TRUE);
	if (retval == 0) {
		/* remove from the debugee list */
		remove_debugee (debugee);
	}

	return retval;
}

/*****************************************************************************/
static long
ptemu_syscall (int pid, unsigned addr UNUSED, unsigned data)
{
	debugee_t *debugee;
	long retval;
	int cont_signo;

	DBPRINT (btl_prt_pt, "\n");

	/* the data argument is the signal with which the debugee has to continue */
	cont_signo = (int) data;
	if ((cont_signo < 0) || (cont_signo > _NSIG)
		|| (cont_signo == PTEMU_SIG0)) {
		return -EIO;
	}

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* set the syscall request from the debugee's tls */
	set_ptrace_field (debugee, PTEMU_SYSCALL, TRUE);

	/* clear the TF flag at the debugee's eflags */
	set_trap_flag (debugee, FALSE);

	/* continue debugee execution */
	retval = continue_debugee (debugee, cont_signo, FALSE);
	return retval;
}

/*****************************************************************************/
static long
ptemu_getregs (int pid, unsigned addr UNUSED, unsigned data)
{
	debugee_t *debugee;
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	void *remote_glst;
	struct user_regs_struct_32 regs;
	int *datap;
	long retval = 0;

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	/* get the remote context */
	memset (&bc32, 0, sizeof (bc32));
	bc32.composition_bits = CTX32_CRITICAL | CTX32_GP | CTX32_SEGMENT;

	btg_retval = BTGENERIC_IA32_CONTEXT_GET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	DBCODE (0, print_btgcontext32 (&bc32));
	assert (btg_retval == BT_STATUS_SUCCESS);

	/* put the register values in the 'data' buffer */

	regs.ebx = bc32.ebx;
	regs.ecx = bc32.ecx;
	regs.edx = bc32.edx;
	regs.esi = bc32.esi;
	regs.edi = bc32.edi;
	regs.ebp = bc32.ebp;
	/* 
	 * in case we're inside interrupted syscall bc32.eax holds
	 * the syscall id, but the debugger expects to see the return
	 * value of the interrupted syscall
	 */
	regs.eax = bc32.eax;

	datap = (int *) (void *) &regs.ds;
	*datap = bc32.ds;
	datap = (int *) (void *) &regs.es;
	*datap = bc32.es;
	datap = (int *) (void *) &regs.gs;
	*datap = bc32.gs;
	datap = (int *) (void *) &regs.fs;
	*datap = bc32.fs;

	regs.orig_eax = get_debugee_blst_field (debugee, syscall_id);
	regs.eip = bc32.eip;

	datap = (int *) (void *) &regs.cs;
	*datap = bc32.cs;

	regs.eflags = bc32.eflags;
	regs.esp = bc32.esp;

	datap = (int *) (void *) &regs.ss;
	*datap = bc32.ss;

	DBPRINT (0, "ebx=%d ecx=%d edx=%d esi=%d edi=%d ebp=%x \n\
		 eax=%x ds=%x __ds=%x es=%x __es=%x gs=%x __gs=%x fs=%x __fs=%x \n\
		 orig_eax=%d eip=%x cs=%x __cs=%x eflags=%x esp=%x ss=%x __ss=%x\n", 
		 regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi, regs.ebp, regs.eax, 
		 regs.ds, regs.__ds, regs.es, regs.__es, regs.gs, regs.__gs, regs.fs, 
		 regs.__fs, regs.orig_eax, regs.eip, regs.cs, regs.__cs, regs.eflags, 
		 regs.esp, regs.ss, regs.__ss);

	/* put the register values in the 'data' buffer */
	if (safe_memcpy (U2P (data), &regs, sizeof (regs)) < 0) {
		retval = -EIO;
	}

	return retval;
}

/*****************************************************************************/
static long
ptemu_setregs (int pid, unsigned addr UNUSED, unsigned data)
{
	debugee_t *debugee;
	int *datap;
	struct user_regs_struct_32 regs;
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	void *remote_glst;

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	/* reset the context struct */
	memset (&bc32, 0, sizeof (bc32));

	/* get the register values from the 'data' buffer */
	if (safe_memcpy (&regs, U2P (data), sizeof (regs)) < 0) {
		return -EIO;
	}

	DBPRINT (0, "ebx=%d ecx=%d edx=%d esi=%d edi=%d ebp=%x \n\
		 eax=%x ds=%x __ds=%x es=%x __es=%x gs=%x __gs=%x fs=%x __fs=%x \n\
		 orig_eax=%d eip=%x cs=%x __cs=%x eflags=%x esp=%x ss=%x __ss=%x\n", 
		 regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi, regs.ebp, regs.eax, 
		 regs.ds, regs.__ds, regs.es, regs.__es, regs.gs, regs.__gs, regs.fs, 
		 regs.__fs, regs.orig_eax, regs.eip, regs.cs, regs.__cs, regs.eflags, 
		 regs.esp, regs.ss, regs.__ss);

	bc32.ebx = regs.ebx;
	bc32.ecx = regs.ecx;
	bc32.edx = regs.edx;
	bc32.esi = regs.esi;
	bc32.edi = regs.edi;
	bc32.ebp = regs.ebp;
	bc32.eax = regs.eax;

	datap = (int *) (void *) &regs.ds;
	bc32.ds = *datap;
	datap = (int *) (void *) &regs.es;
	bc32.es = *datap;
	datap = (int *) (void *) &regs.gs;
	bc32.gs = *datap;
	datap = (int *) (void *) &regs.fs;
	bc32.fs = *datap;

	bc32.eip = regs.eip;

	datap = (int *) (void *) &regs.cs;
	bc32.cs = *datap;

	bc32.eflags = regs.eflags;
	bc32.esp = regs.esp;

	datap = (int *) (void *) &regs.ss;
	bc32.ss = *datap;

	DBCODE (0, print_btgcontext32 (&bc32));

	/* set the remote context with the new value */
	bc32.composition_bits = CTX32_CRITICAL | CTX32_GP | CTX32_SEGMENT;
	btg_retval = BTGENERIC_IA32_CONTEXT_SET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);
	return 0;
}

/*****************************************************************************/
static long
ptemu_getfpregs (int pid, unsigned addr UNUSED, unsigned data)
{
	struct i387_fsave_struct fsave;
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	debugee_t *debugee;
	void *remote_glst;
	long retval = 0;

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	DBPRINT (0, "remote_glst=%p\n", remote_glst);
	/* get the remote context */
	memset (&bc32, 0, sizeof (bc32));
	bc32.composition_bits = CTX32_FLOATING_POINT;
	btg_retval = BTGENERIC_IA32_CONTEXT_GET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);

	DBCODE (0, print_btgcontext32 (&bc32));

	/* convert the fp context to linux format */
	btg2fsave (&bc32, &fsave);

	/* copy the context to the output buffer */
	if (safe_memcpy (U2P (data), &fsave, sizeof (fsave)) < 0) {
		retval = -EIO;
		goto out;
	}

out:
	return retval;
}

/*****************************************************************************/
static long
ptemu_setfpregs (int pid, unsigned addr UNUSED, unsigned data)
{
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	debugee_t *debugee;
	void *remote_glst;
	struct i387_fsave_struct fsave;

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* copy the context to the output buffer */
	if (safe_memcpy (&fsave, U2P (data), sizeof (fsave)) < 0) {
		return -EIO;
	}
	fsave2btg (&bc32, &fsave);
	bc32.composition_bits = CTX32_FLOATING_POINT;

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	/* set the context with the new values */
	btg_retval = BTGENERIC_IA32_CONTEXT_SET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);

	return 0;
}

/*****************************************************************************/
static long
ptemu_getfpxregs (int pid, unsigned addr UNUSED, unsigned data)
{
	struct i387_fxsave_struct fxsave;
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	debugee_t *debugee;
	void *remote_glst;

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	/* get the remote context */
	memset (&bc32, 0, sizeof (bc32));
	bc32.composition_bits = CTX32_FLOATING_POINT | CTX32_XMM;
	btg_retval = BTGENERIC_IA32_CONTEXT_GET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);

	/* get the xmm context */
	btg2fxsave (&bc32, &fxsave);

	/* copy the context to the output buffer */
	if (safe_memcpy (U2P (data), &fxsave, sizeof (fxsave)) < 0) {
		return -EIO;
	}

	return 0;
}

/*****************************************************************************/
static long
ptemu_setfpxregs (int pid, unsigned addr UNUSED, unsigned data)
{
	struct i387_fxsave_struct fxsave;
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	debugee_t *debugee;
	void *remote_glst;
	long retval = 0;

	DBPRINT (btl_prt_pt, "\n");

	/* find the debugee's info */
	debugee = find_debugee (pid);
	if (debugee == NULL) {
		return -ESRCH;
	}

	/* read the new xmm state from user */
	if (safe_memcpy (&fxsave, U2P (data), sizeof (fxsave)) < 0) {
		retval = -EIO;
		goto out;
	}

	/* mxcsr bit 6 and 31-16 must be zero for security reasons */
	fxsave.mxcsr &= 0xffbf;

	/* convert the new fp state to btgeneric format */
	memset (&bc32, 0, sizeof (bc32));
	fxsave2btg (&bc32, &fxsave);

	bc32.composition_bits = CTX32_FLOATING_POINT | CTX32_XMM;

	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	/* set the context with the new values */
	btg_retval = BTGENERIC_IA32_CONTEXT_SET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);
out:
	return retval;
}

/*****************************************************************************/
static long
ptemu_setoptions (int pid, unsigned addr, unsigned data)
{
	long retval;
	const unsigned flag_mask = (unsigned) ~ (PTRACE_O_TRACEFORK |
						PTRACE_O_TRACEVFORK |
						PTRACE_O_TRACECLONE |
						PTRACE_O_TRACEEXEC);

	/* support only FORK, EXEC, CLONE and VFORK */
	DBPRINT (btl_prt_pt, "flags=%x\n", data);
	if (data & flag_mask) {
		retval = -EINVAL;
		goto out;
	}
	retval = SYSCALL (ptrace, (PTRACE_SETOPTIONS, pid, (unsigned long)addr,
				(unsigned long)data));

out:
	DBPRINT (btl_prt_pt, "retval=%ld\n", retval);
	return retval;
}
/*****************************************************************************/
static long
ptemu_geteventmsg (int pid, unsigned addr, unsigned data)
{
	long retval;
	unsigned u;
	unsigned long ul;
	retval = SYSCALL (ptrace, (PTRACE_GETEVENTMSG, pid,
				(unsigned long)addr, (unsigned long) &ul));
	if ( ! IS_SYS_ERROR (retval) ) {
		/* convert to 32 bit value */
		u = GET_LO_DWORD (ul);
		/* copy to user */
		retval = safe_memcpy (U2P(data), &u, sizeof (unsigned));
	}
	if (IS_SYS_ERROR (retval)) {
		DBPRINT (btl_prt_pt, "retval=%ld\n", retval);
	} else {
		DBPRINT (btl_prt_pt, "*addr=%d\n", u);
	}
	return retval;
}

/*****************************************************************************/

/* utility functions */
/*****************************************************************************/
/* modify a bit the ptrace field in the tls of the given debugee 
 * input:  debugee - 
 *         request - the bit to modify. 
 *         set - whether to set or clear the bit
 * 		
 */
/*****************************************************************************/
static void
set_ptrace_field (debugee_t * debugee, unsigned long request, int set)
{
	long retval;
	unsigned long dtpf, dtp;	/* address and value of the ptrace field 
					   in the debugee's tls */

	DBPRINT (btl_prt_pt, "pid=%d request=%lx set=%d\n", debugee->pid, request,
	       set);
	dtpf =
	    (unsigned long) debugee->blst + OFFSETOF (ptrace,
						      btlib_thread_data_t);
	retval = ia64_peekdata (debugee->pid, dtpf, &dtp);
	sys_err_chk (retval, "ptrace(1)");
	DBPRINT (0, "ptrace-field addr=%lx val=%lx\n", dtpf, dtp);
	if (set) {
		dtp |= request;
	} else {
		dtp &= ~request;
	}
	retval = SYSCALL (ptrace, (PTRACE_POKEDATA, debugee->pid, dtpf, dtp));
	sys_err_chk (retval, "ptrace(2)");
#if (DEBUG && 0)
	retval = ia64_peekdata (debugee->pid, dtpf, &dtp);
	sys_err_chk (retval, "ptrace(3)");
	DBPRINT (btl_prt_pt, "ptrace-field val=%lx\n", dtp);
#endif		//DEBUG
}

/*****************************************************************************/
/* set the TF flag in the IA32 context of the given debugee
 * input: debugee -
 * 	  set - wheter to set the flag or clear it
 */
/*****************************************************************************/
static void
set_trap_flag (debugee_t * debugee, int set)
{
	BT_STATUS_CODE btg_retval;
	BTGENERIC_IA32_CONTEXT bc32;
	void *remote_glst;

	DBPRINT (btl_prt_pt, "pid=%d set=%d\n", debugee->pid, set);
	/* get the address of the remote glst */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	bc32.composition_bits = CTX32_CRITICAL;
	btg_retval = BTGENERIC_IA32_CONTEXT_GET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);
	DBPRINT (0, "eflags=%x\n", bc32.eflags);
	if (set) {
		bc32.eflags |= TF_FLAG;
	} else {
		bc32.eflags &= ~TF_FLAG;
	}
	/* set the remote context with the new value */

	btg_retval = BTGENERIC_IA32_CONTEXT_SET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);
#if (DEBUG && 0)
	btg_retval = BTGENERIC_IA32_CONTEXT_GET_REMOTE (I2P (debugee->pid),
							remote_glst, &bc32);
	assert (btg_retval == BT_STATUS_SUCCESS);
	DBPRINT (btl_prt_pt, "eflags(2)=%x\n", bc32.eflags);
#endif		//DEBUG
}

/*****************************************************************************/
/* returns the address in the given context of the requested register
 * input:  addr - the address of the requested register in the user
 * 		 structure
 *         bc32 - address of the IA-32 context (btg format)
 * remarks: the context does not need to be initialized with real data
 * 	   the function only need its address.
 */
/*****************************************************************************/
static unsigned *
get_reg_addr (unsigned addr, BTGENERIC_IA32_CONTEXT * bc32)
{
	__u32 *reg_addr;

	switch (addr >> 2) {	/* shift-right by two == divide by sizeof(int) */
	case EBX:
		reg_addr = &bc32->ebx;
		break;
	case ECX:
		reg_addr = &bc32->ecx;
		break;
	case EDX:
		reg_addr = &bc32->edx;
		break;
	case ESI:
		reg_addr = &bc32->esi;
		break;
	case EDI:
		reg_addr = &bc32->edi;
		break;
	case EBP:
		reg_addr = &bc32->ebp;
		break;
	case EAX:
		reg_addr = &bc32->eax;
		break;
	case DS:
		reg_addr = &bc32->ds;
		break;
	case ES:
		reg_addr = &bc32->es;
		break;
	case FS:
		reg_addr = &bc32->fs;
		break;
	case GS:
		reg_addr = &bc32->gs;
		break;
	case EIP:
		reg_addr = &bc32->eip;
		break;
	case CS:
		reg_addr = &bc32->cs;
		break;
	case EFL:
		reg_addr = &bc32->eflags;
		break;
	case UESP:
		reg_addr = &bc32->esp;
		break;
	case SS:
		reg_addr = &bc32->ss;
		break;
	case ORIG_EAX:
		/* fall through */
	default:
		reg_addr = NULL;
	}
	return (int *) reg_addr;
}

/*****************************************************************************/
/* returns the context flags for the requested register */
/*****************************************************************************/
static unsigned
get_context_flags (unsigned addr)
{
	unsigned ctx_flags;
	switch (addr >> 2) {	/* shift-right by two == divide by sizeof(int) */
	case EDI:
	case ESI:
	case EBX:
	case EDX:
	case ECX:
	case EAX:
	case ORIG_EAX:
		ctx_flags = CTX32_GP;
		break;
	case GS:
	case FS:
	case ES:
	case DS:
		ctx_flags = CTX32_SEGMENT;
		break;
	case EBP:
	case EIP:
	case CS:
	case EFL:
	case UESP:
	case SS:
		ctx_flags = CTX32_CRITICAL;
		break;
	default:
		ctx_flags = 0;
	}
	return ctx_flags;
}

/*****************************************************************************/
/* finds a debugee-info-entry in the thread's list of debugees 
 * input:  thread - 
 * 	   pid - the pid of the debugee
 * output: pointer to the appropriate entry. NULL if not found
 */
/*****************************************************************************/
static debugee_t *
find_debugee (int pid)
{
	debugee_t *debugee = cur_thread_data->debugee_list;

	while (debugee != NULL) {
		if (debugee->pid == pid) {
			break;
		}
		debugee = debugee->next;
	}
	DBPRINT (btl_prt_pt && (debugee == NULL), "debugee=%p\n", debugee);
	return debugee;
}

/*****************************************************************************/
/* add an entry in the thread's list of debugees
 * input: thread -
 * 	  pid - the pid of the new debugee
 * 	  flags - the flags with which to init the new entry
 */
/*****************************************************************************/
static debugee_t *
add_debugee (int pid)
{
	debugee_t *debugee = NULL;

	DBPRINT (btl_prt_pt, "debugee->pid=%d\n", pid);

	/* allocate memory for the new element */
	debugee = (debugee_t *) malloc (sizeof (debugee_t));
	if (debugee == NULL) {
		DBPRINT (TRUE, "malloc failed\n");
		btlib_abort ();
	}

	/* init the new element */
	memset (debugee, 0, sizeof (*debugee));
	debugee->pid = pid;

	/* add the new element to the list */
	debugee->next = cur_thread_data->debugee_list;
	cur_thread_data->debugee_list = debugee;

	return debugee;
}

/*****************************************************************************/
/* remove an entry from the thread's list of debugees
 * input:	thread -
 * 			debugee - pointer to the element to be removed
 */
/*****************************************************************************/
static void
remove_debugee (debugee_t * debugee)
{
	debugee_t *d1, *d2;

	DBPRINT (btl_prt_pt, "debugee->pid=%d\n", debugee->pid);
	/* find the previous element in the debugee list */
	d1 = cur_thread_data->debugee_list;
	d2 = NULL;
	while (d1 != debugee) {
		d2 = d1;
		assert (d1 != NULL);
		d1 = d1->next;
	}
	/* disconect from the list */
	if (d2 == NULL) {	/* d2 now points to the prev element */
		cur_thread_data->debugee_list = debugee->next;
	} else {
		d2->next = debugee->next;
	}
	/* free the memory */
	free (debugee);
}

/*****************************************************************************/
/* get the debugee's signal-hanlder for the given signo 
 * 
 * access to the table usually requires synchronization, but here we only 
 * read one field from the table. this field is aligned to 8 bytes so reading
 * is made atomically.  so, practically - we do not need to sync here.
 */
/*****************************************************************************/
static unsigned
get_debugee_handler (debugee_t * debugee, int signo)
{
	long retval;
	unsigned handler;
	unsigned long table_addr, entry_addr, entry;

	/* get the address of the sighandlers table of the debugee */
	table_addr = get_debugee_blst_field (debugee, sighandlers);

	/* 
	 * get the address of the handler field for the given signo 
	 * entry in the table 
	 */
	entry_addr = table_addr +
		OFFSETOF (handlers, sighand_table) +
		signo * sizeof (struct signal_handler_desc) +
		OFFSETOF (sa.handler, struct signal_handler_desc);

	/* read the first 8 bytes in the entry */
	retval = ia64_peekdata (debugee->pid, entry_addr, &entry);
	sys_err_chk (retval, "ptrace(2)");

	/* the low 4 bytes hold the hanlder */
	handler = (unsigned) (entry & LO_WORD_MASK);
	DBPRINT (btl_prt_pt, "pid=%d signo=%d handler=%x\n",
	         debugee->pid, signo, handler);

	return handler;
}

/*****************************************************************************/
/* check if the debugee has handler for the given signal */
/*****************************************************************************/
static int
debugee_handles_signal (debugee_t * debugee, int signo)
{
	unsigned handler;
	int handled = 0;

	assert ((signo >= 0) && (signo < _NSIG));
	if ((signo == PTEMU_SIG0) || has_permanent_handler (signo)) {
		handled = 1;
	} else if ((signo == 0) || (signo == SIGSTOP) || (signo == SIGKILL)) {
		handled = 0;
	} else {
		/* see if the debugee has a handler for the signal */
		handler = get_debugee_handler (debugee, signo);
		DBPRINT (0, "signo=%d handler=%u\n", signo, handler);

		if ((handler != (unsigned) SIG_DFL) && 
		    (handler != (unsigned) SIG_IGN)) {
			handled = 1;
		}
	}
	DBPRINT (btl_prt_pt, "signo=%d handled=%d\n", signo, handled);

	return handled;
}

/*****************************************************************************/
/* check whether the debugee is inside a long syscall 
 * input:   pid - the pid of the debuggee
 * 	    ptctx - the context of the debuggee (where it was stopped)
 * output:  1 if the debuggee is inside a long syscall, 0 otherwise.
 * remarks: the function checks the instruction at the point where the debuggee
 *          is stopped. if its a syscall break then it concludes that the 
 *          debuggee is inside a long syscall.
 */
/*****************************************************************************/
#define SYSCALL_BREAK_OPCODE 0x1000000000ul
#define PRED_MASK 0x3Ful
#define OPCODE_MASK 0x1ffffffffe0ul

static long
is_inside_long_syscall (debugee_t * debugee, struct pt_all_user_regs *ptctx)
{
	long retval = 0;
	unsigned long ri;
	unsigned long lsf_addr;
        long lsf;


	/* check if the debuggee is inside a long syscall procedure */
	lsf_addr = get_debugee_blst_field (debugee, long_syscall_flag);
	retval = ia64_peekdata (debugee->pid, lsf_addr, &lsf); 
	sys_err_chk (retval, "reading lsf failed");

	DBPRINT (1, "lsf=%ld btl_ls_before_break_inst=%lx cr_iip=%lx\n", 
			lsf, btl_ls_before_break_inst, ptctx->cr_iip);
	if ((lsf != BEFORE_BREAK) || (ptctx->cr_iip < btl_ls_before_break_inst)) {
		/* if the debuggee is not before the invocation of
		 * the long syscall then it is not in a syscall */
		retval = 0;
		goto out;
	}

	/* we know that (iip >= before_break_inst) and that the break instruction
	 * is in slot 0 */
	ri = (ptctx->cr_ipsr >> 41) & 0x3ul;
	DBPRINT (1, "ri=%ld\n", ri);
	assert (ri < 3);

	if (ri ==1) {
		/* debuggee is one instruction pass the break instruction
		 * which means it's inside the long syscall */
		retval = 1;
	} else {
		/* ri==0 - the debuggee did not execute the break inst yet
		 * ri==2 - the debuggee already returned from the syscall */
		retval = 0;
	}

      out:
	DBPRINT (btl_prt_pt, "retval=%lx\n", retval);
	return retval;
}

/*****************************************************************************/
/* stop the given debugee at a consistant IA-32 context */
/*	input:	debugee -
 *		max_count - number of times to repeat trying in case canonization
 *			    is not possible
 *		is_local - whether the child shares vm with the parent ( it is
 *			   possible because this function is also used by
 *			   the local suspension mechanism )
 *	output: 0 on success, 1 if number of tries arrived at max_cout,
 *		-1 on failure
 *	remarks: the debugee must be susended in the 64 bit level when
 *		 calling this function
 *	assumptions: this function is called for a newly attached debugee, not
 *		     for a process which is already being debugged. this is 
 *		     important because in case of BAD_SUSPEND_STATE the debugee
 *		     is resumed and suspended using PTRACE_DETACH/ATTACH. so - 
 *		     when the debugee is resumed, signals that arrive at it will
 *		     not be visible to the debugger. this is only acceptable
 *		     if the debugee is not a 'real' debugee yet.
 *		     it is possible to call this function for a 'real' debugee
 *		     gut then the caller must make sure not to detach from
 *		     the debugee. it is achieved by passing 1 for max_count.
 */
/*****************************************************************************/
long
btlib_stop_debugee (debugee_t * debugee, unsigned max_count, int is_local,
                    int * statusp)
{
	long retval;
	BT_THREAD_SUSPEND_STATE btg_retval;
	struct pt_all_user_regs ptctx;
	BTGENERIC_IA64_CONTEXT bc64;
	int status, pid;
	unsigned try_count = 1;
	void *remote_glst;
	unsigned long rbs_top, rbs_bottom, orig_bspstore, dummy UNUSED;
	unsigned long dpf;

	DBPRINT (btl_prt_pt | btl_prt_thread,
	         "pid=%d max_count=%u is_local=%d\n",
	         debugee->pid, max_count, is_local);

	/* clear the self canonize indication */
	debugee->flags &= ~PTEMU_SELF_CANONIZE;

	/* if debuggee is in-cont-process then put a suspension request
	 * and wait for it stop itself */
	dpf = get_debugee_blst_field (debugee, ptrace);
	if (dpf & PTEMU_IN_CONT_PROCESS) {
		DBPRINT (btl_prt_pt, "PTEMU_IN_CONT_PROCESS\n");

		/* put a suspension request */
		set_ptrace_field (debugee, PTEMU_SUSPENSION_REQ, TRUE);

		/* continue debuggee's execution */
		retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 0ul));
		sys_err_chk (retval, "ptrace cont failed (1)");

		/* wait for it to stop with ptemu-signal */
		retval = wait_for_debugee_to_stop_itself (debugee, &status);
		if (IS_SYS_ERROR (retval)) {
			*statusp = status;
		}
		goto out;
	}
repeat_canonization:

	/* get the thread's 64 bit context */
	retval = SYSCALL (ptrace, (PTRACE_GETREGS, debugee->pid, 0ul,
				(unsigned long) &ptctx));
	sys_err_chk (retval, "ptrace getregs");

	/* keep this value in case we need to change the 64 bit
	 * state to a new state that btgeneric passes to us */
	rbs_top = ptctx.ar[PT_AUR_BSP];

	/* convert the 64 bit state representation into something that
	 * btgeneric understands */
	ptctx2btgctx_64 (&ptctx, &bc64);

	/* 
	 * btgeneric will return the bsp of the current frame. linux expects
	 * it to be the bsp of the next frame. since we don't know how to
	 * convert from current frame to next frame - we cannot deal with
	 * cases where btgeneric changes the bsp. we have to assume that 
	 * btgeneric will not change the bsp, so we keep the bsp value
	 * before the canonization, and compare it with the value after
	 * the canonization to see if it changes. */
	rbs_bottom = bc64.bsp;
	/* btgeneric expects bspstore to be equal to bsp. the conversion
	 * function (ptctx2btgctx_64) does that, but we need to keep the
	 * orig value of bspstore in case we need to change the 64 bit
	 * state to a new state that btgeneric passes to us */

	orig_bspstore = ptctx.ar[PT_AUR_BSPSTORE];
	DBPRINT (btl_prt_pt, "rbs_bottom=%lx rbs_top=%lx\n", rbs_bottom, rbs_top);
	/* get the glst of the debugee */
	remote_glst = get_debugee_glst_address (debugee);
	assert (!IS_SYS_ERROR (remote_glst));

	/* this is an UGLY work around to a problem 2.4 has with 
	 * changing the bsp */
	if (m_is_kernel_2_6) {
		/* force the kernel to write the stacked regsiters to memory in 2.6 */
		retval = SYSCALL (ptrace, (PTRACE_POKEUSR, debugee->pid, PT_AR_BSP, rbs_top - 8));
		sys_err_chk (retval, "ptrace pokeusr 1");
		retval = SYSCALL (ptrace, (PTRACE_POKEUSR, debugee->pid, PT_AR_BSP, rbs_top));
		sys_err_chk (retval, "ptrace pokeusr 2");
	} else {
		/* force the kernel to write the stacked regsiters to memory in 2.4 */
		retval = ia64_peekdata (debugee->pid, rbs_bottom, &dummy);
		sys_err_chk (retval, "ia64_peekdata");
	}


	/* 
	 * suspend the debugee in a canonized IA-32 state 
	 * the last arg to this api functino is prev_suspend_count. this
	 * value is used by btgeneric for sync of processes that try to
	 * suspend the execution of the debugee. 
	 * in linux thread suspension is done by using ptrace - it is not
	 * possible for more than one process to try and suspend another
	 * process. so this argument is useless in linux and should always
	 * set to 0.
	 */
	btg_retval = is_local ? 
		BTGENERIC_CANONIZE_SUSPEND_CONTEXT (remote_glst, &bc64, 0) : 
		BTGENERIC_CANONIZE_SUSPEND_CONTEXT_REMOTE ((void *) (long)debugee->pid,
						 	   remote_glst, &bc64, 0);

	DBPRINT (btl_prt_pt | btl_prt_thread,
			"canonize_context=%u\n", btg_retval);
	DBCODE (0, print_btgcontext64 (&bc64));

	switch (btg_retval) {
	case SUSPEND_STATE_CONSISTENT:
		/* 
		 * thread suspended in a consistent IA32 state 
		 * no changes in IA64Context
		 */
		if (is_inside_long_syscall (debugee, &ptctx)) {

			/* if it is inside a long syscall then stop it */

			DBPRINT (btl_prt_pt | btl_prt_thread, "inside syscall\n");
			set_ptrace_field (debugee, PTEMU_IN_CONT_PROCESS, TRUE);
			set_ptrace_field (debugee, PTEMU_SUSPENSION_REQ, TRUE);
			set_ptrace_field (debugee, PTEMU_CONT, TRUE);
			/* continue debuggee's execution */
			retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 
						(unsigned long)PTEMU_SIG0));
			sys_err_chk (retval, "ptrace cont failed (1)");
			retval = wait_for_debugee_to_stop_itself (debugee, &status);
			if (IS_SYS_ERROR (retval)) {
				*statusp = status;
				return -ESRCH;
			}
			break;
		} else if (get_debugee_blst_field(debugee, is_self_suspended)) {
			/* if it is marked as self suspended then stop it.
			a thread is marking itself as self suspended when waiting
			for the suspension cs (BTL_LOCK_IND_SUSPENSION).
			currently the suspension cs is shared in btlib between core files
			and affinity */
			break;
		}
		/* 
		 * if the debugee is not inside a long syscall or is not marked as self suspended
		 * then treat it as if it cannot be suspended right now
		 *
		 * ------------ fall through ------------  
		 */
	case BAD_SUSPEND_STATE:
		/* 
		 * recoverable error: 
		 * thread can not be suspended in the current state
		 */
		try_count++;
		DBPRINT (btl_prt_pt | btl_prt_thread, "try_count=%u\n", try_count); 
		if ((max_count > 0) && (try_count > max_count)) {
			retval = 1;
			break;
		}
		/* 
		 * make the child continue, yield execution and stop it
		 * again, in the hope that this time - suspension will
		 * succeed
		 */
		retval = SYSCALL (ptrace, (PTRACE_DETACH, debugee->pid, 0ul, 0ul));
		sys_err_chk (retval, "sys_ptrace detach");
		YIELD_THREAD_EXECUTION ();
		retval = SYSCALL (ptrace, (PTRACE_ATTACH, debugee->pid, 0ul, 0ul));
		sys_err_chk (retval, "ptrace attach");
		pid = wait_for_debugee (debugee, &status);
		sys_err_chk ((long) pid, "sys_waitpid");
		assert (pid == debugee->pid);
		/* if the debugee died then we're out of here */
		if (!__WIFSTOPPED (status)) {
			*statusp = status;
			retval = -ESRCH;
			break;
		}
		goto repeat_canonization;
		assert (!"should not get here");

	case SUSPEND_STATE_CANONIZED:
		/* IA32 thread state canonized there are changes in IA64Context */
		assert (is_local == 1);

		btgctx2ptctx_64 (&bc64, &ptctx);
		/* special treatment for bsp */
		assert (bc64.bsp == rbs_bottom);
		ptctx.ar[PT_AUR_BSP] = rbs_top;
		ptctx.ar[PT_AUR_BSPSTORE] = orig_bspstore;

		retval = SYSCALL (ptrace, (PTRACE_SETREGS, debugee->pid, 0ul, 
					(unsigned long) &ptctx));
		sys_err_chk (retval, "ptrace setregs");

		break;
	case SUSPEND_STATE_READY_FOR_CANONIZATION:
		/* 
		 * target thread is ready to canonize IA32 state by itself
		 * there are changes in IA64Context
		 */
		assert (is_local == 0);
		/* set the 64 bit context of the debugee to the given context */
		btgctx2ptctx_64 (&bc64, &ptctx);
		/* special treatment for bsp */
		assert (bc64.bsp == rbs_bottom);
		ptctx.ar[PT_AUR_BSP] = rbs_top;
		ptctx.ar[PT_AUR_BSPSTORE] = orig_bspstore;

		retval = SYSCALL (ptrace, (PTRACE_SETREGS, debugee->pid, 0ul,
				 (unsigned long) &ptctx));
		sys_err_chk (retval, "ptrace setregs");

		retval = sync_self_canonization (debugee);
		if (retval == 0) {
			debugee->flags |= PTEMU_SELF_CANONIZE;
		}
		break;
	case SUSPEND_STATE_INACCESIBLE:
		/* fatal error: thread can not be suspended */
		assert (!"SUSPEND_STATE_INACCESIBLE");
		retval = -1;
		break;
	default:
		assert (!"unkonwn suspension state");
	}
out:
	DBPRINT (btl_prt_pt | btl_prt_thread, "retval=%ld\n", retval);
	return retval;
}

/*****************************************************************************/
/* try to add a child to the debugee list.
 * input:   thread -
 *          child_pid - the pid of the child
 * output:  pointer to the new element in the list, or NULL if no
 *          element was added
 * remarks: the child is assumed to be suspended at the 64 bit level
 */
/*****************************************************************************/
static debugee_t *
handle_new_debugee (int child_pid, int *statusp)
{
	debugee_t *debugee;
	unsigned long ldata;
	long retval;
	int status, is_process = TRUE;

	DBPRINT (btl_prt_pt, "child=%d\n", child_pid);

	/* 
	 * try and get the debugee's tls -
	 * the value that we get here is really meaningless. we only
	 * want to see that we can ptrace the child, because if we
	 * can't then the child is not a debugee of the current
	 * process and we have nothing to do here
	 */
	retval = ia64_peekuser (child_pid, (unsigned long) PT_R13, &ldata);
	if (retval == -ESRCH) {
		DBPRINT (btl_prt_pt, "cannot ptrace child %d\n", child_pid);
		return NULL;
	}
	sys_err_chk (retval, "ia64_peekuser");

	/* 
	 * see that the child is an IA-32 process - if it's not then
	 * we have nothing to do here
	 */
	if (!is_ia32_process (child_pid)) {
		DBPRINT (btl_prt_pt, "not IA-32 process child %d\n", child_pid);
		return NULL;
	}

	/* 
	 * if we got here the the child is an IA-32 process who 
	 * is ptrace by the current process
	 */

	/* add an entry to the debugee list */
	debugee = add_debugee (child_pid);

	/* see if btlib of the debugee is already initialized */
	if ( ! is_debugee_ready (debugee) ) {
		unsigned long is_ptraced;
		retval = ia64_peekdata (child_pid, (unsigned long)&btl_is_ptraced, 
				&is_ptraced);
		sys_err_chk (retval, "ptrace peek data");

		if (is_ptraced == IS_PTRACED_NO) {
			/*
			 * if debugee is not ready then the child process issued
			 * ptrace_tracme and was stopped by the os just before starting
			 * execution. set a global flag at the debuggee and
			 * make it continue so btlib can initialize itself.
			 * at the end of initialization, btlib of the debugee will send
			 * itself sigtrap. we will wait for this event and only then will
			 * we return at the debugger
			 */
			retval = SYSCALL (ptrace, (PTRACE_POKEDATA, child_pid, 
						   (unsigned long)&btl_is_ptraced, 
						   IS_PTRACED_YES_NOT_READY));
			sys_err_chk (retval, "ptrace poke data");
		} else {
			/* a new thread was created at the debuggee. mark its
			 * ptrace field in the tls and wait for it to stop itself
			 * in clone_main */
			set_debugee_blst_field (debugee, ptrace, PTEMU_PTRACED);
			is_process = FALSE;
		}
		retval = SYSCALL (ptrace, (PTRACE_CONT, child_pid, 0ul, 0ul));
		sys_err_chk (retval, "ptrace cont");
		
		int loop = TRUE;
		while (loop) {
			retval = wait_for_debugee (debugee, &status);
			assert (retval == child_pid);
			/* if the debugee terminated then the wait syscall should return
			 * at the debugger */
			if ( IS_SYS_ERROR (retval) || ! __WIFSTOPPED (status) ) {
				DBPRINT (btl_prt_pt, "debugee died\n");
				remove_debugee (debugee);
				*statusp = status;
				return NULL;
			}
			DBPRINT (btl_prt_pt, "WSTOPSIG=%d\n", __WSTOPSIG (status));
			if (__WSTOPSIG (status) == PTEMU_SIG0) {
				loop = FALSE;
			} else {
				/* resend all other signals but ptemu-signal */
				DBPRINT (btl_prt_pt, "resend=%d\n", __WSTOPSIG (status));
				retval = SYSCALL (ptrace, (PTRACE_CONT, child_pid, 0ul, 0ul));
				sys_err_chk (retval, "ptrace cont(2) failed");
				YIELD_THREAD_EXECUTION ();
				retval = SYSCALL (tkill, (debugee->pid, __WSTOPSIG (status)));
				sys_err_chk (retval, "tkill failed");
			}
		} 
		debugee->flags |= PTEMU_NEW;
	}
	/* now that btlib finished initialization the tls pointer is ready */
	debugee->blst = (unsigned long) get_debugee_blst_address (debugee);
	assert (is_debugee_ready (debugee));

	/* TODO : it seems to interfer with the work of btlib_ptemu0_handler.
	 * why do we do it here ? */
	retval = SYSCALL (ptrace, (PTRACE_POKEDATA, child_pid, 
				   (unsigned long)&btl_is_ptraced, 
				   IS_PTRACED_YES));

	/* finish */
	DBPRINT (btl_prt_pt, "added a new debugee %d tls=%lx\n",
		 debugee->pid, debugee->blst);
	return debugee;
}

/*****************************************************************************/
/*
 * how do we know that the debugee is valid ?
 * ------------------------------------------------------------------
 * 1 -	r13 points to a valid area in memory
 * 2 -	at the offset of the pid field from r13 we find the pid of the
 * 	debugee
 */
/*****************************************************************************/
static int
is_debugee_ready (debugee_t * debugee)
{
	long retval;
	unsigned long ldata, addr;
	int pid;
	int is_ready = 0;

	DBPRINT (0, "debugee->blst=%lx\n", debugee->blst);

	if (debugee->blst == 0ul) {
		debugee->blst = (unsigned long) get_debugee_blst_address (debugee);
		if (IS_SYS_ERROR (debugee->blst)) {
			goto out;
		}
	}
	DBPRINT (0, "debugee->blst=%lx\n", debugee->blst);

	addr = debugee->blst + OFFSETOF (tid, btlib_thread_data_t);
	retval = ia64_peekdata (debugee->pid, addr, &ldata);

	DBPRINT (IS_SYS_ERROR (retval), "retval=%ld\n", retval);
	DBPRINT (btl_prt_pt && !IS_SYS_ERROR (retval), "addr=%lx ldata=%lx\n", addr, ldata);

	if (retval == 0) {
		pid = (int) ldata;
		if (pid == debugee->pid) {
			is_ready = 1;
		}
	}
out:
	DBPRINT (btl_prt_pt, "pid=%d return=%d\n", debugee->pid, is_ready);
	return is_ready;
}

/*****************************************************************************/
/*
 * how do we know that a process is an IA-32 process ?
 * --------------------------------------------------
 *
 * we read memory allocated for IA-32 stack.
 * if the operation succeeds then the memory is mapped,
 * that is - this is an IA-32 process
 *
 */
/*****************************************************************************/
static int
is_ia32_process (int pid UNUSED)
{
	long retval = TRUE;
#if 0
	unsigned long tos;

	retval =
	    ia64_peekdata (pid, (unsigned long) (IA32_TASK_SIZE - 8), &tos);

	retval = !retval;	/* make it 0 or 1 */

	DBPRINT (btl_prt_pt, "pid=%d retval=%ld\n", pid, retval);

#endif
	return retval;
}
/*****************************************************************************/
/* 
 * let the debugee canonize itself
 * input :	debugee -
 * 		bc64 - 64 bit context that was returned by
 * 		       btg_canonize_supend_context
 * ouput:	0 on success, -1 on failure
 * remarks:	this function implements a sync mechanism that
 * 		allows the debugee to canonize its own IA-32 context
 * 		when it is stopped at the 64 bit level.
 * 		the algo details follow:
 *
 * preconditions to calling this function:
 * 	1) the debugee is suspended at the 64 bit level
 * 	2) the debugger called btgeneric_canonize_suspend_context_remote
 * 	   and got the 64 bit context that is given in the bc64 param.
 * 	   what btgeneric did in this api function is to prepare things
 * 	   in a way that when it (the debugee) resumes execution with
 * 	   the bc64 context it will canonize its IA-32 context and call
 * 	   BTLIB_REENTER
 * 	3) btlib already set the 64 bit context to the new one
 *
 * algo:
 *
 * 	debugger					debugee
 *	--------					-------
 *
 * 1	create sem set [READY,RESUME]
 * 	initialized to blocking state
 *
 * 2	write the sem key at the debugee's
 * 	tls
 *
 * 3	write a suspenstion request at the
 * 	debugee's tls
 *
 * 5	resume execution of debugee
 *
 * 6	wait on sem[READY]
 *
 * 7							canonize IA-32 context (btgeneric)
 *
 * 8							call BTLIB_REENTER (btgeneric)
 *
 * 9							in btrans_jump - if suspension
 * 							request is pending then 
 * 10							release sem[READY]
 *
 * 11							wait on sem[RESUME]
 *
 * 12	suspend debugee
 *
 * 13	release sem[RESUME]
 *
 * 14							(now when the debugee's execution
 * 							resumes it will find sem[RESUME]
 * 							released and it will continue as
 *							usual - not forgetting to close
 * 							the sem set )
 */
/*****************************************************************************/
static long
sync_self_canonization (debugee_t * debugee)
{
	key_t semkey;
	long retval;
	int semid;
	union semun semsetarg;
	unsigned short int sem_init_array[2] = { 0, 0 };
	struct sembuf semops[2];
	int status;

	DBPRINT (btl_prt_pt, "pid=%d\n", debugee->pid);
	/*
	 * Generate the semaphore set key and create the semaphore set.
	 * The set includes two semaphores.  One on which the master waits.
	 * The other on which the slave waits.
	 */
	semkey = (key_t) SYSCALL (gettid, ());
	semkey += (key_t) (SYSCALL (times, (NULL)) & 0xFFFFFFFFul);
	semid = SYSCALL (semget, (semkey, 2, IPC_CREAT | IPC_EXCL | 0600));
	sys_err_chk ((long) semid, "semget");

	/*
	 * Initialize the semaphore set.
	 * The semaphores are set to 0, so as not to allow anyone to pass through.
	 */
	semsetarg.array = sem_init_array;
	retval = SYSCALL (semctl, (semid, 0, SETALL, semsetarg));
	sys_err_chk (retval, "semctl");

	/* Write the semaphore key to the slave's memory */
	set_debugee_blst_field (debugee, self_canon_semkey, semkey);

	/* write a suspension request to the debugee */
	set_ptrace_field (debugee, PTEMU_SELF_CANONIZE, 1);

	/*
	 * Resume the debugee's execution 
	 * use PTRACE_DETACH and PTRACE_ATTACH to resume/stop
	 * debugee's execution to avoid a deadlock :
	 *  if the debugee receives a signal while the debugger
	 *  waits for it to release a semaphore then the os will
	 *  stop the debugee and wait for the debugger to examine
	 *  its status - but the debugger is waiting for the debugee
	 *  to release the semaphore
	 * there is an assumption here - sync signals are not possible
	 * during state canonization
	 */
	retval = SYSCALL (ptrace, (PTRACE_DETACH, debugee->pid, 0ul, 0ul));
	sys_err_chk (retval, "ptrace detach");

	DBPRINT (btl_prt_pt, "before wait\n");

	/* wait for the completion of the self canonization */
	semops[0].sem_num = SEMREADY;
	semops[0].sem_op = -1;
	semops[0].sem_flg = 0;
	retval = SYSCALL (semop, (semid, semops, 1));
	/* don't let signals interrupt us here */
	do {
		retval = SYSCALL (semop, (semid, semops, 1));
	} while (retval == -EINTR);
	sys_err_chk (retval, "semop");

	DBPRINT (btl_prt_pt, "after wait\n");

	/* Suspend the debugee */
	retval = SYSCALL (ptrace, (PTRACE_ATTACH, debugee->pid, 0ul, 0ul));
	sys_err_chk (retval, "ptrace attach");
	retval = wait_for_debugee (debugee, &status);
	sys_err_chk (retval, "sys_waitpid");
	assert (retval == debugee->pid);
	assert (__WIFSTOPPED (status));

	/* 
	 * relase semaphore to allow the debugee to continue when its 
	 * execution resumes 
	 */
	semops[0].sem_num = SEMRESUME;
	semops[0].sem_op = 1;
	semops[0].sem_flg = 0;
	retval = SYSCALL (semop, (semid, semops, 1));
	sys_err_chk (retval, "semop");

	return 0;
}

/*****************************************************************************/
/* write a memory area of a debugee
 * input:  pid - of the debugee
 * 	   addr - base address of the area to be written
 * 	   buffer - the memory contents will be copied from here
 * 	   size - size of memory area
 */
/*****************************************************************************/
long
btlib_pokedata (int pid, void *addr, const char *buffer, unsigned long size)
{
	unsigned long ldata;
	long retval = 0;
	union {
		const char *cp;
		unsigned *up;
		unsigned long *ulp;
	} cp2ulp;
	cp2ulp.cp = 0ul;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	DBPRINT (0, "pid=%d, addr=%p, buffer=%p, size=%ld\n", pid, addr, buffer,
		 size);

	assert (((unsigned long) addr & 3) == 0);	/* addr aligned at 4 bytes */
	assert ((size & 3) == 0);	/* size aligned at 4 bytes */

	/* 
	 * write 8 bytes at a time. at the end - if a smaller
	 * piece is left then read it in poke_tail.
	 * 
	 * we use the input params as the automatic variables - 
	 * 'addr' points at the next addr to write to
	 * 'buffer' points at the next addr to read from
	 * 'size' is the remaining amout to written
	 *
	 * no need to flush the instruction cache after we write
	 * to the memory. ia32exec.bin will handle it correctly
	 */
	cp2ulp.cp = buffer;
	while (size >= sizeof (long)) {
		retval = SYSCALL (ptrace, (PTRACE_POKEDATA, pid, 
					   (unsigned long) addr, *cp2ulp.ulp));
		if (retval != 0) {
			DBPRINT (btl_prt_pt, "ptrace_pokedata failed %ld\n",
				 retval);
			return retval;
		}
		addr += sizeof (long);
		cp2ulp.ulp++;
		size -= sizeof (long);
	}

	/* poke_tail */
	if (size > 0) {
		assert (size == 4);
		DBPRINT (0, "poke_tail size=%ld datap=%p\n", size, cp2ulp.up);
		/* read a qword from the debugee */
		retval = ia64_peekdata (pid, (unsigned long) addr, &ldata);
		if (retval != 0) {
			DBPRINT (btl_prt_pt, "ia64_peekdata failed %ld\n",
				 retval);
			return retval;
		}
		/* deposit the remining bytes in the qword */
		SET_LO_DWORD (&ldata, *cp2ulp.up);
		/* write the resulted qword back to the debugee */
		retval = SYSCALL (ptrace, (PTRACE_POKEDATA, 
					   pid, (unsigned long) addr, ldata));
		if (retval != 0) {
			DBPRINT (btl_prt_pt, "ptrace_pokedata failed %ld\n",
			       retval);
		}
	}
	return retval;
}

/*****************************************************************************/
/* read a memory area from a debugee
 * input:  pid - of the debugee
 *         addr - base address of the area to be read
 *         buffer - the memory contents will be copied to here
 *         size - size of memory area
 */
/*****************************************************************************/
long
btlib_peekdata (int pid, void *addr, void *buffer, unsigned long size)
{
	long retval;
	union {
		char *cp;
		unsigned long *ulp;
		unsigned *up;
	} cp2up;
	cp2up.cp = 0ul;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	DBPRINT (0, "pid=%d, addr=%p, buffer=%p, size=%ld\n", pid, addr, buffer,
		 size);

	assert (((unsigned long) addr & 3) == 0);	/* addr aligned at 4 bytes */
	assert ((size & 3) == 0);	/* size aligned at 4 bytes */
	/* 
	 * read 8 bytes at a time. at the end - if a smaller
	 * piece is left then read it in peek_tail.
	 * 
	 * we use the input params as the automatic variables - 
	 * 'addr' points at the next addr to read from
	 * 'buffer' points at the next addr to write to
	 * 'size' is the remaining amout to read
	 */
	cp2up.cp = buffer;
	while (size >= sizeof (long)) {
		retval = ia64_peekdata (pid, (unsigned long) addr, cp2up.ulp);
		if (retval != 0) {
			return retval;
		}
		addr += sizeof (long);
		cp2up.ulp++;
		size -= sizeof (long);
	}
	/* peek tail */
	if (size > 0) {
		unsigned long ldata;
		assert (size == 4);
		DBPRINT (0, "peek_tail size=%ld datap=%p\n", size, cp2up.up);
		retval = ia64_peekdata (pid, (unsigned long) addr, &ldata);
		*cp2up.up = (unsigned) ldata;
	}
	return 0;
}

/*****************************************************************************/
/* get 64 bit register in the context of a ptraced process
 * input:  pid - the process
 *         reg - the address of the register as defined in asm/user.h
 *         dest - address where the register's value is to be written
 * output: 0 on success, negative error code on failure
 */
/*****************************************************************************/
long
ia64_peekuser (int pid, unsigned long reg, unsigned long *dest)
{
	long error;
	long retval;

	retval = SYSCALL (ptrace, (PTRACE_PEEKUSR, pid, reg, 0ul));
	__asm__ __volatile__ ("mov %0=r10;;" : "=r"(error));
	if (error == 0) {
		DBPRINT (0, "retval=%ld(%lx)\n", retval, retval);
		*dest = retval;
		retval = 0;
	}
	DBPRINT (error, "peekuser failed:%ld\n", retval);
	return retval;
}

/*****************************************************************************/
/* get 8 bytes of memory contents from a ptraced process
 * input:  pid - the process
 *         addr - the address for reading
 *         dest - address where the memory contents is to be written
 * output: 0 on success, negative error code on failure
 */
/*****************************************************************************/
long
ia64_peekdata (int pid, unsigned long addr, unsigned long *dest)
{
        long retval;
        long error;

        retval = SYSCALL (ptrace, (PTRACE_PEEKDATA, pid, addr, 0ul));
        __asm__ __volatile__ ("mov %0=r10;;" : "=r"(error));
        if (error == 0) {
		DBPRINT (0, "retval=%ld(%lx)\n", retval, retval);
                *dest = retval;
                retval = 0;
        }
	DBPRINT (error, "peekdata failed:%ld\n", retval);
        return retval;
}

/*****************************************************************************/
/* init tia_signal field with cont_signo
 * input: cont_signo - the signal the should restart the debuggee
 */
/*****************************************************************************/
void
btlib_prepare_cont_signo (int cont_signo)
{
	DBPRINT (btl_prt_pt, "cont_signo=%d\n", cont_signo);

	/* init the tia_signal field with cont_signo
	 * and call the handler
	 */
	memset (&cur_thread_data->tia_signal, 0,
		sizeof (cur_thread_data->tia_signal));
	cur_thread_data->tia_signal.siginfo.si_signo = cont_signo;
	cur_thread_data->tia_signal.siginfo.si_code = SI_USER;
}

/*****************************************************************************/
/*  continue the execution of the given debugee with the given signal
 *  this function is used by ptemu_syscall, ptemu_singlestep and ptemu_cont
 */
/*****************************************************************************/
static long
continue_debugee (debugee_t * debugee, int cont_signo, int detach)
{
	long retval;
	int signo, stop_signo;
	int is_self_canonize, is_in_cont_proc;

	stop_signo = debugee->stop_signo;
	debugee->stop_signo = 0;

	DBPRINT (btl_prt_pt, "cont_signo=%d stop_signo=%d\n", 
			cont_signo, stop_signo);

	/* the cont_signo of the debugee should be cleared */
	assert (get_cont_signo (debugee) == 0);

	is_self_canonize = (debugee->flags & PTEMU_SELF_CANONIZE) ? 1 : 0;
	debugee->flags &= ~ PTEMU_SELF_CANONIZE;
	DBPRINT (btl_prt_pt, "is_self_canonize=%d\n", is_self_canonize);

	/* choose the signal to continue with */

			/* in case of self canonization it is imperative to
			 * continue with signal 0, so that the debuggee has
			 * the chance to complete the self-canonization
			 * procedure */
	if (is_self_canonize ||
			/* in case the signal that stopped the debuggee
			 * is the same as the one that resume its execution
			 * then let it continue. it's either towards the end
			 * of btlib signal handler or towards the end of 
			 * btlib_api_ia32_inerrupt */
			(cont_signo == stop_signo) ||
			/* if the debuggee stopped due to PTRACE_SYSCALL and
			 * is resumed with no special signal then it is already
			 * at the point from which it should continue */
			((stop_signo & IA32_SYSCALL_INT) && (cont_signo == 0))) {
		signo = 0;
	} else {
		/* restart the debuggee */
		signo = PTEMU_SIG0;
	}

	DBPRINT (btl_prt_pt, "signo=%d\n", signo);
	if ((is_self_canonize || (signo == PTEMU_SIG0)) && cont_signo) {
		/* in these cases the debuggee will look for cont_signo
		 * in the ptrace field */
		set_cont_signo (debugee, cont_signo);
	}

	/* mark the debuggee with the in-proc-cont flag */
	if ( ! detach ) {
		is_in_cont_proc = 1;
		if (stop_signo & IA32_SYSCALL_INT) {
			/* the only case were we're not sure whether the debuggee
			 * will get to btg-run is when it is before a long syscall */
			long lsf = get_debugee_blst_field (debugee, long_syscall_flag);
			if (lsf != NOT_IN_SYSCALL) {
				is_in_cont_proc = 0;
			}
		}
		DBPRINT (btl_prt_pt, "is_in_cont_proc=%d\n", is_in_cont_proc);
		if (is_in_cont_proc) {
			/* all other cases should end-up either in btg-run
			 * or in restart_btg, which are debuggee's end-points for
			 * the in-cont-proc state. actually, the debuggee might
			 * also end-up in cancellation_signal, but btlib_stop_debugee,
			 * (thru wait_for_debugee_to_stop_itself) will handle
			 * this case gracefully */
			set_ptrace_field (debugee, PTEMU_IN_CONT_PROCESS, TRUE);
		}
	}

	/* mark the debuggee as detached */
	if ( detach && (cur_thread_data->debugee_list == debugee && debugee->next == NULL)) {
		if (signo == PTEMU_SIG0) {
			set_ptrace_field (debugee, PTEMU_DETACH_LAST_THREAD, TRUE);
		} else {
			/* if this is the last debuggee then clear the global flag */
			DBPRINT (btl_prt_pt, "last debuggee\n");
			retval = SYSCALL (ptrace, (PTRACE_POKEDATA, debugee->pid, 
						   (unsigned long)&btl_is_ptraced, 
						   IS_PTRACED_NO));
			/* TODO: what if a signal arrives at the debuggee in the time
			 * between clearing the flag and the actual detach ? */
			sys_err_chk (retval, "clearing btl_is_ptraced failed");
		}
	}

	retval = SYSCALL (ptrace, (detach ? PTRACE_DETACH : PTRACE_CONT, 
				   debugee->pid, 0ul, (unsigned long) signo));

	DBPRINT (btl_prt_pt, "retval=%ld\n", retval);

	return retval;
}

/*****************************************************************************/
/*
 * sigcontext64_to_btrans_context
 *
 * Purpose: convert sigcontext (64) to BTGENERIC_IA64_CONTEXT
 *
 */
/*****************************************************************************/
static void
btgctx2ptctx_64 (BTGENERIC_IA64_CONTEXT * bc64, struct pt_all_user_regs *ptctx)
{
	assert (ptctx != NULL);
	assert (bc64 != NULL);

	assert (bc64->composition_bits ==
		(CTX64_CRITICAL | CTX64_FLOATING_POINT | CTX64_INTEGER |
		 CTX64_IA32_CRITICAL));
	memset (ptctx, 0, sizeof (*ptctx));

	__asm__ ("stf.spill %0=f0%P0" : "=m"(ptctx->fr[0]));
	__asm__ ("stf.spill %0=f1%P0" : "=m"(ptctx->fr[1]));
	memcpy ((void *) &(ptctx->fr[2]), (void *) &bc64->fp_static[0],
		126 * sizeof (struct ia64_fpreg));

	ptctx->gr[0] = 0ul;
	memcpy ((void *) &(ptctx->gr[1]), (void *) &bc64->int_static[0],
		31 * sizeof (long));

	memcpy ((void *) &(ptctx->br[0]), (void *) &bc64->br_regs[0],
		8 * sizeof (long));

	/* IIP will always be aligned (slot in PSR) */
	ptctx->cr_iip = bc64->iip;
	ptctx->cr_ipsr = bc64->ipsr;

	ptctx->nat = bc64->nat_collection;

	ptctx->pr = bc64->predicates;
	ptctx->ar[PT_AUR_FPSR] = bc64->fpsr;
	ptctx->ar[PT_AUR_UNAT] = bc64->br_regs[7];
	ptctx->ar[PT_AUR_LC] = bc64->lc;

	ptctx->ar[PT_AUR_CCV] = bc64->ccv;

	ptctx->ar[PT_AUR_PFS] = bc64->pfs;
	ptctx->ar[PT_AUR_BSPSTORE] = bc64->bspstore;
	/* 
	 * dont put the bsp value in bc64 - btgeneric puts there
	 * the current bsp, while the linux kernel expects a value
	 * as if the user issued 'cover'
	 */
	ptctx->ar[PT_AUR_BSP] = 0ul;

	ptctx->ar[PT_AUR_RSC] = bc64->rsc;
	ptctx->ar[PT_AUR_RNAT] = bc64->rnat;
}

/*****************************************************************************/
/*
 * sigcontext64_to_btrans_context
 *
 * Purpose: convert sigcontext (64) to BTGENERIC_IA64_CONTEXT
 *
 */
/*****************************************************************************/
static void
ptctx2btgctx_64 (struct pt_all_user_regs *ptctx, BTGENERIC_IA64_CONTEXT * bc64)
{

	assert (ptctx != NULL);
	assert (bc64 != NULL);

	memset (bc64, 0, sizeof (*bc64));


	memcpy ((void *) &(bc64->fp_static[0]), (void *) &ptctx->fr[2],
		126 * sizeof (struct ia64_fpreg));
	memcpy ((void *) &(bc64->int_static[0]), (void *) &ptctx->gr[1],
		31 * sizeof (unsigned long));
	memcpy ((void *) &(bc64->br_regs[0]), (void *) &ptctx->br[0],
		8 * sizeof (unsigned long));

	bc64->nat_collection = ptctx->nat;
	/* IIP will always be aligned (slot in PSR) */
	bc64->iip = (ptctx->cr_iip & 0xFFFFFFFFFFFFFFF0UL);
	bc64->ipsr = ptctx->cr_ipsr;

	bc64->predicates = ptctx->pr;
	bc64->fpsr = ptctx->ar[PT_AUR_FPSR];
	bc64->br_regs[7] = ptctx->ar[PT_AUR_UNAT];
	bc64->lc = ptctx->ar[PT_AUR_LC];

	bc64->ccv = ptctx->ar[PT_AUR_CCV];

	bc64->pfs = ptctx->ar[PT_AUR_PFS];
	/* bsp - the os returns the end of the rbs, not the
	   beginning, which is what we want */
	bc64->bsp = (unsigned long)
		get_interrupted_bsp (ptctx->ar[PT_AUR_BSP], ptctx->cfm);

	bc64->bspstore = bc64->bsp;
	bc64->rsc = ptctx->ar[PT_AUR_RSC];
	bc64->rnat = ptctx->ar[PT_AUR_RNAT];

	bc64->composition_bits = CTX64_CRITICAL |
		CTX64_FLOATING_POINT | CTX64_INTEGER | CTX64_IA32_CRITICAL;
}

/*****************************************************************************/
/*****************************************************************************/
void
btlib_init_ptemu_module ()
{
	DBPRINT (btl_prt_pt, "btl_is_ptraced=%ld\n", btl_is_ptraced);
	m_is_kernel_2_6 = is_kernel_2_6 ();
}
/*****************************************************************************/
/*****************************************************************************/
void btlib_emulate_trap ()
{
	long retval;
	DBPRINT (0, "syscall_id=%d\n", cur_thread_data->syscall_id);
	cur_thread_data->tia_signal.siginfo.si_signo = SIGSTOP;
	retval = SYSCALL (tkill, (cur_thread_data->tid, PTEMU_SIG0));
	sys_err_chk (retval, "kill");
	/* reenter btgeneric - the debugger already stopped the debuggee
	 * in a consistent 32 bit state (ptemu_attach) */
	restart_btg (NULL, NULL);
}
/*****************************************************************************/
/* output: a negative error code in case of faiure. 0 on success */
/*****************************************************************************/
static long wait_for_debugee_to_stop_itself (debugee_t * debugee,
                                                    int * statusp)
{
	int debugee_stopped_itself = 0;
	int stop_sig;
	long retval;
	unsigned char signals[SIGRTMAX + 1] = {FALSE};

	while ( ! debugee_stopped_itself ) {
		retval = wait_for_debugee (debugee, statusp);
		assert (retval == debugee->pid);
		/* 
		 * if the debugee terminated then the wait syscall should return
		 * at the debugger
		 */
		if ( ! __WIFSTOPPED (*statusp) ) {
			DBPRINT (btl_prt_pt, "debugee died\n");
			retval = -ESRCH;
			goto out;
		}
		/* we wait for the debuggee to stop itself with the
		 * ptemu-signal. */
		stop_sig = __WSTOPSIG (*statusp);
		DBPRINT (1, "stop_sig=%d\n", stop_sig);
		if (stop_sig == PTEMU_SIG0) {
			debugee_stopped_itself = 1;
		} else {
			if (debugee_handles_signal (debugee, stop_sig)) {
				/* if it has stopped because of another
				 * signal then let it continue - its btlib_signal_handler
				 * will enque the signal and continue with checking the
				 * previous signal */
				retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 
							   (unsigned long)stop_sig));
				sys_err_chk (retval, "ptrace(2)");
			} else {
				signals[stop_sig] = TRUE;
				retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 0ul));
				sys_err_chk (retval, "ptrace(3)");
			}
		}
	}
	/* send all the signals that were lost while waiting for ptemu-signal */
	int j;
	for (j = 0; j <= SIGRTMAX; j ++) {
		if (signals[j] == TRUE) {
			DBPRINT (btl_prt_pt, "resend=%d\n", j);
			/* we don't really care about the return value */
			retval = SYSCALL (tkill, (debugee->pid, j));
			assert ( ! IS_SYS_ERROR (retval) || (retval == -ESRCH));
		}
	}

	retval = 0;
out:
	DBPRINT (btl_prt_pt, "debugee->pid=%d retval=%ld\n", debugee->pid, retval);
	return retval;
}
/*****************************************************************************/
/* output: a negative error code in case of faiure. 0 on success */
/*****************************************************************************/
static long handle_newly_attached_debugee (debugee_t * debugee, int * statusp)
{
	long retval;
	long is_ptraced;


	/* make sure the debuggee is being emulated by the EL */
	assert (is_debugee_ready (debugee));

	/* set the debuggee's per-thread flag */
	set_ptrace_field (debugee, PTEMU_PTRACED, 1);

	/* suspend the thread in consistent 32 bit state */
	retval = btlib_stop_debugee (debugee, INT_MAX, 0, statusp);
	if (IS_SYS_ERROR (retval)) {
		DBPRINT (TRUE, "stop_debugee failed:%ld\n", retval);
		goto out;
	}

	/* check the debuggee's global flag see if this is the
	 * first thread that is being attached to */
	retval = ia64_peekdata (debugee->pid, (unsigned long)&btl_is_ptraced, 
				&is_ptraced);
	sys_err_chk (retval, "peek_data failed");
	DBPRINT (btl_prt_pt, "btl_is_ptraced=%ld\n", is_ptraced);


	/* if non of the debuggee's threads are under debug then
	 * flush its whole instruction cache */
	if (is_ptraced == IS_PTRACED_NO) {
		int is_self_canon = (debugee->flags & PTEMU_SELF_CANONIZE) ? 1 : 0;
		debugee->flags &= ~ PTEMU_SELF_CANONIZE;
		DBPRINT (btl_prt_pt, "is_self_canon=%d\n", is_self_canon);
		if (is_self_canon) {
			/* allow the debuggee to finish the self canonization
			 * procedure */
			set_cont_signo (debugee, PTEMU_SIG0);
			retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 0ul));
			if (IS_SYS_ERROR (retval)) {
				DBPRINT (TRUE, "ptrace_cont failed:%ld\n", retval);
				goto out;
			}
			/* wait for the debuggee to stop */
			retval = wait_for_debugee_to_stop_itself (debugee, statusp);
			if (IS_SYS_ERROR (retval)) {
				goto out;
			}
		}


		/* mark the debuggee as ptraced */
		retval = SYSCALL (ptrace, (PTRACE_POKEDATA, debugee->pid,
					   (unsigned long)&btl_is_ptraced,
					   (unsigned long)IS_PTRACED_YES_NOT_READY));
		sys_err_chk (retval, "pokedata failed");
		/* set the debuggee's per-thread flag */
		set_ptrace_field (debugee, PTEMU_NEWLY_ATTACHED, TRUE);
		/* let the debuggee continue and flush existing translations */
		retval = SYSCALL (ptrace, (PTRACE_CONT, debugee->pid, 0ul, 
					   (unsigned long)PTEMU_SIG0));
		if (IS_SYS_ERROR (retval)) {
			DBPRINT (TRUE, "ptrace_cont failed:%ld\n", retval);
			goto out;
		}
		/* wait for the debuggee to stop */
		retval = wait_for_debugee_to_stop_itself (debugee, statusp);
		if (IS_SYS_ERROR (retval)) {
			goto out;
		}
	}

	debugee->flags &= ~ PTEMU_NEWLY_ATTACHED;
	retval = 0;
out:
	return retval;
}

/*****************************************************************************/
/* return 1 if the current kernel version is 2.6.
 * return 0 if current kernel version could not be retrieved, or it is other
 * than 2.6 */
/*****************************************************************************/
static inline int is_kernel_2_6 ()
{
	int ret = 0;
	long retval;
	char buf[2048];
	int fd;
	char * cp;

	fd = SYSCALL (open, ("/proc/version", O_RDONLY, 0));
	if (IS_SYS_ERROR (fd)) {
		DBPRINT (TRUE, "open failed:%d\n", fd);
		goto out;
	}
	retval = SYSCALL (read, (fd, buf, sizeof(buf) - 1));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "read failed:%d\n", ret);
		goto out;
	}
	*(buf + retval) = '\0';

	cp = strstr (buf, "2.");
	if (cp == NULL) {
		DBPRINT (TRUE, "could not find major version\n");
		goto out;
	}

	DBPRINT (0, "minor=%c\n", *(cp + 2));
	if (*(cp + 2) == '6') {
		ret = 1;
	}
out:
	if ( ! IS_SYS_ERROR (fd) ) {
		retval = SYSCALL (close, (fd));
		DBPRINT (IS_SYS_ERROR (retval), "close failed:%ld\n", retval);
	}
	DBPRINT (0, "ret=%d\n", ret);
	return ret;
}


