/*  
 * btlib_modify_ldt.c
 *  
 * 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>

#include "btlib.h"
#include "btlib_sync.h"
#include "btlib_segment.h"

/* convert sturcture to unsigned long, void * is used to avoid alignment increasing warning */
#define  S2UL(s)      (*(unsigned long *)((void *)&s))

static long read_ldt (unsigned int ptr, unsigned int bytecount);
static long write_ldt (unsigned int ptr, unsigned int bytecount);

ldt_entry ldt[LDT_ENTRIES];

/* 
 * BTLibLinux implementation of modify_ldt system call
 *
 * parameters: 
 *     unsigned int ptr              - 32-bit pointer, address of modify_ldt_s
 *     unsigned int bytecount        - size of memory
 * return value:  
 *     positive number - actual number of bytes reads
 *     0               - success
 *     negative number - error 
 */
long
btlib_modify_ldt (BTGENERIC_IA32_CONTEXT *bc32)
{
	long ret = -ENOSYS;
	int func;
	unsigned int ptr;
	unsigned int bytecount;

	/* func, LDT_READ for reading, LDT_WRITE for modifying/adding new LDT entries */
	func = (int) IA32_SYSCALL_PARAM1 (bc32);
	ptr = (unsigned int) IA32_SYSCALL_PARAM2 (bc32);
	bytecount = (unsigned int) IA32_SYSCALL_PARAM3 (bc32);

	switch (func) {
	case LDT_READ:
		ret = read_ldt (ptr, bytecount);
		break;
	case LDT_READ_DEFAULT:
		/* does it mean anything? */
		DBPRINT (TRUE, "read default ldt!\n");
		break;
	case LDT_WRITE:
	case LDT_WRITE_NEW:
		ret = write_ldt (ptr, bytecount);
		break;
	default:
		break;
	}
	return ret;
}

/* 
 * read ldt into the memory pointed by ptr
 *
 * parameters: 
 *     unsigned int ptr              - 32-bit pointer, given a memory address
 *     unsigned int bytecount        - size of memory
 * return value:  
 *     actual number of bytes reads
 */
static long
read_ldt (unsigned int ptr, unsigned int bytecount)
{
	unsigned long size;
	void *p = U2P (ptr);

	size = LDT_ENTRIES * LDT_ENTRY_SIZE;
	if (bytecount < size) {
		size = bytecount;
	}

	if (safe_memcpy (p, ldt, size) < 0) {
		return -EFAULT;
	}

	return (long) size;
}

/* 
 * modify one ldt entry, we don't flush "shadow part" 
 * (FS_base, GS_base, ES_base) of threads
 * that uses the ldt since we assume that a ldt entry cannot be modified after 
 * being loaded
 *
 * parameters: 
 *     unsigned int ptr       - 32-bit pointer, given a modify_ldt_s structure
 *     unsigned int bytecount - size of modify_ldt_s 
 * return value:  
 *     0 for success, negative number for failure
 */
static long
write_ldt (unsigned int ptr, unsigned int bytecount)
{
	modify_ldt_s ldt_info;
	ldt_entry entry;
	long ret;
	void *p = U2P (ptr);

	if (bytecount != sizeof (modify_ldt_s)) {
		DBPRINT (TRUE, "wrong size!\n");
		return -EINVAL;
	}

	if (safe_memcpy (&ldt_info, p, sizeof (ldt_info)) < 0) {
		return -EFAULT;
	}

	/* 
	 * check the parameters, more strict than what kernel does since we don't 
	 * provide protection 
	 */
	ret = -EINVAL;
	if (ldt_info.entry_number >= LDT_ENTRIES) {
		DBPRINT (TRUE, "invalid entry number!\n");
		goto out;
	}

	/* must be in user space */
	if (ldt_info.base_addr >= IA32_STACK_TOP) {
		DBPRINT (TRUE, "invalid base address!\n");
		goto out;
	}

	/* the segment should never include kernel space */
	if (ldt_info.limit_in_pages == 0) {
		if ((ldt_info.base_addr + ldt_info.limit) >= IA32_MAX_USER_ADDR) {
			DBPRINT (TRUE, "invalid limit!\n");
			goto out;
		}
	} else {
		unsigned limit;
		limit = ldt_info.limit << IA32_PAGE_SHIFT;

		if ((ldt_info.base_addr + limit) >= IA32_MAX_USER_ADDR) {
			DBPRINT (TRUE, "invalid limit!\n");
			goto out;
		}
	}

	DBPRINT (FALSE, "seg_32bit: %d\n", ldt_info.seg_32bit);

	/* should be data segment and expand-up */
	if (ldt_info.contents != 0) {
		DBPRINT (TRUE, "not a data segment!\n");
	}

	/* construct ldt entry according to modify_ldt_s */
	if (ldt_info.base_addr == 0 && ldt_info.limit == 0) {
		/* clear a LDT entry */
		memset (&entry, 0, sizeof (entry));
	} else {
		set_desc (&entry, &ldt_info);
	}

	btlib_acquire_lock (BTL_LOCK_IND_LDT);
	ldt[ldt_info.entry_number] = entry;
	btlib_release_lock (BTL_LOCK_IND_LDT);

	ret = 0;
out:
	return ret;
}

/* 
 * BTLibLinux implementation of sys_set_thread_area system call
 *
 * parameters: 
 *     unsigned int ptr - 32-bit pointer, address of user_desc
 * return value:  
 *     0               - success
 *     negative number - error number
 */
long
btlib_set_thread_area (BTGENERIC_IA32_CONTEXT *bc32)
{
	user_desc u_info, *p;
	int  idx;
	int  i;
	long ret = 0;


	assert (bc32 == cur_thread_data->bc32);  
	p = (user_desc *) U2P (IA32_SYSCALL_PARAM1 (bc32));
	if (safe_memcpy (&u_info, p, sizeof (u_info)) < 0) {
		ret = -EFAULT;
		goto out;
	}

	idx = u_info.entry_number;

	/* index -1 means the syscall should find a free tls/GDT descriptor */
	if (idx == -1) {
		for (i = 0; i < GDT_USER_ENT_NUM; i++) {
			if (S2UL (cur_thread_data->tls_array[i]) == 0) {
				idx = i + GDT_USER_ENT_FIRST;
				break;
			}
		}
		if (i >= GDT_USER_ENT_NUM) {
			ret = -ESRCH;
			goto out;
		}
	}

	if (idx > GDT_USER_ENT_LAST || idx < GDT_USER_ENT_FIRST) {
		ret = -EINVAL;
		goto out;
	}

	if (safe_memcpy (&p->entry_number, &idx, sizeof (idx)) < 0) {
		ret = -EFAULT;
		goto out;
	}

	if (is_udesc_empty (&u_info)) {
		/* it means nothing for IA-32 Exectution Layer */
		memset (&cur_thread_data->tls_array[idx - GDT_USER_ENT_FIRST],
			0, sizeof (desc_struct));
	} else {
		set_desc (&cur_thread_data->tls_array[idx - GDT_USER_ENT_FIRST], 
			  &u_info);
	}

out:
	return ret;
}

/* 
 * BTLibLinux implementation of sys_get_thread_area system call
 *
 * parameters: 
 *     unsigned int ptr - 32-bit pointer, address of user_desc
 * return value:  
 *     0               - success
 *     negative number - error number
 */
long
btlib_get_thread_area (BTGENERIC_IA32_CONTEXT *bc32)
{
	user_desc u_info, *p;
	long ret = 0;


	assert (bc32 == cur_thread_data->bc32);
	p = (user_desc *) U2P (IA32_SYSCALL_PARAM1 (bc32));

	if (safe_memcpy (&u_info, p, sizeof (u_info)) < 0) {
		ret = -EFAULT;
		goto out;
	}

	if ((u_info.entry_number < GDT_USER_ENT_FIRST) || 
	    (u_info.entry_number > GDT_USER_ENT_LAST)) {
		ret = -EINVAL;
		goto out;
	}

	get_desc (&u_info, 
		  &(cur_thread_data->tls_array[u_info.entry_number - GDT_USER_ENT_FIRST]));

	if (safe_memcpy (p, &u_info, sizeof (u_info)) < 0) {
		ret = -EFAULT;
	}
out:
	return ret;
}

/* check segment usage and set corresponding segment base in bc32 accordingly */
long
btlib_check_segment (BTGENERIC_IA32_CONTEXT *bc32)
{
	ldt_entry desc;
	long changed = 0;
	__u32 segbase;



	assert (bc32 == cur_thread_data->bc32);

	assert ((((unsigned long) &(CONTEXT_EXTFIELD ((*bc32), fs_base)) & 0x3) == 0));
	assert ((((unsigned long) &(CONTEXT_EXTFIELD ((*bc32), gs_base)) & 0x3) == 0));
	assert ((((unsigned long) &(CONTEXT_EXTFIELD ((*bc32), es_base)) & 0x3) == 0));

	segbase = 0;
	if (IS_LDT_ENTRY (bc32->fs)) {
		/* fs refers to a ldt segment */
		desc = ldt[DT_INDEX (bc32->fs)];
		if (SEG_BASEADDR (desc) != 0) {
			segbase = SEG_BASEADDR (desc);
		}
	} else if (IS_GDT_TLS_ENTRY (bc32->fs)) {
		int idx = DT_INDEX (bc32->fs);
		segbase = SEG_BASEADDR (cur_thread_data->tls_array[idx - GDT_USER_ENT_FIRST]);
	}

	if (segbase != CONTEXT_EXTFIELD ((*bc32), fs_base)) {
		changed = 1;
		DBPRINT (btl_prt_thread, "fs_base changed: old=%x new=%x\n",
		         CONTEXT_EXTFIELD ((*bc32), fs_base), segbase);
		CONTEXT_EXTFIELD ((*bc32), fs_base) = segbase;
	}

	segbase = 0;
	if (IS_LDT_ENTRY (bc32->gs)) {
		/* gs refers to a ldt segment */
		desc = ldt[DT_INDEX (bc32->gs)];
		if (SEG_BASEADDR (desc) != 0) {
			segbase = SEG_BASEADDR (desc);
		}
	} else if (IS_GDT_TLS_ENTRY (bc32->gs)) {
		int idx = DT_INDEX (bc32->gs);
		segbase = SEG_BASEADDR (cur_thread_data->tls_array[idx - GDT_USER_ENT_FIRST]);
	}
	if (segbase != CONTEXT_EXTFIELD ((*bc32), gs_base)) {
		changed = 1;
		DBPRINT (btl_prt_thread, "gs_base changed: old=%x new=%x\n",
		         CONTEXT_EXTFIELD ((*bc32), gs_base), segbase);
		CONTEXT_EXTFIELD ((*bc32), gs_base) = segbase;
	}

	segbase = 0;
	if (IS_LDT_ENTRY (bc32->es)) {
		/* es refers to a ldt segment */
		desc = ldt[DT_INDEX (bc32->es)];
		if (SEG_BASEADDR (desc) != 0) {
			segbase = SEG_BASEADDR (desc);
		}
	} else if (IS_GDT_TLS_ENTRY (bc32->es)) {
		int idx = DT_INDEX (bc32->es);
		segbase = SEG_BASEADDR (cur_thread_data->tls_array[idx - GDT_USER_ENT_FIRST]);
	}
	if (segbase != CONTEXT_EXTFIELD ((*bc32), es_base)) {
		changed = 1;
		DBPRINT (btl_prt_thread, "es_base changed: old=%x new=%x\n",
			 CONTEXT_EXTFIELD ((*bc32), es_base), segbase);
		CONTEXT_EXTFIELD ((*bc32), es_base) = segbase;
	}

	if (changed) {
		/* if there were changes to the segment descriptors then we
		 * have to set the bt_extended context. in this case we need
		 * to set all the other fields to the correct values
		 */
		SET_DESC_TABLE_REGISTERS (*bc32);
		bc32->composition_bits |= CTX32_BT_EXTENDED;
	}

	return changed;
}

/* user descriptor to segment descriptor */
void
set_desc (desc_struct *desc, user_desc *u)
{
	desc->limit_low = (unsigned short) (u->limit & 0xffff);
	desc->limit_hi = (u->limit & 0xf0000) >> 16;
	desc->base_low = (unsigned short) (u->base_addr & 0xffff);
	desc->base_mid = (u->base_addr & 0xff0000) >> 16;
	desc->base_hi = (u->base_addr & 0xff000000) >> 24;

	/* DT = 1, get TYPE from content and read_exec_only  */
	desc->type = 0x11 | (u->contents << 2) | ((u->read_exec_only ^ 1) << 1);
	desc->dpl = 0x3;
	desc->present = !u->seg_not_present;
	desc->avl = 0x0;
	desc->reserved_0 = 0x0;
	desc->default_big = u->seg_32bit;
	desc->granularity = u->limit_in_pages;

	return;
}

/* segment descriptor to user descriptor except for entry_number */
void
get_desc (user_desc * u, desc_struct * desc)
{
	u->base_addr = SEG_BASEADDR (*desc);
	u->limit = SEG_LIMIT (*desc);
	u->seg_32bit = desc->default_big;
	u->contents = (desc->type >> 2) & 3;
	u->read_exec_only = !((desc->type > 1) & 1);
	u->limit_in_pages = desc->granularity;
	u->seg_not_present = !desc->present;
	u->useable = desc->avl;
	return;
}

inline long
is_udesc_empty (const user_desc * u_info)
{
	return ((u_info->base_addr == 0) &&
		(u_info->limit == 0) &&
		(u_info->contents == 0) &&
		(u_info->read_exec_only) == 1 &&
		(u_info->seg_32bit) == 0 &&
		(u_info->limit_in_pages) == 0 &&
		(u_info->seg_not_present) == 1 && 
		(u_info->useable == 0));
}

