/*  
 * btlib_sig_utils.c
 *  
 *  
 *
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <btlib.h>
#include <btlib_sig_utils.h>

#if DEBUG
static struct _val_name_t {
	unsigned val;
	char *name;
} flag_names[] = {
	{ SA_NOCLDSTOP, "SA_NOCLDSTOP" }, 
	{ SA_SIGINFO, "SA_SIGINFO" }, 
	{ SA_ONSTACK, "SA_ONSTACK"}, 
	{ SA_RESTART, "SA_RESTART"}, 
	{ SA_NODEFER, "SA_NODEFER"}, 
	{ SA_RESETHAND, "SA_RESETHAND"}, 
	{ SA_INTERRUPT, "SA_INTERRUPT"}, 
	{ SA_RESTORER, "SA_RESTORER"}, 
	{ (unsigned) -1, NULL}
};
#endif

/*****************************************************************************/
/* make a given set empty */
/*****************************************************************************/
inline void 
sigemptyset32 (sigset32_t *set)
{
	memset (set, 0, sizeof(sigset32_t));
}

/*****************************************************************************/
/* add the given signal to the given set */
/*****************************************************************************/
inline void
sigaddset32 (sigset32_t *set, int sig)
{
	int s, d, m;

	assert ((sig > 0) && (sig <= _NSIG));

	s = sig - 1;
	d = s >> 5;		// d = s/32
	m = s - (d << 5);	// m = s%32

	set->sig[d] |= (1U << m);
}

/************************************************************************** 
* return 1 if the given signal is a member of the given set
* and 0 otherwise.
**************************************************************************/
inline long
sigismember32 (const sigset32_t *set, int sig)
{
	int s, d, m;

	assert ((sig > 0) && (sig <= _NSIG));

	s = sig - 1;
	d = s >> 5;		// d = s/32
	m = s - (d << 5);	// m = s%32

	return 1 & (set->sig[d] >> m);
}

/*****************************************************************************/
/* return 0 iff the given sigset is empty */
/*****************************************************************************/
inline int
isemptyset32 (sigset32_t *set, unsigned size)
{
	return ((set->sig[0] == 0)
		&& ((size == 1) || set->sig[1] == 0)) ? 1 : 0;
}

/*****************************************************************************/
/* remove the given signal from the given set */
/*****************************************************************************/
inline void
sigdelset32 (sigset32_t *set, int sig)
{
	int s, d, m;

	assert ((sig > 0) && (sig <= _NSIG));

	s = sig - 1;
	d = s >> 5;		// d = s/32
	m = s - (d << 5);	// m = s%32

	set->sig[d] &= ~(1U << m);
}



/******************************************************************************/
/* convert IA-32 sigset (two words) to native sigset */
/******************************************************************************/
inline void
sigset32_2_sigset64 (const sigset32_t *set32, k_sigset_t *set64)
{
	union {
		unsigned m32[2];
		unsigned long m64;
	} mc;

	mc.m32[0] = set32->sig[0];
	mc.m32[1] = set32->sig[1];

	set64->sig[0] = mc.m64;
}

/*****************************************************************************/
/* convert native sigset to IA-32 sigset (two words) */
/*****************************************************************************/
inline void
sigset64_2_sigset32 (const k_sigset_t *set64, sigset32_t *set32)
{
	union {
		sigset32_t m32;
		k_sigset_t m64;
	} mc;

	mc.m64 = *set64;
	*set32 = mc.m32;
}


/*****************************************************************************/
/*****************************************************************************/
inline void
sig_and_mask (sigset32_t *dest, const sigset32_t *src, int size)
{
	int j;
	for (j = 0; j < size; j++) {
		dest->sig[j] &= src->sig[j];
	}
}

/*****************************************************************************/
/*****************************************************************************/
inline void
sig_or_mask (sigset32_t *dest, const sigset32_t *src, int size)
{
	int j;
	for (j = 0; j < size; j++) {
		dest->sig[j] |= src->sig[j];
	}
}

/*****************************************************************************/
/*****************************************************************************/
inline void
sig_andnot_mask (sigset32_t *dest, const sigset32_t *src, int size)
{
	int j;
	for (j = 0; j < size; j++) {
		dest->sig[j] &= ~src->sig[j];
	}
}


/*****************************************************************************/
/* 
 * return 1 if the given signal is asynchronous, and 0 otherwise 
 * input: siginfo - the signal's info as received by btlib's signal handler
 */
/*****************************************************************************/
inline long
is_async_signal (siginfo_t * siginfo)
{
	long retval;

	/* what TODO with SIGXCPU ???? */
	retval = (SI_FROMUSER (siginfo)) ||
		  ((siginfo->si_signo != SIGILL) &&
	     	   (siginfo->si_signo != SIGFPE) &&
		   (siginfo->si_signo != SIGSEGV) &&
		   (siginfo->si_signo != SIGBUS) && 
		   (siginfo->si_signo != SIGTRAP));

	return retval;
}

/*****************************************************************************/
/*
 * unregister btlib's handler for the given signal
 */
/*****************************************************************************/
inline void
unreg_btlib_handler (int signo)
{
	struct sigaction act;
	long retval;

	memset (&act, 0, sizeof (act));
	act.sa_handler = SIG_DFL;
	retval = SYSCALL (rt_sigaction, (signo, &act, NULL, sizeof (k_sigset_t)));
	sys_err_chk (retval, "rt_sigaction");
}

/*****************************************************************************/
/*
 * check if btlib has a permanent handler for the given signal
 */
/*****************************************************************************/
int
has_permanent_handler (int signo)
{
	int permanent, *perm;
	permanent = 0;
	for (perm = m_permanent_handlers; *perm != -1; perm++) {
		assert (perm - m_permanent_handlers <= _NSIG);
		if (*perm == signo) {
			permanent = 1;
			break;
		}
	}
	return permanent;
}


/*****************************************************************************/
/* check if btlib has a half permanent handler for the given signal          */
/* the signals that should trigger core file mechanism and are not with      */
/* permanent handlers are defined with half permanent handlers.              */
/* the difference between permanent and half permanent is that the user can  */
/* change the behaviour for half permanent (see in emulate_sigaction)        */
/*****************************************************************************/
int
has_half_permanent_handler (int signo)
{
	int half_permanent, *perm;
	half_permanent = 0;
	for (perm = m_half_permanent_handlers; *perm != -1; perm++) {
		assert (perm - m_half_permanent_handlers <= _NSIG);
		if (*perm == signo) {
			half_permanent = 1;
			break;
		}
	}
	return half_permanent;
}

/*****************************************************************************/
/*
 * convert signal number to btgeneric's exception codes
 * reference - Intel(R) Itanium(R) Architecture Software Developer's Manual
 * (See a link http://www.intel.com/design/itanium/downloads/24531804.pdf)
 * Volume 2, ch. 8 - Interruption Vector Description
 */
/*****************************************************************************/
inline unsigned
signal2except (siginfo_t *info64)
{
	int except_code = BT_NO_EXCEPT;
	int is_fault = 0;
	int is_trap = 0;

	switch (info64->si_signo) {
	case SIGFPE:
		if (info64->si_isr & 0x880) {
			except_code = BT_EXCEPT_FLT_OVERFLOW;
			is_trap = 1;
		}
		if (info64->si_isr & 0x1100) {
			except_code = BT_EXCEPT_FLT_UNDERFLOW;
			is_trap = 1;
		}
		if (info64->si_isr & 0x2200) {
			except_code = BT_EXCEPT_FLT_INEXACT_RESULT;
			is_trap = 1;
		}
		if (!is_trap) {
			if (info64->si_isr & 0x11) {
				except_code = BT_EXCEPT_FLT_INVALID_OPERATION;
				is_fault = 1;
			}
			if (info64->si_isr & 0x22) {
				except_code = BT_EXCEPT_FLT_DENORMAL_OPERAND;
				is_fault = 1;
			}
			if (info64->si_isr & 0x44) {
				except_code = BT_EXCEPT_FLT_DIVIDE_BY_ZERO;
				is_fault = 1;
			}
		} else {
			/* bit {0} of isr indicates fp-trap */
			assert ((info64->si_isr & 0x01) != 0);
		}
		break;
	case SIGILL:
		except_code = BT_EXCEPT_ILLEGAL_INSTRUCTION;
		break;
	case SIGBUS:
		except_code = BT_EXCEPT_DATATYPE_MISALIGNMENT;
		break;
	case SIGSEGV:
		except_code = BT_EXCEPT_ACCESS_VIOLATION;
		break;
	case SIGTRAP:
		assert (!" should not get SIGTRAP");
		break;
	}
	assert (except_code != BT_NO_EXCEPT);
	return except_code;
}

/*****************************************************************************/
/*
 * Purpose: convert  BTGENERIC_IA32_CONTEXT to sigcontext
 */
/*****************************************************************************/
inline void
btgctx2sigctx_32 (const BTGENERIC_IA32_CONTEXT *bc32,
                  sigcontext32_t *sc, struct _fpstate_ia32 *fpstate)
{
	sc->edi = bc32->edi;
	sc->esi = bc32->esi;
	sc->ebp = bc32->ebp;
	sc->esp = bc32->esp;
	sc->ebx = bc32->ebx;
	sc->edx = bc32->edx;
	sc->ecx = bc32->ecx;
	sc->eax = bc32->eax;
	sc->eip = bc32->eip;
	sc->esp_at_signal = bc32->esp;
	sc->eflags = bc32->eflags;

	sc->ss = bc32->ss;
	sc->cs = bc32->cs;
	sc->ds = bc32->ds;
	sc->es = bc32->es;
	sc->fs = bc32->fs;
	sc->gs = bc32->gs;

	/* this field must be filled only when sc is IA-32 address */
	sc->fpstate = (uintptr_t) NULL;

	/* 
	 * fp state
	 * the new fp state has two parts - 
	 * the first is identical to the old fpstate
	 * (similar to struct i387_fsave_struct) 
	 * and the second is aimed at storing
	 * the new simd regs (similar to struct i387_fxsave_struct)
	 *
	 * to be on the safe side - we save the fp state in both parts
	 */

	btg2fsave (bc32, (struct i387_fsave_struct *) fpstate);
	fpstate->magic = 0;	/* second part is in use */
	btg2fxsave (bc32, (struct i387_fxsave_struct *) (void *)&fpstate->xcwd);
}

/*****************************************************************************/
/*
 * void sigcontext_to_btrans
 *
 * Purpose: convert sigcontext to BTGENERIC_IA32_CONTEXT
 *
 */
/*****************************************************************************/
inline void
sigctx2btgctx_32 (sigcontext32_t *sc, BTGENERIC_IA32_CONTEXT *bc32)
{
	struct _fpstate_ia32 *fpstate;

	fpstate = U2P (sc->fpstate);

	bc32->edi = sc->edi;
	bc32->esi = sc->esi;
	bc32->ebp = sc->ebp;
	bc32->esp = sc->esp;
	bc32->ebx = sc->ebx;
	bc32->edx = sc->edx;
	bc32->ecx = sc->ecx;
	bc32->eax = sc->eax;
	bc32->eip = sc->eip;
	bc32->eflags = sc->eflags;

	bc32->cs = (unsigned) sc->cs;
	bc32->ss = (unsigned) sc->ss;
	bc32->ds = (unsigned) sc->ds;
	bc32->es = (unsigned) sc->es;
	bc32->fs = (unsigned) sc->fs;
	bc32->gs = (unsigned) sc->gs;

	fxsave2btg (bc32, (struct i387_fxsave_struct *)(void *)&fpstate->xcwd);
	fsave2btg (bc32, (struct i387_fsave_struct *)fpstate);

	bc32->composition_bits = CTX32_CRITICAL |
		CTX32_GP | CTX32_SEGMENT | CTX32_FLOATING_POINT | CTX32_XMM;

}

/*****************************************************************************/
/*
 * sigcontext64_to_btrans_context
 *
 * Purpose: convert sigcontext (64) to BTGENERIC_IA64_CONTEXT
 *
 */
/*****************************************************************************/
inline void
sigctx2btgctx_64 (struct sigcontext *sc64, BTGENERIC_IA64_CONTEXT *bc64)
{

	DBCODE (0, print_sigcontext64 (sc64));
	memcpy ((void *) &(bc64->fp_static[0]),
		(void *) &sc64->sc_fr[2], 126 * sizeof (struct ia64_fpreg));
	memcpy ((void *) &(bc64->int_static[0]),
		(void *) &sc64->sc_gr[1], 31 * sizeof (unsigned long));
	memcpy ((void *) &(bc64->br_regs[0]),
		(void *) &sc64->sc_br[0], 8 * sizeof (unsigned long));

	bc64->nat_collection = sc64->sc_nat;
	/* IIP will always be aligned (slot in PSR) */
	bc64->iip = (sc64->sc_ip & 0xFFFFFFFFFFFFFFF0UL);
	bc64->ipsr = (sc64->sc_um | ((sc64->sc_ip & 0x3UL) << 41));

	bc64->predicates = sc64->sc_pr;
	bc64->fpsr = sc64->sc_ar_fpsr;
	bc64->br_regs[7] = sc64->sc_ar_unat;
	bc64->lc = sc64->sc_ar_lc;

	bc64->ccv = sc64->sc_ar_ccv;

	bc64->pfs = sc64->sc_ar_pfs;
	bc64->bsp = (uintptr_t) get_interrupted_bsp (sc64->sc_ar_bsp, sc64->sc_cfm);
	/* this is required for canonizing the IA32 context in IA32EL */
	bc64->bspstore = bc64->bsp;

	bc64->rsc = sc64->sc_ar_rsc;
	bc64->rnat = sc64->sc_ar_rnat;

	bc64->composition_bits = CTX64_CRITICAL |
		CTX64_FLOATING_POINT | CTX64_INTEGER | CTX64_IA32_CRITICAL;
}

/*****************************************************************************/
/* when entering the signal handler of btlib, 
 * the bsp field in the sigcontext holds the bsp value in the
 * signal handler - not the one at the point of interrupt.
 */
/*****************************************************************************/
void *
get_interrupted_bsp (const unsigned long bsp, const unsigned long cfm)
{
	void *int_bsp;
	long regs_in_prev_frame;

	regs_in_prev_frame = cfm & SOF_MASK;
	int_bsp = (regs_in_prev_frame == 0) ?
		  (void *) bsp :
		  ia64_rse_skip_regs ((void *) bsp, -regs_in_prev_frame);
	return int_bsp;
}

/*****************************************************************************/
/* convert the fp regs from linux format to btgeneric's */
/*****************************************************************************/
void
fsave2btg (BTGENERIC_IA32_CONTEXT * bc32, struct i387_fsave_struct *fsave)
{
	bc32->fpu_state.control_word = fsave->cwd;
	bc32->fpu_state.status_word = fsave->swd;
	bc32->fpu_state.tag_word = fsave->twd;
	bc32->fpu_state.instruction_pointer_offset = fsave->fip;
	bc32->fpu_state.instruction_pointer_selector = fsave->fcs;
	bc32->fpu_state.data_pointer_offset = fsave->foo;
	bc32->fpu_state.data_pointer_selector = fsave->fos;
	bc32->fpu_state.unused = 0;	// this field is not used in btg
	memcpy (&bc32->fpu_state.stack, &fsave->st_space,
		sizeof (bc32->fpu_state.stack));
}

/*****************************************************************************/
/* convert the fp regs to linux format */
/*****************************************************************************/
void
btg2fsave (const BTGENERIC_IA32_CONTEXT * bc32, struct i387_fsave_struct *fsave)
{
	fsave->cwd = bc32->fpu_state.control_word;
	fsave->swd = bc32->fpu_state.status_word;
	fsave->twd = bc32->fpu_state.tag_word;
	fsave->fip = bc32->fpu_state.instruction_pointer_offset;
	fsave->fcs = bc32->fpu_state.instruction_pointer_selector;
	fsave->foo = bc32->fpu_state.data_pointer_offset;
	fsave->fos = bc32->fpu_state.data_pointer_selector;
	fsave->status = 0;	// this field has no meaning in userland
	memcpy (&fsave->st_space, &bc32->fpu_state.stack,
		sizeof (fsave->st_space));
}

#define FPREG_NUMBER 8
#define BTGENERIC_FP_BYTES_PER_ENTRY 10
#define FXSAVE_FP_BYTES_PER_ENTRY    16
/*****************************************************************************/
/* fp stack conversion - the function converts fp stack entries from         */
/* btgeneric context format (10 bytes per entry) to fxsave format (16 bytes  */
/* per entry)                                                                */
/*****************************************************************************/
inline void 
st_conversion_fsave_to_fxsave (unsigned *fxsave_st, void *fsave_st)
{
	int i;
	memset (fxsave_st, 0, sizeof(unsigned) * 32);
	for (i = 0; i < FPREG_NUMBER; i++) {
		memcpy ((__u8 *)fxsave_st + i * FXSAVE_FP_BYTES_PER_ENTRY, 
			(__u8 *)fsave_st + i * BTGENERIC_FP_BYTES_PER_ENTRY, 
			(size_t)BTGENERIC_FP_BYTES_PER_ENTRY);
	}
	return;
}


/*****************************************************************************/
/* fp stack conversion - the function converts fp stack entries from         */
/* fxsave format (16 bytes per entry) to btgeneric context format (10 bytes  */
/* per entry)                                                                */
/*****************************************************************************/
inline void 
st_conversion_fxsave_to_fsave (void *fsave_st, unsigned *fxsave_st)
{
	int i;
	memset (fsave_st, 0, sizeof(unsigned) * 20);
	for (i = 0; i < FPREG_NUMBER; i++) {
		memcpy ((__u8 *)fsave_st + i * BTGENERIC_FP_BYTES_PER_ENTRY, 
			(__u8 *)fxsave_st + i * FXSAVE_FP_BYTES_PER_ENTRY, 
			(size_t)BTGENERIC_FP_BYTES_PER_ENTRY);
	}
	return;
}


#define SHIFT_LEFT(x,len) x<<len
#define SHIFT_RIGHT(x,len) x>>len
/*****************************************************************************/
/* tag word conversion - the function takes the tag word and convert it      */
/* to a byte. every pair of bits in the original tag word become 1 bit.      */
/* 11(empty) become 0                                                        */
/* 00(valid), 01(zero), and 10(special) become 1.                            */
/* input: twd - the tag word (16 bits. higher 16 bits are ignored)           */
/* output: conv_twd - the converted tag word (8 bits. higher 8 bits are      */
/* ignored)                                                                  */
/*****************************************************************************/
inline unsigned short
twd_convertion_word_to_byte (__u32 twd)
{
	unsigned short conv_twd = 0;
	unsigned int tmp = twd;
	unsigned int cur_2_bits;
	int i;
	tmp = ~(tmp & (SHIFT_RIGHT (tmp, 1)));
	tmp = tmp & 0x5555;
	// after the 2 previous lines the lower 16 bits of tmp look the following:
	// the even bits cotain the formula: not of the logical and of the 2 
	// corresponding bits (e.g, bit 12 is ~(bit12 and bit13)
	// the uneven bits contain 0
	// tmp - 0X0X0X0X0X0X0X0X where X is the formula representation
	for (i = 0; i < FPREG_NUMBER; i++) {
		cur_2_bits = SHIFT_RIGHT (tmp, 2);
		tmp = SHIFT_RIGHT (tmp, 2);
		conv_twd = (SHIFT_LEFT (conv_twd, 1)) | cur_2_bits;
	}
	return conv_twd;
}

/*****************************************************************************/
/* tag word convertion - the function takes the tag word and convert it      */
/* to a word. every bit is examined. if it is 0, it will become 11(empty).   */
/* otherwise, we will look at the floating point representation and decide   */
/* if it is zero,special or valid by examining the exponent and the          */
/* significand                                                               */
/* input: fxsave - containing the 8 bit tag word and the fp registers        */
/* output: conv_twd - the converted tag word (16 bits)                       */
/*                                                                           */
/* class 	    exponent	    significand	       type	             */
/* -----------------------------------------------------------               */
/* +infinite        0111..11        1000..00           special               */
/* -----------------------------------------------------------               */
/* +normals         0111..10        1111..11           valid                 */
/* +normals             .               .              valid                 */
/* +normals             .               .              valid                 */
/* +normals         0000..01        1000..00           valid                 */
/* -----------------------------------------------------------               */
/* +denormals       0000..00        0111..11           special               */
/* +denormals           .               .              special               */
/* +denormals           .               .              special               */
/* +denormals       0000..00        0000..01           special               */
/* -----------------------------------------------------------               */
/* +zero            0000..00        0000..00           zero                  */
/* -----------------------------------------------------------               */
/* -zero            1000..00        0000..00           zero                  */
/* -----------------------------------------------------------               */
/* -denormals       1000..00        0000..01           special               */
/* -denormals           .               .              special               */
/* -denormals           .               .              special               */
/* -denormals       1000..00        0111..11           special               */
/* -----------------------------------------------------------               */
/* -normals         1000..01        1000..00           valid                 */
/* -normals             .               .              valid                 */
/* -normals             .               .              valid                 */
/* -normals         1111..10        1111..11           valid                 */
/* -----------------------------------------------------------               */
/* -infinite        1111..11        1000..00           special               */
/* -----------------------------------------------------------               */
/* SNan             X111..11        10XX..XX           special               */
/* QNan             X111..11        11XX..XX           special               */
/* Real Indefinite  1111..11        1100..00           special               */
/*****************************************************************************/
inline __u32
twd_convertion_byte_to_word (struct i387_fxsave_struct *fxsave)
{
	unsigned int conv_twd = 0;
	unsigned short cur_twd = fxsave->twd;
	unsigned int two_bits_representation;
	unsigned int low_significand;
	unsigned int high_significand;
	unsigned short exponent;
	int i;

	for (i = 0; i < FPREG_NUMBER; i++) {	// may be valid, zero, or special
		if (cur_twd & 0x1) {		// zero,special,or valid
			low_significand = fxsave->st_space[i * 4];
			high_significand = fxsave->st_space[(i * 4) + 1];
			exponent = (unsigned short) (fxsave->st_space[(i * 4) + 2]);
			if ((exponent & 0x7fff) == 0x0000) { 
				// zero or denormals (special)
				if ((low_significand == 0) && (high_significand == 0)) {
					// zero
					two_bits_representation = FPREG_ZERO;
				} else {	
					// denormals (special)
					two_bits_representation = FPREG_SPECIAL;
				}
			} else if ((exponent & 0x7fff) == 0x7ffff) {
				// special
				two_bits_representation = FPREG_SPECIAL;
			} else	{ 
				//everything else is valid
				two_bits_representation = FPREG_VALID;
			}
		} else {	//must be empty
			two_bits_representation = FPREG_EMPTY;
		}
		conv_twd = SHIFT_LEFT (two_bits_representation, 2 * i) | conv_twd;
		cur_twd = SHIFT_RIGHT (cur_twd, 1);
	}
	return conv_twd;
}

/*****************************************************************************/
/* convert the fp and xmm regs from linux format to btgeneric's              */
/*****************************************************************************/
void
fxsave2btg (BTGENERIC_IA32_CONTEXT *bc32, struct i387_fxsave_struct *fxsave)
{
	bc32->fpu_state.control_word = (unsigned) (fxsave->cwd | 0xffff0000);
	bc32->fpu_state.status_word = (unsigned) (fxsave->swd | 0xffff0000);
	bc32->fpu_state.tag_word = twd_convertion_byte_to_word (fxsave);
	bc32->fpu_state.instruction_pointer_offset = fxsave->fip;
	bc32->fpu_state.instruction_pointer_selector = 
		fxsave->fcs | (unsigned) (fxsave->fop << 16);
	bc32->fpu_state.data_pointer_offset = fxsave->foo;
	bc32->fpu_state.data_pointer_selector = fxsave->fos;
	bc32->fpu_state.unused = 0;	// this field is not used in btg
	st_conversion_fxsave_to_fsave (&bc32->fpu_state.stack, &fxsave->st_space[0]);

	/* xmm regs */
	*(unsigned int *)(void *)&bc32->xmm_and_bt_ext[MXCSR_INDEX_IN_EXT_REGS] = 
		fxsave->mxcsr;
	memcpy (&bc32->xmm_and_bt_ext[XMM0_INDEX_IN_EXT_REGS],
		&fxsave->xmm_space[0], sizeof (fxsave->xmm_space));
}

/*****************************************************************************/
/* convert the fp and xmm regs to linux format */
/*****************************************************************************/
void
btg2fxsave (const BTGENERIC_IA32_CONTEXT *bc32, struct i387_fxsave_struct *fxsave)
{
	fxsave->cwd = (unsigned short) (bc32->fpu_state.control_word & 0xffff);
	fxsave->swd = (unsigned short) (bc32->fpu_state.status_word & 0xffff);
	fxsave->twd = twd_convertion_word_to_byte (bc32->fpu_state.tag_word);
	fxsave->fip = bc32->fpu_state.instruction_pointer_offset;
	fxsave->fop = (unsigned short) 
		((bc32->fpu_state.instruction_pointer_selector & 0xffff) >> 16);
	fxsave->fcs = bc32->fpu_state.instruction_pointer_selector & 0xffff;
	fxsave->foo = bc32->fpu_state.data_pointer_offset;
	fxsave->fos = bc32->fpu_state.data_pointer_selector;

	st_conversion_fsave_to_fxsave (&fxsave->st_space[0], &bc32->fpu_state.stack);
	fxsave->mxcsr = 
		*(unsigned int *)(void *) &bc32->xmm_and_bt_ext[MXCSR_INDEX_IN_EXT_REGS];
	memcpy (&fxsave->xmm_space[0],
		&bc32->xmm_and_bt_ext[XMM0_INDEX_IN_EXT_REGS],
		sizeof (fxsave->xmm_space));
}

#if DEBUG
/*****************************************************************************/
/*****************************************************************************/
void
sprint_flags (unsigned flags, char *flags_str, long str_len)
{
	char *p = flags_str;
	long j = 0;

	while (flag_names[j].name != NULL) {
		if (flags & flag_names[j].val) {
			sprintf (p, "%s ", flag_names[j].name);
			p += strlen (flag_names[j].name) + 1;
			assert (p - flags_str < str_len);
		}
		j++;
	}
	*(p - 1) = '\0';
}

/*****************************************************************************/
/*****************************************************************************/
void
print_sigaction32 (char *header, struct sigaction32 *sa)
{

	int s, i;
	char flags_str[2048];

	fprintf (g_log_file, "%s\n", header);
	if (sa == NULL) {
		fprintf (g_log_file, "\t(nil)\n");
	} else {
		fprintf (g_log_file, "\tsa->handler (%x)\n", sa->handler);
		s = (sizeof (sa->mask) / 4);
		fprintf (g_log_file, "\tsa->mask (");
		for (i = 0; i < s; i++) {
			fprintf (g_log_file, "%s0x%x", i ? " " : "",
				 sa->mask.sig[i]);
		}
		fprintf (g_log_file, ")\n");
		flags_str[0] = '\0';
		sprint_flags ((unsigned) sa->flags, flags_str,
			      sizeof (flags_str));
		fprintf (g_log_file, "\tsa->flags (%x) [%s]\n", sa->flags,
			 flags_str);
		fprintf (g_log_file, "\tsa->restorer (%x)\n", sa->restorer);
	}

}

/*****************************************************************************/
/*****************************************************************************/
void
print_btgcontext32 (BTGENERIC_IA32_CONTEXT *bc32)
{
	int j, k;

	fprintf (g_log_file, "bc32->edi = %x\n", bc32->edi);
	fprintf (g_log_file, "bc32->esi = %x\n", bc32->esi);
	fprintf (g_log_file, "bc32->ebp = %x\n", bc32->ebp);
	fprintf (g_log_file, "bc32->esp = %x\n", bc32->esp);
	fprintf (g_log_file, "bc32->ebx = %x\n", bc32->ebx);
	fprintf (g_log_file, "bc32->edx = %x\n", bc32->edx);
	fprintf (g_log_file, "bc32->ecx = %x\n", bc32->ecx);
	fprintf (g_log_file, "bc32->eax = %x\n", bc32->eax);
	fprintf (g_log_file, "bc32->eip = %x\n", bc32->eip);
	fprintf (g_log_file, "bc32->cs = %x\n", bc32->cs);
	fprintf (g_log_file, "bc32->ss = %x\n", bc32->ss);
	fprintf (g_log_file, "bc32->eflags = %x\n", bc32->eflags);

	fprintf (g_log_file, "bc32->fpu_state.control_word = %x\n",
		 bc32->fpu_state.control_word);
	fprintf (g_log_file, "bc32->fpu_state.status_word = %x\n",
		 bc32->fpu_state.status_word);
	fprintf (g_log_file, "bc32->fpu_state.tag_word = %x\n",
		 bc32->fpu_state.tag_word);
	fprintf (g_log_file,
		 "bc32->fpu_state.instruction_pointer_offset = %x\n",
		 bc32->fpu_state.instruction_pointer_offset);
	fprintf (g_log_file,
		 "bc32->fpu_state.instruction_pointer_selector = %x\n",
		 bc32->fpu_state.instruction_pointer_selector);
	fprintf (g_log_file, "bc32->fpu_state.data_pointer_offset = %x\n",
		 bc32->fpu_state.data_pointer_offset);
	fprintf (g_log_file, "bc32->fpu_state.data_pointer_selector = %x\n",
		 bc32->fpu_state.data_pointer_selector);
	fprintf (g_log_file, "bc32->fpu_state.unused = %x\n",
		 bc32->fpu_state.unused);

	for (j = 0; j < 8; j++) {
		fprintf (g_log_file, "fpreg[%d]: ", j);
		for (k = 9; k >= 0; k--) {
			fprintf (g_log_file, "%02x ",
				 bc32->fpu_state.stack[j * 10 + k]);
		}
		fprintf (g_log_file, "\n");
	}
}

/*****************************************************************************/
/*****************************************************************************/
void
print_btgcontext64 (BTGENERIC_IA64_CONTEXT * bc64)
{
	fprintf (g_log_file, "\tIntGp=%lx\n", bc64->int_static[1]);
	fprintf (g_log_file, "\tIntT0=%lx\n", bc64->int_static[2]);
	fprintf (g_log_file, "\tIntT1=%lx\n", bc64->int_static[3]);
	fprintf (g_log_file, "\tIntS0=%lx\n", bc64->int_static[4]);
	fprintf (g_log_file, "\tIntS1=%lx\n", bc64->int_static[5]);
	fprintf (g_log_file, "\tIntS2=%lx\n", bc64->int_static[6]);
	fprintf (g_log_file, "\tIntS3=%lx\n", bc64->int_static[7]);
	fprintf (g_log_file, "\tIntV0=%lx\n", bc64->int_static[8]);
	fprintf (g_log_file, "\tIntT2=%lx\n", bc64->int_static[9]);
	fprintf (g_log_file, "\tIntT3=%lx\n", bc64->int_static[10]);
	fprintf (g_log_file, "\tIntT4=%lx\n", bc64->int_static[11]);
	fprintf (g_log_file, "\tIntSp=%lx\n", bc64->int_static[12]);
	fprintf (g_log_file, "\tIntTeb=%lx\n", bc64->int_static[13]);
	fprintf (g_log_file, "\tIntT5=%lx\n", bc64->int_static[14]);
	fprintf (g_log_file, "\tIntT6=%lx\n", bc64->int_static[15]);
	fprintf (g_log_file, "\tIntT7=%lx\n", bc64->int_static[16]);
	fprintf (g_log_file, "\tIntT8=%lx\n", bc64->int_static[17]);
	fprintf (g_log_file, "\tIntT9=%lx\n", bc64->int_static[18]);
	fprintf (g_log_file, "\tIntT10=%lx\n", bc64->int_static[19]);
	fprintf (g_log_file, "\tIntT11=%lx\n", bc64->int_static[20]);
	fprintf (g_log_file, "\tIntT12=%lx\n", bc64->int_static[21]);
	fprintf (g_log_file, "\tIntT13=%lx\n", bc64->int_static[22]);
	fprintf (g_log_file, "\tIntT14=%lx\n", bc64->int_static[23]);
	fprintf (g_log_file, "\tIntT15=%lx\n", bc64->int_static[24]);
	fprintf (g_log_file, "\tIntT16=%lx\n", bc64->int_static[25]);
	fprintf (g_log_file, "\tIntT17=%lx\n", bc64->int_static[26]);
	fprintf (g_log_file, "\tIntT18=%lx\n", bc64->int_static[27]);
	fprintf (g_log_file, "\tIntT19=%lx\n", bc64->int_static[28]);
	fprintf (g_log_file, "\tIntT20=%lx\n", bc64->int_static[29]);
	fprintf (g_log_file, "\tIntT21=%lx\n", bc64->int_static[30]);
	fprintf (g_log_file, "\tIntT22=%lx\n", bc64->int_static[31]);
	fprintf (g_log_file, "\tIntNats=%lx\n", bc64->nat_collection);
	fprintf (g_log_file, "\tPreds=%lx\n", bc64->predicates);

	fprintf (g_log_file, "\tfp_static[0]=%lx\n",
		 bc64->fp_static[0].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[0]=%lx\n",
		 bc64->fp_static[0].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[1]=%lx\n",
		 bc64->fp_static[1].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[1]=%lx\n",
		 bc64->fp_static[1].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[2]=%lx\n",
		 bc64->fp_static[2].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[2]=%lx\n",
		 bc64->fp_static[2].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[3]=%lx\n",
		 bc64->fp_static[3].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[3]=%lx\n",
		 bc64->fp_static[3].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[4]=%lx\n",
		 bc64->fp_static[4].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[4]=%lx\n",
		 bc64->fp_static[4].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[5]=%lx\n",
		 bc64->fp_static[5].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[5]=%lx\n",
		 bc64->fp_static[5].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[6]=%lx\n",
		 bc64->fp_static[6].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[6]=%lx\n",
		 bc64->fp_static[6].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[7]=%lx\n",
		 bc64->fp_static[7].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[7]=%lx\n",
		 bc64->fp_static[7].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[8]=%lx\n",
		 bc64->fp_static[8].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[8]=%lx\n",
		 bc64->fp_static[8].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[9]=%lx\n",
		 bc64->fp_static[9].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[9]=%lx\n",
		 bc64->fp_static[9].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[10]=%lx\n",
		 bc64->fp_static[10].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[10]=%lx\n",
		 bc64->fp_static[10].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[11]=%lx\n",
		 bc64->fp_static[11].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[11]=%lx\n",
		 bc64->fp_static[11].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[12]=%lx\n",
		 bc64->fp_static[12].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[12]=%lx\n",
		 bc64->fp_static[12].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[13]=%lx\n",
		 bc64->fp_static[13].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[13]=%lx\n",
		 bc64->fp_static[13].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[14]=%lx\n",
		 bc64->fp_static[14].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[14]=%lx\n",
		 bc64->fp_static[14].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[15]=%lx\n",
		 bc64->fp_static[15].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[15]=%lx\n",
		 bc64->fp_static[15].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[16]=%lx\n",
		 bc64->fp_static[16].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[16]=%lx\n",
		 bc64->fp_static[16].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[17]=%lx\n",
		 bc64->fp_static[17].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[17]=%lx\n",
		 bc64->fp_static[17].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[18]=%lx\n",
		 bc64->fp_static[18].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[18]=%lx\n",
		 bc64->fp_static[18].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[19]=%lx\n",
		 bc64->fp_static[19].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[19]=%lx\n",
		 bc64->fp_static[19].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[20]=%lx\n",
		 bc64->fp_static[20].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[20]=%lx\n",
		 bc64->fp_static[20].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[21]=%lx\n",
		 bc64->fp_static[21].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[21]=%lx\n",
		 bc64->fp_static[21].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[22]=%lx\n",
		 bc64->fp_static[22].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[22]=%lx\n",
		 bc64->fp_static[22].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[23]=%lx\n",
		 bc64->fp_static[23].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[23]=%lx\n",
		 bc64->fp_static[23].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[24]=%lx\n",
		 bc64->fp_static[24].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[24]=%lx\n",
		 bc64->fp_static[24].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[25]=%lx\n",
		 bc64->fp_static[25].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[25]=%lx\n",
		 bc64->fp_static[25].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[26]=%lx\n",
		 bc64->fp_static[26].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[26]=%lx\n",
		 bc64->fp_static[26].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[27]=%lx\n",
		 bc64->fp_static[27].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[27]=%lx\n",
		 bc64->fp_static[27].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[28]=%lx\n",
		 bc64->fp_static[28].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[28]=%lx\n",
		 bc64->fp_static[28].u.bits[0]);
	fprintf (g_log_file, "\tfp_static[29]=%lx\n",
		 bc64->fp_static[29].u.bits[1]);
	fprintf (g_log_file, "\tfp_static[29]=%lx\n",
		 bc64->fp_static[29].u.bits[0]);

	fprintf (g_log_file, "\tBrRp=%lx\n", bc64->br_regs[0]);
	fprintf (g_log_file, "\tBrS0=%lx\n", bc64->br_regs[1]);
	fprintf (g_log_file, "\tBrS1=%lx\n", bc64->br_regs[2]);
	fprintf (g_log_file, "\tBrS2=%lx\n", bc64->br_regs[3]);
	fprintf (g_log_file, "\tBrS3=%lx\n", bc64->br_regs[4]);
	fprintf (g_log_file, "\tBrS4=%lx\n", bc64->br_regs[5]);
	fprintf (g_log_file, "\tBrT0=%lx\n", bc64->br_regs[6]);
	fprintf (g_log_file, "\tBrT1=%lx\n", bc64->br_regs[7]);

	fprintf (g_log_file, "\tApUNAT=%lx\n", bc64->br_regs[7]);
	fprintf (g_log_file, "\tApLC=%lx\n", bc64->lc);
	fprintf (g_log_file, "\tApEC=%lx\n", bc64->ec);
	fprintf (g_log_file, "\tApCCV=%lx\n", bc64->ccv);
	fprintf (g_log_file, "\tApDCR=%lx\n", bc64->dcr);

	// Register stack info
	fprintf (g_log_file, "\tRsPFS=%lx\n", bc64->pfs);
	fprintf (g_log_file, "\tRsBSP=%lx\n", bc64->bsp);
	fprintf (g_log_file, "\tRsBSPSTORE=%lx\n", bc64->bspstore);
	fprintf (g_log_file, "\tRsRSC=%lx\n", bc64->rsc);
	fprintf (g_log_file, "\tRsRNAT=%lx\n", bc64->rnat);

	// Trap Status Information
	fprintf (g_log_file, "\tStIPSR=%lx\n", bc64->ipsr);
	fprintf (g_log_file, "\tStIIP=%lx\n", bc64->iip);
	fprintf (g_log_file, "\tStIFS=%lx\n", bc64->ifs);
}

/*****************************************************************************/
/*****************************************************************************/
unsigned long
load_stacked_reg (struct sigcontext *sc, long num, long *nat)
{
	unsigned long rnat;
	unsigned long rnat_index;
	unsigned long *rnat_addr;
	long slots_in_frame = sc->sc_cfm & SOF_MASK;
	unsigned long *addr_of_first_reg_in_frame = (unsigned long *) sc->sc_ar_bsp;
	unsigned long *addr_of_last_reg_in_frame;
	long slot = num - 32;
	unsigned long *addr_of_requested_reg;
	unsigned long reg_val;

	addr_of_requested_reg =
		ia64_rse_skip_regs (addr_of_first_reg_in_frame, slot);
	addr_of_last_reg_in_frame =
		ia64_rse_skip_regs (addr_of_first_reg_in_frame, slots_in_frame);
	rnat_addr = ia64_rse_rnat_addr (addr_of_requested_reg);
	rnat = (rnat_addr > addr_of_last_reg_in_frame) ? 
		sc->sc_ar_rnat : *rnat_addr;
	//rnat_index = addr_of_requested_reg % 64;
	rnat_index = (unsigned long) addr_of_requested_reg - 
		((unsigned long) addr_of_requested_reg >> 6);
	*nat = (rnat >> rnat_index) & 1;
	reg_val = *nat ? 0 : *addr_of_requested_reg;

	return reg_val;
}

/*****************************************************************************/
/*****************************************************************************/
void
print_stacked_registers (struct sigcontext *sc)
{
	unsigned long reg_val;
	long reg_num, nat;
	long slots_in_frame = sc->sc_cfm & SOF_MASK;
	char buf[256];

	for (reg_num = 32; reg_num < (32 + slots_in_frame); reg_num++) {
		reg_val = load_stacked_reg (sc, reg_num, &nat);
		sprintf (buf, "\nr%ld: %lx", reg_num, reg_val);
		if (nat) {
			strncat (buf, "(NAT)", MIN(sizeof(buf) - 1 - strlen(buf), 5));
		}
		fprintf (g_log_file, "%s", buf);
	}
}

/*****************************************************************************/
/*****************************************************************************/
void
print_sigcontext64 (struct sigcontext *sc)
{

	int i;
	unsigned long pr_mask = 0;

	/* print general registers */
	for (i = 0; i < 32; i++) {
		fprintf (g_log_file, "\ngr[%d] == %lx", i, sc->sc_gr[i]);
	}

	print_stacked_registers (sc);

	/* print fp registers */

	for (i = 0; i < 32; i++) {
		fprintf (g_log_file, "\nfr[%d] == %lx", i,
			 sc->sc_fr[i].u.bits[1]);
		fprintf (g_log_file, "\nfr[%d] == %lx", i,
			 sc->sc_fr[i].u.bits[0]);
	}
	/* print branch registers */
	for (i = 0; i < 8; i++) {
		fprintf (g_log_file, "\nbr[%d] == %lx", i, sc->sc_br[i]);
	}

	fprintf (g_log_file, "\nsc_nat: %lx", sc->sc_nat);
	fprintf (g_log_file, "\nsc_ip : %lx", sc->sc_ip);
	fprintf (g_log_file, "\nsc_um: %lx", sc->sc_um);

	fprintf (g_log_file, "\nsc_pr: %lx", sc->sc_pr);
	fprintf (g_log_file, "\nsc_ar_fpsr: %lx", sc->sc_ar_fpsr);
	fprintf (g_log_file, "\nsc_ar_unat: %lx", sc->sc_ar_unat);
	fprintf (g_log_file, "\nsc_ar_lc: %lx", sc->sc_ar_lc);

	fprintf (g_log_file, "\nsc_ar_ccv: %lx", sc->sc_ar_ccv);

	fprintf (g_log_file, "\nsc_ar_pfs: %lx", sc->sc_ar_pfs);
	fprintf (g_log_file, "\nsc_ar_bsp: %lx", sc->sc_ar_bsp);

	fprintf (g_log_file, "\nsc_ar_rsc: %lx", sc->sc_ar_rsc);
	fprintf (g_log_file, "\nsc_ar_rnat: %lx", sc->sc_ar_rnat);

	fprintf (g_log_file, "\npredicates (0-63):");
	for (i = 0; i < 64; i++) {
		pr_mask = (1 << i);
		if ((i % 8) == 0) {
			fprintf (g_log_file, "\n");
		}

		fprintf (g_log_file, "%d", ((pr_mask & sc->sc_pr) > 0));
	}
	fprintf (g_log_file, "\n");
}

/*****************************************************************************/
/* print the given sigmask in a readable format
 * input:  set - the sigmask (sigset32_t format)
 *         size - size of sigmask in integers 
 */
/*****************************************************************************/
void
print_sigmask (sigset32_t *set, int size, char *header)
{
	static char *sig_names[] = { "",
		"SIGHUP", "SIGINT", "SIGQUIT", "SIGILL", "SIGTRAP",
		"SIGABRT", "SIGBUS", "SIGFPE", "SIGKILL",
		"SIGUSR1", "SIGSEGV", "SIGUSR2", "SIGPIPE", "SIGALRM",
		"SIGTERM", "SIGSTKFLT", "SIGCHLD", "SIGCONT", "SIGSTOP",
		"SIGTSTP", "SIGTTIN", "SIGTTOU", "SIGURG", "SIGXCPU",
		"SIGXFSZ", "SIGVTALRM", "SIGPROF", "SIGWINCH", "SIGIO",
		"SIGPOWR", "SIGSYS"
	};
	char mask[4096];
	char sig_name[32];
	int j;

	sprintf (mask, "%s - [", header);
	for (j = 1; j < SIG_NUM_PER_WORD; j++) {
		if (sigismember32 (set, j)) {
			strncat (mask, " ", MIN(sizeof(mask) - 1 - strlen(mask), 1));
			strncat (mask, sig_names[j], 
				 MIN(sizeof(mask) - 1 - strlen(mask), 
				     strlen(sig_names[j])));
		}
	}
	if (size > 1) {
		for (j = SIG_NUM_PER_WORD; j < (size * SIG_NUM_PER_WORD); j++) {
			if (sigismember32 (set, j)) {
				sprintf (sig_name, " RT%d", j - 32);
				strncat (mask, sig_name,
					 MIN(sizeof(mask) - 1 - strlen(mask), 
					     strlen(sig_name)));
			}
		}
	}
	strncat (mask, " ]", MIN(sizeof(mask) - 1 - strlen(mask), 2));
	DBPRINT (TRUE, "%s\n", mask);
}


void
print_current_sigmask (char *header)
{
	k_sigset_t set64;
	sigset32_t set32;

	long r = SYSCALL (rt_sigprocmask, (SIG_SETMASK, NULL, &set64, sizeof(set64)));
	sys_err_chk (r, "sigprocmask");

	sigset64_2_sigset32 (&set64, &set32);
	print_sigmask (&set32, WORDS_IN_MASK, header);
}
#endif

