/*	
 * btlib_signal.c
 *	
 *	
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include <ia32x_utils.h>

#include <ia32toia64.h>

#include "btlib_segment.h"
#include "btlib_signal.h"
#include "btlib_sig_utils.h"
#include "btlib_sync.h"
#include "btlib_syscall.h"
#include "btlib_asm.h"
#include "btlib_stats.h"
#include "btlib_core_file.h"

extern int g_core_file_policy;

/* 
 * the flags in the following mask should not be applied
 * to btlib handler, even if the translated app specifies them
 */
#define BTLIB_SIGHAND_FLAGS_MASK ~(SA_RESTART | SA_RESTORER | SA_ONSTACK)

/* a list of the signals that require permanent handlers
 * in btlib. last element in the list must be -1.
 * 0) don't add SIGTRAP here. it will give you hard time with
 *    the debugging support stuff
 * 1) try your best to avoid adding SIGCHLD to the list. it
 *    will give you a headache reaping the childern if the
 *    translated process chooses to ignore the signal.
 * 2) don't add signals which default behavior is 'ignore' -  you are going 
 *    to have a very hard time with managing interrupted syscalls.
 */
int m_permanent_handlers[] = { 
        SIGSEGV, SIGBUS, SIGFPE, SIGILL, PTEMU_SIG0, -1 };

int m_half_permanent_handlers[] = { 
	SIGQUIT, SIGTRAP, SIGABRT, SIGSYS, 
	SIGXCPU, SIGXFSZ, -1 };

/* time stamp for a sync signals */
static unsigned long m_time_stamp = 0;

/* long syscall handling */
unsigned long btl_ls_before_break_inst = 0;

/* function prototypes */
static unsigned find_ia32_handler (siginfo_t *siginfo,
                                   struct signal_handler_desc * handler_desc);
static void prepare_sighand_frame (siginfo_t *siginfo,
                                   struct signal_handler_desc *handler_desc,
                                   sigset32_t * old_mask, unsigned int trapno);
static void re_register_oneshot_handlers (int signo);
static int handle_interrupted_syscall (struct sigcontext *ctx);
static int handle_non_sync_signal (siginfo_t *, BTGENERIC_IA64_CONTEXT *);
static int handle_sync_signal (siginfo_t *, BTGENERIC_IA64_CONTEXT *, unsigned int *);
static void cancellation_signal (siginfo_t *siginfo);
static void register_internal_handlers (void);
static inline long check_expansion (siginfo_t *siginfo, unsigned esp);

/****************************************************************************** 
 * setup the stack frame for the IA-32 sig handler
 * input:
 *      siginfo - the siginfo as received by btlib's handler
 *      handler_desc - the discriptor of corresponding IA-32 signal handler
 *      old_mask - old signal mask (comparing with signal mask 
 *                 signal handler will use)
 *      trapno - the trap number (i.e., the interrupt number)
 * output: the eip and esp fields in the IA-32 context change
 *
 * remarks: the function delegate most of the work to two other functions, 
 *          each dealing with another frame type
 ******************************************************************************/
static void
prepare_sighand_frame (siginfo_t *siginfo,
                       struct signal_handler_desc *handler_desc,
                       sigset32_t *old_mask, unsigned int trapno)
{
	int signo;
	handler_info_t hi;
	void *frame;
	unsigned esp;
	int is_info_frame;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (cur_thread_data->sighandlers != NULL);
	assert (siginfo != NULL);
	assert (handler_desc != NULL);

	signo = siginfo->si_signo;

	hi.siginfo = siginfo;
	/* prepare the handler's definitions */
	hi.sa32.handler = handler_desc->sa.handler;
	hi.sa32.restorer = handler_desc->sa.restorer;
	hi.sa32.flags = handler_desc->sa.flags;
	memcpy (&hi.sa32.mask, &handler_desc->sa.mask, sizeof (hi.sa32.mask));
	/* prepare the IA-32 context */
	btgctx2sigctx_32 (cur_thread_data->bc32, &hi.sc32, &hi.fpstate);
	hi.sc32.trapno = trapno;
	/* for SIGSEGV (page fault), cr2 should save the refered memory */
	if ((signo == SIGSEGV) && (trapno == 14) && (siginfo->si_addr != 0)) {
		hi.sc32.cr2 = P2U (siginfo->si_addr);
	}
	memcpy (&hi.mask, old_mask, sizeof (hi.mask));
	hi.sc32.oldmask = hi.mask.sig[0];
	DBCODE (0,
		print_sigmask (&hi.mask, WORDS_IN_MASK, "prepare_sighand_frame"));

	/*
	 * If we're in a long system call then setup the IA-32 context
	 * according to the "interrupted" behaviour of the system call.
	 * the context at this point 
	 * 1) eip points pass the int 0x80 instruction
	 * 2) orig_eax holds the syscall id
	 * 3) eax holds the return value of the syscall
	 */
	if (cur_thread_data->syscall_id >= 0) {
		char need_restart = 0;

		DBPRINT (btl_prt_sig, "interrupted syscall_id=%d eax=%d eip=%x\n",
		         cur_thread_data->syscall_id, hi.sc32.eax, hi.sc32.eip);

		assert (is_long_syscall (cur_thread_data->syscall_id));

		need_restart =
			syscall_need_restart (cur_thread_data->syscall_id) && 
			(hi.sa32.flags & SA_RESTART);

		if (need_restart) {
			/* reexceute syscall */
			hi.sc32.eax = cur_thread_data->syscall_id;
			hi.sc32.eip -= 2;
		} else {
			/* finish syscall */
			hi.sc32.eax = (unsigned) -EINTR;
		}
	}

	/* prepare the frame */
	is_info_frame = (hi.sa32.flags & SA_SIGINFO) ? 1 : 0;
	esp = allocate_sighand_frame (&hi.sc32, hi.sa32.flags, is_info_frame);
	frame = U2P (esp);
	write_sighand_frame (&hi, frame, is_info_frame);

	/* update the context */
	cur_thread_data->bc32->esp = esp;
	cur_thread_data->bc32->eip = hi.sa32.handler;
}

/****************************************************************************** 
 * deque a pending signal and raise it
 * input: no input. the functino uses the thread's state for its operation
 * output: none
 ******************************************************************************/
void
btlib_deliver_signals ()
{
	/* get the thread's specific data */
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	DBPRINT (btl_prt_sig, "\n");

	restart_btg (NULL, NULL);
	assert (!"should not get here");
}

/*******************************************************************************
 * btlib_signal_handler()
 *
 * The handler is NOT reentrant. signals should be masked for it !!
 ******************************************************************************/
void
btlib_signal_handler (int signo, siginfo_t *siginfo, struct sigcontext *sc64)
{
	int action;
	BTGENERIC_IA64_CONTEXT bc64;
	unsigned int trapno;
	long retval;

	DBPRINT (btl_prt_sig, "signo=%d code=%d ip=%lx\n",
	         siginfo->si_signo, siginfo->si_code, sc64->sc_ip);

	STATCODE (btl_stats, btlib_update_stat (BTL_STAT_SIG, COUNT, 0ul));
	STATCODE (btl_stats && is_async_signal (siginfo),
		  btlib_update_stat (BTL_STAT_ASYNC_SIG, COUNT, 0ul));
	STATCODE (btl_stats && !is_async_signal (siginfo),
		  btlib_update_stat (BTL_STAT_SYNC_SIG, COUNT, 0ul));
	
	if (cur_thread_data->tid == 0) {
		cur_thread_data->tid = SYSCALL (gettid, ());
	}

	/* 
	 * if the thread is in the middle of permissions check then ignore
	 * the synchronous signal
	 */
	if (cur_thread_data->ignore_exception && !is_async_signal (siginfo)) {
		DBPRINT (btl_prt_sig, "ignore_exception\n");
#if PTEMU_ENABLED
		/* 
		 * if the process is ptraced then it should send itself
		 * ptemu_signal. this signal will be caught by btlib of the debugger
		 */
		if (cur_thread_data->ptrace & PTEMU_PTRACED) {
			cur_thread_data->ptrace |= PTEMU_DISCARD;
			retval = SYSCALL (tkill, (cur_thread_data->tid, PTEMU_SIG0));
			sys_err_chk (retval, "kill");
		}
#endif				/* PTEMU_ENABLED */
		longjmp (cur_thread_data->ignore_exception_jmpbuf, signo);
	}

	/* convert sigcontext to btgeneric format */
	sigctx2btgctx_64 (sc64, &bc64);

	/* treat sync signals differently than async ones */
	if (is_async_signal (siginfo)) {
		/*
		 * if the thread has not been initialed, it is unsafe to call 
		 * btg_canonize_context. Just enque the signal, then wait the initializtion 
		 * to be finished.
		 */
		if (cur_thread_data->is_run_context_ready != cur_thread_data->tid) {
			/* atomic inc */
			__sync_fetch_and_add_di (&m_time_stamp, 1ul);

			/* enque the signal for later delivery */
			btlib_que_enque (cur_thread_data->pending_que, siginfo,
					m_time_stamp);

			action = CONTINUE;
		} else {
			/* we can get a signal before a new thread had the chance
			 * to update the address of the long-syscall-flag */
			if (cur_thread_data->long_syscall_flag == NULL) {
				retval = iti_long_syscall_interface (
						&cur_thread_data->long_syscall_flag, NULL);
				assert (retval == 0);
			}

			DBPRINT (btl_prt_sig, "IN_SYSCALL=%d\n",
			         (sc64->sc_flags & IA64_SC_FLAG_IN_SYSCALL) ? 1 : 0);
			trapno = 0;
			/* atomic inc */
			__sync_fetch_and_add_di (&m_time_stamp, 1ul);

			/* enque the signal for later delivery */
			
			btlib_que_enque (cur_thread_data->pending_que, siginfo,
					 m_time_stamp);

			action = (*cur_thread_data->long_syscall_flag != NOT_IN_SYSCALL) ?
				handle_interrupted_syscall (sc64) :
			 	handle_non_sync_signal (siginfo, &bc64);
		}
	} else {
		DBPRINT (btl_prt_sig, "addr=%p isr=%lx\n",
		         siginfo->si_addr, siginfo->si_isr);
		action = handle_sync_signal (siginfo, &bc64, &trapno);
	}
	
	assert ((action == CONTINUE) || (action == SIGNAL) ||
		(action == IA32_SYSCALL) || (action == DELIVER) || 
		(action == RUN));
	/* 
	 * if the signal should be delivered to the translated process then
	 * we need to copy the siginfo to tia_signal
	 */
	if (action == SIGNAL) {
		memcpy (&cur_thread_data->tia_signal.siginfo,
			siginfo, sizeof (cur_thread_data->tia_signal.siginfo));
		cur_thread_data->tia_signal.trapno = trapno;
	}

	/* 
	 * if the needed action is other than to continue then we restart
	 * from the first nesting level
	 */
	if (action != CONTINUE) {
		unsigned long *plabel;
		cur_thread_data->longjmp_status = action;
		/* set new ip to do_longjmp */
		plabel = (unsigned long *) do_longjmp;
		sc64->sc_ip = PLABEL_FUNC (plabel);
		sc64->sc_gr[1] = PLABEL_GP (plabel);
	}
#if PTEMU_ENABLED
	/* 
	 * if the process is ptraced then it should send itself
	 * ptemu_signal. this signal will be caught by btlib of the debugger
	 */
	if (cur_thread_data->ptrace & PTEMU_PTRACED) {
		/* 
		 * this is the only place where it is possible for a debugee to send 
		 * itself the ptemu-signal when the IA-32 context is not canonized. 
		 * when this is the case - the signal must be discarded
		 */
		if (action == SIGNAL) {
			cur_thread_data->ptrace &= ~PTEMU_DISCARD;
		} else {
			cur_thread_data->ptrace |= PTEMU_DISCARD;
		}
		retval = SYSCALL (tkill, (cur_thread_data->tid, PTEMU_SIG0));
		sys_err_chk (retval, "kill");
	}
#endif		/* PTEMU_ENABLED */

	DBPRINT (btl_prt_sig, "action=%d\n", action);
}

/*******************************************************************************
 * void do_longjmp
 *
 * Purpose: this is where control resumes after the wrapping 64 bit signal
 *          returns. After excution of __siglongjmp, control will return to btrans_jump
 *          which will continue according to the current->longjmp_status.
 * 
 *******************************************************************************/
void
do_longjmp ()
{
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	DBPRINT (0, "longjmp_status=%d\n", cur_thread_data->longjmp_status);
	longjmp (cur_thread_data->btgeneric_run_context, cur_thread_data->longjmp_status);
}

/************************************************************************** 
* handle synchronous signals. these are supposed to be a result of hw interrupt
* input: 
*       siginfo - the signal info as received at btlib's handler
*       bc64 - the 64 bit context at the point of the signal
*       trapno - trap number (interrupt number)
* output:
*	returns the action to be taken as the response for the
*	signal. **MUST** be one that restart btgeneric from the 
*	first nesting level.
* remarks: the order of calling the api functions must be
*           btg_ia32_canonize_context
*           btg_convert_exception
*           btg_run_exit
*
**************************************************************************/
static int
handle_sync_signal (siginfo_t *siginfo, BTGENERIC_IA64_CONTEXT *bc64, unsigned int *trapno)
{
	int action = -1;
	BT_EXCEPTION_RECORD excpt_rec;
	unsigned excpt_code;
	int discard = 0;

	assert (siginfo != NULL);
	assert (bc64 != NULL);

	/* convert the signal info to btgeneric exception code */
	excpt_code = signal2except (siginfo);
	excpt_rec.exception_code = excpt_code;
	excpt_rec.ia64_iipa = 0ul;	/* not supported in Linux yet */
	excpt_rec.ia64_isr = siginfo->si_isr;
	assert (excpt_rec.exception_code != BT_NO_EXCEPT);
	DBPRINT (btl_prt_sig, "code=%u isr=%lx\n",
	         excpt_rec.exception_code, excpt_rec.ia64_isr);

	if (btl_btg_initialized) {
		/* 
		 * canonize the context and
		 * let btgeneric decide how to treat this signal 
		 */
		excpt_code = BTGENERIC_IA32_CANONIZE_CONTEXT (NULL, bc64, &excpt_rec);
		DBPRINT (0, "\nexcpt_code=%d\n", excpt_rec.exception_code);
		discard = (excpt_code == BT_NO_EXCEPT);
		switch (excpt_code) {
		case BT_EXCEPT_ACCESS_VIOLATION:
		case BT_EXCEPT_RNX_ACCESS_VIOLATION:
			*trapno = 14;
			break;
		case BT_EXCEPT_FLT_DENORMAL_OPERAND:
		case BT_EXCEPT_FLT_DIVIDE_BY_ZERO:
		case BT_EXCEPT_FLT_INEXACT_RESULT:
		case BT_EXCEPT_FLT_INVALID_OPERATION:
		case BT_EXCEPT_FLT_OVERFLOW:
		case BT_EXCEPT_FLT_UNDERFLOW:
		case BT_EXCEPT_FLT_STACK_CHECK:
			*trapno = 16;
			break;
		case BT_EXCEPT_FLOAT_MULTIPLE_FAULTS:
		case BT_EXCEPT_FLOAT_MULTIPLE_TRAPS:
			*trapno = 19;
			break;
		case BT_NO_EXCEPT:
			break;
		default:
			assert (!"never get here!");
			break;
		}
	
		if (discard) {
			/* 
			 * if the signal is discarded - we need
			 * to resotre btlib's handlers that were declared as ONESHOT
			 * because as far as the translated process is concered, the
			 * signal never occured
			 */
			re_register_oneshot_handlers (siginfo->si_signo);
			action = RUN;
		} else {
			if (check_expansion (siginfo, cur_thread_data->bc32->esp)) {
				action = RUN;
			} else {
				/* 
				 * if the signal is not discarded then raise it.
				 * no need to see if the signal is masked because
				 * it is not possible to mask signals that are the 
				 * result of hw interrupt
				 */

				action = SIGNAL;
			}
		}
	} else {
		long retval;
		retval = check_expansion (siginfo, 0);
		if (retval) {
			action = CONTINUE;
		} else {
			DBPRINT (1, "unhandled signal before btgeneric is initialized\n");
			btlib_abort_all();
		}
	}	
	return action;
}

/************************************************************************** 
 * handle asynchronous signals. 
 * input: 
 *      siginfo - the signal info as received at btlib's handler
 *      bc64 - the 64 bit context at the point of the signal
 * output:
 *      returns the action to be taken as the response for the	signal 
 *
 * remarks: because the signal is not synchronous, it may be harmful
 *          to stop btgeneric and restart it from the first nesting level.
 *          therefore, a simple call to CANONIZE_CONTEXT is not adequate here.
 *          instead, we first call CHECK_SUSPEND_CONTEXT, which returns an
 *          indication whether btgeneric can be stopped and restarted from
 *          another point or not.
 **************************************************************************/
static int
handle_non_sync_signal (siginfo_t *siginfo UNUSED, BTGENERIC_IA64_CONTEXT *bc64)
{
	int btg_ret;
	unsigned excpt_code;
	BT_EXCEPTION_RECORD excpt_rec;
	int action = -1;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (siginfo != NULL);
	assert (bc64 != NULL);

	/* see if suspension is possible now */
	btg_ret = BTGENERIC_CHECK_SUSPEND_CONTEXT (NULL, bc64);
	DBPRINT (btl_prt_sig, "suspend-status=%d\n", btg_ret);
	/* there are 3 possible return values here:
	 * - CANONIZED SUSPEND_STATE_CONSISTENT, 
         *      IA-32 state is consistent, probably inside btlib (may hold
	 *      a cs or unfreed memory), let the execution continue.
	 * - SUSPEND_STATE_READY_FOR_CANONIZATION - 
	 *      suspension is possible - canonize the IA-32 context
	 *      and raise the signal
	 * - BAD_SUSPEND_STATE 
	 *      cannot suspend execution - continue execution
	 */
	switch (btg_ret) {
	case SUSPEND_STATE_READY_FOR_CANONIZATION:
		/* suspension is possible */
		excpt_rec.exception_code = BT_NO_EXCEPT;
		excpt_rec.ia64_iipa = excpt_rec.ia64_isr = 0ul;
		/* canonize the IA-32 context */
		excpt_code =
		    BTGENERIC_IA32_CANONIZE_CONTEXT (NULL, bc64, &excpt_rec);
		assert (excpt_code == BT_NO_EXCEPT);
		action = DELIVER;
		break;
	case BAD_SUSPEND_STATE:
		/* ask ia32exec.bin to pass control to ia32x asap */
		BTGENERIC_EXIT_SIMULATION_REQUEST ();
		/* fall through */
	case SUSPEND_STATE_CONSISTENT:
		/* suspension is NOT possible */
		action = CONTINUE;
		break;
	case SUSPEND_STATE_CANONIZED:
		/* should never get such state because we just check the state */
		assert (!"never get here!");
		break;
	case SUSPEND_STATE_INACCESIBLE:
		assert (!"suspend state is not available");
		break;
	default:
		assert (!"unknown state");
		break;
	}

	return action;
}

/******************************************************************************
 * registers internal signal handlers - permanent/half permanent 
 * input:   none
 * output:  none
 * remarks: the function runs over the list of signals that
 *          require permanent handler/half permanent handler  and register a handler
 *          for all of them the function runs over the list of signals that
 *          require permanent handler and register a handler
 *          for all of them
 ******************************************************************************/
static void
register_internal_handlers ()
{
	struct sigaction sa;
	long retval;
	int *perm;

	memset (&sa, 0, sizeof (sa));
	sa.sa_flags = SA_SIGINFO;

	/* 
	 * prepare the handler's definitions 
	 * the second argument need to be other than SIG_IGN and SIG_DFL
	 */
	sa.sa_handler = (PLABEL_PTR_TYPE) __btlib_signal_handler;
	sigfillset (&sa.sa_mask);

	/* register all permanent handlers */
	for (perm = m_permanent_handlers; *perm != -1; perm++) {
		assert (perm - m_permanent_handlers <= _NSIG);
		DBPRINT (0, "signo=%d\n", *perm);
		retval = SYSCALL (rt_sigaction, (*perm, &sa, NULL, sizeof (k_sigset_t)));
		sys_err_chk (retval, "sigaction");
	}


	/* register all half permanent handlers */
	for (perm = m_half_permanent_handlers; *perm != -1; perm++) {
		struct sigaction hsa;
		retval = SYSCALL (rt_sigaction, 
				  (*perm, NULL, &hsa, sizeof (k_sigset_t)));
		sys_err_chk (retval, "sigaction");

		
		/* if current signal handler is SIG_IGN, don't register signal handler 
		   because core file should not be generated */
		if (hsa.sa_handler == SIG_IGN) {
			continue;
		}

		retval = SYSCALL (rt_sigaction, (*perm, &sa, NULL, sizeof (k_sigset_t)));
		sys_err_chk (retval, "sigaction");
	}
}

/**************************************************************************
 * free the given sighand-table 
 * input:   table
 * output:  none
 * remarks:
 ****************************************************************************/
void
btlib_try_free_sighand_table (btlib_thread_data_t *tls)
{
	int ref_count;
	int common_sighandlers;

	common_sighandlers =
		(tls->sighandlers->cs == cur_thread_data->sighandlers->cs) ? 1 : 0;
	if (common_sighandlers) {
		btlib_acquire_lock (BTL_LOCK_IND_SIGHAND);
	} else {
		btlib_enter_critical_section (tls->sighandlers->cs, 1);
	}
	tls->sighandlers->ref_count--;
	ref_count = tls->sighandlers->ref_count;
	assert (ref_count >= 0);
	if (common_sighandlers) {
		btlib_release_lock (BTL_LOCK_IND_SIGHAND);
	} else {
		btlib_leave_critical_section (tls->sighandlers->cs);
	}

	if (ref_count == 0) {
		btlib_free_critical_section (tls->sighandlers->cs);
		free (tls->sighandlers);
		tls->sighandlers = NULL;
	}
}

/**************************************************************************
 * creates a new sighand-table and initialize it
 * input:   none
 * output:  table's address on success, 0 on failure
 * remarks:
 ****************************************************************************/
sighand_table *
btlib_create_sighand_table ()
{
	sighand_table *table = NULL;

	DBPRINT (0, "\n");

	/* allocate memory for the table */
	table = (sighand_table *) malloc (sizeof (sighand_table));
	if (table == NULL) {
		DBPRINT (TRUE, "malloc failed\n");
		btlib_abort ();
	}

	/* allocate memory for the table's critical section object */
	table->cs = btlib_create_critical_section ();
	btlib_init_critical_section (table->cs);

	assert (SIG_DFL == 0);	/* just to make sure we set everything to 
				   SIG_DFL at the next line */
	memset (&table->handlers, 0, sizeof (table->handlers));
	table->ref_count = 1;

	return table;
}

/**************************************************************************
* do the actual work for sigaction syscalls
* input: 
*       signo - the signal no.
*       sa - the requested sigaction
*       osa - output param, will hold the old sigaction
*       size - size of sigset32_t used in this call
* output: the syscall return value
****************************************************************************/
long
emulate_sigaction (int signo, sigaction32_t *sa32, sigaction32_t *osa32, int size)
{
	sighand_table *sighand;
	long retval;
	struct sigaction sa;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	DBPRINT (btl_prt_sig, "signo=%d\n", signo);
	DBCODE (0, print_sigaction32 ("sa32", sa32));

	/* verify parameters */
	if ((signo < 1) || (signo > size * SIG_NUM_PER_WORD) ||
	    ((sa32 != NULL) && ((signo == SIGKILL) || (signo == SIGSTOP)))) {
		return -EINVAL;
	}

	if (!sa32 && !osa32) {
		/* 
		 * in this case we do nothing but returning 0 if the signal is 
		 * valid and -EINVAL if not. since the signal validity was just
		 * checked in the previous paragraph, all we have to do here is
		 * to return 0. 
		 */
		return 0;
	}

	/* PTEMU_SIG0 is used for debugging - don't let the user use it
	 * if we are under a debugger */
	if (signo == PTEMU_SIG0) {
		DBPRINT (TRUE, "user attempts to use ptemu signal\n");
	}

	sighand = cur_thread_data->sighandlers;
	assert (sighand != NULL);

	/* the handlers table may be shared among threads. we need to 
	 * sync the access to it */
	btlib_acquire_lock (BTL_LOCK_IND_SIGHAND);

	if (osa32) {
		/* copy the current definitions of the signal dispostion to
		 * the caller's struct */
		*osa32 = sighand->handlers[signo].sa;
		DBCODE (0, print_sigaction32 ("osa32", osa32));
	}

	if (sa32) {
		int half_perm = has_half_permanent_handler (signo);
		int perm = has_permanent_handler (signo);

		/* change the signals disposition for btlib - 
		   the disposition is not changed for permanent signals */
		if (!perm) {
			/* prepare the signal handler details */
			memset (&sa, 0, sizeof(sa));

			/* for signals with half permanent signal handler, BTLib 
			   must intercept SIG_DFL to enable core file */
			if ((sa32->handler == (unsigned) SIG_IGN) ||  
			    ((sa32->handler == (unsigned) SIG_DFL) && !half_perm)) {
				sa.sa_handler = (__sighandler_t)U2P(sa32->handler);
			} else {
				sigfillset (&sa.sa_mask);
				sa.sa_handler = (PLABEL_PTR_TYPE) __btlib_signal_handler;
				sa.sa_flags = (sa32->flags | SA_SIGINFO) &
					BTLIB_SIGHAND_FLAGS_MASK;
			}
			retval = SYSCALL (rt_sigaction, 
					  (signo, &sa, NULL, sizeof (k_sigset_t)));
			sys_err_chk (retval, "rt_sigaction");
			DBPRINT (0, "signum (%d) registered handler\n", signo);
		}

		/* keep the IA-32 handler definition.
		   according to POSIX, never set SIGSTOP/SIGKILL */
		sigdelset32 (&sa32->mask, SIGSTOP);
		sigdelset32 (&sa32->mask, SIGKILL);
		sighand->handlers[signo].sa = *sa32;
		
		btlib_release_lock (BTL_LOCK_IND_SIGHAND);
		if (!perm) {
			return 0;
		}


		/* POSIX Issue 6:
		   1266 Setting a signal action to SIG_IGN for a signal that is pending 
			shall cause the
		   1267 pending signal to be discarded, whether or not it is blocked.
		 */
		assert (has_permanent_handler (signo));
		if ((sa32->handler == P2U (SIG_IGN)) ||
		    ((sa32->handler == P2U (SIG_DFL)) &&
		     (SIGNAL_DEFAULT_IGNORE (signo)))) {
			btlib_que_remove_signo (cur_thread_data->pending_que, signo);
			btlib_que_remove_signo (btl_shared_que, signo);
		}
	} else {
		btlib_release_lock (BTL_LOCK_IND_SIGHAND);
	}


	return 0;
}

/**************************************************************************
 * send a signal to the translated process
 * input:   the tia_signal field in the thread's  data structure is the 
 *          input for the functions
 * output:  void
 * remarks: this is memrely a wrapper function to btlib_run_ia32_handler
 *          for cases where we need to raise the signal from the first 
 *          nesting level
 ***************************************************************************/
void
btlib_raise_signal ()
{
	siginfo_t *siginfo = NULL;
	int signo;

	/* get the thread's specific data */
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (cur_thread_data->sighandlers != NULL);
	DBPRINT (0, "\n");

	siginfo = &cur_thread_data->tia_signal.siginfo;
	signo = siginfo->si_signo;
	assert ((signo >= 1) && (signo <= _NSIG));

	btlib_run_ia32_handler (siginfo, NULL,
				cur_thread_data->tia_signal.trapno);
	assert (!"should not get here");
}

/************************************************************************** 
* re register a one-shot signal
* input: thread - 
*        signo - 
* output: none
**************************************************************************/
static void
re_register_oneshot_handlers (int signo)
{
	struct sigaction act;
	long retval;

	DBPRINT (0, "signo=%d\n", signo);
	btlib_acquire_lock (BTL_LOCK_IND_SIGHAND);
	/* 
	 * register the handler according to the definitions found
	 * in the thread's data struct
	 */
	if (cur_thread_data->sighandlers->handlers[signo].sa.flags & SA_ONESHOT) {
		act.sa_handler = (PLABEL_PTR_TYPE) __btlib_signal_handler;
		act.sa_flags =
			cur_thread_data->sighandlers->handlers[signo].sa.flags;
		act.sa_flags &= BTLIB_SIGHAND_FLAGS_MASK;
		memcpy (&act.sa_mask,
			&cur_thread_data->sighandlers->handlers[signo].sa.mask,
			sizeof (act.sa_mask));
		retval = SYSCALL (rt_sigaction, (signo, &act, NULL, sizeof (k_sigset_t)));
		sys_err_chk (retval, "rt_sigaction");
	}
	btlib_release_lock (BTL_LOCK_IND_SIGHAND);
}

/**************************************************************************
 * find the IA-32 handler and make preparations for calling it
 * input: 
 *      siginfo - description of the signal
 *      bc32 - the FULL IA-32 context
 * output: the IA-32 handler
 ***************************************************************************/
static unsigned
find_ia32_handler (siginfo_t *siginfo, struct signal_handler_desc *handler_desc)
{
	int signo;
	unsigned handler;

	assert (siginfo != NULL);
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (cur_thread_data->sighandlers != NULL);

	btlib_acquire_lock (BTL_LOCK_IND_SIGHAND);

	/* find the handler of the translated process */
	signo = siginfo->si_signo;
	handler = cur_thread_data->sighandlers->handlers[signo].sa.handler;
	DBPRINT (0, "signo=%d,handler=0x%x\n", signo, handler);

	/* if there is no handler then return */
	if ((handler == (unsigned) SIG_IGN) || (handler == (unsigned) SIG_DFL)) {
		btlib_release_lock (BTL_LOCK_IND_SIGHAND);
		return (unsigned) handler;
	}

	/* if we got here then there is a registered signal handler */
	assert ((handler != (unsigned) SIG_DFL) && (handler != (unsigned) SIG_IGN));
	memcpy (handler_desc, &cur_thread_data->sighandlers->handlers[signo],
		sizeof (*handler_desc));

	/* 
	 * if the handler was registerd as ONESHOT then 
	 * clear the handler's info from the table and
	 * unregister the btlib's handler (this is necessary
	 * since we may get to this point from BtlIA32Interrupt
	 * and not only as the result of a signal) 
	 */
	if (cur_thread_data->sighandlers->handlers[signo].sa.flags & SA_ONESHOT) {
		cur_thread_data->sighandlers->handlers[signo].sa.handler = 0;

		/* unregister btlib's handling  (if not permanent/half permanent) */
		if (!has_permanent_handler (signo) && !has_half_permanent_handler(signo)) {
			unreg_btlib_handler (signo);
		}
	}
	btlib_release_lock (BTL_LOCK_IND_SIGHAND);

	return handler;
}

/************************************************************************** 
* implement the sigerturn IA-32 syscalls
* intput:
*  cur_bc32 - the IA-32 context at the point where the syscall
*             was called
*  is_rt - whether the frame is for rt signals or not, allowing
*          this functino to support both sigreturn and rt_sigreturn
*          syscalls
*  output: none
*  remarks: the function sets the IA-32 context according to the stack frame 
*           of the signal handler and restart btgeneric from the first 
*           nesting level.
**************************************************************************/
void
btlib_sigreturn (BTGENERIC_IA32_CONTEXT *cur_bc32, int is_rt)
{
	frame_info_t frame_info;
	BTGENERIC_IA32_CONTEXT bc32;

	assert (cur_bc32 != NULL);
	/* get the thread's specific data */
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (cur_thread_data->sighandlers != NULL);

	DBPRINT (0, "is_rt=%d\n", is_rt);

	/* extract info from the signal frame */
	read_sighand_frame (cur_bc32->esp, &frame_info, is_rt);

	/* convert the context to btgeneric format */
	sigctx2btgctx_32 (frame_info.ctx, &bc32);
	DBPRINT (0, "es=%x fs=%x gs=%x\n", bc32.es, bc32.fs, bc32.gs);

	/* apply the mask to btlib */
	DBCODE (0,
		print_sigmask (&frame_info.mask, WORDS_IN_MASK, "btlib_sigreturn"));
	set_sigmask (&frame_info.mask);

	DBPRINT (FALSE, "sigreturn from eip = %x, esp = %x\n", bc32.eip, bc32.esp);

	/* restart btgeneric */
	cur_thread_data->syscall_id = IA32_SYSCALL_NONE; /* needed for restart_btg */
	restart_btg (&bc32, NULL);
	assert (!"should not get here");
}

/******************************************************************************
 * if there are pending signals then try to deliver them
 * otherwise, restart btgeneric
 * input : new_bc32 - if not null then this parameter is used for
 *                    setting a new IA-32 context for BTGeneric 
 *         handler_mask - signal mask used in IA-32 signal handler 
 * output: none
 ******************************************************************************/
void
restart_btg (BTGENERIC_IA32_CONTEXT *new_bc32, sigset32_t *oldmask)
{
	int pend;
	siginfo_t siginfo;
	BT_STATUS_CODE btg_ret;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	if (new_bc32 != NULL) {
		/*
		 * btlib_api_ia32_reenter/btlib_sigreturn are the only two cases 
		 * that segment registers can be changed, both of them call 
		 * restart_btg with a non NULL new_bc32 
		 */
		new_bc32->composition_bits = 
			CTX32_GP | CTX32_CRITICAL | CTX32_SEGMENT |
			CTX32_FLOATING_POINT | CTX32_XMM; 
		btg_ret = BTGENERIC_IA32_CONTEXT_SET (NULL, new_bc32);
		assert (btg_ret == BT_STATUS_SUCCESS);

		btlib_check_segment (cur_thread_data->bc32);
	}

	/* see if debugger is waiting for us to stop *before* we remove a 
	   signal from the que */
	FINISH_CONT_PROCESS ();

	/* try to get a pending signal */
	pend = btlib_que_deque (&siginfo);

	if (pend) {

		/* deliver signal */
		DBPRINT (0, "signo=%d\n", siginfo.si_signo);

#if PTEMU_ENABLED
		if (cur_thread_data->ptrace & PTEMU_PTRACED) {
			/*
			 * notify the debugger of the signal -
			 * the debugger will look at the tia_siganal field
			 * in the tls, for the signo
			 */
			cur_thread_data->tia_signal.siginfo.si_signo =
				siginfo.si_signo;
			SYSCALL (tkill, (cur_thread_data->tid, PTEMU_SIG0));
		}
#endif	/* PTEMU_ENABLED */

		btlib_run_ia32_handler (&siginfo, oldmask, 0);
		assert (!"should not get here");
	}

	/* 
	 * if we got here then there were no pending signals.
	 * interrupted syscalls should have been handled at 
	 * earlier stages
	 */
	assert (cur_thread_data->syscall_id == IA32_SYSCALL_NONE);

	longjmp (cur_thread_data->btgeneric_run_context, RUN);
}

/************************************************************************** 
 * forces a cancellation signal
 * input:   siginfo - the siginfo to be send to the process
 * output:  none
 * remarks: the function kills the current process by sending it
 *          the given signal
 *          the currnet process is killed without cleanup work, actually -
 *          there is no much point in doing the cleanup work in this case
 *          because the signal handler is changed to the default action.
 *          the next thread that will get this signal will not have the
 *          chance to do the clean up work anyway
 **************************************************************************/
static void
cancellation_signal (siginfo_t *siginfo)
{
	long retval;
	struct sigaction act;
	int signo = siginfo->si_signo;
	int core_limit;
	struct rlimit rlp;
	int status;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	DBPRINT (btl_prt_sig, "signo=%d ptrace=%lx\n", signo,
		 cur_thread_data->ptrace);
	
	status = sigsetjmp(cur_thread_data->btgeneric_run_context, 0);
	if (status == 0) {
		if (should_core_dump(signo)) {
			core_file(siginfo);
		}
	} else {
		/* cleanup */
		int pid;
		char core_file_name[PATH_MAX];
		pid = SAFE_SYSCALL (getpid, ());
		if (g_core_file_policy == DONT_SET_CORE_BIT) {
			sprintf(core_file_name, "./core.%d", pid); 
		} else { /* g_core_file_policy == SET_CORE_BIT */
			sprintf(core_file_name, "./core.%d.32", pid);	
		}
		if (status != -1) {
			SAFE_SYSCALL (close, (status));
			SAFE_SYSCALL (unlink, (core_file_name));
		}
		btlib_release_lock (BTL_LOCK_IND_SUSPENSION);	
	}
	if (g_core_file_policy == DONT_SET_CORE_BIT) {
		rlp.rlim_cur = rlp.rlim_max = 0;
		core_limit = SAFE_SYSCALL (setrlimit, (RLIMIT_CORE, &rlp));
	}

	/* unregister btlib's handler for the signal */
	memset (&act, 0, sizeof (act));
	act.sa_handler = SIG_DFL;
	retval = SYSCALL (rt_sigaction, (signo, &act, NULL, sizeof (k_sigset_t)));
	sys_err_chk (retval, "sigaction");

#if PTEMU_ENABLED
	if (cur_thread_data->ptrace & PTEMU_PTRACED) {
		/* inform the debugger */
		cur_thread_data->tia_signal.siginfo.si_signo = signo | PTEMU_CANCEL;
		retval = SYSCALL (tkill, (cur_thread_data->tid, PTEMU_SIG0));
		sys_err_chk (retval, "kill(1)");
	} else
#endif				// PTEMU_ENABLED
	{
		/* let btgeneric prepare stats on IA32-EL run */
		BTGENERIC_NOTIFY_EXIT ();
		/* the following api allows btg to print the rolling-trace records */
		DBCODE (TRUE, BTGENERIC_EXCEPTION_DEBUG_PRINT ());
		DBCODE (TRUE, btlib_print_rolling_trace_array ());
		/* send it (==suicide) */
		retval = KILL_THREAD_SIGNAL (cur_thread_data->tid, signo);
		sys_err_chk (retval, "kill(2)");
	}

	assert (!"should not get here");
}

/************************************************************************** 
 * find the IA-32 handler and run it
 * input  : siginfo - the signal's description
 *          oldmask - old signal mask, valid only if getting here by 
 *                    sigsuspend system call
 * output : none
 * remarks: if no handler is registered by the translated process
 *          then the function emulates the OS behavior - either 
 *          ignore the signal or terminate the process
 **************************************************************************/
void
btlib_run_ia32_handler (siginfo_t *siginfo, sigset32_t *oldmask, unsigned int trapno)
{
	unsigned   handler;
	int        signo = siginfo->si_signo;
	long       retval;
	sigset32_t mask;
	struct signal_handler_desc handler_desc;
	enum { IGNORE, TERMINATE, RAISE } action;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (siginfo != NULL);

	DBPRINT (btl_prt_sig, "signo=%d\n", siginfo->si_signo);
	/* find the ia32 handler for the given signal */
	handler = find_ia32_handler (siginfo, &handler_desc);
	DBPRINT (btl_prt_sig, "find_prepare=%x\n", handler);


	if (handler == (unsigned) SIG_DFL) {
		action = SIGNAL_DEFAULT_IGNORE (signo) ? IGNORE : TERMINATE;
	} else if (handler == (unsigned) SIG_IGN) {
		/* sync signals cannot be ignored */
		action = is_async_signal (siginfo) ? IGNORE : TERMINATE;
	} else {
		mask = cur_thread_data->sigmask;

		/* prepare the stack frame for the IA-32 handler */
		prepare_sighand_frame (siginfo, &handler_desc, 
				       oldmask ? oldmask : &mask, trapno);

		assert (!sigismember32 (&mask, SIGKILL) && !sigismember32 (&mask, SIGSTOP));

		/* 
		 * set the signal mask for the handler. it is
		 * (current-mask | handler's-specific-mask | signo )
		 */
		/* add the handler's specific mask */
		sig_or_mask (&mask, &handler_desc.sa.mask, WORDS_IN_MASK);
		if (!(handler_desc.sa.flags & SA_NOMASK)) {
			sigaddset32 (&mask, signo);
		}

		/* apply the mask to btlib  */
		retval = set_sigmask (&mask);
		assert (retval == 0);

		/* clear the fp pending exceptions in case of numeric exception */
		if ((signo == SIGFPE) && (SI_FROMKERNEL(siginfo))) {
			cur_thread_data->bc32->fpu_state.status_word &=
				~FPSW_EXCEPTIONS_MASK;
		}

		/* the eip/esp for the handler were already set by 
		   prepare_sighand_frame */
		action = RAISE;
	}
	/* at this point info about interrupted syscall is 
	   already stored in the handler's description struct */
	cur_thread_data->syscall_id = IA32_SYSCALL_NONE;

	DBPRINT (0, "action=%d\n", action);

	switch (action) {
	    case IGNORE:
		restart_btg (NULL, NULL);
		assert (!"should not get here");
	    case TERMINATE:
		cancellation_signal (siginfo);
		assert (!"should not get here");
	    case RAISE:
		longjmp (cur_thread_data->btgeneric_run_context, RUN);
	    default:
		assert (!"unknown action");
	}
	assert (!"should not get here");
}

/************************************************************************** 
 * initialize the signal module
 **************************************************************************/
void
btlib_init_signal_module ()
{
	long ret RELEASE_UNUSED;

	/* register internal handlers signal handlers */
	register_internal_handlers ();

	/* initialize the shared que */
	btlib_que_init (&btl_shared_que);

	/* get the address of the long syscall break inst */
	ret = iti_long_syscall_interface (NULL, &btl_ls_before_break_inst);
	assert (ret == 0);
}

/************************************************************************** 
 * convert execption codes used by btgeneric to their
 * corresponding siginfo settings
 **************************************************************************/
void
except2signal (BTGENERIC_IA32_CONTEXT *bc32, unsigned ic, siginfo_t *siginfo, int *trapno)
{
	switch (ic) {
	case BT_EXCEPT_ACCESS_VIOLATION:
	case BT_EXCEPT_RNX_ACCESS_VIOLATION:
		siginfo->si_signo = SIGSEGV;
		siginfo->si_code = SI_KERNEL;
		break;
	case BT_EXCEPT_DATATYPE_MISALIGNMENT:
		siginfo->si_signo = SIGBUS;
		siginfo->si_code = BUS_ADRALN;
		break;
	case BT_EXCEPT_ARRAY_BOUNDS_EXCEEDED:
		siginfo->si_signo = SIGSEGV;
		siginfo->si_code = SI_KERNEL;
		*trapno = 5;
		break;
	case BT_EXCEPT_FLT_DIVIDE_BY_ZERO:
		siginfo->si_signo = SIGFPE;
		siginfo->si_code = FPE_FLTDIV;
		*trapno = 16;
		break;
	case BT_EXCEPT_FLT_INEXACT_RESULT:
		siginfo->si_signo = SIGFPE;
		siginfo->si_code = FPE_FLTRES;
		*trapno = 16;
		break;
	case BT_EXCEPT_FLT_INVALID_OPERATION:
	case BT_EXCEPT_FLT_STACK_CHECK:
		siginfo->si_signo = SIGFPE;
		siginfo->si_code = FPE_FLTINV;
		*trapno = 16;
		break;
	case BT_EXCEPT_FLT_OVERFLOW:
		siginfo->si_signo = SIGFPE;
		siginfo->si_code = FPE_FLTOVF;
		*trapno = 16;
		break;
	case BT_EXCEPT_FLT_UNDERFLOW:
	case BT_EXCEPT_FLT_DENORMAL_OPERAND:
		siginfo->si_signo = SIGFPE;
		siginfo->si_code = FPE_FLTUND;
		*trapno = 16;
		break;
	case BT_EXCEPT_INT_DIVIDE_BY_ZERO:
		siginfo->si_signo = SIGFPE;
		siginfo->si_code = FPE_INTDIV;
		*trapno = 0;
		break;
	case BT_EXCEPT_INT_OVERFLOW:
		siginfo->si_signo = SIGFPE;
		siginfo->si_code = FPE_INTOVF;
		*trapno = 4;
		break;
	case BT_EXCEPT_PRIV_INSTRUCTION:
		siginfo->si_signo = SIGSEGV;
		siginfo->si_code = SI_KERNEL;
		*trapno = 13;
		break;
	case BT_EXCEPT_ILLEGAL_INSTRUCTION:
		siginfo->si_signo = SIGILL;
		siginfo->si_code = ILL_ILLOPN;
		*trapno = 6;
		break;
	case BT_EXCEPT_FLOAT_MULTIPLE_FAULTS:
	case BT_EXCEPT_FLOAT_MULTIPLE_TRAPS:
		/*
		 * SIMD exception. should be translated to the corresponding
		 * si_code recognized by Linux. we check mxcsr and pick the
		 * si_code for the exception with the highest priority.
		 * for priority of SIMD exceptions please refer to - 
		 * Intel Architecture Software Developer's Manual 
		 *      Volume 3  - System Programming
		 *      Chapter 5 - Interrupt and Exception Handling
		 *      Section 7 - Priority Among Simultaneous Exceptions and
		 *                  Interrupts
		 */
		siginfo->si_signo = SIGFPE;
		{
			unsigned mxcsr =
			    bc32->xmm_and_bt_ext[MXCSR_INDEX_IN_EXT_REGS];
			unsigned pending = mxcsr & MXCSR_EXCEPTION_MASK;
			unsigned masked = (mxcsr & MXCSR_EXCEPTION_MASKS) >>
			    MXCSR_EXCEPTION_MASK_POS;
			unsigned live = pending & ~masked;

			if (live & MXCSR_IE_STATUS_MASK) {	
				/* invalid operation */
				siginfo->si_code = FPE_FLTINV;
			} else if (live & MXCSR_ZE_STATUS_MASK) {	
				/* divide by zero */
				siginfo->si_code = FPE_FLTDIV;
			} else if (live & MXCSR_DE_STATUS_MASK) {	
				/* denormal */
				siginfo->si_code = FPE_FLTUND;
			} else if (live & MXCSR_UE_STATUS_MASK) {	
				/* underflow */
				siginfo->si_code = FPE_FLTUND;
			} else if (live & MXCSR_OE_STATUS_MASK) {	
				/* overflow */
				siginfo->si_code = FPE_FLTOVF;
			} else if (live & MXCSR_PE_STATUS_MASK) {	
				/* inexact result */
				siginfo->si_code = FPE_FLTRES;
			} else {
				siginfo->si_code = SI_KERNEL;
			}
		}
		*trapno = 19;
		break;
	case BT_EXCEPT_BREAKPOINT:
	case BT_EXCEPT_SINGLE_STEP:
		siginfo->si_signo = SIGTRAP;
		siginfo->si_code = TRAP_BRKPT;
		break;
	case BT_EXCEPT_STACK_OVERFLOW:
		siginfo->si_signo = SIGSEGV;
		siginfo->si_code = SI_KERNEL;
	case BT_EXCEPT_GUARD_PAGE:
	default:
		assert (!"unknown exception code");
	}
}

static void 
get_perm_mask (sigset32_t *mask)
{
	static     sigset32_t perm_mask;
	static int perm_mask_initialized = 0;

	if (!perm_mask_initialized) {
		int   i;

		memset (&perm_mask, 0, sizeof (perm_mask));
		for (i = 0; m_permanent_handlers[i] != -1; i++) {
			sigaddset32 (&perm_mask, m_permanent_handlers[i]);
		}	
		perm_mask_initialized = 1;
	}
	memcpy (mask, &perm_mask, sizeof (sigset32_t));
}


/****************************************************************************
    set the signal mask for btlib
 ****************************************************************************/
long
set_sigmask (sigset32_t *mask32)
{
	k_sigset_t mask64;
	sigset32_t perm_mask, actual_mask;
	long retval;

	assert (mask32 != NULL);

	actual_mask = *mask32;

	/* never mask signal with permanent signal handler */
	get_perm_mask (&perm_mask);

#if DEBUG
{
	int i;
	for (i = 1; i <  64; i++) {
		if (sigismember32 (&perm_mask, i) && sigismember32 (mask32, i)) {
			DBPRINT (1, "try to block signals with permanent handler!\n");
		}
	}
}
#endif

	sig_andnot_mask (&actual_mask, &perm_mask, 1);

	/* convert the mask to 64 bit format */
	sigset32_2_sigset64 (&actual_mask, &mask64);

	/* call the syscall */
	retval = SYSCALL (rt_sigprocmask, (SIG_SETMASK, &mask64, NULL, 
					   sizeof (k_sigset_t)));

	if (IS_SYS_ERROR (retval)) {
		goto out;
	}
	cur_thread_data->sigmask = *mask32;
out:
	return retval;
}

/*****************************************************************************/
/*****************************************************************************/
static int
handle_interrupted_syscall (struct sigcontext *sc64)
{
	long ret = CONTINUE;

	assert (*cur_thread_data->long_syscall_flag != NOT_IN_SYSCALL);

	ret = CONTINUE;
	switch (*cur_thread_data->long_syscall_flag) {
	case CHECK_QUE:
		*cur_thread_data->long_syscall_flag = NOT_IN_SYSCALL;
		ret = IA32_SYSCALL;
		break;
	case AFTER_CHECK_QUE:
		*cur_thread_data->long_syscall_flag = SIG_PENDING;
		break;
	case BEFORE_BREAK:
		DBPRINT (btl_prt_sig, "ip=%lx before_break_inst=%lx\n",
			 sc64->sc_ip, btl_ls_before_break_inst);
		if (sc64->sc_ip < btl_ls_before_break_inst) {
			/* clear p1 */
			sc64->sc_pr &= ~2;
		}
		break;
	case SIG_PENDING:
	case AFTER_SYSCALL:
	case SYSCALL_COMPLETED:
		/* do nothing */
		break;
	default:
		DBPRINT (TRUE, "unkown *lsf value=%ld\n", 
				*cur_thread_data->long_syscall_flag);
		btlib_abort_all ();
		assert (!"not reached");
	}
	DBPRINT (btl_prt_syscall | btl_prt_sig, "ret=%ld\n", ret);
	return ret;
}

/*****************************************************************************************/
/* we need to expand a memory region (most likely the stack) when the following are true:*/
/* 1> SIGSEGV (segmentation fault) is the signal that brought us here                    */
/* 2> the access is to an address less than 32 bytes lower than esp                      */
/* 3> the memory region (most likely the stack) that the memory access is just below it, */
/*    can be expanded, aka has MAP_GROWSDOWN attribute                                   */
/*****************************************************************************************/
static inline long check_expansion (siginfo_t *siginfo, unsigned esp)
{
	struct memory_region closest_mem_region;
	long found;
	long retval = 0;
	int limit_criteria;
	unsigned new_stack_bottom;
	unsigned size; 
	if (esp) {
		limit_criteria = ((unsigned long)(esp - 32) <= (unsigned long)siginfo->si_addr);
	} else {
		limit_criteria = 1;
	}

	if (siginfo->si_signo == SIGSEGV && limit_criteria) {
		
		DBPRINT(0, "\nsegmentation fault below the stack esp=%ld addr=%ld\n",
			(unsigned long)esp,
			(unsigned long)siginfo->si_addr);
		
		found = iti_closest_mem_region((unsigned long)siginfo->si_addr, &closest_mem_region);
		if (found && (closest_mem_region.flags & MAP_GROWSDOWN)) { 
				new_stack_bottom = PAGE_START ((unsigned long)siginfo->si_addr);
				size = closest_mem_region.address - new_stack_bottom;
			DBPRINT(0, "esp=%lx addr=%lx size=%x\n",
					(unsigned long)esp, (unsigned long)siginfo->si_addr, size);
			legacy_mmap_args_t args = {
				(unsigned) (unsigned long) new_stack_bottom, 
				(unsigned) size,
				(unsigned) closest_mem_region.prot,
				(unsigned) MAP_PRIVATE | MAP_ANONYMOUS | MAP_FIXED | MAP_GROWSDOWN,
				(unsigned) -1,
				(unsigned) 0
			};
			retval = ILP32_WRAPPER1 (IA32_SYSCALL_MMAP, (unsigned long) &args); 
			sys_err_chk (retval, "mmap");
		}
	}
	return retval;
}
