/*  
 * btlib_signal_syscalls.c
 *  
 * wrappers to IA-32 syscalls related to signals 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include <ia32x_utils.h>
#include "btlib_sync.h"
#include "btlib_sig_utils.h"
#include "btlib_signal.h"
#include "btlib_syscall.h"

struct ia32_sighand_frame {
	unsigned prev_frame_ptr;
	int signo;
	sigcontext32_t sc32;
	struct _fpstate_ia32 fpstate;
	sigset32_t mask;
	char legacy_restorer[LEGACY_RESTORER_SIZE];
};

struct ia32_info_sighand_frame {
	unsigned prev_frame_ptr;
	int signo;
	unsigned si32_ptr;
	unsigned uc32_ptr;
	struct siginfo32 si32;
	struct ucontext32 uc32;
	struct _fpstate_ia32 fpstate;
	char legacy_restorer[LEGACY_RESTORER_SIZE];
};

const unsigned char RESTORER_CODE[LEGACY_RESTORER_SIZE] = {
	IA32_POP_EAX,		/* glibc and gdb expect esp to be 4 bytes more 
				 * than usual */
	IA32_MOV_EAX_IMM,
	(unsigned char) IA32_SYSCALL_SIGRETURN, 0x0, 0x0, 0x0,
	IA32_INT,
	0x80,

};

const unsigned char RT_RESTORER_CODE[LEGACY_RESTORER_SIZE] = {
	IA32_MOV_EAX_IMM,
	(unsigned char) IA32_SYSCALL_RT_SIGRETURN, 0x0, 0x0, 0x0,
	IA32_INT,
	0x80,
	0x00
};

static int fill_siginfo32 (struct siginfo32 *info32, siginfo_t *info64, sigcontext32_t *sc);
static int calc_current_ss_flags (unsigned esp);
static void lsigact_2_sigact (legacy_sigaction32_t *osa, sigaction32_t *sa);
static void sigact_2_lsigact (sigaction32_t *sa, legacy_sigaction32_t *osa);
static void sig_set_mask (sigset32_t *dest, sigset32_t *src, int size);
#if DEBUG
static void UNUSED print_info_frame (struct ia32_info_sighand_frame *info_frame);
static void UNUSED print_sigcontext32 (sigcontext32_t *sc);
#endif

/*****************************************************************************/
/* setup a stack frame for a signal handler
 * input:	hi- handler's info
 *		frame - address of memory where the frame should be setup
 *		i - type of frame (is SA_INFO?)*/
/*****************************************************************************/
void
write_sighand_frame (handler_info_t *hi, void *framep, int is_info)
{
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (cur_thread_data->sighandlers != NULL);
	assert (hi != NULL);
	DBPRINT (0, "framep=%p is_info=%d\n", framep, is_info);

	if (is_info) {
		struct ia32_info_sighand_frame *frame =
		    (struct ia32_info_sighand_frame *) framep;

		frame->signo = hi->siginfo->si_signo;
		/* initialize SA_SIGINFO fields */
		frame->si32_ptr = P2U (&frame->si32);
		fill_siginfo32 (&frame->si32, hi->siginfo, &hi->sc32);
		/* initialize the context */
		frame->uc32_ptr = P2U (&frame->uc32);
		frame->uc32._flags = 0;
		frame->uc32._link = 0;
		frame->uc32._stack.ss_sp = cur_thread_data->alt_stack.ss_sp;
		frame->uc32._stack.ss_flags =
		    calc_current_ss_flags (hi->sc32.esp);
		frame->uc32._stack.ss_size = cur_thread_data->alt_stack.ss_size;
		memcpy (&frame->uc32._sigmask, &hi->mask,
			sizeof (frame->uc32._sigmask));
		frame->uc32._mcontext = hi->sc32;
		memcpy ((void *) &frame->fpstate, (void *) &hi->fpstate,
			sizeof (struct _fpstate_ia32));
		frame->uc32._mcontext.fpstate = P2U (&frame->fpstate);
		/* setup the return code */
		if (hi->sa32.flags & SA_RESTORER) {
			frame->prev_frame_ptr = hi->sa32.restorer;
		} else {
			frame->prev_frame_ptr = P2U (btl_rt_restorer);
			/* 
			 * although below RESTORER codes are not used by btlib, 
			 * gdb still uses them as flag.
			 */
			memcpy (&frame->legacy_restorer[0], RT_RESTORER_CODE,
				LEGACY_RESTORER_SIZE);
		}
	} else {
		struct ia32_sighand_frame *frame =
		    (struct ia32_sighand_frame *) framep;

		frame->signo = hi->siginfo->si_signo;
		/* initialize the context */
		frame->sc32 = hi->sc32;
		frame->sc32.oldmask = hi->mask.sig[0];
		memcpy ((void *) &frame->fpstate, (void *) &hi->fpstate,
			sizeof (struct _fpstate_ia32));
		frame->sc32.fpstate = P2U (&frame->fpstate);
		/* store the mask */
		sig_set_mask (&frame->mask, &hi->mask, WORDS_IN_MASK);
		/* setup the return code */
		if (hi->sa32.flags & SA_RESTORER) {
			frame->prev_frame_ptr = hi->sa32.restorer;
		} else {
			frame->prev_frame_ptr = P2U (btl_restorer);
			/* 
			 * although below RESTORER codes are not used by btlib, 
			 * gdb still uses them as flag.
			 */
			memcpy (&frame->legacy_restorer[0], RESTORER_CODE,
				LEGACY_RESTORER_SIZE);
		}

	}
}

/*****************************************************************************/
/*
 * fill the siginfo32_t in the info_frame pointed to by fp from the 64 bit siginfo 
 * see siginfo32_t definitions in btlib_signal.h for more information
 */
/*****************************************************************************/
static int
fill_siginfo32 (struct siginfo32 *info32, siginfo_t *info64, sigcontext32_t *sc)
{

	info32->si_signo = info64->si_signo;
	info32->si_errno = info64->si_errno;
	info32->si_code = info64->si_code;

	switch (info32->si_signo) {
	case SIGPOLL:
		info32->_sifields._sigpoll._band = info64->si_band;
		info32->_sifields._sigpoll._fd = info64->si_fd;
		break;
	case SIGILL:
	case SIGFPE:
	case SIGBUS:
		info32->_sifields._sigfault._addr = (unsigned) sc->eip;
		break;
	case SIGSEGV:
		info32->_sifields._sigfault._addr = P2U (info64->si_addr);
		break;
	case SIGCHLD:
		info32->_sifields._sigchld._pid = (int) info64->si_pid;
		info32->_sifields._sigchld._uid = (int) info64->si_uid;
		info32->_sifields._sigchld._status = info64->si_status;
		info32->_sifields._sigchld._utime = (unsigned) info64->si_utime;
		info32->_sifields._sigchld._stime = (unsigned) info64->si_stime;
	default:
		info32->_sifields._kill._pid = (unsigned) info64->si_pid;
		info32->_sifields._kill._uid = (unsigned) info64->si_uid;
		info32->_sifields._rt._sigval.sival_int = info64->si_int;
		info32->_sifields._rt._sigval.sival_ptr = P2U (info64->si_ptr);
	}
	return info32->si_signo;
}

/************************************************************************** 
* read the stack frame of the signal handler
* 
* see the header of read_info_frame() for details
**************************************************************************/
void
read_sighand_frame (unsigned esp, frame_info_t *frame_info, int is_info)
{
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	/* glibc and gdb expect 
	 * esp points one 'sizeof(int)' above the pointer
	 * to the previous frame, in case of sigreturn, and 
	 * 2*sizeof(int) in case of rt_sigreturn */
	if (is_info) {
		/* esp point right above the prev-frame-ptr */
		struct ia32_info_sighand_frame *frame;
		frame = U2P (esp - (unsigned) sizeof (int));
		frame_info->ctx = &frame->uc32._mcontext;
		sig_set_mask (&frame_info->mask,
			      &frame->uc32._sigmask, WORDS_IN_MASK);
	} else {
		/* esp points 2 integers above the prev-frame-ptr */
		struct ia32_sighand_frame *frame;
		frame = U2P (esp - 2 * (unsigned) sizeof (int));
		frame_info->ctx = &frame->sc32;
		sig_set_mask (&frame_info->mask, &frame->mask, WORDS_IN_MASK);
	}
	DBPRINT (0, "frame_info->ctx=%p\n", frame_info->ctx);
}

/*****************************************************************************/
/* allocate memory for a sig frame on the stack of the IA-32 signal handler.
 * input:	sc32 - the sigcontext
 *		flags - the flags of the signal handler
 *		frame_size - the size of the frame
 * output: address of the frame
 */
/*****************************************************************************/
unsigned
allocate_sighand_frame (sigcontext32_t *sc32, int flags, int is_info)
{
	unsigned long esp;
	unsigned long frame_size;

	frame_size = is_info ? sizeof (struct ia32_info_sighand_frame) :
	    sizeof (struct ia32_sighand_frame);
	/* preparation for stack expansion patch */
	int __attribute__ ((unused)) grows_down = 1;

	/* get the address of the stack frame */
	esp = (unsigned long) sc32->esp;
	if ((flags & SA_ONSTACK) && (cur_thread_data->alt_stack.ss_sp != 0)
	    && (cur_thread_data->alt_stack.ss_flags != SS_DISABLE)) {
		grows_down = 0;
		if (!is_using_alt_stack (sc32->esp)) {
			esp =
			    cur_thread_data->alt_stack.ss_sp +
			    cur_thread_data->alt_stack.ss_size;
		}
	}
	esp = ((esp - frame_size) & -8ul);
	assert ((esp & 0xfffffffful) == esp);

	return (unsigned) esp;
}

/**************************************************************************
* implement the signal IA-32 syscall (for backwards compatibility)
* int signal(int sig, __sighandler_t handler)
* input:  bc32 - the IA-32 context as received at the sigaction syscall
* output: the syscall return value
****************************************************************************/
long
btlib_signal (BTGENERIC_IA32_CONTEXT *bc32)
{
	int signo;
	sigaction32_t sa, osa;
	long retval;

	signo = (int) IA32_SYSCALL_PARAM1 (bc32);
	sa.handler = (unsigned) IA32_SYSCALL_PARAM2 (bc32);
	sa.flags = SA_ONESHOT | SA_NOMASK;

	retval = emulate_sigaction (signo, &sa, &osa, 1);
	if (IS_SYS_ERROR (retval)) {
		return retval;
	}
	return (long) (osa.handler);
}

/**************************************************************************
 * implement the sigpending syscall
 * input:  bc32 - the IA-32 context 
 *         is_rt - whether sigpending or rt_sigpending was called
 * output: EFAULT in case of illegal pointer to set
 *
***************************************************************************/
long
btlib_sigpending (BTGENERIC_IA32_CONTEXT *bc32, int is_rt)
{
	sigset32_t que_set32, shared_que_set32, set32;
	sigset32_t *uset32;
	k_sigset_t set64;
	int        sigset_size, nsig_words;
	long       retval = 0;


	assert (bc32 != NULL);
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	if (is_rt) {
		sigset_size = IA32_SYSCALL_PARAM2 (bc32);
		if (sigset_size != sizeof (sigset32_t)) {
			retval = -EINVAL;
			goto out;
		}
	} else {
		sigset_size = sizeof (legacy_sigset32_t);
	}
	nsig_words = (int) (sigset_size / sizeof (unsigned));
	uset32 = U2P (IA32_SYSCALL_PARAM1 (bc32)); 

	/* get the signals that are pending in the kernel */
	retval = SYSCALL (rt_sigpending, (&set64, sizeof (set64)));
	if (IS_SYS_ERROR (retval)) {
		goto out;
	}
	sigset64_2_sigset32 (&set64, &set32);

	/* add the signals that are pending in the que and are currently masked */

	/* get the set of signals in the que - both private and shared */
	btlib_que_sigpending (cur_thread_data->pending_que, nsig_words, &que_set32);
	btlib_que_sigpending (btl_shared_que, nsig_words, &shared_que_set32);
	sig_or_mask (&que_set32, &shared_que_set32, nsig_words);

	/* remove the non-blocked ones */
	sig_and_mask (&que_set32, &cur_thread_data->sigmask, nsig_words);

	/* add it to the set we got from the kernel */
	sig_or_mask (&set32, &que_set32, nsig_words);

	/* copy the mask back to the user */
	if (safe_memcpy (uset32, &set32, (unsigned long)sigset_size) < 0) {
		retval = -EFAULT;
	}

out:
	return retval;
}

/* implement the signal sigprocmask, rt_sigprocmask */
long 
btlib_sigprocmask (BTGENERIC_IA32_CONTEXT *bc32, int is_rt)
{
	
	int         how; 
	long        ret;
	void        *set32, *oset32;
	sigset32_t  new_set32;
	int         sigset_size;

	assert (bc32 != NULL);
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	how = (int) IA32_SYSCALL_PARAM1 (bc32);
	set32 = U2P (IA32_SYSCALL_PARAM2 (bc32));
	oset32 = U2P (IA32_SYSCALL_PARAM3 (bc32));

	if (is_rt) {
		sigset_size = IA32_SYSCALL_PARAM4 (bc32);
		if (sigset_size != sizeof (sigset32_t)) {
			ret = -EINVAL;
			goto out;
		}
	} else {
		sigset_size = sizeof (legacy_sigset32_t);
	}

	/* only if "how" is valid or "newset" is NULL, copy old signal mask */
	if ((how != SIG_BLOCK) && (how != SIG_UNBLOCK) && (how != SIG_SETMASK) && 
	    (set32 != NULL)) {
		ret = -EINVAL;
		goto out;
	}

	if (oset32 != NULL) {
		if (safe_memcpy (oset32, &cur_thread_data->sigmask, 
				 (unsigned long) sigset_size) < 0) {
			ret = -EFAULT;
			goto out;
		}
	}
	
	if (set32 == NULL) {
		ret = 0;
		goto out;
	}


	memset (&new_set32, 0, sizeof (new_set32)); 
	if (safe_memcpy (&new_set32, set32, (unsigned long) sigset_size) < 0) {
		ret = -EFAULT;
		goto out;
	}

	if (how == SIG_UNBLOCK) {
		k_sigset_t  new_set64;

		/* it's ok for non rt sigprocmask because the high part is zero */
		sigset32_2_sigset64 (&new_set32, &new_set64);
		ret = SYSCALL (rt_sigprocmask, 
			       (how, &new_set64, NULL, sizeof (new_set64)));

		if (!IS_SYS_ERROR (ret)) {
			sig_andnot_mask (&cur_thread_data->sigmask, &new_set32, 
					 (int) (sigset_size / sizeof (unsigned)));
		}
		goto out;
	} else if (how == SIG_BLOCK) {
		sig_or_mask (&new_set32, &cur_thread_data->sigmask, 
			     (int) (sigset_size / sizeof (unsigned)));
	} else {
		if (!is_rt) {
			new_set32.sig[1] = cur_thread_data->sigmask.sig[1];
		}
	}
	
	sigdelset32 (&new_set32, SIGSTOP);
	sigdelset32 (&new_set32, SIGKILL);
	
	ret = set_sigmask (&new_set32);

out:
	return ret;

}

/***************************************************************************
* implement the rt_sigaction IA-32 syscall
* int sigaction(int sig,const struct sigaction *act,struct sigaction *oact)
* 
* input:  bc32 - the IA-32 context as received at the sigaction syscall
* output: the syscall return value
****************************************************************************/
long
btlib_rt_sigaction (BTGENERIC_IA32_CONTEXT *bc32)
{
	int signo;
	sigaction32_t *sa32, *osa32;
	long retval = 0;

	assert (bc32 != NULL);

	/* extract parameters from IA-32 context */
	signo = (int) IA32_SYSCALL_PARAM1 (bc32);
	sa32 = U2P (IA32_SYSCALL_PARAM2 (bc32));
	osa32 = U2P (IA32_SYSCALL_PARAM3 (bc32));

	retval = emulate_sigaction (signo, sa32, osa32, WORDS_IN_MASK);
	return retval;
}

/***************************************************************************
* implement the sigaction IA-32 syscall
*  int sigaction(int sig,const struct legacy_sigaction *act,
*                struct legacy_sigaction *oact)
* input:  bc32 - the IA-32 context as received at the sigaction syscall
* output: the syscall return value
****************************************************************************/
long
btlib_sigaction (BTGENERIC_IA32_CONTEXT *bc32)
{
	int signo;
	long retval;
	sigaction32_t local_sigact, local_old_sigact;
	sigaction32_t *local_sigact_p = NULL, *local_old_sigact_p = NULL;
	legacy_sigaction32_t *user_sigact_p, *user_old_sigact_p;
	legacy_sigaction32_t local_legacy_sigact, local_legacy_old_sigact;

	assert (bc32 != NULL);

	/* extract parameters from IA-32 context */
	signo = (int) IA32_SYSCALL_PARAM1 (bc32);
	user_sigact_p = U2P (IA32_SYSCALL_PARAM2 (bc32));
	user_old_sigact_p = U2P (IA32_SYSCALL_PARAM3 (bc32));

	/* convert legacy_sigaction to new sigaction */
	if (user_sigact_p != NULL) {
		if (safe_memcpy (&local_legacy_sigact, user_sigact_p,
				 sizeof (local_legacy_sigact)) < 0) {
			retval = -EFAULT;
			goto out;
		}
		lsigact_2_sigact (&local_legacy_sigact, &local_sigact);
		local_sigact_p = &local_sigact;
	}
	if (user_old_sigact_p != NULL) {
		local_old_sigact_p = &local_old_sigact;
	}
	/* call the sigaction wrapper */
	retval = emulate_sigaction (signo, local_sigact_p, local_old_sigact_p, 1);
	if (IS_SYS_ERROR (retval)) {
		goto out;
	}

	/* convert new sigaction to legacy_sigaction */
	if (user_old_sigact_p != NULL) {
		sigact_2_lsigact (&local_old_sigact, &local_legacy_old_sigact);
		if (safe_memcpy (user_old_sigact_p, &local_legacy_old_sigact,
				 sizeof (local_legacy_old_sigact)) < 0) {
			retval = -EFAULT;
			goto out;
		}
	}

out:
	return retval;
}

/**************************************************************************
 * implement the IA-32 sigaltstack syscall
 *  int sigaltstack(const stack_t *ss, stack_t *oss);
 * input: bc32
 * output: 0 on success, <0 on failure
 ***************************************************************************/
long
btlib_sigaltstack (BTGENERIC_IA32_CONTEXT *bc32)
{
	stack32_t *uss, *uoss;
	int ret = 0;

	assert (bc32 != NULL);
	/* get the thread's specific data */
	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	DBPRINT (0, "\n");
	/* extract parameters from IA-32 bc32 */
	uss = U2P (IA32_SYSCALL_PARAM1 (bc32));
	uoss = U2P (IA32_SYSCALL_PARAM2 (bc32));
	DBPRINT (0, "uss=%p uoss=%p\n", uss, uoss);

	/* copy the current info before changing it */
	if (uoss != NULL) {
		stack32_t oss;
		oss.ss_size = cur_thread_data->alt_stack.ss_size;
		oss.ss_sp = cur_thread_data->alt_stack.ss_sp;
		oss.ss_flags =
		    calc_current_ss_flags (cur_thread_data->bc32->esp);
		if (safe_memcpy (uoss, &oss, sizeof (*uoss)) < 0) {
			ret = -EFAULT;
			goto out;
		}
	}

	/* now do the changes */
	if (uss != NULL) {
		stack32_t ss;
		if (safe_memcpy (&ss, uss, sizeof (ss)) < 0) {
			ret = -EFAULT;
			goto out;
		}
		/* POSIX Issue 6 -
		   41769 ERRORS
		   41770 The sigaltstack ( ) function shall fail if:
		   41771 [EINVAL] The ss argument is not a null pointer, and the ss_flags member pointed to by ss
		   41772 contains flags other than SS_DISABLE.
		 */
		if (is_using_alt_stack (cur_thread_data->bc32->esp)) {
			ret = -EPERM;
			goto out;
		}
		/*
		   41773 [ENOMEM] The size of the alternate stack area is less than MINSIGSTKSZ.
		 */
		if ((ss.ss_flags & ~SS_DISABLE) != 0) {
			ret = -EINVAL;
			goto out;
		}
		/*
		   41774 [EPERM] An attempt was made to modify an active stack.
		 */
		if ((ss.ss_flags != SS_DISABLE) && (ss.ss_size < MINSIGSTKSZ32)) {
			ret = -ENOMEM;
			goto out;
		}

		/* everything is good, update current */
		if (ss.ss_flags == SS_DISABLE) {
			cur_thread_data->alt_stack.ss_flags = SS_DISABLE;
		} else {
			cur_thread_data->alt_stack = ss;
		}
	}

      out:
	return ret;
}

/*****************************************************************************/
/* True if we are on the alternate signal stack */
/*****************************************************************************/
int
is_using_alt_stack (unsigned esp)
{
	unsigned alt_stack_hi = cur_thread_data->alt_stack.ss_sp +
	    cur_thread_data->alt_stack.ss_size;
	unsigned alt_stack_lo = cur_thread_data->alt_stack.ss_sp;
	return ((esp >= alt_stack_lo) && (esp < alt_stack_hi)) ? 1 : 0;
}

/*****************************************************************************/
/* return the alternate-stack flags */
/*****************************************************************************/
static int
calc_current_ss_flags (unsigned esp)
{
	int flags;
	if ((cur_thread_data->alt_stack.ss_size == 0) ||
	    (cur_thread_data->alt_stack.ss_flags == SS_DISABLE)) {
		flags = SS_DISABLE;
	} else if (is_using_alt_stack (esp)) {
		flags = SS_ONSTACK;
	} else {
		flags = 0;
	}
	return flags;
}

/*****************************************************************************/
/* convert from legacy_sigaction32_t to sigaction32_t */
/*****************************************************************************/
static void
lsigact_2_sigact (legacy_sigaction32_t *lsa, sigaction32_t *sa)
{
	sa->handler = lsa->handler;
	sa->flags = lsa->flags;
	sa->restorer = lsa->restorer;
	sa->mask.sig[0] = lsa->mask;
	sa->mask.sig[1] = 0;
}

/*****************************************************************************/
/* convert from sigaction32_t to legacy_sigaction32_t */
/*****************************************************************************/
static void
sigact_2_lsigact (sigaction32_t *sa, legacy_sigaction32_t *lsa)
{
	lsa->handler = sa->handler;
	lsa->flags = sa->flags;
	lsa->restorer = sa->restorer;
	lsa->mask = sa->mask.sig[0];
}

/*****************************************************************************/
/* do some basic operations on signal masks:
 * sig_set_mask:=  (dest = src)
 * sig_andnot_mask := (dest &= ~src)
 * sig_or_mask := (dest != src)
 */
/*****************************************************************************/
static void
sig_set_mask (sigset32_t *dest, sigset32_t *src, int size)
{
	int j;
	for (j = 0; j < size; j++) {
		dest->sig[j] = src->sig[j];
	}
}

#if DEBUG

/*****************************************************************************/
/*****************************************************************************/
static void
print_info_frame (struct ia32_info_sighand_frame *info_frame)
{
	fprintf (g_log_file, "\ninfo_frame:");
	fprintf (g_log_file, "\nprev_frame_ptr\t\t\%x\n:",
		 info_frame->prev_frame_ptr);
	fprintf (g_log_file, "\nsig\t\t\t\%x\n:", info_frame->signo);
	fprintf (g_log_file, "\nsi32_ptr\t\t\%x\n:", info_frame->si32_ptr);
	fprintf (g_log_file, "\npuc\t\t\%x\n:", info_frame->uc32_ptr);
}

/*****************************************************************************/
/*****************************************************************************/
static void
print_sigcontext32 (sigcontext32_t * sc)
{

	fprintf (g_log_file, "sc->edi = %x\n", sc->edi);
	fprintf (g_log_file, "sc->esi = %x\n", sc->esi);
	fprintf (g_log_file, "sc->ebp = %x\n", sc->ebp);
	fprintf (g_log_file, "sc->esp = %x\n", sc->esp);
	fprintf (g_log_file, "sc->ebx = %x\n", sc->ebx);
	fprintf (g_log_file, "sc->edx = %x\n", sc->edx);
	fprintf (g_log_file, "sc->ecx = %x\n", sc->ecx);
	fprintf (g_log_file, "sc->eax = %x\n", sc->eax);
	fprintf (g_log_file, "sc->eip = %x\n", sc->eip);
	fprintf (g_log_file, "sc->cs = %x\n", sc->cs);
	fprintf (g_log_file, "sc->esp_at_signal = %x\n", sc->esp_at_signal);
	fprintf (g_log_file, "sc->ss = %x\n", sc->ss);
	fprintf (g_log_file, "sc->eflags = %x\n", sc->eflags);
	fprintf (g_log_file, "sc->fpstate = %x\n", sc->fpstate);

	long i, j;
	struct _fpstate_ia32 *fpstate;
	fpstate = U2P (sc->fpstate);

	fprintf (g_log_file, "fpstate->cw = %x\n", fpstate->cw);
	fprintf (g_log_file, "fpstate->sw = %x\n", fpstate->sw);
	fprintf (g_log_file, "fpstate->tag = %x\n", fpstate->tag);
	fprintf (g_log_file, "fpstate->ipoff = %x\n", fpstate->ipoff);
	fprintf (g_log_file, "fpstate->cssel = %x\n", fpstate->cssel);
	fprintf (g_log_file, "fpstate->dataoff = %x\n", fpstate->dataoff);
	fprintf (g_log_file, "fpstate->datasel = %x\n", fpstate->datasel);
	for (i = 0; i < 8; i++) {
		fprintf (g_log_file, "_st[%ld]: ", i);
		for (j = 0; j < 10; j++) {
			fprintf (g_log_file, "%x ",
				 *((unsigned char *) (fpstate->_st) + i * 8 +
				   j));
		}
		fprintf (g_log_file, "\n");
	}

	fprintf (g_log_file, "fpstate->xcwd = %x\n", fpstate->xcwd);
	fprintf (g_log_file, "fpstate->xswd = %x\n", fpstate->xswd);
	fprintf (g_log_file, "fpstate->xtwd = %x\n", fpstate->xtwd);
	fprintf (g_log_file, "fpstate->xfip = %x\n", fpstate->xfip);
	fprintf (g_log_file, "fpstate->xfop = %x\n", fpstate->xfop);
	fprintf (g_log_file, "fpstate->xfcs = %x\n", fpstate->xfcs);
	fprintf (g_log_file, "fpstate->xfoo = %x\n", fpstate->xfoo);
	fprintf (g_log_file, "fpstate->xfos = %x\n", fpstate->xfos);
	fprintf (g_log_file, "fpstate->mxcsr = %x\n", fpstate->mxcsr);

	for (i = 0; i < 8; i++) {
		fprintf (g_log_file, "_fxsr_st[%ld]: ", i);
		for (j = 0; j < 10; j++) {
			fprintf (g_log_file, "%x ",
				 *((unsigned char *) (fpstate->_fxsr_st) +
				   i * 8 + j));
		}
		fprintf (g_log_file, "\n");
	}
}
#endif				// DEBUG
