/*  
 * btlib_sync.c
 *  
 * centralized sync routines
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include <ia32x_utils.h>
#include "btlib.h"
#include "btlib_sync.h"
#include "btlib_signal.h"
#include "btlib_asm.h"
#include <ia32toia64.h>

cs_handle cs_array[BTL_MAX_LOCK_IND];

/* 
 * the cs module cannot use a cs object for sync so it uses a simple 
 * spin-lock instead
 */
static mutex_opaque m_cs_sync;

/*****************************************************************************/
/* init the sync module */
/*****************************************************************************/
void
btlib_init_sync_module ()
{
	long ret;

	btlib_init_cs_module ();

	cs_array[BTL_LOCK_IND_THREAD_SCALABILITY] = btlib_create_critical_section ();
	cs_array[BTL_LOCK_IND_SHARED_QUE] = btlib_create_critical_section ();
	cs_array[BTL_LOCK_IND_THREAD] = btlib_create_critical_section ();
	cs_array[BTL_LOCK_IND_FREE_STACK] = btlib_create_critical_section ();
	cs_array[BTL_LOCK_IND_INTERNAL_MEM] = btlib_create_critical_section ();
	cs_array[BTL_LOCK_IND_LDT] = btlib_create_critical_section ();
	cs_array[BTL_LOCK_IND_SUSPENSION] = btlib_create_critical_section ();

	btlib_init_critical_section (cs_array[BTL_LOCK_IND_THREAD_SCALABILITY]);
	btlib_init_critical_section (cs_array[BTL_LOCK_IND_SHARED_QUE]);
	btlib_init_critical_section (cs_array[BTL_LOCK_IND_THREAD]);
	btlib_init_critical_section (cs_array[BTL_LOCK_IND_FREE_STACK]);
	btlib_init_critical_section (cs_array[BTL_LOCK_IND_INTERNAL_MEM]);
	btlib_init_critical_section (cs_array[BTL_LOCK_IND_LDT]);
	btlib_init_critical_section (cs_array[BTL_LOCK_IND_SUSPENSION]);

	ret = mutex_init (&m_cs_sync);
	sys_err_chk (ret, "mutex_init failed");

}

/*****************************************************************************/
/* acquire all the sync objects used inside btlib */
/*****************************************************************************/
void
btlib_freeze ()
{
	int j;
	long retval;

	for (j = 0; j < BTL_MAX_LOCK_IND; j++) {
		btlib_acquire_lock (j);
	}
	retval = iti_before_fork ();
	sys_err_chk (retval, "iti_before_fork failed");
	retval = lcs_before_fork ();
	sys_err_chk (retval, "lcs_before_fork failed");
}

/*****************************************************************************/
/* release all the sync objects used inside btlib */
/*****************************************************************************/
void
btlib_unfreeze (long is_parent)
{
	int j;
	long retval;

	retval = lcs_after_fork (is_parent);
	sys_err_chk (retval, "lcs_after_fork failed");
	retval = iti_after_fork (is_parent);
	sys_err_chk (retval, "iti_after_fork failed");
	for (j = BTL_MAX_LOCK_IND - 1; j >= 0; j--) {
		btlib_release_lock (j);
	}
}

/*****************************************************************************/
/* acquire a sync object */
/*****************************************************************************/
void
btlib_acquire_lock (int lock_ind)
{
	long ret;

	DBPRINT (0, "cur_thread_data=%p lock_ind=%d\n", cur_thread_data,
	         lock_ind);
#if DEBUG
	/* enforce order between lsync objects */
	if (cur_thread_data) {
		assert ((cur_thread_data->locks >> lock_ind) <= 1);
		cur_thread_data->locks |= (1 << lock_ind);
	}
#endif

	switch (lock_ind) {
	case BTL_LOCK_IND_THREAD_SCALABILITY:
	case BTL_LOCK_IND_SHARED_QUE:
	case BTL_LOCK_IND_THREAD:
	case BTL_LOCK_IND_FREE_STACK:
	case BTL_LOCK_IND_INTERNAL_MEM:
	case BTL_LOCK_IND_LDT:
		btlib_enter_critical_section (cs_array[lock_ind], 1);
		break;
	case BTL_LOCK_IND_SUSPENSION:
		cur_thread_data->is_self_suspended = 1;
		btlib_enter_critical_section (cs_array[lock_ind], 1);
		cur_thread_data->is_self_suspended = 0;
		break;
	case BTL_LOCK_IND_CS:
		ret = mutex_lock (&m_cs_sync);
		sys_err_chk (ret, "mutex_lock failed");
		break;
	case BTL_LOCK_IND_SIGHAND:
		btlib_enter_critical_section (cur_thread_data->sighandlers->cs, 1);
		break;
	default:
		assert (!"unknown lsync object");
	}
}

/*****************************************************************************/
/* release a sync object */
/*****************************************************************************/
void
btlib_release_lock (int lock_ind)
{
	long ret;
	DBPRINT (0, "cur_thread_data=%p lock_ind=%d\n", cur_thread_data,
	         lock_ind);
#if DEBUG
	if (cur_thread_data) {
		cur_thread_data->locks &= ~(1 << lock_ind);
	}
#endif
	switch (lock_ind) {
	case BTL_LOCK_IND_THREAD_SCALABILITY:
	case BTL_LOCK_IND_SHARED_QUE:
	case BTL_LOCK_IND_THREAD:
	case BTL_LOCK_IND_FREE_STACK:
	case BTL_LOCK_IND_INTERNAL_MEM:
	case BTL_LOCK_IND_LDT:
	case BTL_LOCK_IND_SUSPENSION:
		btlib_leave_critical_section (cs_array[lock_ind]);
		break;
	case BTL_LOCK_IND_CS:
		ret = mutex_unlock (&m_cs_sync);
		sys_err_chk (ret, "mutex_unlock failed");
		break;
	case BTL_LOCK_IND_SIGHAND:
		btlib_leave_critical_section (cur_thread_data->sighandlers->cs);
		break;
	default:
		assert (!"unknown sync object");
	}
}

#if DEBUG
/*****************************************************************************/
/* check whether the given thread is holding a sync object inside btlib */
/*****************************************************************************/
long
btlib_is_inside_lock (btlib_thread_data_t * tls)
{
	long retval;

	if (who_holds_critical_section (cs_array[BTL_LOCK_IND_THREAD_SCALABILITY]) 
		== tls->__r13) {
		retval = BTL_LOCK_IND_THREAD_SCALABILITY;

	} else if (who_holds_critical_section (cs_array[BTL_LOCK_IND_SHARED_QUE]) == 
		tls->__r13) {
		retval = BTL_LOCK_IND_SHARED_QUE;

	} else if (who_holds_critical_section (cs_array[BTL_LOCK_IND_THREAD]) == 
		   tls->__r13) {
		retval = BTL_LOCK_IND_THREAD;

	} else if (who_holds_critical_section (cs_array[BTL_LOCK_IND_FREE_STACK]) == 
		   tls->__r13) {
		retval = BTL_LOCK_IND_FREE_STACK;

	} else if (who_holds_critical_section (tls->sighandlers->cs) == 
		   tls->__r13) {
		retval = BTL_LOCK_IND_SIGHAND;

	} else if (get_mutex_owner (&m_cs_sync) == tls->__r13) {
		retval = BTL_LOCK_IND_CS;

	} else if (who_holds_critical_section
		   (cs_array[BTL_LOCK_IND_LDT]) == tls->__r13) {
		retval = BTL_LOCK_IND_LDT;

	} else if (who_holds_critical_section (cs_array[BTL_LOCK_IND_INTERNAL_MEM]) == 
		   tls->__r13) {
		retval = BTL_LOCK_IND_INTERNAL_MEM;

	} else if (who_holds_critical_section (cs_array[BTL_LOCK_IND_SUSPENSION]) == 
		   tls->__r13) {
		retval = BTL_LOCK_IND_SUSPENSION;

	} else {
		retval = 0;
	}

	return retval;
}
#endif	// DEBUG

