/*	
 * btlib_thread.c
 *	
 * 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include <ia32x_utils.h>

#include <ia32toia64.h>

#include "btlib_sig_utils.h"
#include "btlib_signal.h"
#include "btlib_asm.h"
#include "btlib_segment.h"
#include "btlib_sync.h"
#include "btlib_ptemu.h"
#include "btlib_stats.h"

/* The cloning data struct */
typedef struct btlib_clone_data_t {
	BTGENERIC_IA32_CONTEXT start_context;
	btlib_thread_desc *td;
	k_sigset_t sigmask;
	long personality;
} btlib_clone_data_t;

void __btlib_exit (int err_code, void *stack, unsigned long len);
extern size_t __default_stack_size;

/* are we part of a thread group */
static int m_thread_group = 0;
/* thread list, protected by BTL_LOCK_IND_THREAD */
btlib_thread_data_t *m_used_thread_list = NULL;
static long m_kernel_support_nptl = 0;


/* variables used for the affinity and shared mem handling */
#if SHMEM_PROC_AFFINITY || SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
static cpu_set_t m_orig_process_affinity_mask = {{0}};
static int m_did_clone = 0;
static int m_did_shared_mem = 0;
int btl_narrowed_affinity = 0;
int btl_did_ordered_code =0;
#endif
#if SHMEM_PROC_AFFINITY && SHMEM_ORDER_SOLUTION
#error: should not be both defined
#endif


static void before_thread_exit (void);
static void list_add (btlib_thread_data_t *elem, btlib_thread_data_t **list);
static void list_del (btlib_thread_data_t *elem, btlib_thread_data_t **list);
static void *allocate_thread_data (btlib_thread_desc **tdp);
static int  clone_main (void *arg);
static void clone_signals (unsigned, btlib_thread_data_t *);
static long do_fork (btlib_thread_data_t *parent RELEASE_UNUSED,
                     unsigned clone_flags,
                     unsigned ia32_tos,
                     int *parent_tidptr,
                     int *child_tidptr, 
                     BTGENERIC_IA32_CONTEXT *bc32);
static void set_sig_inherit (void);
static void cleanup_thread (btlib_thread_data_t *tbr, int free_thread_stack);
static void init_thread_data (btlib_thread_desc *td,
                              btlib_thread_data_t *blst, void *glst,
                              void *stack_lowest_address,
                              unsigned long stack_size);

#if ALL_THREAD_AFFINITY
static int select_one_cpu (cpu_set_t *cur_mask, cpu_set_t *new_mask);
static int get_current_cpu (void);
static int get_random_cpu (cpu_set_t *cpu_mask);
static void end_thread_scalability_interval (unsigned long cur_timestamp_cntr);
static unsigned long m_num_threads_in_sim = 0;
static unsigned long m_mthread_sim_tm_strt = 0;
static unsigned long m_total_mthread_dur = 0;
static unsigned long m_interval_start_tm = 0;
static unsigned long m_try_for_ordered_code = 0;
#if THREAD_SCALABILIY_WITH_IDLE_TIME
static unsigned long get_cur_num_secs_idle (void);
static unsigned long m_sec_idle_at_istrt = 0;
#endif

#define CYCLES_IN_MILI_SECOND (btl_cycles_in_second/1000)
/* half of the interval is multi-threaded => thread scalability */
#define INVERSE_FRACTION_OF_INTERVAL_FOR_THREAD_SCALABILITY 2
/* one quarter of the interval is idle time => worthwhile to switch
   to ordered code in order to remove affinity */
#define INVERSE_FRACTION_OF_IDLETIME_FOR_THREAD_SCALABILITY 4
#if DEBUG
static int is_only_one_cpu_allowed (cpu_set_t * cpu_mask);
unsigned long btl_start_time = 0;
#endif /* DEBUG */
#endif /* ALL_THREAD_AFFINITY */

#if DEBUG
static void print_thread_list (void);
#endif	// DEBUG

#if SHMEM_PROC_AFFINITY
static int select_lowest_cpu (cpu_set_t *cur_mask, cpu_set_t *new_affinity_mask);
#endif	/* SHMEM_PROC_AFFINITY */

#if SHMEM_PROC_AFFINITY || SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
static long set_affinity_to_all_threads (cpu_set_t * new_affinity_mask);
#endif

/*****************************************************************************/
/* utility functions for CPU masks.	*/
/*****************************************************************************/
static inline int __CPU_IS_EQUAL (cpu_set_t *set1, cpu_set_t *set2)
{
	unsigned int __i;
	for (__i = 0; __i < sizeof (cpu_set_t) / sizeof (__cpu_mask); ++__i)
		if (set1->__bits[__i] != set2->__bits[__i])
			return 0;
	return 1;
}
static inline int __CPU_IS_ZERO (cpu_set_t *set)
{
	unsigned int __i;
	for (__i = 0; __i < sizeof (cpu_set_t) / sizeof (__cpu_mask); ++__i)
		if (set->__bits[__i] != 0)
			return 0;
	return 1;
}
#if DEBUG
static inline void __CPU_PRINT (char *msg, cpu_set_t *set)
{
	char set_str[4096];
	char * setp = set_str;
	unsigned int cpu;

	for (cpu = 0; cpu < 8 * sizeof (*set) - 1; cpu++) {
		if (__CPU_ISSET (cpu, set)) {
			sprintf (setp, "%u ", cpu);
			setp += strlen (setp);
		}
	}
	fprintf (g_log_file, "%s: cpu_mask=[ %s]\n", msg, set_str);
}
#endif
/*****************************************************************************/
		
 
 
/*****************************************************************************/
/* 
 * initialize the thread module. must be called before any exported function 
 * is called
 */
/*****************************************************************************/
void
btlib_init_thread_module ()
{
	long retval;
	retval = SYSCALL (set_tid_address, (0));
	m_kernel_support_nptl = (retval == -ENOSYS) ? FALSE : TRUE;
	btlib_init_tls_module ();
}

/**************************************************************************/
/*
 * main_thread_init
 *
 * Purpose:	
 *      1. Allocate thread specific struct and initialize it.
 *      2. Allocate glst memory for BTGeneric and call BTGeneric_Thread_Init.
 */
/**************************************************************************/
long
btlib_main_thread_init_tsd ()
{
	btlib_thread_data_t *blst;
	btlib_thread_desc *td;

	/* install btlib's thread descriptor 
	 * when we got here r13 already points at the main-thread's tcb.
	 * the thread's stack arranged by glibc so that just below r13
	 * there is enough space for 'struct pthread', which btlib does
	 * not need. instead - btlib installs its own thread descriptor
	 * just below r13. it is much smaller than 'struct pthread' so
	 * there should be no problem with using the space for that purpose
	 */
	td = __current - 1;
	assert ((uintptr_t) td == ROUNDUP ((uintptr_t) td, sizeof (void *)));
	assert (sizeof(btlib_thread_desc) <= TLS_PRE_TCB_SIZE);

	/* allocate blst */
	blst = malloc (sizeof (*blst));
	if (blst == NULL) {
		DBPRINT (TRUE, "malloc failed\n");
		return -ENOMEM;
	}
	DBPRINT (0, "r13=%p td=%p blst=%p\n", __current, td, blst);

	/* init the thread desc */
	td->glst = 0ul;
	td->blst = blst;

	return 0;
}

/**************************************************************************/
/**************************************************************************/
long
btlib_main_thread_init (const btrans_binary_map_t *app_info, long personality)
{
	BTGENERIC_IA32_CONTEXT bc32;
	BT_STATUS_CODE btg_ret;
	btlib_thread_desc *td = __current - 1;
	void *glst;
	long ret RELEASE_UNUSED;
	k_sigset_t mask64;
	sigset32_t mask32;

	/* allocate glst */
	glst = btlib_mem_alloc (INITIAL_BTLIB_ADDRESS,
				(unsigned long) btl_glst_size,
				PROT_READ | PROT_WRITE | PROT_EXEC);
	if (IS_SYS_ERROR ((long) glst)) {
		DBPRINT (TRUE, "mmap failed: %ld\n", (long)glst);
		return -ENOMEM;
	}
	DBPRINT (0, "cur_thread_data=%p glst=%p\n", cur_thread_data, glst);
	assert (glst != NULL);

	/* blst is already initialized by btlib_main_thread_init_tsd */
	init_thread_data (td, td->blst, glst, 0, 0ul);

	btlib_que_init (&cur_thread_data->pending_que);

	/* store the tid */
	cur_thread_data->tid = SYSCALL (gettid, ());

	/* the first element in the list. 
	 * since adding to the list is always at the beginning, iterating
	 * the list can be done by starting with m_used_thread_list backwards
	 */
	m_used_thread_list = cur_thread_data;
	cur_thread_data->next = NULL;
	cur_thread_data->prev = NULL;

	/* create a signal-handelr table */
	cur_thread_data->sighandlers = btlib_create_sighand_table ();
	assert (cur_thread_data->sighandlers != NULL);

	/* set the inherited mask and disposition */
	set_sig_inherit ();


	
	ret = SYSCALL (rt_sigprocmask, 
		       (SIG_SETMASK, NULL, &mask64, sizeof(k_sigset_t)));
	sys_err_chk (ret, "rt_sigprocmask");
	sigset64_2_sigset32 (&mask64, &mask32);
	set_sigmask (&mask32);
	

	/* get the address of the long syscall flag */
        ret = iti_long_syscall_interface (&cur_thread_data->long_syscall_flag, NULL);
	assert (ret == 0);
	DBPRINT (0, "long_syscall_flag=%p\n", cur_thread_data->long_syscall_flag);

	/* initialize the thread */
	btg_ret = BTGENERIC_THREAD_INIT (td->glst, app_info->start_stack, 0ul);
	assert (btg_ret == BT_STATUS_SUCCESS);

	/* set the start context */
	memset (&bc32, 0, sizeof (BTGENERIC_IA32_CONTEXT));

	bc32.fpu_state.control_word = FPCW_INIT_VAL;
	bc32.fpu_state.tag_word = FPTW_INIT_VAL;
	bc32.eflags = EFLAGS_INIT_VAL;
	*(unsigned int *) (void *) &bc32. xmm_and_bt_ext[MXCSR_INDEX_IN_EXT_REGS] = 
		MXCSR_INIT_VAL;

	/* same with the kernel */
	bc32.ds = __USER_DS;
	bc32.es = __USER_DS;
	bc32.ss = __USER_DS;
	bc32.cs = __USER_CS;

	bc32.esp = (unsigned) app_info->start_stack;
	bc32.eip = (unsigned) app_info->entry_point;
	SET_DESC_TABLE_REGISTERS (bc32);
	bc32.composition_bits =
		CTX32_CRITICAL | CTX32_GP | CTX32_FLOATING_POINT | CTX32_XMM |
		CTX32_SEGMENT | CTX32_BT_EXTENDED;

	btg_ret = BTGENERIC_IA32_CONTEXT_SET (NULL, &bc32);
	assert (btg_ret == BT_STATUS_SUCCESS);

#if PTEMU_ENABLED
	/* see if the new app was invoked by a debugger */
	if (btl_is_ptraced != IS_PTRACED_NO) {
		cur_thread_data->ptrace = PTEMU_PTRACED;
	}
#endif	// PTEMU_ENABLED

	/* record personality */
#if ! RELEASE
	if (btl_always_rie) {
		personality |= READ_IMPLIES_EXEC;
	}
#endif
	ret = iti_personality_interface (&cur_thread_data->personality);
	assert (ret == 0);
	*cur_thread_data->personality = personality;
	
	return 0;
}

/******************************************************************************/
/*
 *  btlib_cleanup_thread()
 */
/******************************************************************************/
void
btlib_cleanup_thread (int tbr_pid)
{
	btlib_thread_data_t *tbr;
	int found = 0;

	DBPRINT (btl_prt_thread, "tbr_pid=%d\n", tbr_pid);

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (cur_thread_data->tid != tbr_pid);

	/* find the thread to be removed */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	for (tbr = m_used_thread_list; tbr != NULL; tbr = tbr->next) {
		if (tbr->tid == tbr_pid) {
			found = 1;
			break;
		}
	}
	btlib_release_lock (BTL_LOCK_IND_THREAD);

	/* the thread should be in the used-thread-list only if it
	   terminated abnormaly. otherwise - in btlib_thread_exit it would
	   have been removed from the list */

	if (!found) {
		return;
	}
	assert (tbr->is_exited == 0);

	/* thread exited abnormaly */
	DBPRINT (btl_prt_thread, "abnormal\n");

	/* if it's holding a critical section then we are doomed */
	if (is_holding_critical_section (tbr)) {
		DBPRINT (btl_prt_thread, "holding cs - abort!!\n");
		btlib_abort_all ();
	}
	/* notify btgeneric so it can do its checks */
	BTGENERIC_THREAD_ABORTED (tbr_pid);

	/* free the thread data */
	cleanup_thread (tbr, TRUE);
}

/******************************************************************************/
/* free resources held by the given thread
 * input : tbr, thread to be free 
 *         free_thread_stack, whether to free thread stack
 * output: none
 */
/******************************************************************************/
static void
cleanup_thread (btlib_thread_data_t *tbr, int free_thread_stack)
{
	btlib_try_free_sighand_table (tbr);
	btlib_que_free (&tbr->pending_que);	/* free pending que */

	if (cur_thread_data->stackblock != 0) {
		btlib_deallocate_stack (tbr->stackblock, &tbr->tid,
					free_thread_stack);
	}
}

/******************************************************************************/
static int
clone_main (void *arg)
{
	long retval;
	BT_STATUS_CODE btg_ret;
	btlib_clone_data_t *clone_params = (btlib_clone_data_t *)arg;
	unsigned RELEASE_UNUSED flags =
		IA32_SYSCALL_PARAM1 (&clone_params->start_context);

	if (!m_kernel_support_nptl ) {
		/* switch thread data to the new thread data */
		__current = clone_params->td + 1;
	} else {
		assert (__current - 1 == clone_params->td);
	}

	/* store the new thread id */
	cur_thread_data->tid = SYSCALL (gettid, ());

	sigset64_2_sigset32 (&clone_params->sigmask, &cur_thread_data->sigmask);

	/* get the address of the long syscall flag */
	retval = iti_long_syscall_interface (&cur_thread_data->long_syscall_flag, NULL);
	assert (retval == 0);

	DBPRINT (0, "long_syscall_flag=%p\n", cur_thread_data->long_syscall_flag);

	/* get the address of the thread personality */
	retval = iti_personality_interface (&cur_thread_data->personality);
	assert (retval == 0);
	*cur_thread_data->personality = clone_params->personality;
	
	/* we are child now, eax should be 0 */
	clone_params->start_context.esp =
		IA32_SYSCALL_PARAM2 (&clone_params->start_context);
	clone_params->start_context.eax = 0;

	/* 
	 * add current thread to used threa list, should be done in child to 
	 * avoid cases that child exits before parent adds it to the list
	 */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	list_add (cur_thread_data, &m_used_thread_list);
#if ALL_THREAD_AFFINITY
	btl_num_threads++;
#endif
	btlib_release_lock (BTL_LOCK_IND_THREAD);

	/* add FS actual offset and GS actual offset to the calling sequence */
	retval = (long) BTGENERIC_THREAD_INIT (clone_params->td->glst,
					       (__u32) clone_params->start_context.esp, 
					       (__u64) 0);

	assert (retval == (long) BT_STATUS_SUCCESS);

	CHECK_DESC_TABLE_REGISTERS (clone_params->start_context);
	clone_params->start_context.composition_bits = 
		CTX32_CRITICAL | CTX32_GP | CTX32_SEGMENT |
		CTX32_FLOATING_POINT | CTX32_XMM | CTX32_BT_EXTENDED;

	btg_ret = BTGENERIC_IA32_CONTEXT_SET (NULL, &clone_params->start_context);

	retval = btlib_check_segment (cur_thread_data->bc32);

	/* 
	 * if the base address of one of the segments changed then 
	 * it must be that clone was called with clone-settls
	 */
	assert ((retval == 0) || (flags & CLONE_SETTLS));

	assert (btg_ret == BT_STATUS_SUCCESS);

	if (!m_kernel_support_nptl) {
		/* for kernel without NPTL support, all signals are blocked in 
		   the parent before clone, we should recover original signal
		   mask */
		retval = set_sigmask (&cur_thread_data->sigmask);
		assert (retval == 0);
	}

	/* clone_params is not needed anymore */
	free (clone_params);

        /* a new thread needs to be stopped with sigstop */
        if (cur_thread_data->ptrace) {
                assert (cur_thread_data->ptrace == PTEMU_PTRACED);
                memset (&cur_thread_data->tia_signal, 0,
                        sizeof (cur_thread_data->tia_signal));
                cur_thread_data->tia_signal.siginfo.si_signo = SIGSTOP;
                retval = SYSCALL (tkill, (cur_thread_data->tid, PTEMU_SIG0));
                sys_err_chk (retval, "kill");
        }

	
	/* run the new thread */
	DBPRINT (btl_prt_thread, "run\n");
	cur_thread_data->longjmp_status = RUN;
	btrans_jump ();

	assert (!"should not get here");
	return 0;
}

/**************************************************************************/
/* set the sighand table for the new thread
 * input : clone_flags - the flags for the clone syscall
 *         parent - the parent's thread data struct
 *         child - the child's thread data struct
 * output: none
 */
/**************************************************************************/
static void
clone_signals (unsigned clone_flags, btlib_thread_data_t *child)
{
	struct signal_handler_table *cur_sighand;

	assert (child != NULL);

	cur_sighand = cur_thread_data->sighandlers;

	btlib_acquire_lock (BTL_LOCK_IND_SIGHAND);
	if (clone_flags & CLONE_SIGHAND) {
		cur_sighand->ref_count++;
		child->sighandlers = cur_sighand;
	} else {
		child->sighandlers = btlib_create_sighand_table ();
		assert (child->sighandlers != NULL);

		memcpy (child->sighandlers->handlers, cur_sighand->handlers, 
			sizeof (struct signal_handler_desc) * (_NSIG + 1));
	}
	btlib_release_lock (BTL_LOCK_IND_SIGHAND);

	__sync_synchronize ();
}

/**************************************************************************/
/* IA-32 set_tid_address syscall implementation */
/**************************************************************************/
long
btlib_set_tid_address (const BTGENERIC_IA32_CONTEXT *bc32)
{
	long retval;

	/* abort execution if set_tid_address is not supported */
	retval = SYSCALL (set_tid_address, (U2P (IA32_SYSCALL_PARAM1 (bc32))));
	sys_err_chk (retval, "set_tid_address");

	return retval;
}

#if ALL_THREAD_AFFINITY
/******************************************************************************
 btlib_record_explicit_affinity_request - records the affinity mask that was
 set due to the user's (IA-32 thread) explicit request. This mask will override
 the mask seleceted by the thread-affinity or shared-mem-process-affinity 
 solutions.This mask will be inherited in case of clone or fork.

 This function is called only after the set_affinity ia32-sys-call was passed
 to the wrapper and returned successfully.
*******************************************************************************/
void
btlib_record_explicit_affinity_request (int target_tid)
{
	long retval;

	/* Note: the user may request affinity mask that inlcudes more 
	   processors that are actually available. The kernel ANDs the 
           available cpus with the user's requested cpus and sets the actaul 
           affinity mask accordingly. This is the affinity mask
           we'll record here and not the user's initial request */
	cpu_set_t current_affinity_mask;

	/* check if the affinity mask is for current thread,
	   if not, we do nothing about it for now... */
	if (target_tid == cur_thread_data->tid) {
		memset (&current_affinity_mask, 0, sizeof (current_affinity_mask));
		retval = SYSCALL (sched_getaffinity, 
				  (target_tid, 
				   (unsigned int)sizeof (current_affinity_mask),
				   &current_affinity_mask.__bits[0]));
		sys_err_chk (retval, "sys_sched_getaffinity");
		DBPRINT (btl_prt_thread,
			 "tid=%d explicit affinity mask requested=%lx\n",
			 target_tid, current_affinity_mask.__bits[0]);
		cur_thread_data->explicit_affinity_mask = current_affinity_mask;

		if (m_did_clone == 0) {
			/* this is the main thread that asks for explicit 
			   affinity_mask (before any clones), this will be 
			   recorded as the process's orig mask */
			m_orig_process_affinity_mask =
				cur_thread_data->explicit_affinity_mask;
		}
	}
	return;
}

/**************************************************************************
get_current_cpu - attempts to read from /proc/self/stat and retrieve
 the current processor the thread is executing on.
Return Value:
 Success: the processoor num the thread is currently executin on
 Failure: -1
**************************************************************************/
#define STAT_FILE "/proc/self/stat"
#define STAT_STR_LENGTH 512ul
int
get_current_cpu ()
{
	int stat_file;
	char stat_str[STAT_STR_LENGTH];
	int current_processor = -1, retval;

	/* read from /proc/self/stat file, which is the processor we're executing on */
	stat_file = SYSCALL (open, (STAT_FILE, O_RDONLY, 0));
	if (IS_SYS_ERROR (stat_file)) {
		DBPRINT (btl_prt_thread, "Error: open failed: %d\n", stat_file);
		return -1;
	} else {
		retval = SYSCALL (read, (stat_file, stat_str, STAT_STR_LENGTH));
		if (IS_SYS_ERROR (retval)) {
			DBPRINT (btl_prt_thread,
				 "Error: failed to read from %s: %d\n", 
				 STAT_FILE, retval);
			SYSCALL (close, (stat_file));
			return -1;
		} else {
			char * p = strchr (stat_str, ')');
			int j = 1;
			while ((p != NULL) && (j < 37)) {
				p = strchr (p + 1, ' ');
				j ++;
			}
			if (j == 37) {
				p = strchr (p + 1, ' ');
				if (p != NULL) {
					*p = '\0';
					current_processor = atol (p + 1);
				}
			}

			DBPRINT (btl_prt_thread, "current processor - %d\n",
				 current_processor);
		}
		SYSCALL (close, (stat_file));
	}

	return current_processor;
}

/**************************************************************************
get_random_cpu - randomly selects a CPU number out of all the allowed CPUs
                 according to the input cpu_mask
Return Value:
 Success: the selected CPU number
 Failure: -1
**************************************************************************/
int
get_random_cpu (cpu_set_t *cpu_mask)
{
	int num_of_allowed_cpus, cpu_X_out_of, randnum;
	unsigned int cpu;

	/* how many cpus are allowed according to cpu mask */
	num_of_allowed_cpus = 0;
	for (cpu = 0; cpu < 8 * sizeof (*cpu_mask) - 1; cpu++) {
		if (__CPU_ISSET (cpu, cpu_mask)) {
			num_of_allowed_cpus++;
		}
	}

	/* now get a random number */
	{
		struct timeval tv;
		long ret = SYSCALL (gettimeofday, (&tv, NULL));
		if (IS_SYS_ERROR (ret)) {
			randnum = 0;
		} else {
			randnum = tv.tv_sec;
		}
	}

	/* so we'll use the X'th cpu out of these allowed */
	cpu_X_out_of = randnum % num_of_allowed_cpus;
	for (cpu = 0; cpu < 8 * sizeof (*cpu_mask) - 1; cpu++) {
		if (__CPU_ISSET (cpu, cpu_mask)) {
			if (cpu_X_out_of == 0) {
				DBPRINT (btl_prt_thread, "randomly selected cpu %d\n", cpu);
				return cpu;
			}
		}
		cpu_X_out_of--;
	}

	/* if we get here, some error happened... */
	DBPRINT (btl_prt_thread, "Error: couldn't select a cpu\n");
	return -1;
}

/**************************************************************************
select_one_cpu - This function implements the heuristic for selecting only
 one cpu for all the process's threads to run on, out of all allowed cpus 
 according to the cpu_mask.
Return Value:
 Success: the new cpu-affinity-mask that contains only one cpu
 Failure: 0 (empty affinity-mask)
**************************************************************************/
static int select_one_cpu (cpu_set_t *cur_mask, cpu_set_t *new_mask)
{
	int cpu;

	DBPRINT (btl_prt_thread, "allowed cpu_mask=%lx\n", cur_mask->__bits[0]);

	/* try getting current cpu we're executing on */
	cpu = get_current_cpu ();
	if (cpu < 0) {
		cpu = get_random_cpu (cur_mask);
	}
	DBPRINT (btl_prt_thread, "new_mask=%lx\n", new_mask->__bits[0]);
	if (cpu >= 0 ) {
		__CPU_SET (cpu, new_mask);
	}
	assert (is_only_one_cpu_allowed (new_mask)); 
	assert (cpu >= 0);
	return (cpu >= 0) ? 0 : -1;

}

#if DEBUG
/**************************************************************************
is_only_one_cpu_allowed - make sure that the input cpu_mask specifies only
one cpu.
Return Value:
 Success: 1
 Failure: 0
**************************************************************************/
int
is_only_one_cpu_allowed (cpu_set_t *cpu_mask)
{
	unsigned int cpus_allowed;
	unsigned long cpu;

	cpus_allowed = 0;

	DBCODE (0, __CPU_PRINT ("is_only_one_cpu_allowed", cpu_mask));
	for (cpu = 0; cpu < 8 * sizeof (*cpu_mask) - 1; cpu++) {
		if (__CPU_ISSET (cpu, cpu_mask)) {
			cpus_allowed++;
		}
	}

	if (cpus_allowed == 1)
		return 1;

	return 0;
}
#endif /* DEBUG */
#endif /* ALL_THREAD_AFFINITY */
/******************************************************************************/
/* 
 * implement the IA-32 clone syscall 
 *	param1 - clone flags
 *	param2 - child stack
 *	param3 - parent tid pointer
 *	param4 - new TLS area for new thread (child), user_desc pointer
 *	param5 - child tid pointer
 *
 * NOTES: some clone flags must be handled inside BTLib
 *	CLONE_SETTLS : 
 *	1. It requests that the very first userland instruction of the new thread 
 *         can access thread local data by thread pointer. 
 *	2. No limitation for the parent thread. 
 *	3. BTLib must handle this by pasinging its own TLS pointer to kernel 
 *         because IA-32 kernel implements TLS using GDT.
 *
 *	CLONE_PARENT_SETTID/CLONE_CHILD_SETTID : 
 *	The parent_tidptr must be filled with new thread's tid before both 
 *	parent and child are back to userland. There's no way for BTLib
 *	to guarantee that, so it must pass kernel tid pointer of translated 
 *	process.
 *
 *	CLONE_CHILD_CLEARTID :
 *	The kernel takes child_tidptr as a futex, record child_tidptr somewhere.
 *	When the thread exits, the kernel is responsible to clear it, and wake 
 *	up thread waiting on it. BTLib should passing kernel its own 
 *	child_tidptr because we want to use it for BTLib thread memory 
 *	management. And for child_tidptr of translated process, BTLib should 
 *	should simulate kernel behavior
 *
 *	xlin2: usage of CLONE_CHILD_CLEARTID changed because there is a race
 *             condition: the thread waiting on the futex can be waken up even
 *             if the child thread call exit_group
 */
/******************************************************************************/
long
btlib_clone (BTGENERIC_IA32_CONTEXT *bc32)
{
	void *stacktop = 0;
	unsigned ia32_stack = 0;
	btlib_clone_data_t  *clone_params;
	btlib_thread_desc   *ctd;
	btlib_thread_data_t *child;
	btlib_thread_data_t *parent;
	unsigned clone_flags;
	long retval;
	void *new_r13;
	void *parent_tidptr = NULL, *child_tidptr = NULL;
	k_sigset_t mask, oldmask;
#if ALL_THREAD_AFFINITY
	cpu_set_t current_affinity_mask, new_affinity_mask;
	int tid;
#endif	/* ALL_THREAD_AFFINITY */

	/* Locate the thread data for the current thread. */
	parent = cur_thread_data;
	assert (parent->tid == SYSCALL (gettid, ()));

	/* get the syscall params */
	clone_flags = IA32_SYSCALL_PARAM1 (bc32);
	ia32_stack = IA32_SYSCALL_PARAM2 (bc32);
	DBPRINT (btl_prt_thread, "clone_flags=0x%x ia32_stack=0x%x\n",
		 clone_flags, ia32_stack);

	/* don't handle shared pid */
	if (clone_flags & CLONE_PID) {
		DBPRINT (TRUE, "someone is requesting a shared pid!\n");
		return -EPERM;
	}
	/* if the memory is not shared then a special treatment is required */
	if (!(clone_flags & CLONE_VM)) {
#if DEBUG
		if (clone_flags & CLONE_SETTLS) {
			assert (!"Don't support!\n");
		}
#endif
		retval = do_fork (parent, clone_flags, ia32_stack,
				  U2P (IA32_SYSCALL_PARAM3 (bc32)),
				  U2P (IA32_SYSCALL_PARAM5 (bc32)), bc32);

		return retval;
	}
	/* verify that we run on an adequate kernel */
	if (IS_NPTL_FLAGS (clone_flags)) {
		if (!m_kernel_support_nptl) {
			DBPRINT (TRUE, "CLONE_THREAD is called on an old kernel\n");
			return -EINVAL;
		}
	}
	

	/* allocate new stack and thread data for the child thread. */
	stacktop = allocate_thread_data (&ctd);
	if (IS_SYS_ERROR (stacktop)) {
		return (long) stacktop;
	}
	assert (ctd != NULL);
	child = ctd->blst;
	new_r13 = ctd + 1;

	/* allocate pending que for new thread */
	btlib_que_init (&child->pending_que);

	child->tid = 0;
	/* 
	 * param4 is a pointer to user_desc (TLS area of new thread) 
	 * if CLONE_SETTLS is specified 
	 */
	if (clone_flags & CLONE_SETTLS) {
		/* 
		 * set child's TLS here because if validating user_desc fails, 
		 * the 64 bit clone should not be called!
		 */
		int idx;
		user_desc u_info;

		retval = safe_memcpy (&u_info,
				      U2P (IA32_SYSCALL_PARAM4 (bc32)),
				      sizeof (u_info));
		if (retval != 0) {
			return -EFAULT;
		}

		if (is_udesc_empty (&u_info)) {
			return -EINVAL;
		}

		idx = u_info.entry_number;

		if (idx < GDT_USER_ENT_FIRST || idx > GDT_USER_ENT_LAST) {
			return -EINVAL;
		}
		set_desc (&child->tls_array[idx - GDT_USER_ENT_FIRST], &u_info);
	} else {
		/* child should have same tls array with parent if not specifying 
		   CLONE_SETTLS */
		memcpy (child->tls_array, cur_thread_data->tls_array, 
			sizeof (desc_struct) * GDT_USER_ENT_NUM);
	}

	if (clone_flags & CLONE_THREAD) {
		m_thread_group = 1;
	}

	clone_signals (clone_flags, child);

	/*
	 * the clone-params struct will be released in clone_main. it
	 * is very rare that more than one such a struct will be used
	 * simultanuously. therefore, there is no benefit in using the
	 * memory pools here, only an overhead
	 */
	clone_params = malloc (sizeof (*clone_params));
	if (clone_params == NULL) {
		DBPRINT (TRUE, "malloc failed\n");
		cleanup_thread (child->stackblock, TRUE);
		return -ENOMEM;
	}

	clone_params->start_context = *bc32;
	clone_params->td = ctd;
	sigset32_2_sigset64 (&cur_thread_data->sigmask, &clone_params->sigmask);

	if (!(clone_flags & CLONE_UNTRACED)) {
		clone_params->td->blst->ptrace = cur_thread_data->ptrace;
	}
	if ((clone_flags & CLONE_PARENT_SETTID) || 
	    (clone_flags & CLONE_CHILD_SETTID)) {
		/* CLONE_VM must be specified upon here, so CLONE_PARENT_SETTID and
		   CLONE_CHILD_SETTID is exactly the same */
		parent_tidptr = U2P (IA32_SYSCALL_PARAM3 (bc32));
	}

	if (clone_flags & CLONE_CHILD_CLEARTID) {
		child_tidptr = U2P (IA32_SYSCALL_PARAM5 (bc32));
	}


	DBPRINT (btl_prt_thread,
		 "stackblock=%p stacksize=%lx flags=%x parent_tidptr=%p, new_r13=%p child_tidptr=%p\n",
		 child->stackblock, child->user_stack_size, clone_flags,
		 parent_tidptr, new_r13, child_tidptr);

#if ALL_THREAD_AFFINITY
	/* affinity solution for inter-thread ordering hazard */
	if (m_did_clone == 0 
#if !RELEASE
	    && btl_no_affinity == 0
#endif
        ) {
		/* this is the first time clone is called */
		m_did_clone = 1;

		/* Before narrowing the thread's affinity mask, we must verify that
		   (1) shared-mem-process-affinity or shared-mem-ordering solution  or switch to
		       ordered code wasn't already applied
		   (2) explicit-mask is still 0, we don't want to override user's explicit request */
		if ((m_did_shared_mem == 0 && btl_did_ordered_code == 0) && 
		    __CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask)) {

			/* just before calling clone, we'll reduce the affinity mask of the 
			   thread to allow only one CPU. The affinity mask will be inherited by the
			   child thread */
			tid = SYSCALL (gettid, ());
			memset (&current_affinity_mask, 0, sizeof (current_affinity_mask));
			memset (&new_affinity_mask, 0, sizeof (new_affinity_mask));
			retval = SYSCALL (sched_getaffinity, 
					  (tid, 
					   (unsigned int)sizeof(current_affinity_mask),
					   &current_affinity_mask.__bits[0]));
			sys_err_chk (retval, "sys_sched_getaffinity");

			/* we'll save the original affinity mask of this process, so in case
			   of fork we'll be able to use it */
			m_orig_process_affinity_mask = current_affinity_mask;

			/* the heuristic for selecting one CPU out of all the allowed CPUs
			   is contained in the "select_one_cpu" function */
			retval = select_one_cpu (&current_affinity_mask, &new_affinity_mask);
			assert (retval == 0);

			retval = SYSCALL (sched_setaffinity, 
					  (tid, 
					   (unsigned int)sizeof (new_affinity_mask),
					   &new_affinity_mask.__bits[0]));
			sys_err_chk (retval, "sys_sched_setaffinity");
			btl_narrowed_affinity = 1;
		}
	}
#if DEBUG
	if (__CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask) && btl_narrowed_affinity) {
		tid = SYSCALL (gettid, ());
		memset (&current_affinity_mask, 0, sizeof (current_affinity_mask));
		retval = SYSCALL (sched_getaffinity, (tid, (unsigned int)
				  sizeof (current_affinity_mask),
				  &current_affinity_mask.__bits[0]));
		sys_err_chk (retval, "sys_sched_getaffinity");
		assert (is_only_one_cpu_allowed (&current_affinity_mask));
	}
#endif	/* DEBUG */
#endif	/* ALL_THREAD_AFFINITY */


	/* BTLib always uses thread local data */
	if (m_kernel_support_nptl) {
		clone_flags |= CLONE_SETTLS;
	} else { 
		/* 
		 * there is no kernel NPTL support
		 *
		 * The child will inherit the parent's blocked signal mask.
		 * so block all signals before calling clone(). this way we
		 * make sure no signals arrive at the child before it is ready
		 */
		sigfillset (&mask);
		retval = SYSCALL (rt_sigprocmask, 
				  (SIG_SETMASK, &mask, &oldmask, 
				  sizeof(k_sigset_t)));
		sys_err_chk (retval, "rt_sigprocmask(1)");
		clone_params->sigmask = oldmask;
	}

	/* pass personality to child */
	assert (cur_thread_data->personality != NULL);
	clone_params->personality = *cur_thread_data->personality;
	
	/* create the child thread */
	retval = __btlib_clone (clone_main,
				child->stackblock,
				(unsigned long) (child->user_stack_size),
				(unsigned long) clone_flags,
				(void *) clone_params,
				parent_tidptr, new_r13, child_tidptr);

	if (!m_kernel_support_nptl) {
		/* restore the sigmask */
		long r = SYSCALL (rt_sigprocmask, 
				  (SIG_SETMASK, &oldmask, NULL, 
				  sizeof(k_sigset_t)));
		sys_err_chk (r, "rt_sigprocmask(2)");
	}

	DBPRINT (retval && btl_prt_thread, "child pid=%ld<%lx>\n", retval, retval);
	if (IS_SYS_ERROR (retval)) {
		DBPRINT (TRUE, "clone failed=%ld\n", retval);
		/* undo allocate_thread_data */
		cleanup_thread (child, TRUE);
	}

	return retval;
}

/**************************************************************************/
/* 
 * group termination procedure
 * called when the translated process calls the exit_group syscall
 */
/**************************************************************************/
void
btlib_group_exit (const BTGENERIC_IA32_CONTEXT *bc32)
{
	int exit_code = IA32_SYSCALL_PARAM1 (bc32);

	BLOCK_ALL_SIGNALS ();

	/* if this syscall is called in a thread that is not a part
	   of a thread group ( as in linuxthreads ) then it should be
	   treated as exit() */
	DBPRINT (btl_prt_thread, "m_thread_group=%d\n", m_thread_group);
	if (!m_thread_group) {
		before_thread_exit ();
		__btlib_exit (exit_code, cur_thread_data->stackblock,
			      __default_stack_size);
	}

	STATCODE (btl_stats, btlib_print_stats ());

	/* let btgeneric prepare stats on IA32-EL run */
	BTGENERIC_NOTIFY_EXIT ();

	SYSCALL (exit_group, (exit_code));
	RESTORE_SIGNALS ();	/* dummy - should never be reached */
}

/**************************************************************************/
/* 
 * normal termination procedure
 * called when the translated process calls the IA-32 exit() syscall
 */
/**************************************************************************/
void
btlib_thread_exit (BTGENERIC_IA32_CONTEXT *bc32)
{
	int exit_code;
	int signo;
	siginfo_t siginfo;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	assert (bc32 == cur_thread_data->bc32);

	assert (!btlib_is_inside_lock (cur_thread_data));
	assert (!is_holding_critical_section (cur_thread_data));

	/* get syscall params */
	exit_code = IA32_SYSCALL_PARAM1 (bc32);

	DBPRINT (0, "exiting %d\n", exit_code);

	/* we dont want to get signals from this point and further */
	BLOCK_ALL_SIGNALS ();

	/* see if we have pending signals */
	signo = btlib_que_deque (&siginfo);
	if (signo != 0) {
		/* the signal is not blocked. it needs to be handled. */

		/* 1 - set the IA-32 context to repeat the call to exit() */
		bc32->eip -= 2;
		bc32->eax = IA32_SYSCALL_EXIT;
		cur_thread_data->syscall_id = IA32_SYSCALL_NONE;
		/* 2 - deliver the signal. continue with signals masked -
		       it shouldn't matter much */
		btlib_run_ia32_handler (&siginfo, NULL, 0);
		assert (!"should never get here");
	}

	before_thread_exit ();

	/* all signals must be blocked! */
	__btlib_exit (exit_code, cur_thread_data->stackblock,
		      __default_stack_size);

	RESTORE_SIGNALS ();	/* dummy - never reached */
}

/******************************************************************************/
/* implement the IA-32 fork syscall */
/******************************************************************************/
long
btlib_fork (BTGENERIC_IA32_CONTEXT *bc32)
{
	long retval;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	/* do the actual work */
	retval = do_fork (cur_thread_data, SIGCHLD, bc32->esp, NULL, NULL, bc32);

	return retval;
}

/**************************************************************************/
/**************************************************************************/
static long
do_fork (btlib_thread_data_t *parent RELEASE_UNUSED,
	 unsigned clone_flags,
	 unsigned ia32_tos,
	 int *parent_tidptr, 
         int *child_tidptr, 
         BTGENERIC_IA32_CONTEXT *bc32)
{
	long retval;
	btlib_thread_data_t *p = NULL;
	BT_STATUS_CODE btg_retval = BT_STATUS_UNSUCCESSFUL;
#if ALL_THREAD_AFFINITY
	cpu_set_t current_affinity_mask;
	long aff_retval;
#endif	/* ALL_THREAD_AFFINITY */

	assert (parent->tid == SYSCALL (gettid, ()));
	DBPRINT (btl_prt_thread, "clone_flags=0x%x ia32_tos=0x%x\n",
		 clone_flags, ia32_tos);

	/* 
	 * freeze ia32exec.bin to ensure that its state is
	 * consistent in the child process. 
	 */
	BTGENERIC_FREEZE ();

	/* 
	 * freeze btlib
	 * **********
	 * DONT EVER move it before the call to BTGENERIC_FREEZE
	 * **********
	 * (see the header of btlib_enter_critical_section)
	 */

	/* acquire all the sync objects used by btlib */
	btlib_freeze ();

	/* mask all the signals, so the child dont get one before it is ready */
	BLOCK_ALL_SIGNALS ();

#if ALL_THREAD_AFFINITY
	/* before we do actual fork...
	   if we narrowed the thread's affinity mask to allow only one processor
	   (due to threads\process affinity solution), we don't want this to be
	   inherited by the new process we'll craete.
	   So first change the affinity mask to be the original mask as before we 
	   narrowed it.
	   After the fork work is completed, then the narrowed affinity mask will 
	   be restored for the parent process only.

	   BUT, in case the current thread explicitly called set-affinity, the 
	   affinity mask he requested overrides the affinity-solution selected mask 
	   and the process's original affinity mask, therefore we'll let the child 
	   process inherit the current affinity mask. */
	if (btl_narrowed_affinity && 
	    __CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask)) {
		memset (&current_affinity_mask, 0, sizeof (current_affinity_mask));
		aff_retval = SYSCALL (sched_getaffinity, 
				      (cur_thread_data->tid, 
				       (unsigned int)sizeof (current_affinity_mask),
				       &current_affinity_mask.__bits[0]));
		sys_err_chk (aff_retval, "sys_sched_getaffinity");

		aff_retval = SYSCALL (sched_setaffinity, 
				      (cur_thread_data->tid,
				      (unsigned int) sizeof(m_orig_process_affinity_mask),
				      &m_orig_process_affinity_mask.__bits[0]));
		sys_err_chk (aff_retval, "sys_sched_setaffinity");
		
	}
#endif  /* ALL_THREAD_AFFINITY */

	assert ((clone_flags & CLONE_VM) == 0);
	assert ((clone_flags & CLONE_SETTLS) == 0);
	retval = SYSCALL (clone2, 
			  ((unsigned long) clone_flags,
			  (unsigned long) NULL, 0l,
			  (unsigned long) parent_tidptr,
			  (unsigned long) child_tidptr, 
			  (unsigned long) NULL));

	if (retval) {

		/************** we are in the parent */
		DBPRINT (0, "in parent: child tid=%ld\n", retval);

#if ALL_THREAD_AFFINITY
		/* just before returning, we must narrow again the affinity mask.
		   This is done only for the parent */
		if (btl_narrowed_affinity && 
		    __CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask)) {
			aff_retval = SYSCALL (sched_setaffinity, 
					      (cur_thread_data->tid,
					       (unsigned int)sizeof(current_affinity_mask),
					       &current_affinity_mask.__bits[0]));
			sys_err_chk (aff_retval, "sys_sched_setaffinity");
			
		}
#endif	/* ALL_THREAD_AFFINITY */

		/* restore the signals and release the handlers table */
		RESTORE_SIGNALS1 ();

		/* release all the sync objects used by btlib */
		btlib_unfreeze (1l);

		/* btgeneric may now continue as usual */
		BTGENERIC_UNFREEZE ();

		/* return the child's tid */
		DBCODE (0, print_thread_list ());

		return retval;
	}

	/************** we are in the child process */

	/* 
	 * the child has a copy of its parent memory space, so their
	 * thread data structures reside at the same address
	 */
	assert (cur_thread_data == parent);
	cur_thread_data->tid = SYSCALL (gettid, ());

	/* free all critical sections that may be locked by btlib */
	btlib_unfreeze (0l);

	/* 
	 * let btgeneric continue - it cannot do anything anyway
	 * because the current process is the only thread in the new
	 * app
	 */
	BTGENERIC_UNFREEZE ();

	/* change the thread's identity */
	btg_retval = BTGENERIC_CHANGE_THREAD_IDENTITY ();
	assert (btg_retval == BT_STATUS_SUCCESS);

	/* 
	 * remove all the elements in the thread list 
	 * ( no need to lock it since we are the only thread )
	 */
	p = m_used_thread_list;
	while (p != NULL) {
		btlib_thread_data_t *next = p->next;
		if ((p != cur_thread_data) && (p->stackblock != NULL)) {
			list_del (p, &m_used_thread_list);
			BTGENERIC_THREAD_ABORTED (p->tid);
			cleanup_thread (p, TRUE);
		}
		p = next;
	}
#if ALL_THREAD_AFFINITY
	btl_num_threads = 1;
#endif


	cur_thread_data->sighandlers->ref_count = 1;
	cur_thread_data->is_run_context_ready = cur_thread_data->tid;

	/* change the esp of the current process to the given stack */
	if (ia32_tos != 0) {
		bc32->esp = ia32_tos;
	}
#if ALL_THREAD_AFFINITY
	/*clear the global flags related to the affinity-ordering-solution */
	__CPU_ZERO (&m_orig_process_affinity_mask);
	m_did_clone = 0;
	btl_narrowed_affinity = 0;
#endif				/*ALL_THREAD_AFFINITY */
#if SHMEM_PROC_AFFINITY || SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
	m_did_shared_mem = 0;
#endif

	/* it's a single threaded app now */
	m_thread_group = 0;

	/* and it is not ptraced */
	cur_thread_data->ptrace = 0ul;
	btl_is_ptraced = IS_PTRACED_NO;

	/* from here on the current thread may receive signals */
	RESTORE_SIGNALS ();

	/* return success indication */
	return 0;
}

#if SHMEM_PROC_AFFINITY || SHMEM_ORDER_SOLUTION || ALL_THREAD_AFFINITY
/**************************************************************************
set_affinity_to_all_threads - set the affinity mask of all the threads that
 didn't explicitly request a specific affinity mask

Return Value:
 Success: 0
 Failure: -1
**************************************************************************/
static long
set_affinity_to_all_threads (cpu_set_t *new_affinity_mask)
{
	long retval;
	btlib_thread_data_t *susp_thread;

	/* go over the existing threads list */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	for (susp_thread = m_used_thread_list; susp_thread != NULL;
	     susp_thread = susp_thread->next) {
		if (__CPU_IS_ZERO (&susp_thread->explicit_affinity_mask)) {
			/* this thread didn'e explicitly call set-affinity to 
			   request an affinity mask we can chagne his affinity 
			   mask */
			retval = SYSCALL (sched_setaffinity, 
					  (susp_thread->tid,
					   (unsigned int)sizeof (*new_affinity_mask),
					   &new_affinity_mask->__bits[0]));
			sys_err_chk (retval, "sys_sched_setaffinity");
		}
	}			/*for */

	btlib_release_lock (BTL_LOCK_IND_THREAD);

	return 0;
}
#endif // SHMEM_PROC_AFFINITY || SHMEM_ORDER_SOLUTION

#if SHMEM_PROC_AFFINITY
/**************************************************************************
select_lowest_cpu - This function will find the lowest allowed cpu in the
 input cpu_mask.
Return Value:
 Success: the new cpu-affinity-mask that contains only the lowest allowed cpu
 Failure: 0 (empty affinity-mask)
**************************************************************************/
static int
select_lowest_cpu (cpu_set_t * cur_mask, cpu_set_t * new_mask)
{
	int cpu, found = -1;

	__CPU_ZERO (new_mask);
	for (cpu = 0; cpu < (int)(8 * sizeof (cur_mask) - 1); cpu++) {
		if (__CPU_ISSET (cpu, cur_mask)) {
			__CPU_SET (cpu, new_mask);
			found = cpu;
			break;
		}
	}

	/* getting here means a problem occurred */
	DBPRINT (found < 0, "FAILED cur_mask=%lx\n", cur_mask->__bits[0]);
	assert (found >= 0);
	return found;
}



/**************************************************************************
apply_shared_memory_affinity_solution - Attempt to move all application's 
 threads to the lowest processor allowed. For any thread that explicitly
 called set-affinity to request an affinity mask, we will not override their
 request but leave theit affinity mask as is.

 In order to minimize the ordering hazard, in case of multi-threaded apps, we'll
 also attempt to suspend all threads to ensure they are rescheduled to the new
 processor asap.

Return Value:
**************************************************************************/
void
apply_shared_memory_affinity_solution ()
{
	long retval;
	cpu_set_t current_affinity_mask, new_affinity_mask;

	DBPRINT (btl_prt_mem || btl_prt_thread, "apply_shared_memory_affinity_solution\n");
	if (__sync_bool_compare_and_swap_si(&m_did_shared_mem, 0, 1)
#if ! RELEASE
	    && btl_no_affinity == 0
#endif
        ) {
		btl_narrowed_affinity = 1;

		if (m_did_clone) {
			/* find the lowest processor allowed according to the process's 
			   original affinity mask */
			retval = select_lowest_cpu (&m_orig_process_affinity_mask, 
						    &new_affinity_mask);
			assert (retval == 0);

			btlib_acquire_lock (BTL_LOCK_IND_SUSPENSION);
			
			/* suspend all the other threads of the current process */
			retval = suspend_all_threads ();
			assert (retval == 0);

			/* for each thread that didn't explicitly call set-affinity for 
			   himself, set the new affinity mask */
			retval = set_affinity_to_all_threads (&new_affinity_mask);
			assert (retval == 0);

			/* resume all threads */
			retval = resume_all_threads ();
			assert (retval == 0);

			btlib_release_lock (BTL_LOCK_IND_SUSPENSION);
			
		} else { 
			/* app is single threaded */
			/* change the affinity mask unless there was an explicit 
			   request by the user */
			if (__CPU_IS_ZERO (&cur_thread_data->explicit_affinity_mask)) {
				/* find the process's current affinity mask */
				memset (&current_affinity_mask, 0, 
					sizeof (current_affinity_mask));
				retval = SYSCALL (sched_getaffinity, 
						  (cur_thread_data->tid, 
						   (unsigned int)sizeof(current_affinity_mask),
						   &current_affinity_mask.__bits[0]));
				sys_err_chk (retval, "sys_sched_getaffinity");

				/* store the process's original affinity mask to be used in 
				   case of fork */
				m_orig_process_affinity_mask = current_affinity_mask;

				/* find the lowest processor allowed according to the process's 
				   current affinity mask */
				retval = select_lowest_cpu (&current_affinity_mask, 
							    &new_affinity_mask);
				assert (retval == 0);

				/* set the current process's new affinity mask */
				retval = SYSCALL (sched_setaffinity, 
						  (cur_thread_data->tid, 
						  (unsigned int) sizeof (new_affinity_mask),
						  &new_affinity_mask.__bits[0]));
				sys_err_chk (retval, "sys_sched_setaffinity");
				DBPRINT (btl_prt_thread,
					 "single threaded app - affinity mask set to %lx\n",
					 new_affinity_mask.__bits[0]);
			}	/* else */
		}	/* m_did_clone */
	}	/* m_did_shared_mem */
	return;
}

#elif SHMEM_ORDER_SOLUTION
void
apply_shared_memory_ordering_solution ()
{
	EMULATION_POLICY emulation_policy;	

	DBPRINT (btl_prt_mem || btl_prt_thread, "\n");

	if (btl_is_ptraced == IS_PTRACED_YES) { 
		/* no sense in trying under debugger - required suspension will fail */
		DBPRINT (btl_prt_mem || btl_prt_thread, 
			 "apply_shared_memory_ordering_solution not applied under debugger\n");
		return;
	}

	if (m_did_shared_mem == 0 && btl_did_ordered_code == 0 
		/* since can't put call to BTGENERIC_SET_EMULATION_POLICY in critical 
		   section, and since the call can theoretically fail => the two boolean's 
		   m_did_shared_mem and m_did_ordered_code are not set atomically 
		   (using __sync_bool_compare_and_swap_si before the call, since in that 
		   case if the call fails another thread might miss calling it again.  
		   having themnot set atomicaly merely means that two threads might call 
		   BTGENERIC_SET_EMULATION_POLICY, but since the change always goes in 
		   the same direction (to USE_ORDERED_LD_ST) and since there is a critical 
		   section in BTGENERIC_SET_EMULATION_POLICY - there is no correctness 
		   problem). */
#if ! RELEASE
	    && btl_no_affinity == 0
#endif
	) {
		emulation_policy.memory_order_policy = USE_ORDERED_LD_ST;
		emulation_policy.strict_state_policy = POLICY_UNCHANGED;

		BTGENERIC_SET_EMULATION_POLICY (&emulation_policy);
		assert (emulation_policy.memory_order_policy == USE_ORDERED_LD_ST);
		btl_did_ordered_code = m_did_shared_mem = 1;
#if ALL_THREAD_AFFINITY
		if (__sync_bool_compare_and_swap_si(&btl_narrowed_affinity, 1, 0)) {
			assert (! __CPU_IS_ZERO (&m_orig_process_affinity_mask));
			/* give all the threads (that didn't explicitly change their affinity) 
			   the original process affinity */
			set_affinity_to_all_threads (&m_orig_process_affinity_mask);
		}
#endif
	}

	return;
}
#endif


/******************************************************************************
 This function will try to suspend all running threads, not including current. 
 If it fails to suspend an existing thread, it will still continue to the next 
 threads and attempt to suspend them.

 Note: After suspending other threads and before resuming them, calling 
       printf\DBPRINT is not allowed! This is because they might be holding 
       a lock that prinf needs this will cause a deadlock!!

Return Value:
 Success: 0 (even if not all threads were successfully suspended)
 Failure: -1
**************************************************************************/
long
suspend_all_threads (void)
{
	long retval;
	int status;
	btlib_thread_data_t *susp_thread;

	/* go over the existing threads list and try to suspend them */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	for (susp_thread = m_used_thread_list; susp_thread != NULL;
	     susp_thread = susp_thread->next) {
		if (susp_thread->tid == cur_thread_data->tid) {
			/* don't suspend current thread... */
			continue;
		}

		/* now attach to target thread and suspend it in the 64 bit level */
		retval = SYSCALL (ptrace, (PTRACE_ATTACH, susp_thread->tid, 0ul, 0ul));
		sys_err_chk (retval, "PTRACE_ATTACH failed");
		if (!IS_SYS_ERROR (retval)) {
			/* wait for the target thread to actually stop */
			do {
				DBPRINT (btl_prt_thread | btl_prt_pt,
						"before-wait: pid=%d options=%x\n",
						susp_thread->tid, WUNTRACED | __WALL);
				retval = SYSCALL (wait4, (susp_thread->tid, &status, 
							  WUNTRACED | __WALL, NULL));
				DBPRINT (btl_prt_thread | btl_prt_pt,
						"after-wait: ret=%ld status=%d\n", retval, status);
			} while (retval == -EINTR);
			sys_err_chk (retval, "waitpid");
			assert ((int) retval == susp_thread->tid);

			if (__WIFSTOPPED (status)) {
				/* finally, the target thread is suspended, 
				   mark it as such */
				susp_thread->is_suspended = 1;
			}
		} /* else */
	} /* for */

	btlib_release_lock (BTL_LOCK_IND_THREAD);

	return 0;
}

/**************************************************************************
 This function will try to resume all threads
 that were suspended in the "suspend_all_threads" function.

Return Value:
 Success: 0 (all threads that were suspended were successfully resumed)
 Failure: -1
**************************************************************************/
long
resume_all_threads (void)
{
	long retval;
	btlib_thread_data_t *susp_thread;

	/* go over the existing threads list, look for suspended threads 
	   and try to resume them */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	for (susp_thread = m_used_thread_list; susp_thread != NULL;
		susp_thread = susp_thread->next) {
		if (susp_thread->is_suspended) {
			/* ok, found a suspended thread */

			/* make sure a thread will not resume itself */
			assert (susp_thread->tid != cur_thread_data->tid);

			retval = SYSCALL (ptrace, (PTRACE_DETACH, 
						   susp_thread->tid, 0ul, 0ul));
			sys_err_chk (retval, "PTRACE_DETACH failed");

			/* mark that the thread is no longer suspended */
			susp_thread->is_suspended = 0;
		}
	} /* for */

	btlib_release_lock (BTL_LOCK_IND_THREAD);

	return 0;
}

/**************************************************************************/
/* 
 * suspends thread's execution in a consistent state
 * input:  tid - the thread to suspend
 *         max_count - max tries allowed
 * output: 0 on success ( thread suspended in a consistent state),
 *         -1 otherwise.
 */
/**************************************************************************/
long
btlib_suspend_thread (int tid, unsigned max_count)
{
	long retval;
	int status;
	btlib_thread_data_t *susp_thread;
	debugee_t debugee;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));
	DBPRINT (btl_prt_thread, "tid=%d, max_count=%d\n", tid, max_count);

	/* a thread should not suspend itself */
	assert (tid != cur_thread_data->tid);

	/* find the thread's data struct of the target thread */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);

	susp_thread = m_used_thread_list;
	while (susp_thread != NULL) {
		if (susp_thread->tid == tid) {
			break;
		}
		susp_thread = susp_thread->next;
	}

	btlib_release_lock (BTL_LOCK_IND_THREAD);
	assert (susp_thread != NULL);
	assert (susp_thread->tid == tid);

	/* attach to target thread and suspend it in the 64 bit level */
	retval = SYSCALL (ptrace, (PTRACE_ATTACH, tid, 0ul, 0ul));
	if (IS_SYS_ERROR (retval)) {
		DBPRINT (btl_prt_thread, "attach failed=%ld\n", retval);
		return -1;
	}
	do {
		DBPRINT (btl_prt_thread | btl_prt_pt,
				"before-wait: pid=%d options=%x\n",
				tid, WUNTRACED | __WALL);
		retval = SYSCALL (wait4, (tid, &status, WUNTRACED | __WALL, NULL));
		DBPRINT (btl_prt_thread | btl_prt_pt, "after-wait: ret=%ld status=%d\n",
				retval, status);
	} while (retval== -EINTR);
	sys_err_chk (retval, "waitpid");
	assert ((int) retval == tid);
	if (!__WIFSTOPPED (status)) {
		DBPRINT (btl_prt_thread, "target thread terminated\n");
		return -1;
	}

	/* 
	 * use the thread suspension mechanism of ptrace emulation
	 * to stop the thread in a consistent IA-32 state
	 */
	debugee.pid = susp_thread->tid;
	debugee.blst = (unsigned long) susp_thread;
	retval = btlib_stop_debugee (&debugee, max_count, 1, &status);
	/* if stopping the debugee is not possible then let the debugee go */
	if (retval != 0) {
		retval = SYSCALL (ptrace, (PTRACE_DETACH, tid, 0ul, 0ul));
		sys_err_chk (retval, "ptrace-detach");
		retval = -1;
	}

	DBPRINT (btl_prt_thread, "retval=%ld\n", retval);
	return retval;
}

/**************************************************************************/
/* resume execution of the given thread
 * input:  tid - the thread to resume
 * output: 0 on success, non zero value otherwise
 */
/**************************************************************************/
long
btlib_resume_thread (int tid)
{
	long retval;

	DBPRINT (btl_prt_thread, "tid=%d\n", tid);
	/* a thread should not resume itself */
	assert (tid != SYSCALL (gettid, ()));

	retval = SYSCALL (ptrace, (PTRACE_DETACH, tid, 0ul, 0ul));
	DBPRINT (IS_SYS_ERROR (retval), "retval=%ld\n", retval);
	return retval;
}

/**************************************************************************/
/* set the signal hanlders according to the inherited handlers.
 * input:   thread -
 * output:  none
 * remarks: after an execve - a process inherits from its parent the 
 *          SIG_IGN dispositions and the signal mask.
 */
/**************************************************************************/
static void
set_sig_inherit ()
{
	int j;
	long retval;
	struct sigaction act;

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	DBPRINT (0, "\n");
	/* change all the inherited SIG_IGN signals */
	for (j = 1; j < _NSIG; j++) {
		retval = SYSCALL (rt_sigaction, (j, NULL, &act, sizeof (k_sigset_t)));
		sys_err_chk (retval, "rt_sigaction");
		if (act.sa_handler == SIG_IGN) {
			DBPRINT (0, "ignored %d\n", j);
			cur_thread_data->sighandlers->handlers[j].sa.handler =
				(unsigned) SIG_IGN;
		}
	}
}

/*****************************************************************************/
/* init thread data */
/*****************************************************************************/
static void
init_thread_data (btlib_thread_desc *td,
                  btlib_thread_data_t *blst,
                  void *glst, 
                  void *stackblock, 
                  unsigned long user_stack_size)
{
	DBPRINT (0, "glst=%p size=%x\n", glst, btl_glst_size);

	/* init some of the thread's fields */
	blst->__r13 = (unsigned long)(td + 1);
	blst->stackblock = stackblock;
	blst->user_stack_size = user_stack_size;
	blst->syscall_id = IA32_SYSCALL_NONE;
	blst->bc32 = (BTGENERIC_IA32_CONTEXT *) ((uintptr_t) glst + btl_bc32);

	/* init the thread desc */
	td->glst = glst;
	td->blst = blst;
	DBPRINT (btl_prt_thread, "td=%p glst=%p blst=%p bc32=%p\n",
		 td, td->glst, td->blst, cur_thread_data->bc32);
}

/*****************************************************************************/
/*****************************************************************************/
static void
list_add (btlib_thread_data_t *elem, btlib_thread_data_t **list)
{
	elem->next = *list;
	elem->prev = NULL;

	if (*list != NULL) {
		(*list)->prev = elem;
	}

	*list = elem;
}

/*****************************************************************************/
/*****************************************************************************/
static void
list_del (btlib_thread_data_t *elem, btlib_thread_data_t **list)
{
	if (*list == elem) {
		/* elem is head */
		*list = elem->next;
	}

	if (elem->next != NULL) {
		elem->next->prev = elem->prev;
	}
	if (elem->prev != NULL) {
		elem->prev->next = elem->next;
	}
}

/*****************************************************************************/
/* 
 * allocate memory for 64 bit thread stack, btlib thread data, and GLST. 
 * initialize btlib thread data
 */
/*****************************************************************************/
static void *
allocate_thread_data (btlib_thread_desc **tdp)
{
	btlib_thread_data_t *blst;
	void *glst;
	void *stackblock, *bos;	/* top and bottom of 64 bit stack */
	void *td;
	int ret;
	size_t stack_size;

	/* 
	 * allocate the stack - both the thread_data and glst
	 * will reside on the memory chunk allocated here
	 */

	ret = btlib_allocate_stack (&stackblock, &stack_size, &td);

	if (IS_SYS_ERROR (ret)) {
		DBPRINT (btl_prt_thread, "btlib_allocate_stack failed=%d\n", ret);
		return (void *) (uintptr_t) ret;
	}
	assert ((uintptr_t) td == ROUNDUP ((uintptr_t) td, sizeof (void *)));

	bos = stackblock + stack_size - 1;

	/* allocate thread_data */
	bos -= sizeof (btlib_thread_data_t);
	ALIGN_STACK (bos);
	blst = (btlib_thread_data_t *) bos;

	memset (blst, 0, sizeof (btlib_thread_data_t));
	DBPRINT (0, "thread_data at %p\n", blst);

	/* allocate glst */
	bos -= btl_glst_size;
	ALIGN_STACK (bos);
	glst = bos;

	init_thread_data (td, blst, glst, stackblock, 
			  (unsigned long) (glst - stackblock));

	DBPRINT (0, "returns	%p\n", bos);

	*tdp = td;
	return bos;
}

/**************************************************************************/
static void
before_thread_exit ()
{
	/* 
	 * if this is the last thread then call this api function
	 * which prepares stats on IA32-EL run
	 */
	if ((cur_thread_data->next == NULL) && (cur_thread_data->prev == NULL)) {
		STATCODE (btl_stats, btlib_print_stats ());
		BTGENERIC_NOTIFY_EXIT ();
	}

	/* notify btgeneric */
	BTGENERIC_THREAD_TERMINATED ();

	/* mark as exited */
	cur_thread_data->is_exited = 1;

	/* last print to log */
	DBPRINT (btl_prt_thread, "bye\n");

	/* remove current thread from used thread list */
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	list_del (cur_thread_data, &m_used_thread_list);
#if ALL_THREAD_AFFINITY
	btl_num_threads--;
#endif
	btlib_release_lock (BTL_LOCK_IND_THREAD);

	/* signal handler table and pending que will never be used, but thread 
           stack should not be freed */
	cleanup_thread (cur_thread_data, FALSE);
}

#if DEBUG
/************************************************************************** 
* print the thread list starting from the current thread
**************************************************************************/
static void
print_thread_list ()
{
	btlib_thread_data_t *p;
	char list[4096];
	char th[32];

	assert (cur_thread_data->tid == SYSCALL (gettid, ()));

	sprintf (list, "(%d) thread list = (%d)", cur_thread_data->tid,
		 cur_thread_data->tid);
	btlib_acquire_lock (BTL_LOCK_IND_THREAD);
	for (p = m_used_thread_list; p != NULL; p = p->next) {
		sprintf (th, "->(%d)", p->tid);
		strncat (list, th, MIN(sizeof(list) - 1 - strlen(list), strlen(th)));
	}
	btlib_release_lock (BTL_LOCK_IND_THREAD);
	strncat (list, "\n", MIN(sizeof(list) - 1 - strlen(list), 2));
	fprintf (g_log_file, list);
}
#endif	// DEBUG


#if ALL_THREAD_AFFINITY

#if THREAD_SCALABILIY_WITH_IDLE_TIME
#define UPTIME_FILE "/proc/uptime"
#define UPTIME_STR_LENGTH 512ul
/* 
get_cur_num_secs_idle

read the current number of idle secs from /proc/uptime
*/
static unsigned long get_cur_num_secs_idle ()
{
	int uptime_file;
	char uptime_str[UPTIME_STR_LENGTH];
	int  retval;
	unsigned long idle_time;

	/* read from /proc/uptime file, to get the current idle time (in seconds) */
	uptime_file = SYSCALL (open, (UPTIME_FILE, O_RDONLY, 0));
	if (IS_SYS_ERROR (uptime_file)) {
		DBPRINT (btl_prt_thread, "Error: open of %s failed: %d\n", 
			 UPTIME_FILE, uptime_file);
		return 0;
	} else {
		idle_time = 0;
		
		retval = SYSCALL (read, (uptime_file, uptime_str, UPTIME_STR_LENGTH));
		if (IS_SYS_ERROR (retval)) {
			DBPRINT (btl_prt_thread,
				 "Error: failed to read from %s: %d\n", 
				 UPTIME_FILE, retval);
		} else {
			char * p = strchr (uptime_str, ' '), *start_idle_time_str;
			if (p) {
				start_idle_time_str = p+1;
				p = strchr (p, '.');
				if (p) {
					*p = '\0';
					idle_time = atol (start_idle_time_str);
				}
			}
		}
		SYSCALL (close, (uptime_file));
	}

	return idle_time;
}
#endif


/*
    end_thread_scalability_interval

    End a thread scalability measurement interval, and start a new measurement 
    interval at the appropriate point
*/
static void end_thread_scalability_interval (unsigned long cur_timestamp_cntr)
{
	if (m_mthread_sim_tm_strt &&
	    (cur_timestamp_cntr - m_mthread_sim_tm_strt) < (btl_cyc_in_thrd_scl_intvl)) { 
		/* if the last recorded multi-threaded start occured not more than an interval ago
		   move interval_start_time to be the last m_mthread_sim_tm_strt
		   in order to include that multi-threaded time in the new interval.
		   note that in this case some of the idle time may not be recorded, but this is
		   the proces that is paid in order not to obtain the idle time at each 
		   multi-thread start */
		m_interval_start_tm = m_mthread_sim_tm_strt;
#if THREAD_SCALABILIY_WITH_IDLE_TIME
		m_sec_idle_at_istrt = get_cur_num_secs_idle ();
#endif
	}
	else { /* will restart interval next time a multi threaded interval starts */
		m_interval_start_tm = 0;
	}
	m_total_mthread_dur = 0;
}



/*
    btlib_record_thread_enter_simulation

    Maintain count of how many threads are in simulation session. When number 
    becomes 2, this is the start of a multi-threaded interval, record the 
    timestamp of the start of the multi-threaded interval. These are used to
    compute whether the application suffers from thread scalaibility 
    (see RecordThreadExitSimulation)
*/
void btlib_record_thread_enter_simulation ()
{
   unsigned long cur_timestamp_cntr;

	btlib_acquire_lock (BTL_LOCK_IND_THREAD_SCALABILITY);
	m_num_threads_in_sim++;
	
	if (m_num_threads_in_sim == 2) { /* start of multi thread simulation */
		cur_timestamp_cntr = __btlib_read_timestamp_counter ();
		DBPRINT (btl_prt_thread, "multi-thread start at elapsed %d interval %d\n",
			 (unsigned)((cur_timestamp_cntr - btl_start_time) / 
				CYCLES_IN_MILI_SECOND),
			 (unsigned)((cur_timestamp_cntr - m_interval_start_tm) / 
				CYCLES_IN_MILI_SECOND));
					 
		
		if (cur_timestamp_cntr - m_interval_start_tm >= (btl_cyc_in_thrd_scl_intvl)) {
			/* since so much time has passed since last multi-threaded, 
			   start new interval */
			m_total_mthread_dur = 0;
			m_interval_start_tm = cur_timestamp_cntr;
#if THREAD_SCALABILIY_WITH_IDLE_TIME
			/* note that idle time is only obtained at each interval start */
			m_sec_idle_at_istrt = get_cur_num_secs_idle ();
#endif
		}
		m_mthread_sim_tm_strt = cur_timestamp_cntr;	   
	}
	btlib_release_lock (BTL_LOCK_IND_THREAD_SCALABILITY);
}


/*
    btlib_record_thread_exit_simulation

    Maintain count of how many threads are in simulation session. 
    When there are still more than 1 thread in a simulation session, this is an 
    extension of a multi-threaded interval, a check is done to see if the  time 
    in a multi-threaded interval has execeeded the threshold for considering the
    processs to suffer from thread scalability (currently ~5sec out of a ~10sec 
    interval) 
    When number of threads becomes 1, this is the end of a multi-threaded interval
*/


void btlib_record_thread_exit_simulation ()
{
	unsigned long cur_mthread_dur;
	unsigned long cur_timestamp_cntr;
	unsigned long released_lock = 0;
	

	btlib_acquire_lock (BTL_LOCK_IND_THREAD_SCALABILITY);
	m_num_threads_in_sim--;

	if (m_num_threads_in_sim >= 1 && m_mthread_sim_tm_strt != 0) {
		cur_timestamp_cntr = __btlib_read_timestamp_counter ();
		cur_mthread_dur = cur_timestamp_cntr - m_mthread_sim_tm_strt;

		if ((m_try_for_ordered_code == 0) &&
		    ((cur_mthread_dur + m_total_mthread_dur)
		     * INVERSE_FRACTION_OF_INTERVAL_FOR_THREAD_SCALABILITY 
			 >= btl_cyc_in_thrd_scl_intvl)) {
			/* multi-threaded threshold execeeded */
			EMULATION_POLICY emulation_policy;
#if THREAD_SCALABILIY_WITH_IDLE_TIME
			if (m_sec_idle_at_istrt == 0 ||
			    ((get_cur_num_secs_idle() - m_sec_idle_at_istrt)
				   * INVERSE_FRACTION_OF_IDLETIME_FOR_THREAD_SCALABILITY
				   >= NUM_SECS_IN_THREAD_SCALABILITY_INTERVAL)) {
				/* there is enough idle percentage to warrant the price of 
				   ordered code to get thread scalability. 
				   (0 ==  m_sec_idle_at_istrt means the idle time
				  cannot be obtained, so it is assumed there is enough) */
				
#endif

				/* critical section must be released before calling 
				   BTGENERIC_SET_EMULATION_POLICY */
				m_try_for_ordered_code = 1;
				btlib_release_lock (BTL_LOCK_IND_THREAD_SCALABILITY);
				released_lock = 1;
			
				assert (! __CPU_IS_ZERO (&m_orig_process_affinity_mask));
#if THREAD_SCALABILIY_WITH_IDLE_TIME
				DBPRINT (btl_prt_thread, "idle %u\n", 
					 (unsigned)(get_cur_num_secs_idle() - m_sec_idle_at_istrt)
					 	* 1000);
#endif
				emulation_policy.memory_order_policy = USE_ORDERED_LD_ST;
				emulation_policy.strict_state_policy = POLICY_UNCHANGED;

				BTGENERIC_SET_EMULATION_POLICY(&emulation_policy);
				assert (emulation_policy.memory_order_policy == 
					USE_ORDERED_LD_ST);
				if (emulation_policy.memory_order_policy == USE_ORDERED_LD_ST) {
					/* give all the threads (that didn't explicitly change 
					   their affinity) the original process affinity */
					set_affinity_to_all_threads (&m_orig_process_affinity_mask);
					btl_narrowed_affinity = 0;
					btl_did_ordered_code = 1;
		
				} else { 
					/* somehow couldn't switch to ordered code, 
					   restart interval measurement */
					end_thread_scalability_interval (cur_timestamp_cntr);
					m_try_for_ordered_code = 0;
				}
#if THREAD_SCALABILIY_WITH_IDLE_TIME
			} else { /*not enough idle time - discard all info about this interval */
				m_mthread_sim_tm_strt = 0;
				end_thread_scalability_interval (cur_timestamp_cntr);
			}
#endif
		} else {
			if ((cur_timestamp_cntr - m_interval_start_tm) 
				  >= (btl_cyc_in_thrd_scl_intvl)) {
				/* the time interval for measurement has elapsed */
				DBCODE (btl_prt_thread, 
					m_total_mthread_dur += cur_mthread_dur);
				DBPRINT (btl_prt_thread, 
					 "ending interval at elapsed %u interval %u (total %u)\n",
					 (unsigned)((cur_timestamp_cntr - btl_start_time) / 
						CYCLES_IN_MILI_SECOND),
					 (unsigned)((cur_timestamp_cntr - m_interval_start_tm) /
						CYCLES_IN_MILI_SECOND),
					 (unsigned)(m_total_mthread_dur / CYCLES_IN_MILI_SECOND));
				end_thread_scalability_interval (cur_timestamp_cntr);
			} else if (m_num_threads_in_sim == 1) {
				/* entering single thread mode*/
				m_total_mthread_dur += cur_mthread_dur;
				/* will restart multi thread duration next time there are more 
				   than 1 threads entered simulation */
				m_mthread_sim_tm_strt = 0;
			}
		}
	} 
	if (released_lock == 0) {
		btlib_release_lock (BTL_LOCK_IND_THREAD_SCALABILITY);
	}
	assert(((long)m_num_threads_in_sim) >= 0);
	
}
#endif




