/* 
 * btlib_tls.c
 *
 * allocate tls. this module is taken almost as-is from the sources of nptl
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include <ia32x_utils.h>

#include "btlib_sync.h"
#include "btlib_stats.h"

/* Default stack size.  */
#define ARCH_STACK_DEFAULT_SIZE	(32 * 1024 * 1024)
/* Required stack pointer alignment at beginning.  */
#define STACK_ALIGN		16

/* Minimum size for a thread.  We are free to choose a reasonable value.  */
#define THREAD_STACK_MIN	(1024ul * 1024ul)

/* this is how we find the tcb when we have the address of the thread desc */
#define TLS_TPADJ(pd) ((void *)pd + TLS_PRE_TCB_SIZE)

/* this is how we find the tcb when we have the address of the stackblock */
#define TLS_OFFSET_FROM_STACKBLOCK(stackblock, size) \
		(void *) ((((uintptr_t) (stackblock) + (size) - __static_tls_size) \
						& ~__static_tls_align_m1) 		\
						- TLS_PRE_TCB_SIZE);

/* Default stack size.  */
size_t __default_stack_size;
#define STACK_SIZE __default_stack_size

/* Size and alignment of static TLS block.  */
static size_t __static_tls_size;
static size_t __static_tls_align_m1;

/*****************************************************************************/
/*****************************************************************************/
int
btlib_init_tls_module ()
{
	struct rlimit limit;

	/* this is the stack size needed for btlib */
	STACK_SIZE = THREAD_STACK_MIN;

	/* Determine the default allowed stack size. */
	if (SYSCALL (getrlimit, (RLIMIT_STACK, &limit)) != 0 || 
	    limit.rlim_cur == RLIM_INFINITY) {
		/* The system limit is not usable. Use an architecture-specific
		   default. */
		limit.rlim_cur = ARCH_STACK_DEFAULT_SIZE;
	}

	if (limit.rlim_cur < THREAD_STACK_MIN) {
		DBPRINT (TRUE, "limit.rlim_cur=%ld < THREAD_STACK_MIN=%ld\n",
		         limit.rlim_cur, THREAD_STACK_MIN);
		btlib_abort_all ();
	}

	/* Get the size of the static and alignment requirements for the TLS
	   block.  */
	size_t static_tls_align;
	get_tls_static_info (&__static_tls_size, &static_tls_align);

	/* Make sure the size takes all the alignments into account.  */
	if (STACK_ALIGN > static_tls_align) {
		static_tls_align = STACK_ALIGN;
	}
	__static_tls_align_m1 = static_tls_align - 1;

	__static_tls_size = ROUNDUP (__static_tls_size, static_tls_align);

	DBPRINT (0, "__static_tls_size=%lx __static_tls_align_m1=%lx\n",
	       __static_tls_size, __static_tls_align_m1);
	return 0;
}

/******************************************************************************
 * allocate stack
 * after creation the stack layout is - 
 *
 *        -----------------------------------------------------------------
 *        |
 *        |		TLS AREA INITIALIZED BY LD-LINUX (__static_tls_size)
 *        |
 *        |  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  
 *        |		TCB
 *        -------------------------------------------------------------------
 *        |		BTLIB'S THREAD DESCRIPTOR 
 * tls -> ----------------- stacktop  ---------------------------------------
 *        |
 *        |
 *        |
 *        |
 *        |
 *        |
 *        |
 *        |		USER STACK (*stacksize)
 *        |
 *        |
 *        |
 *        |
 *        |
 *        |
 *        |
 *        |
 *        |
 *        ----------------- *stack -------------------------------------------
 *   
 *****************************************************************************/
int
btlib_allocate_stack (void **stack, size_t *stacksize, void **tls)
{
	void *stackblock = NULL;
	void *stacktop = NULL;
	void *td = NULL;
	size_t size = STACK_SIZE;
	long retval;

	/* get a free stack */
	stackblock = btlib_mem_alloc (INITIAL_BTLIB_ADDRESS,
				      size, PROT_READ | PROT_WRITE | PROT_EXEC);
	if (IS_SYS_ERROR ((long) stackblock)) {
		return (int) (long) stackblock;
	}

	td = TLS_OFFSET_FROM_STACKBLOCK (stackblock, size);

	/* Allocate the DTV for this thread. */
	DBPRINT (0, "TLS_TPADJ(td)=%p\n", TLS_TPADJ (td));
	retval = allocate_and_init_dtv (TLS_TPADJ (td));
	if (IS_SYS_ERROR (retval)) {
		/* Something went wrong */
		/* Free the stack memory we just allocated.     */
		(void) SYSCALL (munmap, ((unsigned long)stackblock, size));
		return -ENOMEM;
	}

	stacktop = (char *) (td - 1);
	*stack = stackblock;
	*stacksize = stacktop - *stack;
	*tls = td;
	DBPRINT (0, "td=%p stackblock=%p stacksize=%lx stacktop=%p\n",
	       td, stackblock, *stacksize, stacktop);
	return 0;
}

/******************************************************************************/
/******************************************************************************/
int
btlib_deallocate_stack (void *stackblock, UNUSED int *clear_tid, int remove_stack)
{
	size_t size = STACK_SIZE;
	void *td = NULL;
	int ret;

	td = TLS_OFFSET_FROM_STACKBLOCK (stackblock, size);

	STATCODE (btl_stats, btlib_update_stat (BTL_STAT_STACK_FREE, COUNT, 0ul));
	DBPRINT (btl_prt_thread, "td=%p tid=%d\n",
	       td, ((btlib_thread_desc *) td)->blst->tid);

	/* Get rid of the TLS block we allocated.  */
	deallocate_dtv (TLS_TPADJ (td));

	/* remove the stack */
	if (remove_stack) {
		ret = btlib_mem_free (stackblock, size);
		DBPRINT (IS_SYS_ERROR (ret), "munmap failed\n");
	}
	return 0;
}

