/* 
 * btlib_uaccess.c
 *
 * check memory permissions and provide safe read/write
 *
 */
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "btlib.h"
#include "btlib_asm.h"
#include <ia32x_utils.h>
/*****************************************************************************/
/* call memcpy but don't let SIGSEGV and SIGBUS crash the process */
/* return: 0 on success, -EFAULT on failure
 */
/*****************************************************************************/
long
safe_memcpy (void *dest, void *src, unsigned long len)
{
	long ret;
	long status;

	DBPRINT (btl_prt_pt | btl_prt_mem, "dest=%p src=%p len=%lx\n",
	       dest, src, len);

	/* mark the thread as using the special signal handler */
	cur_thread_data->ignore_exception = 1;

	/* do the copy */
	status = sigsetjmp (cur_thread_data->ignore_exception_jmpbuf, 1);
	if (status == 0) {
		memcpy (dest, src, len);
		ret = 0;
	} else {
		ret = -EFAULT;
	}

	DBPRINT (btl_prt_pt | btl_prt_mem, "ret=%ld\n", ret);

	/* stop using the special signal handler */
	cur_thread_data->ignore_exception = 0;

	return ret;
}
/*****************************************************************************/
/* strncpy for un-trusted source. destination is trusted.
 * returns - in case of success: number of bytes that were copied
 *		on failure - negative error code
 * note - the function does NOT copy the null char at the end of the string
 *	even when the given length allows it */
/*****************************************************************************/
long safe_stringcopy (char * dest, char * src, unsigned long len)
{
	long ret;
	char dummy;
	/* get the size of the src */
	char * p = src;
	unsigned long local_buf_size = len;
	int cont_check = 1;
	unsigned long size = 0;
	while (cont_check) {
		unsigned long safe_size, size_in_cur_page, next_page, cur_size;
		ret = safe_memcpy (&dummy, p, 1ul);
		if (IS_SYS_ERROR (ret)) {
			goto out;
		} else {
			next_page = PAGEALIGN ((unsigned long)p + 1, PAGE_SIZE);
			size_in_cur_page = next_page - (unsigned long)p;
			safe_size = MIN (local_buf_size, size_in_cur_page);
			cur_size = strnlen (p, safe_size);
			size += cur_size;

			if (cur_size < safe_size) {
				cont_check = 0;
			} else {
				if (safe_size == local_buf_size) {
					cont_check = 0;
				} else {
					/* need to read from next page too */
					local_buf_size -= size_in_cur_page;
					assert ((long)local_buf_size > 0);
					p = (char *)next_page;
				}
			}
		}
	}
	DBPRINT (0, "size=%lu\n", size);
	assert (size <= len);
	ret = safe_memcpy (dest, src, size);
	if ( ! IS_SYS_ERROR (ret) ) {
		ret = size;
	}
out:
	return ret;
}

