/*
 * linux/fs/binfmt_elf.c
 *
 * These are the functions used to load ELF format executables as used
 * on SVr4 machines.	Information on the format may be found in the book
 * "UNIX SYSTEM V RELEASE 4 Programmers Guide: Ansi C and Programming Support
 * Tools".
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "../btlib.h"
#undef _GNU_SOURCE
#define _GNU_SOURCE
#include <ia32x_utils.h>
#include <ia32toia64.h>


#ifndef IA32_SYSCALL_MMAP
#define IA32_SYSCALL_MMAP	90
#endif				/*	*/

#ifndef IA32_SYSCALL_MMAP2
#define IA32_SYSCALL_MMAP2	192
#endif				/*	*/

#ifndef IA32_SYSCALL_MPROTECT
#define IA32_SYSCALL_MPROTECT	125
#endif				/*	*/

#ifndef IA32_TASK_SIZE
#define IA32_TASK_SIZE		IA32_MAX_USER_ADDR
#endif				/*	*/

#ifndef IA32_PAGE_SIZE
#define IA32_PAGE_SIZE		0x1000ul
#endif				/*	*/

#ifndef IA32_STACK_TOP
#define IA32_STACK_TOP		IA32_TASK_SIZE
#endif				/*	*/

#ifndef MAX_ARG_PAGES
#define MAX_ARG_PAGES		32
#endif				/*	*/

#ifndef INITIAL_IA32_STACK_SIZE
#define INITIAL_IA32_STACK_SIZE	(MAX_ARG_PAGES*IA32_PAGE_SIZE)
#endif				/*	*/

#ifndef IA32_CLOCKS_PER_SEC
#define IA32_CLOCKS_PER_SEC	100
#endif				/*	*/

#ifndef ELF_EXEC_PAGE_SIZE
#define ELF_EXEC_PAGE_SIZE	IA32_PAGE_SIZE
#endif				/*	*/

#ifndef ELF_MAP_BASE_ADDR
#define ELF_MAP_BASE_ADDR	(IA32_TASK_SIZE / 3)
#endif				/*	*/

#ifndef ELF_HWCAP
#define ELF_HWCAP		0
#endif				/*	*/

#ifndef ELF_PLATFORM
#define ELF_PLATFORM		NULL
#endif				/*	*/

#define MANDATORY_AT_NUM 14
#define NUMBER_OF_GROUPS 64
 
int g_dont_core_suid_own = 0;
extern char * BTGeneric_search_path[];
enum {LEB_FILE_NAME, LEB_FILE_DESC};

#if SUID_BTLIB
static long compute_creds (int bin_name);
#else
#define SUID_LIB_NAME	"suid_libia32x.so"
static int invoke_suid_btlib_if_needed (char * file_name, char ** argv, char ** envp);
#endif
static long process_belongs_to_group (gid_t group);
static long check_exec_permissions (char *file, int file_name_or_fd);
static inline void *
ia32_mmap (void *start, size_t length, int prot, int flags, int fd, off_t offset)
{
	legacy_mmap_args_t args = {
		(unsigned) (unsigned long) start,
		(unsigned) length,
		(unsigned) prot,
		(unsigned) flags, (unsigned) fd, (unsigned) offset
	};
	return (void *) ILP32_WRAPPER1 (IA32_SYSCALL_MMAP, (unsigned long) &args);
}

static inline int
ia32_mprotect (void *addr, size_t len, int prot)
{
	return ILP32_WRAPPER3 (IA32_SYSCALL_MPROTECT, (unsigned long) addr, len, prot);
}

static long
check_ia32_elf_file (const char *file, int file_name_or_fd, int interp, int *fdp, 
                     Elf32_Ehdr ** ehdrp, Elf32_Phdr ** phdrp)
{
	int fd;
	Elf32_Ehdr *ehdr;
	Elf32_Phdr *phdr;
	const char *errstring = NULL;
	ssize_t ehdr_size = sizeof(Elf32_Ehdr);
	ssize_t phdr_size;
	
	DBPRINT (0, "file='%s' file_name_or_fd=%d\n", file, file_name_or_fd);

	/* Open the file. */
	if (file_name_or_fd == LEB_FILE_NAME) {
		fd = interp ? 
			ILP32_WRAPPER3 (IA32_SYSCALL_OPEN, file, O_RDONLY, 0) :
			SYSCALL (open, (file, O_RDONLY, 0));
		if (IS_SYS_ERROR (fd)) {
			DBPRINT (TRUE, "open failed: %d\n", fd);
			return fd;
		}
	} else {
		fd = atol (file);
	}

	ehdr = (Elf32_Ehdr *)malloc (sizeof(Elf32_Ehdr));
	if (SYSCALL (pread, (fd, ehdr, sizeof(Elf32_Ehdr), (off_t)0) != ehdr_size)) {
		errstring = "Cannot read ELF file";
		goto failed;
	}
	if (ehdr == NULL
		|| ehdr->e_machine != EM_386
		|| memcmp (&ehdr->e_ident[EI_MAG0], ELFMAG, (size_t)SELFMAG) != 0) {
		errstring = "Wrong ELF file";
		goto failed;
	}
	if (ehdr->e_type != ET_DYN && (interp || ehdr->e_type != ET_EXEC)) {
		errstring = "Not an ELF executable file";
		goto failed;
	}

	phdr = (Elf32_Phdr *)malloc (sizeof(Elf32_Phdr) * ehdr->e_phnum);
	phdr_size = sizeof(Elf32_Phdr) * ehdr->e_phnum;
	if (SYSCALL (pread, (fd, phdr, sizeof(Elf32_Phdr) * ehdr->e_phnum, 
			     (off_t)ehdr->e_phoff) != phdr_size)) {
		errstring = "Cannot read ELF file";
		goto failed;
	}	
	if (phdr == NULL) {
		errstring = "Cannot get ELF program header";
		goto failed;
	}
	
	*fdp = fd;
	*ehdrp = ehdr;
	*phdrp = phdr;
	return 0;

failed:
	DBPRINT (TRUE, "%s: %s\n", errstring, file);
	SYSCALL (close, (fd));
	return -ENOEXEC;
}

#ifndef MAP_COPY
# define MAP_COPY	MAP_PRIVATE
#endif		/*	*/
static int
map_elf32_binary (int fd, Elf32_Ehdr * ehdr, Elf32_Phdr * phdr, int interp,
                  Elf32_Addr * l_addrp, Elf32_Addr * l_phdrp,
                  btrans_binary_map_t * app_info, long personality)
{
	const Elf32_Phdr *ph;
	unsigned long l_map_end, l_map_start, l_addr, l_phdr;
	size_t maplength, filelength = 0;
	const char *errstring = NULL;
	long errcode;

	/* Scan the program header table, collecting its load commands. */
	struct loadcmd {
		const Elf32_Phdr *phdr;
		Elf32_Addr mapstart, mapend, dataend, allocend;
		Elf32_Off mapoff;
		int prot;
	} loadcmds[ehdr->e_phnum], *c;
	size_t nloadcmds = 0;
	int has_holes = 0;

	l_addr = 0;
	l_phdr = 0;
	for (ph = phdr; ph < &phdr[ehdr->e_phnum]; ++ph) {
		switch (ph->p_type) {
		case PT_PHDR:
			l_phdr = ph->p_vaddr;
			break;
		case PT_LOAD:
			/* A load command tells us to map in part of the file.
			   We record the load commands and process them all later. */
			if ((ph->p_align & (ELF_EXEC_PAGE_SIZE - 1)) != 0) {
				errstring = "ELF load command alignment not page-aligned";
			failed:
				DBPRINT (TRUE, "%s\n", errstring);
				return -1;
			}
			if (((ph->p_vaddr - ph->p_offset) & (ph->p_align - 1)) != 0) {
				errstring = "ELF load command address/offset not properly aligned";
				goto failed;
			}
			c = &loadcmds[nloadcmds++];
			c->mapstart = ph->p_vaddr & ~(ph->p_align - 1);
			c->mapend = ((ph->p_vaddr + ph->p_filesz +
					ELF_EXEC_PAGE_SIZE - 1) & ~(ELF_EXEC_PAGE_SIZE - 1));
			c->dataend = ph->p_vaddr + ph->p_filesz;
			c->allocend = ph->p_vaddr + ph->p_memsz;
			c->mapoff = ph->p_offset & ~(ph->p_align - 1);

			/* Determine whether there is a gap between the last segment
			   and this one. */
			if (nloadcmds > 1 && c[-1].mapend != c->mapstart) {
				has_holes = 1;
			}
			c->prot = 0;
			if (ph->p_flags & PF_R) {
				c->prot |= PROT_READ;
				/* 
				 * we will call ia32_mmap() below, but personality in current 
				 * thread structure has not been initialized in this phase, 
				 * so we have to judge personality here.
				 */
				if (RX_PERSON(personality)) {
					c->prot |= PROT_EXEC;
				}
			}
			if (ph->p_flags & PF_W) {
				c->prot |= PROT_WRITE;
			}
			if (ph->p_flags & PF_X) {
				c->prot |= PROT_EXEC;
			}
			c->phdr = ph;
			filelength = ph->p_offset + ph->p_filesz;
			break;
		default:
			break;
			
		}
	}
	if (nloadcmds == 0) {
		/* This only happens for a bogus object that will be caught with
		   another error below. But we don't want to go through the
		   calculations below using NLOADCMDS - 1. */
		errstring = "object file has no loadable segments";
		goto failed;
	}

	/* Now process the load commands and map segments into memory. */
	c = loadcmds;

	/* Length of the sections to be loaded. */
	maplength = loadcmds[nloadcmds - 1].allocend - c->mapstart;
	if (ehdr->e_type == ET_DYN) {

		/* Remember which part of the address space this object uses. */

		/* Double the base address if the dynamic linker is run as
		   an executable. */
		unsigned long base = (interp ? 
			ELF_MAP_BASE_ADDR : (2 * ELF_MAP_BASE_ADDR));
		l_map_start = (unsigned long) ia32_mmap ((void *) base,
							 maplength, c->prot,
							 MAP_COPY | MAP_FILE,
							 fd, (off_t)c->mapoff);
		if (IS_SYS_ERROR ((void *) l_map_start)) {
			errcode = (long)l_map_start;
map_error:
			errstring = "failed to map segment from shared object";
failed_errno:
			DBPRINT (TRUE, "%s (%ld)\n", errstring, errcode);
			return -1;
		}
		if (l_map_start > 0xffffffffUL) {
			errstring = "failed to map segment into 32 bit address";
			goto failed;
		}
		l_map_end = l_map_start + maplength;
		l_addr = l_map_start - c->mapstart;
		if (has_holes) {
			/* Change protection on the excess portion to disallow all access;
			   the portions we do not remap later will be inaccessible as if
			   unallocated. Then jump into the normal segment-mapping loop to
			   handle the portion of the segment past the end of the file
			   mapping. */
			ia32_mprotect ((caddr_t)(l_addr + c->mapend),
				       (unsigned long)(loadcmds[nloadcmds - 1].allocend - 
						       c->mapend), 
				       PROT_NONE);
		}
		goto postmap;
	}

	/* Remember which part of the address space this object uses. */
	l_map_start = c->mapstart + l_addr;
	l_map_end = l_map_start + maplength;

	while (c < &loadcmds[nloadcmds]) {
		if (c->mapend > c->mapstart) {
			/* Map the segment contents from the file. */
			errcode = (long)ia32_mmap ((void *) (l_addr + c->mapstart), 
						   (unsigned long)(c->mapend - c->mapstart), 
						   c->prot, MAP_FIXED | MAP_COPY | MAP_FILE, fd, 
						   (off_t)c->mapoff);
			if (IS_SYS_ERROR (errcode))
				goto map_error;
		}
postmap:
		if ((l_phdr == 0) && 
		    ((Elf32_Off) c->mapoff <= ehdr->e_phoff) &&
		    ((size_t) (c->mapend - c->mapstart + c->mapoff) >= 
			ehdr->e_phoff + ehdr->e_phnum * sizeof (Elf32_Phdr))) {
			l_phdr = (c->mapstart + ehdr->e_phoff - c->mapoff);
		}

		if (c->allocend > c->dataend) {
			/* Extra zero pages should appear at the end of this segment,
			   after the data mapped from the file.	*/
			unsigned long zero, zeroend, zeropage;
			zero = l_addr + c->dataend;
			zeroend = l_addr + c->allocend;
			zeropage = ((zero + ELF_EXEC_PAGE_SIZE - 1) & 
				   ~(ELF_EXEC_PAGE_SIZE - 1));

#if 0
			/* FIXME: For some reason, IA32 EL wants to explicitly zero
			   the paritial page if it has been mapped in
			   before. */
			if (zeroend < zeropage)
				/* All the extra data is in the last page of the segment.
				   We can just zero it. */
				zeropage = zeroend;

#endif	
			if (zeropage > zero) {
				/* Zero the final part of the last page of the segment.	*/
				if ((c->prot & PROT_WRITE) == 0) {
					/* Dag nab it.	*/
					errcode = ia32_mprotect (
						(caddr_t) (zero & ~ (ELF_EXEC_PAGE_SIZE - 1)),
					 	ELF_EXEC_PAGE_SIZE, 
						c->prot | PROT_WRITE);

					if (IS_SYS_ERROR (errcode)) {
						errstring = "cannot change memory protections";
						goto failed_errno;
					}
				}
				memset ((void *) zero, '\0', zeropage - zero);
				if ((c->prot & PROT_WRITE) == 0) {
					ia32_mprotect (
						(caddr_t) (zero & ~ (ELF_EXEC_PAGE_SIZE - 1)), 
						ELF_EXEC_PAGE_SIZE, c->prot);
				}
			}
			if (zeroend > zeropage) {
				/* Map the remaining zero pages in from the zero fill FD. */
				caddr_t mapat;
				mapat = ia32_mmap ((caddr_t) zeropage, zeroend - zeropage,
						   c->prot, 
						   MAP_ANON | MAP_PRIVATE | MAP_FIXED,
						   -1, (off_t)0);
				if (IS_SYS_ERROR (mapat)) {
					errcode = (long)mapat;
					errstring = "cannot map zero-fill pages";
					goto failed_errno;
				}
			}
		}
		++c;
	}
	if (l_phdr == 0) {
		errstring = "program header is not in a loadable segment";
		goto failed;
	} else {
		l_phdr += l_addr;
	}
	if (interp) {
		app_info->dl_base_addr = l_addr;
		app_info->dl_size = filelength;
	} else {
		/* FIXME: What if there are more than one text segement and one
		   data segement? */
		app_info->start_code = l_addr + loadcmds[0].phdr->p_vaddr;
		app_info->end_code = l_addr + loadcmds[0].dataend;
		app_info->start_data = l_addr + loadcmds[nloadcmds - 1].phdr->p_vaddr;
		app_info->end_data = l_addr + loadcmds[nloadcmds - 1].dataend;
		app_info->start_brk = app_info->brk = l_addr + loadcmds[nloadcmds - 1].allocend;
	}

	if (l_addrp) {
		*l_addrp = l_addr;
	}
	if (l_phdrp) {
		*l_phdrp = l_phdr;
	}
	return 0;
}

static long
setup_elf32_stack (int argc, char **argv, char **envp, unsigned phdr, unsigned phnum,
		unsigned base, unsigned entry, btrans_binary_map_t * app_info,
		int stack_executable, long personality)
{
	typedef struct {
		int a_type;
		int a_val;
	} auxv_t;
	void *stack_pointer;
	size_t len;
	int envc, auxn, auxc, i;
	char *info_start, *info_end, *ia32_info;
	char *platform, *ia32_platform;
	auxv_t *auxv;
	unsigned int *ia32_envp, *ia32_argv;
	int *ia32_argc;
	int prot;
	const char *errstring = NULL;
	long retval;
	
	/* setup prot of ia32 stack */
	prot = PROT_READ | PROT_WRITE;
	if ((stack_executable == IA32_STACK_EXEC_ENABLE) ||
		((stack_executable == IA32_STACK_EXEC_DEFAULT) && RX_PERSON(personality))) {
		prot |= PROT_EXEC;
	}

	/* Allocate the initial stack.	*/
	stack_pointer = (void *) (IA32_STACK_TOP - INITIAL_IA32_STACK_SIZE);
	retval = (long)ia32_mmap (
			stack_pointer, INITIAL_IA32_STACK_SIZE,
			prot,
		 	MAP_PRIVATE | MAP_ANONYMOUS | MAP_FIXED | MAP_GROWSDOWN, -1,
			(off_t)0);

	if (IS_SYS_ERROR (retval)) {
		errstring = "failed to allocate the initial statck";
		DBPRINT (TRUE, "%s (%ld)\n",  errstring, retval);
		return retval;
	}

	assert ((void *)retval == stack_pointer);
	assert ((unsigned long)stack_pointer < 0xffffffff);

	/* Initialize the stack. */
	memset (stack_pointer, 0, INITIAL_IA32_STACK_SIZE);

	/* Put a null pointer at top of stack  */
	stack_pointer = (void *) (IA32_STACK_TOP - 4);

	/* Copy the binary name. */
	len = strlen (argv[1]) + 1;
	stack_pointer -= len;
	memcpy (stack_pointer, argv[1], len);

	/* Copy the ia32 argv and env strings from the real stack to the
	   emulated stack. argv[0-1] are not the part of the ia32
	   process. */
	info_start = argv[1] + len;

	/* Count the env strings */
	for (envc = 0; envp[envc] != NULL; envc++);
	if (envc) {
		info_end = envp[envc - 1] + strlen (envp[envc - 1]) + 1;
	} else if (argc == 0) {
		info_end = info_start;
	} else {
		info_end = argv[argc + 1] + strlen (argv[argc + 1]) + 1;
	}
	len = info_end - info_start;
	if (len) {
		stack_pointer -= len;
		memcpy (stack_pointer, info_start, len);
	}

	/* Remember where we copied ia32 information block. */
	ia32_info = stack_pointer;

	/* Set up auxilary strings. */
	auxc = MANDATORY_AT_NUM;
	platform = ELF_PLATFORM;
	if (platform) {
		len = strlen (platform) + 1;
		stack_pointer -= len;
		memcpy (stack_pointer, platform, len);
		ia32_platform = stack_pointer;
		auxc ++;
	}
	if (g_is_at_secure != 0) {
		auxc ++;
	}


	/* We want to start the stack at the 16byte boundary. */
	stack_pointer -= (auxc * sizeof (auxv_t) /* Auxialiary vector */
			 + (envc + 1) * sizeof (unsigned int) /* Envp */
			 + (argc + 1) * sizeof (unsigned int) /* Argv */
			 + sizeof (int));	/* Argc */
	stack_pointer = (void *) ((unsigned long) stack_pointer & ~15UL);

	/* Set up argc.	*/
	ia32_argc = (int *) stack_pointer;
	*ia32_argc = argc;

	/* Set up argv.	*/
	app_info->arg_start = (unsigned long) ia32_info;
	ia32_argv = ia32_argc + 1;
	for (i = 0; i < argc; i++) {
		ia32_argv[i] = (unsigned int) (unsigned long) ia32_info;
		len = strnlen (ia32_info, INITIAL_IA32_STACK_SIZE) + 1;
		if (len > INITIAL_IA32_STACK_SIZE) {
			errstring = "illegal environmental string";
bad_string:
			DBPRINT (TRUE, "%s\n", errstring);
			return -EINVAL;
		}
		ia32_info += len;
	}
	ia32_argv[i] = 0;

	/* Set up envp.	*/
	app_info->arg_end = app_info->env_start = (unsigned long) ia32_info;
	ia32_envp = ia32_argv + argc + 1;
	for (i = 0; i < envc; i++) {
		ia32_envp[i] = (unsigned int) (unsigned long) ia32_info;
		len = strnlen (ia32_info, INITIAL_IA32_STACK_SIZE) + 1;
		if (len > INITIAL_IA32_STACK_SIZE) {
			errstring = "illegal argument string";
			goto bad_string;
		}
		ia32_info += len;
	}
	ia32_envp[i] = 0;
	app_info->env_end = (unsigned long) ia32_info;

	/* Set up the auxiliar vector. */
	auxn = 0;
	auxv = (auxv_t *) (ia32_envp + envc + 1);
	auxv[auxn].a_type = AT_HWCAP;
	auxv[auxn++].a_val = ELF_HWCAP;
	auxv[auxn].a_type = AT_PAGESZ;
	auxv[auxn++].a_val = ELF_EXEC_PAGE_SIZE;
	auxv[auxn].a_type = AT_CLKTCK;
	auxv[auxn++].a_val = IA32_CLOCKS_PER_SEC;
	auxv[auxn].a_type = AT_PHDR;
	auxv[auxn++].a_val = phdr;
	auxv[auxn].a_type = AT_PHENT;
	auxv[auxn++].a_val = sizeof (Elf32_Phdr);
	auxv[auxn].a_type = AT_PHNUM;
	auxv[auxn++].a_val = phnum;
	auxv[auxn].a_type = AT_BASE;
	auxv[auxn++].a_val = base;
	auxv[auxn].a_type = AT_FLAGS;
	auxv[auxn++].a_val = 0;
	auxv[auxn].a_type = AT_ENTRY;
	auxv[auxn++].a_val = entry;
	auxv[auxn].a_type = AT_UID;
	auxv[auxn++].a_val = (int) SYSCALL (getuid, ());
	auxv[auxn].a_type = AT_EUID;
	auxv[auxn++].a_val = (int) SYSCALL (geteuid, ());
	auxv[auxn].a_type = AT_GID;
	auxv[auxn++].a_val = (int) SYSCALL (getgid, ());
	auxv[auxn].a_type = AT_EGID;
	auxv[auxn++].a_val = (int) SYSCALL (getegid, ());
	if (platform) {
		auxv[auxn].a_type = AT_PLATFORM;
		auxv[auxn++].a_val = (int) (long) ia32_platform;
	}
	if (g_is_at_secure) {
		auxv[auxn].a_type = AT_SECURE;
		auxv[auxn++].a_val = (int) g_at_secure;
	}
	auxv[auxn].a_type = AT_NULL;
	auxv[auxn++].a_val = 0;
	assert (auxn == auxc);
	app_info->start_stack = (unsigned long) stack_pointer;
	return 0;
}

/* argv[1] holds the fd of the binary to be loaded
   argv[2, ... ,(argc-2)] is the argument-vector for the binary. */
long
load_elf32_binary (int argc, char **argv, char **envp,
                   btrans_binary_map_t *app_info,
		   long *personality)
{
	int exec_fd = -1, interp_fd = -1;
	Elf32_Ehdr *exec_ehdr = NULL, *interp_ehdr = NULL;
	Elf32_Phdr *exec_phdr = NULL, *interp_phdr = NULL;
	int i, dynamic;
	long ret;
	char *interp = NULL;
	const char *errstring = NULL;
	Elf32_Addr entry, phdr, exec_bias, interp_bias;
	int file_name_or_fd;
	char buff[PATH_MAX];
	int stack_executable;
	int found;

	/* init personality */
	*personality = 0;

	/* We need first to set the personality to LINUX32, so we get the
	   right root in case the system uses altroot. */
	ret = SYSCALL (personality, ((unsigned long)PER_LINUX32));
	if (IS_SYS_ERROR (ret)) {
		errstring = "failed to set LINUX32 personality";
		DBPRINT (TRUE, "%s (%ld)\n", errstring, ret);
		goto done;
	}
	if (g_is_at_execfd) {
		file_name_or_fd = LEB_FILE_DESC;
		ret = sprintf(buff, "%ld", g_at_execfd);
		assert(ret > 0);
	} else {
#if SUID_BTLIB
		file_name_or_fd = LEB_FILE_DESC;
#else
		ret = invoke_suid_btlib_if_needed (argv[1], &argv[0], envp);
		if (IS_SYS_ERROR (ret)) {
			goto done;
		}
		/* if we got here then suid_btlib is not needed and we can
		 * continue loading the binary */
		file_name_or_fd = LEB_FILE_NAME;
#endif
		strncpy (buff, argv[1], sizeof(buff) - 1);
	}

	/* check whether we're allowed to exec this binary */
	ret = check_exec_permissions (buff, file_name_or_fd);
	if (IS_SYS_ERROR (ret)) {
		return ret;
	}

	/* Check if the file is an ia32 ELF executable.	*/
	ret = check_ia32_elf_file (buff, file_name_or_fd, 0, &exec_fd, 
				   &exec_ehdr, &exec_phdr);
	if (ret) {
		goto done;
	}
	/* read the full path to the binary */
	{
		char proc_fd_name[64];
		sprintf (proc_fd_name, "/proc/self/fd/%d", exec_fd);
		assert (sizeof (btl_ia32_bin_full_path) < INT_MAX);
		ret = SYSCALL (readlink, (proc_fd_name, btl_ia32_bin_full_path, 
					(int)sizeof (btl_ia32_bin_full_path)));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "realink failed: %ld\n", ret);
#if SUID_BTLIB
			/* nothing we can do more */
			btl_ia32_bin_full_path[0] = '\0';
#else
			/* it won't be the full path in all cases - but this is
			 * the max we can do here */
			strncpy (btl_ia32_bin_full_path, argv[1],
					sizeof (btl_ia32_bin_full_path) - 1);
#endif
		}
		char * name = btl_ia32_bin_name = btl_ia32_bin_full_path;
		while ((name = strchr (name + 1, '/')) != NULL) {
			btl_ia32_bin_name = name + 1;
		}
		DBPRINT (0, "btl_ia32_bin_full_path='%s'\n", btl_ia32_bin_full_path);
		DBPRINT (0, "btl_ia32_bin_name='%s'\n", btl_ia32_bin_name ?: "(nil)");
	}
	
	/* check if stack is executable */
	found = 0;
	stack_executable = IA32_STACK_EXEC_DEFAULT;
	for (i = 0; i < exec_ehdr->e_phnum; i++) {
		if (exec_phdr[i].p_type == PT_GNU_STACK) {
			found = 1;
			if (exec_phdr[i].p_flags & PF_X) {
				stack_executable = IA32_STACK_EXEC_ENABLE;
			} else {
				stack_executable = IA32_STACK_EXEC_DISABLE;
			}
			break;
		}
	}
#if !SUID_BTLIB
	/* always unset READ_IMPLIES_EXEC for SUID binaries */
	if (!found) {
		*personality |= READ_IMPLIES_EXEC;
	}
#endif

	dynamic = -1;
	for (i = 0; i < exec_ehdr->e_phnum; i++) {
		if (exec_phdr[i].p_type == PT_INTERP) {
			dynamic = i;
			break;
		}
	}
	if (dynamic != -1) {
		unsigned long interp_size = (size_t)(exec_phdr[dynamic].p_filesz); 
		if (interp_size == 0 || interp_size > PATH_MAX) {
			errstring = "Illegal interpreter size";
			DBPRINT (TRUE, "%s: %s\n", errstring, argv[1]);
			ret = -ENOEXEC;
			goto done;
		}
		/* Get the dynamic linker. */
		interp = (char *) malloc (interp_size);
		if (interp == NULL) {
			errstring = "Cannot allocat memory for dynamic linker";
			ret = -ENOMEM;
			DBPRINT (TRUE, "%s: %s\n", errstring, argv[1]);
			goto done;
		}
		ret = SYSCALL (pread, (exec_fd, interp, 
				       interp_size, 
				       (off_t)exec_phdr[dynamic].p_offset));
		if (ret != (long)interp_size) {
			errstring = "Cannot read ELF file";
			DBPRINT (TRUE, "%s: %s\n", errstring, argv[1]);
			goto done;
		}
		if (interp[interp_size - 1] != '\0') {
			errstring = "Illegal interpreter string";
			DBPRINT (TRUE, "%s: %s\n", errstring, argv[1]);
			ret = -ENOEXEC;
			goto done;
		}

		/* Check if the dynamic linker is an ia32 ELF DSO.	*/
		ret = check_ia32_elf_file (interp, LEB_FILE_NAME, 1, &interp_fd, 
					   &interp_ehdr, &interp_phdr);
		if (ret) {
			goto done;
		}
	}
	if (map_elf32_binary (exec_fd, exec_ehdr, exec_phdr, 0, &exec_bias, 
			      &phdr, app_info, *personality)) {
failed_map:
		ret = -EIO;
		goto done;
	}
	if (dynamic != -1) {
		if (map_elf32_binary (interp_fd, interp_ehdr, interp_phdr, 1, 
				      &interp_bias, NULL, app_info, *personality)) {
			goto failed_map;
		}
		app_info->entry_point = interp_bias + interp_ehdr->e_entry;
	} else {
		app_info->entry_point = exec_bias + exec_ehdr->e_entry;
	}
	entry = exec_bias + exec_ehdr->e_entry;

	ret = setup_elf32_stack (argc - 2, argv, envp, phdr, exec_ehdr->e_phnum,
				 interp_bias, entry, app_info, stack_executable,
				 *personality);
	if (ret) {
		goto done;
	}

#if SUID_BTLIB
	/* if we run with setuid perms then set the creds */
	ret = compute_creds (exec_fd);
	if (IS_SYS_ERROR(ret)) {
		DBPRINT (TRUE, "compute_creds failed:%ld\n", ret);
		goto done;
	}
#endif

done:
	if (interp) {
		free (interp);
	}
	if (exec_ehdr) {
		free (exec_ehdr);
	}
	if (exec_phdr) {
		free (exec_phdr);
	}
	if (interp_ehdr) {
		free (interp_ehdr);
	}
	if (interp_phdr) {
		free (interp_phdr);
	}
	if (exec_fd != -1) {
		SYSCALL (close, (exec_fd));
	}
	if (interp_fd != -1) {
		SYSCALL (close, (interp_fd));
	}
	return ret;
}

/*****************************************************************************/
/* check whether the binary can be emulated                                  */
/*****************************************************************************/
static long check_exec_permissions (char *file, int file_name_or_fd)
{
	long ret;
	struct stat stat_buf;
	unsigned ruid, euid, suid;
	unsigned rgid, egid, sgid;
	int is_exec = 0;
	int is_setuid;

	/* get the user's details */
	ret = SYSCALL (getresuid, (&ruid, &euid, &suid));
	if (ret < 0) {
		DBPRINT (TRUE, "getresuid failed:%ld\n", ret);
		return ret;
	}
	ret = SYSCALL (getresgid, (&rgid, &egid, &sgid));
	if (ret < 0) {
		DBPRINT (TRUE, "getresgid failed:%ld\n", ret);
		return ret;
	}
	
	/* get the details of the file */
	if (file_name_or_fd == LEB_FILE_NAME) {
		ret = SYSCALL (stat, (file, &stat_buf));
	} else {
		int bin_fd = atol (file);
		ret = SYSCALL (fstat, (bin_fd, &stat_buf));
	}
	if (ret < 0) {
		DBPRINT (TRUE, "stat failed:%ld\n", ret);
		return ret;
		goto out;
	}

	/* check the exec permissions */
	DBPRINT (0, "st_uid=%d st_gid=%d st_mode=%o\n",
		stat_buf.st_uid, stat_buf.st_gid, stat_buf.st_mode);
	if ((stat_buf.st_mode & 0111) == 0) {
		is_exec = 0;
	} else if (euid == 0) {
		is_exec = 1;
	} else {
		ret = process_belongs_to_group (stat_buf.st_gid);
		if (ret < 0) {
			DBPRINT (TRUE, "process_belongs_to_group failed:%ld\n", ret);
			return ret;
		}	
		if (euid == stat_buf.st_uid) {
			is_exec = stat_buf.st_mode & S_IXUSR;
		} else if (ret) {
			is_exec = stat_buf.st_mode & S_IXGRP;
		} else {
			is_exec = stat_buf.st_mode & S_IXOTH;
		}
	}

	if ( ! is_exec ) {
		DBPRINT (TRUE, "file not executable "
			     "st_mode=0%0o euid=%d egid=%d st_uid=%d st_gid=%d\n",
			     stat_buf.st_mode, euid, egid, stat_buf.st_uid, stat_buf.st_gid);
		ret = -EACCES;
		goto out;
	}

	/* get the s bits */
 	is_setuid = stat_buf.st_mode & S_ISUID;

	/* don't core dump in case of setuid binary when the user that runs the
        binary is not the owner of the binary */
	if (is_setuid && (euid != stat_buf.st_uid)) {
		g_dont_core_suid_own = 1;
	}

out:
	DBPRINT (IS_SYS_ERROR (ret), "ret=%ld\n", ret);
	return ret;
}

/*****************************************************************************/
/* check whether the running process belongs to a certain group              */
/*****************************************************************************/
static long process_belongs_to_group (gid_t group)
{
	int number_of_groups;
	int i;
	gid_t group_list[NUMBER_OF_GROUPS], *groups = group_list;

	number_of_groups = SYSCALL (getgroups, (NUMBER_OF_GROUPS, group_list));
	if (number_of_groups == -EINVAL) {
		number_of_groups = SYSCALL (getgroups, (0, NULL));
		if (number_of_groups < 0) {
			DBPRINT (TRUE, "getgroups failed:%d\n", number_of_groups);
			return number_of_groups;
		}
		groups = malloc (number_of_groups * sizeof (gid_t));
		if (groups == NULL) {
			DBPRINT (TRUE, "malloc failed\n");
			return -ENOMEM;
		}
		if ((number_of_groups = SYSCALL (getgroups, (number_of_groups, groups))) < 0) {
			DBPRINT (TRUE, "getgroups failed:%d\n", number_of_groups);
			free (groups);
			return number_of_groups;	
		}	
	} else if (number_of_groups < 0) {
		DBPRINT (TRUE, "getgroups failed:%d\n", number_of_groups);
		return number_of_groups;
	}
	
	for (i = 0; i < number_of_groups; i++) {
		if (group == groups[i]) {
			/* the running process belongs to the group */
			return 1;
		}	
	}
	
	/* if we arrived here, the running process does not belong to the group */
	if (groups != group_list) {
		free (groups);
	}
	return 0;
}


#if SUID_BTLIB
/*****************************************************************************/
/* set the crdentials for the process.
 * input:  bin_fd - the file desc of the binary
 * output: 0 on success, a negative error value on failure
 *****************************************************************************/
static long compute_creds (int bin_fd)
{
	struct stat buf;
	long ret;
	unsigned ruid, euid, suid;
	unsigned rgid, egid, sgid;
	unsigned new_ruid, new_euid, new_suid;
	unsigned new_rgid, new_egid, new_sgid;
	int is_setuid, is_setgid;

	DBPRINT (0, "bin_fd=%d\n", bin_fd);

	/* get the user's details */
	ret = SYSCALL (getresuid, (&ruid, &euid, &suid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "getresuid failed:%ld\n", ret);
		return ret;
	}
	ret = SYSCALL (getresgid, (&rgid, &egid, &sgid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "getresgid failed:%ld\n", ret);
		return ret;
	}
	DBPRINT (0, "ruid=%d euid=%d suid=%d rgid=%d egid=%d sgid=%d\n",
		 ruid, euid, suid, rgid, egid, sgid);

 	/* get the details of the file */
	ret = SYSCALL (fstat, (bin_fd, &buf));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "stat failed:%ld bin_fd=%d\n", ret, bin_fd);
		return ret;
	}

 	/* get the s bits */
 	is_setuid = buf.st_mode & S_ISUID;
 	is_setgid = ((buf.st_mode & (S_ISGID | S_IXGRP)) == (S_ISGID | S_IXGRP));

	/* we should get to suid-btlib only if the binary is setuid */
	assert (is_setuid || is_setgid);

	/* real remains the same */
	new_ruid = ruid;
	new_rgid = rgid;

	/* effective uid/gid depends on the set?id bit of the binary */
	new_euid = is_setuid ? buf.st_uid : ruid;
	new_egid = is_setgid ? buf.st_gid : rgid;

	/* forget the original root permissions of btlib */
	new_suid = new_euid;
	new_sgid = new_egid;

	DBPRINT (0, "new_ruid=%d new_euid=%d new_suid=%d "
		    "new_rgid=%d new_egid=%d new_sgid=%d\n",
		 new_ruid, new_euid, new_suid, new_rgid, new_egid, new_sgid);

	/* set the credentials */
	ret = SYSCALL (setresgid, (new_rgid, new_egid, new_sgid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "setresgid failed:%ld\n", ret);
		return ret;
	}

	ret = SYSCALL (setresuid, (new_ruid, new_euid, new_suid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "setresuid failed:%ld\n", ret);
		return ret;
	}

	/* make sure we have the right credentials */
	ret = SYSCALL (getresuid, (&ruid, &euid, &suid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "getresuid failed:%ld\n", ret);
		return ret;
	}
	ret = SYSCALL (getresgid, (&rgid, &egid, &sgid));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "getresgid failed:%ld\n", ret);
		return ret;
	}

	if ((ruid != new_ruid) ||
	    (euid != new_euid) ||
	    (suid != new_suid) ||
	    (rgid != new_rgid) ||
	    (egid != new_egid) ||
	    (sgid != new_sgid)) {
		DBPRINT (TRUE, "compute_creds failed\n");
		return -1;
	}


	return 0;
}
#else // if ! SUID_BTLIB
/*****************************************************************************/
/* if the binary is loaded then switch to suid_btlib */
/* input:  file_name - the name of the binary to run
 *         argv, envp - the arguments that btlib's main() got from the system
 * output: negative value in case of error. 0 if no need to change to 
 *         suid_btlib. does not return if suid_btlib is invoked */
/*****************************************************************************/
static int invoke_suid_btlib_if_needed (char * file_name, char ** argv, char ** envp)
{
	int is_setuid, is_setgid;
	struct stat buf;
	int ret;

	DBPRINT (0, "file_name='%s' argv[0]='%s' argv[1]='%s'\n",
		 file_name, argv[0], argv[1]);

	/* get the setuid/setgid bits of the file */
	ret = SYSCALL (stat, (file_name, &buf));
	if (IS_SYS_ERROR (ret)) {
		DBPRINT (TRUE, "stat failed:%d\n", ret);
		goto out;
	}
	is_setuid = buf.st_mode & S_ISUID;
	is_setgid = buf.st_mode & S_ISGID;

	DBPRINT (0, "is_setuid=%d is_setgid=%d\n", is_setuid, is_setgid);

	/* if it's setuid then invoke suid_btlib */
	if (is_setuid || is_setgid) {
		int flags;
		char ** search_path = BTGeneric_search_path;
		char buff[256];
		int elf_exec_fileno;
		char argv1[32];

		elf_exec_fileno = SYSCALL (open, (file_name, O_RDONLY, 0));
		if (IS_SYS_ERROR (elf_exec_fileno)) {
			goto out;
		}
		
		/* clear the close-on-exec flag */
		flags = SYSCALL (fcntl, (elf_exec_fileno, F_GETFD, 0ul));
		if (IS_SYS_ERROR (flags)) {
			DBPRINT (TRUE, "getfd failed: %d\n", ret);
			goto out;
		}
		flags &= ~FD_CLOEXEC;
		ret = SYSCALL (fcntl, (elf_exec_fileno, F_SETFD, (unsigned long)flags));
		if (IS_SYS_ERROR (ret)) {
			DBPRINT (TRUE, "fcntl failed: %d", ret);
			goto out;
		}
		/* search the suid-btlib, then invoke it */
		while (strlen(*search_path) != 0) {
			strncpy (buff, *search_path, (size_t)255);
			strncat (buff, "/"SUID_LIB_NAME, 
				 MIN(sizeof(buff) - strlen(buff) - 1, strlen(SUID_LIB_NAME) + 1));

			ret = SYSCALL (stat, (buff, &buf));
			if (ret == 0){
				break;
			} else {
				ret = -1;
				search_path++;
			}	
		}
		if (ret == -1){
			DBPRINT (TRUE, "%s could not be found.\n", SUID_LIB_NAME);
			goto out;
		}
		/* replace file_name with the fd */
		sprintf (argv1, "%d", elf_exec_fileno);
		argv[1] = argv1;

		/* invoke suid_btlib */
		DBPRINT (0, "buff='%s' argv[1]='%s'\n", buff, argv[1]);
		ret = SYSCALL (execve, (buff, argv, envp));
		DBPRINT (TRUE, "execve failed: %d\n", ret);
		goto out;
	}

	ret = 0;
out:
	return ret;
}
#endif

