/*  
 * btload.c
 *  
 * 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Doron Ziffrony
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
 

#include "os_dep_linux.h"
#include "os_dep_pub.h"
#include "new_pe.h"
#include "btload_error.h"
#include "btload.h"
#include "btload_pub.h"
#include "../btlib.h"


/*************************************************************************
 * Func:   image_read	
 * Inputs: fd - File descriptor of the file from which to read.	
 *         offset - File offset from which to read. Zero means current.
 *         size. Amount of data to read	
 * Output: buffer - Where to read to.
 * Return Value:  Success or Fail code (see BTload_error.h)
 * Functionality: Position the file to the appropriate location (if not	
 *                reading the next bunch of data) and read the data.
 *************************************************************************/

FUNC BT_U64 image_read(int fd, BT_U64 offset, BT_U64 size, void * buffer)
{
	BT_U64	rc;	/* Return value from system calls */

	/* Check if we need repositioning */
	if (offset > 0) {
		rc = OS_LSEEK(fd, (__off_t)offset, SEEK_SET);
		if (rc != offset) {
			return BTLOAD_FILE_SEEK_ERR;
		}
	}

	/* Read from the file */
	rc = OS_READ(fd, buffer, size);
	if (rc != size) {
		return BTLOAD_FILE_READ_ERR;
	}
		
	return BTLOAD_OK;
}

/************************************************************************
 * Func:   verify_coff_header
 * Inputs: coff_header_ptr. Pointer to the coff header.	
 * Output: None.
 * Return Value:  Success or Fail code (see BTload_error.h)
 * Functionality: Verify that we have a valid 64 bit image file.
 *************************************************************************/
FUNC BT_U64 verify_coff_header(coff_header_type * coff_header_ptr)
{

	/* Verify that we are looking at an 64 bit image file */
	if (coff_header_ptr->machine_type != PE_MACHINE_TYPE_IA64) {
		return BTLOAD_NON_IA64_ERR;
	}

	if (coff_header_ptr->optional_header_size != ((sizeof(opt_header_type)) + 
		(sizeof(pe_directory_entry_type) * PE_NUMOF_DIRECTORIES))) {
		return BTLOAD_BAD_HEADER_ERR;
	}

	return BTLOAD_OK;
}

/*************************************************************************
 * Func:	verify_optional_header
 * Inputs: opt_header_ptr. Pointer to the optional header.
 * Output: None.
 * Return Value:	Success or Fail code (see BTload_error.h)
 * Functionality: Verify that we have a valid BTGeneric image file.
 *************************************************************************/
FUNC BT_U64 verify_optional_header(opt_header_type * opt_header_ptr)
{
	/* Verify that we are looking at an 64 bit image file - PE32+ */
	if (opt_header_ptr->magic != PE_32PLUS_SIGNATURE) {
		return BTLOAD_NON_PE32PLUS_ERR;
	}

	/* Verify that the directories data is in the file */
	if (opt_header_ptr->number_of_directories !=
		PE_NUMOF_DIRECTORIES) {
		return BTLOAD_BAD_HEADER_ERR;
	}
	
	return BTLOAD_OK;
}

/*************************************************************************
 * Func:   load_section_headers
 * Inputs: num_of_sections - Number of sections in the image.
 *         image_size - Size of the image.
 *         image_load_addr - Requested image load address.
 *         fd - File descriptor for image file.	
 * Output: alloced_mem_size - Actual size of allocated memory
 *         sections_table_ptr - Returned pointer to the sections table.
 * Return Value:  Success or Fail code (see BTload_error.h)
 * Functionality: Compute the section table load address, allocate memory
 *                and load the sections table onto it.
 *************************************************************************/
FUNC BT_U64 load_section_headers(BT_U64 num_of_sections,
                                 BT_U64 image_size,
                                 BT_U64 image_load_addr,
                                 int    fd,
                                 BT_U64 * alloced_mem_size,
                                 psection_header_type * sections_table_ptr)
{
	BT_U64	load_address;
	BT_U64	section_table_size;
	BT_U64	rc;
	BT_U64	page_size;
	BT_U64	returned_size;

	/* 
	 * compute load address: image load address + 2 * image_size 
	 * and round to the start of the next page. 
	 */
	page_size = OS_GET_PAGE_SIZE();
	load_address = (image_load_addr + 2 * image_size);
	load_address = load_address - load_address % page_size + page_size;
	section_table_size = sizeof(section_header_type) * num_of_sections;
	rc = OS_GET_MEMORY(load_address,
			   section_table_size,
			   (BT_U64)(OS_PROT_READ | OS_PROT_WRITE),
			   &returned_size, 
			   (void **) sections_table_ptr);
	
	if (rc != BTLOAD_OK) {
		return rc;
	}
	
	/* load the section headers to memory (no need to seek) */
	rc = image_read(fd, 0ul, section_table_size, *sections_table_ptr);
	*alloced_mem_size = returned_size;

	return rc;
}


/*************************************************************************
 * Func:  load_relocations
 * Inputs: reloc_sect_header_ptr - Pointer to the .reloc section header.
 *         fd - File descriptor of mapped file.
 *         target_address - Address to which the image is loaded.
 * Output: head_reloc_table_ptr - A return pointer to the head of the
 *         relocations table.
 * Return Value:  Success or Fail code (see BTload_error.h)
 * Functionality: Loads the relocation table to memory to to its 
 *                load address (taken from the section header).
 *************************************************************************/
FUNC BT_U64 load_relocations(psection_header_type reloc_sect_header_ptr,
                             int fd,
                             BT_U64 target_address,
                             fixup_block_type ** head_reloc_table_ptr)
{
	BT_U64 rc;
	BT_U64 ret_size;
	void * returned_addr;

	rc = OS_MAP_FILE(reloc_sect_header_ptr->virtual_address + target_address,
			 (BT_U64)reloc_sect_header_ptr->total_section_size,
			 (BT_U64)reloc_sect_header_ptr->pointer_to_raw_data,
			 (BT_U64)OS_PROT_READ, 0, fd, &ret_size, &returned_addr);
	
	if (rc == BTLOAD_OK) {
		*head_reloc_table_ptr = (fixup_block_type *) returned_addr;
	}

	return rc;
}


/*************************************************************************
 * Func: load_section
 * Inputs: sect_header_ptr - Pointer to the section header.
 *         fd - File descriptor of mapped file.
 *         taget_address - Image target load address.
 * Output: None.
 * Return Value: Success or Fail code (see BTload_error.h)
 * Functionality: Loads the given section to memory to to its load
 *                address (taken from the section header).
 *************************************************************************/
FUNC BT_U64 load_section(psection_header_type sect_header_ptr,
                         int fd,
                         BT_U64 target_address,
                         long enforce_write_permission)
{
	BT_U64 rc = BTLOAD_OK;
	BT_U64 ret_size;
	BT_U64 anonymous_size;
	int    shared;
	int    prot;
	void * returned_addr;

	prot = 0;
	prot |= (sect_header_ptr->characteristics & PE_SECTION_MEM_READ	? OS_PROT_READ : 0);
	prot |= (sect_header_ptr->characteristics & PE_SECTION_MEM_WRITE ? OS_PROT_WRITE : 0);
	prot |= (sect_header_ptr->characteristics & PE_SECTION_MEM_EXECUTE ? OS_PROT_EXEC : 0);
	
	shared = (((prot & OS_PROT_WRITE) == 0) && ! enforce_write_permission);
#if 0
	/* workaround for gdb bug: cannot write to a MAP_SHARED memory */
	shared = FALSE;
#endif 

	if (enforce_write_permission) {
		prot |= OS_PROT_WRITE;
	}

	if (sect_header_ptr->size_of_raw_data > 0) {
		rc = OS_MAP_FILE(sect_header_ptr->virtual_address + target_address,
				 (BT_U64)sect_header_ptr->size_of_raw_data,
				 (BT_U64)sect_header_ptr->pointer_to_raw_data,
				 (BT_U64)prot, shared, fd, &ret_size, &returned_addr);
		
		if (rc != BTLOAD_OK) {
			return rc;
		}
	}

	if (sect_header_ptr->size_of_raw_data < sect_header_ptr->total_section_size) {
		rc = OS_GET_MEMORY((sect_header_ptr->virtual_address +
				   sect_header_ptr->size_of_raw_data+
				   target_address),
				   (BT_U64)(sect_header_ptr->total_section_size - 
				   sect_header_ptr->size_of_raw_data),
				   (BT_U64) prot,
				   &anonymous_size,
				   &returned_addr);
	}
	
	return rc;
}

/*************************************************************************
 * Func: load_pe_binary
 * Inputs: path - Path of BTGeneric - to be loaded to memory.
 *         p_target_address - Pointer to the pointer to location of BTGeneric load.
 *                            In APP mode, the init is not used (address given 
 *                            by R.T. functions).
 * Output: get_api_info_address - Return a pointer to BTGeneric GetAPIInfo function
 *         p_load_size - Return the size of the loaded BTGeneric.
 * Return Value:  Success or Fail code (see BTload_error.h)
 * Functionality: Load BTGeneric to memory, perform fixup to the code, if necessary, 
 *                search the symbol table for the value of the value of the symbol 
 *                BTGeneric_API_table and return it to the caller.
 *************************************************************************/
FUNC BT_U64 load_pe_binary(char * path, void ** p_target_address, 
                           void ** get_api_info_address, BT_U64 * p_load_size, 
                           unsigned long rsrc[2])
{
	int	BTGenericFd;		/* File descriptor of the BTGeneric image */
	int	map_handle;		/* Handle of the mapping (mainly for NT) */
	BT_U64 rc;			/* Return value from system calls */
	unsigned i, j;
	coff_header_type coff_header;	/* File COFF header */
	opt_header_type optional_header;	/* File optional header */
	pe_directory_entry_type *directories;	/* Data directories header */
	psection_header_type
		sect_header_ptr,	/* Pointer to the "current" section header */
		reloc_sect_header_ptr=NULL, /* Pointer to the relocations section */
		sect_header_head;	/* Pointer to the head of the section header table */
	BT_U64 sect_header_mem_size; /* Size of allocated memory for the section headers */
	BT_U32 pe_offset;	/* Location of the "PE" signature in the file */
	void	*map_base;	/* base address in memory of PE header */
	pe_signature_type 
        pe_signature;		/* Contains the PE signature */
	BT_U8	relocate;	/* Do we need to apply fixups */
	BT_U8	api_found;	/* Found the exported extry for the API table */
	pe_directory_entry_type *
		reloc_dir_ptr;	/* Relocation directory pointer */
	fixup_block_type 
		* reloc_base_end,	/* End of relocation section */
		* reloc_page_entry;	/* Current relocation page entry */
	BT_U32 end_section_addr;	/* End address of section */
	reloc_entry_type *
		reloc_entry_ptr;	/* Pointer to the "current" relocation entry */
	BT_U64 imm64;			/* the imm64 value in IMM64 relocations */
	bool   is_imm64_reloc = 0;	/* the flag to indicate if this is IMM64 relocation */
	BT_U8	divided_section;	/* Section has relocation and needs to be treated on
					a page group basis */
	BT_S64 delta_address;		/* Delta between target_address and DLL preferred address */
	export_directory_type * export_table_ptr;	/* Pointer to the export table */
	void * page_ptr;		/* Pointer to the page with fixups */
	BT_U64 * fixup_ptr;		/* and a pointer to the fixed item */
	BT_U32 * name_pointer_ptr;	/* Pointer to the export names pointers */
	BT_U16 * address_index_ptr;	/* In the export ordinals table */
	BT_U32 * export_address_ptr;	/* Value of the exported symbol */
	BT_U64 pages_to_read_only;	/* Number of consecutive pages to turn to read only after relocation */
	BT_U64 page_size;		/* Machine's page size */
	BT_U64 previous_page_addr;	/* Address of previous page to see if we can join this page to the list */
	BT_U64 first_page_addr;		/* Address of first page in the list to turn to read only */
	BT_U64 loaded_image_size;	/* Size of loaded image (net) */
	void * target_address = *p_target_address;	/* The desired load address */

	/* Make sure target address is a multiple of 64KB */
	if ((BT_U64)target_address & 0xffff) {
		return BTLOAD_INVALID_ADDR_ERR;
	}
	
	/* Open the EXE file of BTGeneric */
	BTGenericFd = OS_OPEN(path, OS_RDONLY);
	if (BTGenericFd < 0) {
		DBPRINT(TRUE, "Unable to open the BTGeneric file\n");
		return BTLOAD_FILE_OPEN_ERR;
	}

	/* Load the PE file header at the BTGENERIC start address */
    /* PROT_EXEC is specified so that header section will be taken into account by performance analysis tools */
	map_base = (void*)SYSCALL (mmap, (BTGENERIC_LOAD_ADDRESS, 0x10000ul, PROT_READ|PROT_EXEC, 
				MAP_PRIVATE, BTGenericFd, 0ul));
	if (BTGENERIC_LOAD_ADDRESS != (unsigned long)map_base) {
		DBPRINT(TRUE, "Problem - Can not map the PE header.\n");
		return BTLOAD_PE_HEADER_LOAD_ERROR;
	}

	/* Read the PE signature, and make sure this is a PE file */
	map_base += PE_OFFSETOF_SIGNATURE_ADDRESS;
	pe_offset = *(BT_U32*)(map_base);

	map_base = (void *)((uintptr_t)pe_offset + BTGENERIC_LOAD_ADDRESS);

	pe_signature = *(BT_U32*)(map_base);
	map_base += sizeof(pe_signature);

	if (pe_signature != PE_SIGNATURE) {
		DBPRINT(TRUE, "Loaded file is not an EXE file (Bad PE signature)\n");
		return BTLOAD_BAD_IMAGE_ERR;
	}
		
	/* Read the COFF header and verify this is a valid BTGeneric image. */
	coff_header = *(coff_header_type *)(map_base);
	map_base += sizeof(coff_header);

	rc = verify_coff_header(&coff_header);
	if (rc != BTLOAD_OK) {
		DBPRINT(TRUE, "This not a valid BTGeneric image: (1)\n");
		return rc;
	}

	/* Read the optional header (we verified that it exists) and check it */
	optional_header = *(opt_header_type*)(map_base);
	map_base += sizeof(optional_header);

	rc = verify_optional_header(&optional_header);
	if (rc != BTLOAD_OK) {
		DBPRINT(TRUE, "This not a valid BTGeneric image: (2)\n");
		return rc;
	}
	
	/* Read the directories data */
	directories = (pe_directory_entry_type *)map_base;
	map_base += (sizeof(directories) * PE_NUMOF_DIRECTORIES);


	/*
	* Check whether we need to fixup (relocate) the executable image.
	* This happens when the actual load address (most of the times it will
	* also be the desired load address that was given as a parameter to this function)
	* is different from the load module's preffered load address
	*/
	if (target_address == 0) {
	target_address = (void*)optional_header.load_address;
	} 

	delta_address = (BT_S64)target_address - (BT_S64)optional_header.load_address;
	if (delta_address == 0) {
		reloc_base_end = 0;
		relocate = FALSE;
	}
	else {
		relocate = TRUE;
		reloc_dir_ptr = directories + PE_BASE_REALLOC_DIRECTORY_IDX;
		reloc_base_end = (void *) ((BT_U64)reloc_dir_ptr->rva + 
					   (BT_U64)reloc_dir_ptr->size + 
					   (BT_U64)target_address);

		if (!reloc_dir_ptr->rva || !reloc_base_end) {
			/* No relocation data is available. Can't continue */
			return BTLOAD_NO_RELOCATION_ERR;
		}
	}

	/* Allocate memory in which to place the section headers and load them to memory */
	sect_header_head = map_base;
	sect_header_mem_size = sizeof(section_header_type) * 
		(BT_U64)coff_header.number_of_sections;
	map_base += sect_header_mem_size;

	/* Create the file mapping (in NT). In Linux this is a dummy) */
	 rc = OS_CREATE_MAP(
		BTGenericFd,
		(BT_U64)sect_header_head[0].pointer_to_raw_data,
		(BT_U64)sect_header_head[coff_header.number_of_sections - 1].pointer_to_raw_data + 
			sect_header_head[coff_header.number_of_sections - 1].size_of_raw_data,
		&map_handle);
	 
	/* Locate the .reloc section and map the file area to memory above the sections table. */
	if (relocate) {
		sect_header_ptr = sect_header_head;
		
		for (i = 0 ; i < coff_header.number_of_sections; sect_header_ptr++, i++) {
			if (strncmp(sect_header_ptr->section_name, ".reloc", 
				    sizeof(".reloc")) == 0) {
				break;
			}
		}
		
		reloc_sect_header_ptr = sect_header_ptr;


		/* Load the relocation section to memory (obtained dynamically) */
		rc = load_relocations(reloc_sect_header_ptr, map_handle, 
				      (BT_U64)target_address, &reloc_page_entry);
		if (rc != BTLOAD_OK) {
			DBPRINT(TRUE, "Unable to load relocations to memory.\n");
			return rc;
		}
	}

	/*
	* Scan the sections.	For each section, verify that we need to load it
	* (it is not discardable, like the .reloc section, for example).
	* If it is a R/W section, it is unsharable, and consequently it can
	* be loaded/mapped to memory with no further thought.
	* After loading/mapping, it is checked whether the section need any relocation
	* fixups, which are then promptly performed.
	* If the section is a RO section (like .text, usually), it is necessary to
	* distinguish the pages that have no relocation requirements from those that do.
	* Those pages that have no relocation requirements can be loaded/mapped as shared
	* pages, to be shared by all IA-32 Execution Layer clients.	
	* Those that do are loaded/ampped as private pages and are
	* then fixed up following the relocation information.
	*/

	page_size = OS_GET_PAGE_SIZE();
	sect_header_ptr = sect_header_head;
	loaded_image_size = 0;
	for (j = 0; j < coff_header.number_of_sections; sect_header_ptr++, j++) {
		end_section_addr = sect_header_ptr->virtual_address +
			sect_header_ptr->total_section_size;
		
		if ((strncmp(sect_header_ptr->section_name, ".reloc", sizeof(".reloc")) == 0) ||
		    (strncmp(sect_header_ptr->section_name, ".comment", sizeof(".comment")) == 0)) {
			continue;
		}
		if (strncmp(sect_header_ptr->section_name, ".rsrc", sizeof(".rsrc")) == 0) {
			rsrc[0] = (unsigned long)target_address + sect_header_ptr->virtual_address;
			rsrc[1] = sect_header_ptr->total_section_size;
		}

		
		if (relocate && (reloc_page_entry->page_rva < sect_header_ptr->virtual_address)) {
			DBPRINT(TRUE, "Problem - relocation entries not belonging to any section.\n");
			return BTLOAD_RELOC_ENTRIES_ERR;
		}

		loaded_image_size += sect_header_ptr->size_of_raw_data;
		
		if (!relocate || (reloc_page_entry->page_rva > end_section_addr)) {
			/* No relocation entries for the current section - we can load it directly */
			rc = load_section(sect_header_ptr, map_handle, (BT_U64)target_address, 0l);
			if (rc != BTLOAD_OK) {
				DBPRINT(TRUE, "Problems loading section # %d\n", j);
				return rc;
			}
		}
		else {
			unsigned long section_base_addr = sect_header_ptr->virtual_address + (uintptr_t)target_address;
			unsigned long section_size = sect_header_ptr->size_of_raw_data;
			unsigned long pages_in_section= section_size / OS_GET_PAGE_SIZE();
			
			/*
			 * The section has relocations.
			 * If a R/W section, allocate it normaly. If R/O, need to divide section to
			 * pages with relocateable data and those without it.
			 */
			divided_section =
				((sect_header_ptr->characteristics & PE_SECTION_MEM_WRITE) == 0);
			pages_to_read_only = 0;
			
			if (!divided_section) {
				rc = load_section(sect_header_ptr, map_handle, (BT_U64)target_address, 0l);
			}
			else {
				rc = load_section(sect_header_ptr, map_handle, (BT_U64)target_address, 1l);
			}
			
			if (rc != BTLOAD_OK) {
				DBPRINT(TRUE, "Problems loading section # %d\n", j);
				return rc;
			}

			/*
			 * Insert relocations. Currently, only DIR64 and IMM64 reloc types are supported.
			 * Before the actual handling of relocations, there is a special treatment
			 * of divided sections.	We need to make the pages that we mended with relocation
			 * fixups non-writable again.
			 */
			previous_page_addr = 0;
			first_page_addr = 0;
			is_imm64_reloc = 0;
			while (reloc_page_entry->page_rva < end_section_addr) {
				page_ptr = (void *)(reloc_page_entry->page_rva + (BT_U64)target_address);

				reloc_entry_ptr =
					(reloc_entry_type *)((BT_U64)reloc_page_entry + sizeof(fixup_block_type));
				for (i = 0 ;
				     i < (reloc_page_entry->size_of_block - sizeof(fixup_block_type)) /
						sizeof(reloc_entry_type);
				     i++) {
					
					switch (GET_RELOCATION_TYPE(*reloc_entry_ptr)) {
					case PE_RELOC_TYPE_ABSOLUTE:
						break;

					case PE_RELOC_TYPE_IA64_IMM64:
						is_imm64_reloc = 1;
						
						/* get the addr of the inst */
						fixup_ptr = (BT_U64 *)((GET_RELOCATION_OFFSET(*reloc_entry_ptr)
								+ (BT_U64)page_ptr));
						fixup_ptr = (BT_U64 *)((BT_U64)fixup_ptr & 0xfffffffffffffffcUL);
						
						/* get the content of imm64 reloc data */
						imm64 = ((fixup_ptr[1] & I_MASK) << I_SHIFT)
							| ((fixup_ptr[1] & IMM41_1_MASK) << IMM41_1_SHIFT)
							| ((fixup_ptr[1] & IC_MASK) >> IC_SHIFT)
							| ((fixup_ptr[1] & IMM5C_MASK) >> IMM5C_SHIFT)
							| ((fixup_ptr[1] & IMM9D_MASK) >> IMM9D_SHIFT)
							| ((fixup_ptr[1] & IMM7B_MASK) >> IMM7B_SHIFT)
							| ((fixup_ptr[0] & IMM41_2_MASK) >> IMM41_2_SHIFT);
						/* relocate */
						imm64 += delta_address;
						
						/* clean the old imm64 value */
						fixup_ptr[0] &= ~INST_LOW64_MASK;
						fixup_ptr[1] &= ~INST_HIGHT64_MASK;

						/* write back new imm64 value */
						fixup_ptr[0] |= ((imm64 << IMM41_2_SHIFT) & IMM41_2_MASK);
						fixup_ptr[1] |= ((imm64 >> I_SHIFT) & I_MASK)
								| ((imm64 >> IMM41_1_SHIFT) & IMM41_1_MASK)
								| ((imm64 << IC_SHIFT) & IC_MASK)
								| ((imm64 << IMM5C_SHIFT) & IMM5C_MASK)
								| ((imm64 << IMM9D_SHIFT) & IMM9D_MASK)
								| ((imm64 << IMM7B_SHIFT) & IMM7B_MASK);
						break;

					case PE_RELOC_TYPE_DIR64:
						fixup_ptr = (BT_U64 *)(GET_RELOCATION_OFFSET(*reloc_entry_ptr)
									+ (BT_U64)page_ptr);
						*fixup_ptr += delta_address;
						break;

					default:
						DBPRINT(TRUE, "Image includes unsupported relocation type %d\n",
							GET_RELOCATION_TYPE(*reloc_entry_ptr));
						return BTLOAD_UNSUPP_RELOC_ERR;
						break;
					}

					reloc_entry_ptr++;
				}

				/*
				 * Advance to the next relocation page entry.
				 * If it appears that we reached the end of the relocations 
				 * 	section, turn off further relocations.
				 */
				reloc_page_entry = (fixup_block_type *)((BT_U64)reloc_page_entry
							+ reloc_page_entry->size_of_block);
				if (reloc_page_entry >= reloc_base_end) {
					relocate = FALSE;
					break;
				}
			}

			/*
			 * Deal with the changing of protection of the last relocated page in the section.
			 * This is done for divided sections only.
			*/
			if (divided_section || is_imm64_reloc) {
				/* return the section to read-only */
				int is_executable = sect_header_ptr->characteristics & PE_SECTION_MEM_EXECUTE;
				rc = OS_PAGES_TO_NON_WRITABLE(section_base_addr, pages_in_section, is_executable);
				
				if (rc != BTLOAD_OK) {
					DBPRINT(TRUE, "Unable to turn protection of pages to read only (2)\n");
					return rc;
				}
			}
		}
	}

	/*
	 * Check to see whether we still have relocation entries that were not used to fix the code.
	 * This is not right.
	 */
	if (relocate) {
		DBPRINT(TRUE,
			"There are still relocation entries that weren't used to fixup the code\n");
		return BTLOAD_MORE_RELOC_ERR;
	}

	
	/*
	 * Unmap the relocations section
	 * 
	 * if no reloacations are needed for the module then 
	 * reloc_sect_header_ptr will never get a real value, so we must
	 * check it before freeing the memory
	 */
	if (reloc_sect_header_ptr != NULL) {
		rc = OS_FREE_MEMORY((BT_U64)reloc_sect_header_ptr->virtual_address + (BT_U64)target_address,
				    (BT_U64)reloc_sect_header_ptr->size_of_raw_data);
		
		if (rc != BTLOAD_OK) {
			DBPRINT(TRUE,
				"Problems releasing relocations section memory: Error %d\n", (int) rc);
			return rc;
		}
	}

	/*
	 * It is possible now to safely release the section header memory
	 *
	 * Search for the location of the API table entry: BtgGetAPIInfo and retrieve the value.
	 * Assume that the export data is in memory (it is supposed to be a part
	 * of a section that's already loaded).
	 */
	export_table_ptr = (export_directory_type *)(((pe_directory_entry_type *)(directories
			       	+ PE_EXPORT_DIRECTORY_IDX))->rva + (BT_U64)target_address);
	name_pointer_ptr = (BT_U32 *)((BT_U32)export_table_ptr->names_rva + (BT_U64)target_address);
	address_index_ptr = (BT_U16 *)((BT_U32)export_table_ptr->ordinals_rva
					+ (BT_U64)target_address);
	export_address_ptr = (BT_U32 *)((BT_U32)export_table_ptr->functions_rva
					+ (BT_U64)target_address);
	api_found = FALSE;
	
	for (i = 0 ; i < export_table_ptr->number_of_names; name_pointer_ptr++, i++) {
		if (strncmp("BtgGetApiInfo", (char *)(*name_pointer_ptr + (BT_U64)target_address), 
			    sizeof("BtgGetApiInfo")) == 0) {
			api_found = TRUE;
			break;
		}
	}

	if (!api_found) {
		DBPRINT(TRUE, "Unable to locate BtgGetApiInfo in the export table\n");
		return BTLOAD_NO_API_TABLE_ERR;
	}
	else {
		address_index_ptr = (BT_U16 *)((BT_U32)export_table_ptr->ordinals_rva
					+ (BT_U64)target_address + i*sizeof(BT_U16));
		export_address_ptr = (BT_U32 *)((BT_U32)export_table_ptr->functions_rva
					+ (BT_U64)target_address + *address_index_ptr*sizeof(BT_U32));
		
		*get_api_info_address = (void *)(*export_address_ptr + (BT_U64)target_address);
		*p_target_address = target_address;
		*p_load_size = loaded_image_size;
		
		j = OS_CLOSE (BTGenericFd);
		assert (j == 0);
		
		return BTLOAD_OK;
	}
}

