/*  
 * os_dep_linux
 *  
 * 
 * 
*/
 /*
 *
 * Copyright (C) 2005       Yoav Zach <yoav.zach@intel.com>
 * Copyright (C) 2005       Eric Lin <eric.lin@intel.com>
 * Copyright (C) 2005       Bruce Chen <bruce.chen@intel.com>
 * Copyright (C) 2005       Dror Shenkar <dror.shenkar@intel.com>
 * Copyright (C) 2005       Doron Ziffrony
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
 

#include "os_dep_linux.h"
#include "btload_error.h"
#include "os_dep_pub.h"



/*************************************************************************/
/* Func:   linux_get_page_size                                           */
/* Inputs: None.                                                         */
/* Output: None.                                                         */
/* Return Value:  Page size of the machine.                              */
/* Functionality: Return the computer's page size.                       */
/*************************************************************************/

FUNC BT_U64 linux_get_page_size(void)
{

	/*
	 * Although actual page size is smaller, we want to prevent holes 
	 * between the sections and on the disk, the sections are 64KB aligned.
	 */
	return (BT_U64) 65536;
}

/*************************************************************************/
/* Func:   linux_allocate_memory                                         */
/* Inputs: page_address -  Address where memory is allocated.            */
/*         size -  Amount of memory to allocate (we round it to be a     */
/*                 multiplier of the page size. )                        */
/*         protection -  Protection of the allocated segment             */
/* Output: returned_size -  Actual amount of memory allocated.           */
/*         mem_ptr -  Returned pointer to the allocated memory.          */
/* Return Value:  Success or Fail code (see BTload_error.h)              */
/* Functionality: Allocate memory from the system using anonymous mmap() */
/*                call.                                                  */
/*************************************************************************/

FUNC BT_U64 linux_allocate_memory(BT_U64 page_address,
                                  BT_U64 size, 
				  BT_U64 protection,
                                  BT_U64 * returned_size,
                                  void   ** mem_ptr)
{
	BT_U64  page_size;
	int     dummy_fd = 0;
	off_t   dummy_offset = 0;
	int     protect;


	/*
	 * Verify page_address to be page alligned, and then round the 
	 * requested size to be a multiply of the page size.
 	 */

	page_size = (BT_U64) linux_get_page_size();
	if (page_address % page_size) {
		return BTLOAD_INVALID_ADDR_ERR;
	}

	if (size % page_size) {
		*returned_size = page_size * (1 + size/page_size);
	} else {
		*returned_size = size;
	}

	switch (protection) {
	case OS_PROT_READ:
		protect = PROT_READ;
		break;

	case (OS_PROT_WRITE | OS_PROT_READ):
		protect = (PROT_READ | PROT_WRITE);
		break;

	case (OS_PROT_EXEC | OS_PROT_READ):
		protect = (PROT_READ | PROT_EXEC);
		break;

	case (OS_PROT_WRITE | OS_PROT_READ | OS_PROT_EXEC):
		protect = (PROT_READ | PROT_EXEC | PROT_WRITE);
		break;

	default:
		return BTLOAD_INVALID_PROTECTION_ERR;
	}

	*mem_ptr = (void *)SYSCALL (mmap, (page_address,
					   *returned_size,
					   protect,
					   MAP_PRIVATE | MAP_FIXED | MAP_ANONYMOUS,
					   dummy_fd, dummy_offset));
	if (IS_SYS_ERROR ((long)*mem_ptr)) {
		return BTLOAD_MEM_ALLOC_FAIL_ERR;
	} else {
		return BTLOAD_OK;
	}
}


/*************************************************************************/
/* Func:   linux_free_memory                                             */
/* Inputs: page_address -  Address where memory is allocated.            */
/*         size -  Amount of memory to free (must be a multiple of page  */
/*                 size).                                                */
/* Output: None.                                                         */
/* Return Value:  Success or Fail code (see BTload_error.h)              */
/* Functionality: Free memory previously allocated by                    */
/*                linux_allocate_memory()                                */
/*************************************************************************/

FUNC BT_U64 linux_free_memory(BT_U64 page_address,
                              BT_U64 size)
{
	size_t  page_size;
	int     rc;

	/*
	 * Verify page_address to be page alligned, and then round the 
	 * requested size to be a multiply of the page size.
	 */

	page_size = linux_get_page_size();
	if (page_address % page_size) {
        	return BTLOAD_INVALID_ADDR_ERR;
    	}

	if (size % page_size) {
        	return BTLOAD_BAD_MEM_SIZE_ERR;
	}
    
	rc = SYSCALL (munmap, (page_address, size));
	if (rc < 0) {
		return BTLOAD_MEM_FREE_FAIL_ERR;
	} else {
		return BTLOAD_OK;
	}
}

/*************************************************************************/
/* Func:   linux_map_file_to_memory                                      */
/* Inputs: page_address -  Address where memory is allocated.            */
/*         size - Amount of memory to allocate (we round it to be a      */
/*                multiplier of the page size.                           */
/*         file_offset - Start mapping file at offset ...                */
/*         protection - Protection of mapped pages                       */
/*         shared -  Private or shared mapping?                          */
/*         fd - File descriptor of file                                  */
/* Output: returned_size - Actual amount of memory allocated.            */
/*         mem_ptr - Returned pointer to the allocated memory.           */
/* Return Value:  Success or Fail code (see BTload_error.h)              */
/* Functionality: Allocate memory from the system using anonymous mmap() */
/*                call.                                                  */
/*************************************************************************/

FUNC  BT_U64 linux_map_file_to_memory(BT_U64 page_address,
                                      BT_U64 size, 
                                      BT_U64 file_offset,
                                      BT_U64 protection,
                                      int   shared,
                                      int    fd,
                                      BT_U64 * returned_size,
                                      void   ** mem_ptr)
{
	BT_U64  page_size;

	/*
	 * Verify page_address to be page alligned, and then round the 
         * requested size to be a multiply of the page size.
 	 */

	page_size = (BT_U64) linux_get_page_size();
	if (page_address % page_size) {
		return BTLOAD_INVALID_ADDR_ERR;
	}

	if (size % page_size) {
		*returned_size = page_size * (1 + size/page_size);
	} else {
		*returned_size = size;
	}

#if DEBUG
	shared = 0;
#endif

	*mem_ptr = (void *)SYSCALL (mmap, (page_address,
					   *returned_size,
					   (int)protection,
					   (shared ? MAP_SHARED : MAP_PRIVATE) | MAP_FIXED,
					   fd, (off_t) file_offset));
	if (IS_SYS_ERROR((long)*mem_ptr)) {
		return BTLOAD_MEM_ALLOC_FAIL_ERR;
	} else {
		return BTLOAD_OK;
	}
}


/*******************************************************************************/
/* Func:   linux_pages_to_non_writable                                         */
/* Inputs: page_address - Address of first page in range to turn NON-WRITABLE. */
/*         num_pages - Amount of memory (in pages) to change to non-writable.  */
/*         is_executable - Indicate if the pages are executable.               */
/* Output: None.                                                               */
/* Return Value:  Success or Fail code (see BTload_error.h)                    */
/* Functionality: Change the protection of the specified pages to non-writable.*/
/*******************************************************************************/

FUNC  BT_U64 linux_pages_to_non_writable(BT_U64 page_address, BT_U64 num_pages, 
                                         int is_executable)
{
	int     rc;
	BT_U64  page_size;
	unsigned int   prot = is_executable? (PROT_READ | PROT_EXEC) : PROT_READ;

	/* Verify page_address to be page alligned. */
	page_size = (BT_U64) linux_get_page_size();
	if (page_address % page_size) {
		return BTLOAD_INVALID_ADDR_ERR;
	}


	rc = SYSCALL (mprotect, (page_address, page_size*num_pages, prot));
	if (IS_SYS_ERROR (rc)) {
		return BTLOAD_PROTECTION_CHANGE_ERR;
	}

	return BTLOAD_OK;
}


/*************************************************************************/
/* Func:   linux_create_map_file                                         */
/* Inputs: fd -  handle to the opened file.                              */
/*         low_offset - Offset in file we start to map (raw data offset  */
/*                      of the first section)                            */
/*         high_offset -  Offset in file we end mapping (raw data offset */
/*                        of the end of the last section).               */
/* Output: map_handle -  Handle of the mapping.                          */
/* Return Value:  Success or Fail code (see BTload_error.h)              */
/* Functionality: This is a dummy routine.                               */
/*************************************************************************/

FUNC BT_U64 linux_create_map_file(int fd, BT_U64 low_offset UNUSED, 
                                  BT_U64 high_offset UNUSED, int *map_handle)
{
	*map_handle = fd;
	return BTLOAD_OK;
}

