
/* IA32EXEC_OS_API HEADER FILE */
/* =========================== */

/* defines interface-related structures and apis         */
 
/* comments / names terminology:                         */
/*    btgeneric = IA32Exec.bin, the OS-independent part  */
/*    btlib     = the OS-interface library               */
/*
 *
 * Copyright (C) 2005       Yigal Zemach <yigal.zemach@intel.com>
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#ifndef IA32EXEC_OS_API_H
#define IA32EXEC_OS_API_H

#ifndef NODEBUG
#define OVERRIDE_TIA 1
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define MEM_READ    0x1
#define MEM_WRITE   0x2
#define MEM_EXECUTE 0x4
#define IS_MEM_ACCESSIBLE(permission) ((permission) != 0)

#define R13_FREE (-0x1)
#define R13_USED (-0x2)

#define BLOCK 0x0
#define CHECK 0x1

#define ACCESS_LOCK_OBJECT_SIZE 24


/* defines for initial memory allocation for either code or data. */

#define INITIAL_DATA_ALLOC      ((void *)0x1)
#define INITIAL_CODE_ALLOC      ((void *)0x2)

/* IA-32 interrupts */

#define IA32_DIVIDE_ERR_INTR        0
#define IA32_DEBUG_INTR             1
#define IA32_BREAKPOINT_INTR        3
#define IA32_OVERFLOW_INTR          4
#define IA32_BOUND_INTR             5
#define IA32_INV_OPCODE_INTR        6
#define IA32_DEV_NA_INTR            7
#define IA32_DOUBLE_FAULT_INTR      8
#define IA32_INV_TSS_INTR          10
#define IA32_NO_SEG_INTR           11
#define IA32_STK_SEG_FAULT_INTR    12
#define IA32_GEN_PROT_FAULT_INTR   13
#define IA32_PAGE_FAULT_INTR       14
#define IA32_MATH_FAULT_INTR       16
#define IA32_ALIGN_CHECK_INTR      17
#define IA32_MACHINE_CHECK_INTR    18
#define IA32_SIMD_INTR             19

/* BT exception codes used to communicate exception information between btgeneric and btlib:     */
/* code range 0-255 is reserved for IA-32 interrupt vector numbers. the vector numbers are       */
/* used to specify software interrupts (intn) only; the CPU-detected exceptions should be        */
/* encoded with a corresponding BT_EXCEPT_* value.                                               */
/* BT_NO_EXCEPT is a special code indicating either false exception or                           */
/* external interrupt (suspention)                                                               */
/* BT_EXCEPT_UNKNOWN code used to specify all exceptions unknown to IA-32 execution layer        */
/* all values in the bt_exception_code enumerator should be in the range                         */
/* [0, BT_MAX_EXCEPTION_CODE]; other values are reserved for internal use.                       */
enum bt_exception_code {
    BT_MAX_INTERRUPT_NUMBER = 0xFF,
    BT_NO_EXCEPT,
    BT_EXCEPT_UNKNOWN,
    BT_EXCEPT_ACCESS_VIOLATION,
    BT_EXCEPT_RNX_ACCESS_VIOLATION,
    BT_EXCEPT_DATATYPE_MISALIGNMENT,
    BT_EXCEPT_ARRAY_BOUNDS_EXCEEDED,
    BT_EXCEPT_FLT_DENORMAL_OPERAND,
    BT_EXCEPT_FLT_DIVIDE_BY_ZERO,
    BT_EXCEPT_FLT_INEXACT_RESULT,
    BT_EXCEPT_FLT_INVALID_OPERATION,
    BT_EXCEPT_FLT_OVERFLOW,
    BT_EXCEPT_FLT_UNDERFLOW,
    BT_EXCEPT_FLT_STACK_CHECK,
    BT_EXCEPT_INT_DIVIDE_BY_ZERO,
    BT_EXCEPT_INT_OVERFLOW,
    BT_EXCEPT_PRIV_INSTRUCTION,
    BT_EXCEPT_ILLEGAL_INSTRUCTION,
    BT_EXCEPT_FLOAT_MULTIPLE_FAULTS,
    BT_EXCEPT_FLOAT_MULTIPLE_TRAPS,
    BT_EXCEPT_STACK_OVERFLOW,
    BT_EXCEPT_GUARD_PAGE,
    BT_EXCEPT_BREAKPOINT,
    BT_EXCEPT_SINGLE_STEP
};

#define BT_MAX_EXCEPTION_CODE 0xFFF

typedef __u32  BT_EXCEPTION_CODE;

/* structure that represents interruption context in addition to thread CONTEXT_64 */
typedef struct bt_exception_record {
    BT_EXCEPTION_CODE exception_code;/* BT exception code                                          */
    __u64 ia64_iipa;                 /* interruption instruction previous address. 0 if unknown    */
    __u64 ia64_isr;                  /* interruption status register. UNKNOWN_ISR_VALUE if unknown */
} BT_EXCEPTION_RECORD;

#define UNKNOWN_ISR_VALUE ((__u64)(-1))

/* BT status codes used to communicate error information between btgeneric and btlib: */
enum bt_status_code {
    BT_STATUS_SUCCESS = 0,
    BT_STATUS_UNSUCCESSFUL,
    BT_STATUS_NO_MEMORY,
    BT_STATUS_ACCESS_VIOLATION,
    BT_STATUS_BAD_VERSION
};
typedef __u32  BT_STATUS_CODE;

/* API_VERSION type */
typedef struct{
           __u32 major_version;
           __u32 minor_version;
} API_VERSION;
/* current API major & minor versions */
#define MAJOR_VERSION_NUMBER 5
#define MINOR_VERSION_NUMBER 3

/* btg_flush_ia32_instruction_cache reason codes */
enum bt_flush_reason {
    BT_FLUSH_FORCE = 0, /* code modification */
    BT_FLUSH_FREE,      /* memory release    */
    BT_FLUSH_ALLOC,     /* memory allocation */
    BT_FLUSH_PROTECT    /* permission change */
}; 
typedef __u32  BT_FLUSH_REASON;

/* BT object handle (process, etc.) */
typedef void * BT_HANDLE;

/* static information: constants that can be queried once and are guaranteed to stay constant for the    */
/*                     entire process life-time                                                          */
typedef union bt_static_info {
    struct {
        __u32     jmp_buf_size;          /* size of buffer for setjmp / longjmp */
        __u16     number_of_logical_cpus; /* number of logical processors        */
        __u16     reserved;
    }       genuine;
    __u64     alignment8_forcing_dummy;
} BT_STATIC_INFO;

#define BT_STATIC_INFO_JMP_BUF_SIZE(_BT_STATIC_INFORMATION) \
    ((_BT_STATIC_INFORMATION).genuine.jmp_buf_size)

#define BT_STATIC_INFO_LOGICAL_CPUS_NUMBER(_BT_STATIC_INFORMATION) \
    ((_BT_STATIC_INFORMATION).genuine.number_of_logical_cpus)

#define BT_STATIC_INFO_QUERY_JMP_BUF_SIZE           0x1
#define BT_STATIC_INFO_QUERY_LOGICAL_CPUS_NUMBER    0x2

/* a pointer to plabel.                                  */
/* this type will be used to define the API_Table        */

typedef void(*PLABEL_PTR_TYPE)();

/* API table entry is simply a plable pointer...             */
typedef PLABEL_PTR_TYPE API_TABLE_ELEMENT_TYPE;
#define GET_PLABEL_PTR_FROM_API_TABLE_ELEMENT(TABLE_ELEM) (TABLE_ELEM)
#define SET_PLABEL_PTR_TO_API_TABLE_ELEMENT(TABLE_ELEM,PLBL_PTR) (TABLE_ELEM=PLBL_PTR)

/* API table TYPE is simply an array of API TABLE entries... */
typedef API_TABLE_ELEMENT_TYPE * API_TABLE_TYPE;
#define GET_PLABEL_PTR_FROM_API_TABLE(API_TABLE,INDEX) (GET_PLABEL_PTR_FROM_API_TABLE_ELEMENT(API_TABLE[INDEX]))
#define SET_PLABEL_PTR_TO_API_TABLE(API_TABLE,INDEX,PLBL_PTR) (SET_PLABEL_PTR_TO_API_TABLE_ELEMENT((API_TABLE[INDEX]),PLBL_PTR))

/*////////////////////////////////////////////////////////// */
/*/  IA-32 CONTEXT as passed between btgeneric and btlib /// */
/*////////////////////////////////////////////////////////// */
#define IA32_FP_STACK_BYTE_SIZE         80

#define CTX32_CRITICAL          (0x10001L) 
#define CTX32_GP                (0x10002L) 
#define CTX32_SEGMENT           (0x10004L) 
#define CTX32_FLOATING_POINT    (0x10008L) 
#define CTX32_DEBUG             (0x10010L) 
#define CTX32_XMM               (0x10020L) 
#define CTX32_BT_EXTENDED       (0x10040L) 


/*  IA-32 execution layer specific context */

typedef struct _BT_CONTEXT32_EXT {
    __u32   fs_base;
    __u32   gs_base;
    __u32   es_base; 
    __u32   gdt_base;
    __u32   idt_base;
    __u16   gdt_limit;
    __u16   idt_limit;
    __u16   ldt_selector;
} BT_CONTEXT32_EXT;
/* mask for significant buts in segment registers                            */
/* (bits beyond this mask are ignored in any segment registers comparison)   */
#define BT_CONTEXT32_SEGREG_MASK   0xffff

#define BT_MAXIMUM_SUPPORTED_EXTENSION      (512 + sizeof(BT_CONTEXT32_EXT))
#define BTEXT_INDEX_IN_EXT_REGS             512

typedef struct floating_point_state32 {
    __u32   control_word;
    __u32   status_word;
    __u32   tag_word;
    __u32   instruction_pointer_offset;
    __u32   instruction_pointer_selector;
    __u32   data_pointer_offset;
    __u32   data_pointer_selector;
    __u8    stack[IA32_FP_STACK_BYTE_SIZE];
    __u32   unused;
} FLOATING_POINT_STATE32;

typedef struct BTGENERIC_IA32_CONTEXT {
    __u32   composition_bits;
    __u32      dr0;
    __u32      dr1;
    __u32      dr2;
    __u32      dr3;
    __u32      dr6;
    __u32      dr7;
    FLOATING_POINT_STATE32 fpu_state;
    __u32   gs, fs, es, ds;
    __u32   edi;
    __u32   esi;
    __u32   ebx;
    __u32   edx;
    __u32   ecx;
    __u32   eax;
    __u32   ebp;
    __u32   eip;
    __u32   cs;
    __u32   eflags;
    __u32   esp;
    __u32   ss;
    __u8    xmm_and_bt_ext[BT_MAXIMUM_SUPPORTED_EXTENSION];
} BTGENERIC_IA32_CONTEXT;

/*** (void*) is used for some compilers (pointer casting causes alignment increasing) ***/
#define CONTEXT_EXTFIELD(context, name) \
        (((BT_CONTEXT32_EXT*)((void*)(&context.xmm_and_bt_ext[BTEXT_INDEX_IN_EXT_REGS])))->name)

/*//////////////////////////////////////////////////////////// */
/*/  64-bit CONTEXT as passed between btgeneric and btlib  /// */
/*//////////////////////////////////////////////////////////// */

#define CTX64_CRITICAL              (0x80001L) 
#define CTX64_STATIC_FLOATING_POINT (0x80002L)
#define CTX64_ROTATE_FLOATING_POINT (0x80004L)
#define CTX64_FLOATING_POINT        (0x80006L)
#define CTX64_INTEGER               (0x80008L)
#define CTX64_DEBUG                 (0x80010L)
#define CTX64_IA32_CRITICAL         (0x80020L)  

typedef struct BTGENERIC_IA64_CONTEXT {
    __u32 composition_bits;
    __u32 unused4[35];
    struct ia64_fpreg fp_static[30];
    struct ia64_fpreg fp_rotate[96];
    __u64 fpsr;
    __u64 int_static[31];
    __u64 nat_collection, predicates;
    __u64 br_regs[8];
    __u64 unat, lc, ec, ccv, dcr; 
    __u64 pfs, bsp, bspstore, rsc, rnat;
    __u64 ipsr, iip, ifs;
    __u64 fcr, eflags, csd, ssd, cflags, fsr, fir, fdr, unused8fdr;
} BTGENERIC_IA64_CONTEXT;

/* SSE related constants */

#define MXCSR_IE_STATUS_MASK        0x1
#define MXCSR_DE_STATUS_MASK        0x2
#define MXCSR_ZE_STATUS_MASK        0x4
#define MXCSR_OE_STATUS_MASK        0x8
#define MXCSR_UE_STATUS_MASK        0x10
#define MXCSR_PE_STATUS_MASK        0x20

#define MXCSR_INIT_VALUE            0x1f80

#define MXCSR_EXCEPTION_MASK_POS    7
#define MXCSR_INDEX_IN_EXT_REGS     24
#define XMM0_INDEX_IN_EXT_REGS      160

#define MXCSR_EXCEPTION_MASK        0x3f

#define MXCSR_INVALID_OPERATION_EXCEPTION_MASK            0x80
#define MXCSR_DENORMALIZED_OPERAND_EXCEPTION_MASK         0x100
#define MXCSR_ZERO_DIVIDE_EXCEPTION_MASK                  0x200
#define MXCSR_OVERFLOW_EXCEPTION_MASK                     0x400
#define MXCSR_UNDERFLOW_EXCEPTION_MASK                    0x800
#define MXCSR_PRECISION_EXCEPTION_MASK                    0x1000

#define MXCSR_EXCEPTION_MASKS  (  MXCSR_INVALID_OPERATION_EXCEPTION_MASK        \
                                | MXCSR_DENORMALIZED_OPERAND_EXCEPTION_MASK     \
                                | MXCSR_ZERO_DIVIDE_EXCEPTION_MASK              \
                                | MXCSR_OVERFLOW_EXCEPTION_MASK                 \
                                | MXCSR_UNDERFLOW_EXCEPTION_MASK                \
                                | MXCSR_PRECISION_EXCEPTION_MASK )


/* btg_canonize_suspend_context return values  */
enum bt_thread_suspend_state {
    SUSPEND_STATE_CONSISTENT, /* thread suspended in a consistent IA32 state;                            */
                              /* no changes in IA64Context                                               */

    SUSPEND_STATE_CANONIZED,  /* IA32 thread state canonized;                                            */
                              /* there are changes in IA64Context                                        */

    SUSPEND_STATE_READY_FOR_CANONIZATION, /* target thread is ready to canonize IA32 state by itself.    */
                                          /* IA64Context has been changed so that resumed thread         */
                                          /* will cannonize IA32 state and exit simulation               */

    BAD_SUSPEND_STATE,         /* recoverable error:                                                     */
                               /* thread can not be suspended in the current state                       */

    SUSPEND_STATE_INACCESIBLE  /* fatal error: thread can not be suspended                               */
};
    
typedef __u32 BT_THREAD_SUSPEND_STATE;

/* EMULATION_POLICY and following #defines used to set ordering and state policy for emulation in */
/* call to SET_EMULATION_POLICY    */
typedef struct emulation_policy_s {
    __u8  memory_order_policy;
    __u8  strict_state_policy;
} EMULATION_POLICY;


#define INVALID_ORDERING              0x0
#define NO_ORDERING                   0x1
#define PRESERVE_LOAD_STORE_ORDERING  0x2
#define USE_ORDERED_LD_ST             0x4
#define STATE_POLICY_REGULAR          0x0
#define STATE_POLICY_PRESERVE_EFLAGS  0x1
#define POLICY_UNCHANGED              0xFF

#ifdef VTUNE 
/* vtune stuff */

typedef enum {
    VTUNE_CALL_ID_CREATED=1,
    VTUNE_CALL_ID_DELETED,
    VTUNE_CALL_ID_FLUSH,
    VTUNE_CALL_ID_ENTER,
    VTUNE_CALL_ID_EXIT,
    VTUNE_CALL_ID_EVENT,
    VTUNE_BTGENERIC_LOADED
} VTUNE_CALL_ID_TYPE;

typedef enum {
    VTUNE_COLD_BLOCK=1,
    VTUNE_HOT_BLOCK,
    VTUNE_HOT_LOOP_BLOCK,
    VTUNE_COLD_WRITABLE_BLOCK,
    VTUNE_HOT_WRITABLE_BLOCK,
    VTUNE_HOT_LOOP_WRITABLE_BLOCK
} VTUNE_BLOCK_CATEGORY_TYPE;

typedef struct  {
    __u64 name;                           /* length is smaller then 256 chars         */
    VTUNE_BLOCK_CATEGORY_TYPE type;     /* vtune++ should be aware of this type, for breakdown */
    __u64 start;                          /* index, i.e. guaranteed not to repeat     */
    __u64 size;                           /* in bytes                                 */
    __u32 IA32start;                      /* IA32 address                             */
    __u64 traversal;                      /* an ID                                    */
    /*char *assembly;                   // if needed */
    __u64 reserved;                       /* unused                                   */
} VTUNE_BLOCK_TYPE;

typedef enum {
    ESTIMATED_TIME,
    CODE_SIZE,
    INST_COUNT
} VTUNE_EVENT_TYPE;
#endif 
/*////////////////////////////////////////////////////////////////////////////////////////////////////////// */
/* API FUNCTIONS: functions used for btgeneric <-> btlib communication and control flow. all functions are   */
/*                called indirectly through API table entries. the one and only exception is the first call  */
/*                of btlib to btgeneric which "boots" the process: get_api_info() function, which is exposed */
/*                as a symbol that can be accessed in the load of btgeneric.                                 */
/*                get_api_info() may fail due to incompatibility of versions.                                */
/*                after get_api_info() succeeds, the tables are set and each component can call the entries  */
/*                that the other component implements.                                                       */
/*////////////////////////////////////////////////////////////////////////////////////////////////////////// */
/*  BTGENERIC APIS:  functions called by btlib and implemented by btgeneric.                                 */
/*////////////////////////////////////////////////////////////////////////////////////////////////////////// */

/*////////////////////////////////////////////////////////////////////////////////////////////////////////// */
/* pointer to "get_api_info" func */
typedef BT_STATUS_CODE (*PTR_GET_API_INFO_FUNC)(API_VERSION *array_of_api_versions, __u32 num_of_api_versions, __s32 *index_of_accepted_api_version, API_TABLE_TYPE **btgeneric_api_table);
#define BTGENERIC(api_name) (*(btl_place_holder_table[IDX_BTGENERIC_##api_name]))
#define BTGENERIC_(TYPE,api_name) (*(TYPE (*)())(btl_place_holder_table[IDX_BTGENERIC_##api_name]))


/* BTGENERIC APIS */
/* ============== */

#define  BTGENERIC_START(api_version_used, btlib_api_table, btgeneric_address, btgeneric_end, glst_offset, btgeneric_tls_size_p, btgeneric_context_offset_p)\
         BTGENERIC_(BT_STATUS_CODE,START)((API_VERSION*)(api_version_used), (API_TABLE_TYPE*)(btlib_api_table), (void*)(btgeneric_address), (void*)(btgeneric_end), (int)(glst_offset), (__u32*)(btgeneric_tls_size_p), (__u32*)(btgeneric_context_offset_p))
#define  IDX_BTGENERIC_START 0

#define  BTGENERIC_THREAD_INIT(glst_memory, ia32_stack_pointer, FS_offset)\
         BTGENERIC_(BT_STATUS_CODE,THREAD_INIT)((void *)(glst_memory), (__u32)(ia32_stack_pointer), (__u64)(FS_offset))
#define  IDX_BTGENERIC_THREAD_INIT 1

#define  BTGENERIC_RUN()\
         BTGENERIC(RUN)()
#define  IDX_BTGENERIC_RUN 2

#define  BTGENERIC_EXIT_SIMULATION_REQUEST()\
         BTGENERIC(EXIT_SIMULATION_REQUEST)()
#define  IDX_BTGENERIC_EXIT_SIMULATION_REQUEST 3

#define  BTGENERIC_THREAD_TERMINATED()\
         BTGENERIC(THREAD_TERMINATED)()
#define  IDX_BTGENERIC_THREAD_TERMINATED 4

#define  BTGENERIC_THREAD_ABORTED(thread_id)\
         BTGENERIC(THREAD_ABORTED)((__u64)(thread_id))
#define  IDX_BTGENERIC_THREAD_ABORTED 5

#define  BTGENERIC_PROCESS_TERM()\
         BTGENERIC(PROCESS_TERM)()
#define  IDX_BTGENERIC_PROCESS_TERM 6

/*       Unused index 7 */

#define  BTGENERIC_IA32_CONTEXT_SET(glst_memory, context)\
         BTGENERIC_(BT_STATUS_CODE,IA32_CONTEXT_SET)((void *)(glst_memory), (const BTGENERIC_IA32_CONTEXT *)(context))
#define  IDX_BTGENERIC_IA32_CONTEXT_SET 8

#define  BTGENERIC_IA32_CONTEXT_GET(glst_memory, context)\
         BTGENERIC_(BT_STATUS_CODE,IA32_CONTEXT_GET)((void *)(glst_memory), (BTGENERIC_IA32_CONTEXT *)(context))
#define  IDX_BTGENERIC_IA32_CONTEXT_GET 9

#define  BTGENERIC_IA32_CONTEXT_SET_REMOTE(process_handle, glst_memory, context)\
         BTGENERIC_(BT_STATUS_CODE,IA32_CONTEXT_SET_REMOTE)((BT_HANDLE)(process_handle), (void *)(glst_memory), (const BTGENERIC_IA32_CONTEXT *)(context))
#define  IDX_BTGENERIC_IA32_CONTEXT_SET_REMOTE 10

#define  BTGENERIC_IA32_CONTEXT_GET_REMOTE(process_handle, glst_memory, context)\
         BTGENERIC_(BT_STATUS_CODE,IA32_CONTEXT_GET_REMOTE)((BT_HANDLE)(process_handle), (void *)(glst_memory), (BTGENERIC_IA32_CONTEXT *)(context))
#define  IDX_BTGENERIC_IA32_CONTEXT_GET_REMOTE 11

#define  BTGENERIC_IA32_CANONIZE_CONTEXT(glst_memory, ia64context, exception_record_p)\
         BTGENERIC_(BT_EXCEPTION_CODE,IA32_CANONIZE_CONTEXT)((void *)(glst_memory), (const BTGENERIC_IA64_CONTEXT *)(ia64context), (const BT_EXCEPTION_RECORD *)(exception_record_p))
#define  IDX_BTGENERIC_IA32_CANONIZE_CONTEXT 12

#define  BTGENERIC_CANONIZE_SUSPEND_CONTEXT_REMOTE(process_handle, glst_memory, ia64context, prev_suspend_count)\
         BTGENERIC_(BT_THREAD_SUSPEND_STATE,CANONIZE_SUSPEND_CONTEXT_REMOTE)((BT_HANDLE)(process_handle), (void *)(glst_memory), (BTGENERIC_IA64_CONTEXT *)(ia64context), (__u64)(prev_suspend_count))
#define  IDX_BTGENERIC_CANONIZE_SUSPEND_CONTEXT_REMOTE 13

#define  BTGENERIC_REPORT_LOAD(location, size, name, name_size)\
         BTGENERIC(REPORT_LOAD)((__u64 *)(location), (__u32)(size), (__u8 *)(name), (__u32)(name_size))
#define  IDX_BTGENERIC_REPORT_LOAD 14

#define  BTGENERIC_REPORT_UNLOAD(location, size, name)\
         BTGENERIC(REPORT_UNLOAD)((__u64 *)(location), (__u32)(size), (__u8 *)(name))
#define  IDX_BTGENERIC_REPORT_UNLOAD 15

#define  BTGENERIC_NOTIFY_CHANGE_PERMISSION_REQUEST(page_start, num_pages, permissions)\
         BTGENERIC(NOTIFY_CHANGE_PERMISSION_REQUEST)((void *)(page_start), (__u32)(num_pages), (__u64)(permissions))
#define  IDX_BTGENERIC_NOTIFY_CHANGE_PERMISSION_REQUEST 16

#define  BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE(address, size, reason)\
         BTGENERIC(FLUSH_IA32_INSTRUCTION_CACHE)((void *)(address), (__u32)(size), (BT_FLUSH_REASON)(reason))
#define  IDX_BTGENERIC_FLUSH_IA32_INSTRUCTION_CACHE 17

#define  BTGENERIC_DEBUG_SETTINGS(argc, argv)\
         BTGENERIC_(BT_STATUS_CODE,DEBUG_SETTINGS)((int)(argc), (char **)(argv))
#define  IDX_BTGENERIC_DEBUG_SETTINGS 18

#define  BTGENERIC_CHECK_SUSPEND_CONTEXT(glst_memory, ia64context)\
         BTGENERIC_(BT_THREAD_SUSPEND_STATE,CHECK_SUSPEND_CONTEXT)((void *)(glst_memory), (const BTGENERIC_IA64_CONTEXT *)(ia64context))
#define  IDX_BTGENERIC_CHECK_SUSPEND_CONTEXT 19

#define  BTGENERIC_EXCEPTION_DEBUG_PRINT()\
         BTGENERIC(EXCEPTION_DEBUG_PRINT)()
#define  IDX_BTGENERIC_EXCEPTION_DEBUG_PRINT 20

#define  BTGENERIC_NOTIFY_EXIT()\
         BTGENERIC(NOTIFY_EXIT)()
#define  IDX_BTGENERIC_NOTIFY_EXIT 21

#define  BTGENERIC_CANONIZE_SUSPEND_CONTEXT(glst_memory, ia64context, prev_suspend_count)\
         BTGENERIC_(BT_THREAD_SUSPEND_STATE,CANONIZE_SUSPEND_CONTEXT)((void *)(glst_memory), (BTGENERIC_IA64_CONTEXT *)(ia64context), (__u64)(prev_suspend_count))
#define  IDX_BTGENERIC_CANONIZE_SUSPEND_CONTEXT 22

#define  BTGENERIC_NOTIFY_PREPARE_EXIT()\
         BTGENERIC(NOTIFY_PREPARE_EXIT)()
#define  IDX_BTGENERIC_NOTIFY_PREPARE_EXIT 23

#define  BTGENERIC_FREEZE()\
         BTGENERIC(FREEZE)()
#define  IDX_BTGENERIC_FREEZE 24

#define  BTGENERIC_UNFREEZE()\
         BTGENERIC(UNFREEZE)()
#define  IDX_BTGENERIC_UNFREEZE 25

#define  BTGENERIC_CHANGE_THREAD_IDENTITY()\
         BTGENERIC_(BT_STATUS_CODE,CHANGE_THREAD_IDENTITY)()
#define  IDX_BTGENERIC_CHANGE_THREAD_IDENTITY 26

#define  BTGENERIC_SET_EMULATION_POLICY(policy)\
         BTGENERIC(SET_EMULATION_POLICY)((EMULATION_POLICY*)(policy))
#define  IDX_BTGENERIC_SET_EMULATION_POLICY 27

#ifdef OVERRIDE_TIA
#define  BTGENERIC_USE_OVERRIDE_TIA(ovr_tia_size, ovr_tia_buffer)\
         BTGENERIC(USE_OVERRIDE_TIA)((unsigned int)(ovr_tia_size), (unsigned char *)(ovr_tia_buffer))
#define  IDX_BTGENERIC_USE_OVERRIDE_TIA 28

#endif /* OVERRIDE_TIA */


#define  NO_OF_BTGENERIC_APIS 29

/* BTLIB APIS */
/* ========== */

extern   __u64 btlib_api_get_thread_id(void);
#define  IDX_BTLIB_GET_THREAD_ID 0

extern   void btlib_api_ia32_reenter(BTGENERIC_IA32_CONTEXT * ia32context);
#define  IDX_BTLIB_IA32_REENTER 1

extern   void btlib_api_ia32_lcall(BTGENERIC_IA32_CONTEXT * ia32context, __u32 return_address, __u32 target_address);
#define  IDX_BTLIB_IA32_LCALL 2

extern   void btlib_api_ia32_interrupt(BTGENERIC_IA32_CONTEXT * ia32context, BT_EXCEPTION_CODE exception_code, __u32 return_address);
#define  IDX_BTLIB_IA32_INTERRUPT 3

extern   void btlib_api_ia32_jmp_ia64(BTGENERIC_IA32_CONTEXT * ia32context, __u32 return_address, __u32 target_address);
#define  IDX_BTLIB_IA32_JMP_IA64 4

extern   void btlib_api_ia32_sysenter(BTGENERIC_IA32_CONTEXT * ia32context);
#define  IDX_BTLIB_IA32_SYSENTER 5

extern   void btlib_api_lock_signals(void);
#define  IDX_BTLIB_LOCK_SIGNALS 6

extern   void btlib_api_unlock_signals(void);
#define  IDX_BTLIB_UNLOCK_SIGNALS 7

extern   void * btlib_api_memory_alloc(void * start_address, __u32 size, __u64 prot);
#define  IDX_BTLIB_MEMORY_ALLOC 8

extern   BT_STATUS_CODE btlib_api_memory_free(void  * start_address, __u32 size);
#define  IDX_BTLIB_MEMORY_FREE 9

extern   __u32 btlib_api_memory_page_size(void);
#define  IDX_BTLIB_MEMORY_PAGE_SIZE 10

extern   __u64 btlib_api_memory_change_permissions(void * start_address, __u32 size, __u64 prot);
#define  IDX_BTLIB_MEMORY_CHANGE_PERMISSIONS 11

extern   __u64 btlib_api_memory_query_permissions(void * address, void ** p_region_start, __u32 * p_region_size);
#define  IDX_BTLIB_MEMORY_QUERY_PERMISSIONS 12

extern   BT_STATUS_CODE btlib_api_memory_read_remote(BT_HANDLE process_handle, void * base_address, void * buffer, __u32 requested_size);
#define  IDX_BTLIB_MEMORY_READ_REMOTE 13

extern   BT_STATUS_CODE btlib_api_memory_write_remote(BT_HANDLE process_handle, void * base_address, const void * buffer, __u32 requested_size);
#define  IDX_BTLIB_MEMORY_WRITE_REMOTE 14

/*       Unused index 15 */

/*       Unused index 16 */

extern   BT_STATUS_CODE btlib_api_suspend_thread(__u64 thread_id, __u32 try_counter);
#define  IDX_BTLIB_SUSPEND_THREAD 17

extern   BT_STATUS_CODE btlib_api_resume_thread(__u64 thread_id);
#define  IDX_BTLIB_RESUME_THREAD 18

extern   BT_STATUS_CODE btlib_api_init_access_lock(void * lock);
#define  IDX_BTLIB_INIT_ACCESS_LOCK 19

extern   BT_STATUS_CODE btlib_api_lock_access(void * lock, __u64 flag);
#define  IDX_BTLIB_LOCK_ACCESS 20

extern   void btlib_api_unlock_access(void * lock);
#define  IDX_BTLIB_UNLOCK_ACCESS 21

extern   void btlib_api_invalidate_access_lock(void * lock);
#define  IDX_BTLIB_INVALIDATE_ACCESS_LOCK 22

extern   __u64 btlib_api_query_static_info(BT_STATIC_INFO * static_info_p, __u64 requested_fields);
#define  IDX_BTLIB_QUERY_STATIC_INFO 23

extern   void btlib_api_yield_thread_execution(void);
#define  IDX_BTLIB_YIELD_THREAD_EXECUTION 24

extern   void btlib_api_flush_ia64_instruction_cache(void* address, __u32 length);
#define  IDX_BTLIB_FLUSH_IA64_INSTRUCTION_CACHE 25

#define  IDX_BTLIB_SETJMP 26

#define  IDX_BTLIB_LONGJMP 27

extern   void btlib_api_abort(void);
#define  IDX_BTLIB_ABORT 28

/*#ifndef NODEBUG // - to be uncommented */
extern   void btlib_api_debug_print(__u8 * buffer);
#define  IDX_BTLIB_DEBUG_PRINT 29


#define  IDX_BTLIB_VTUNE_CODE_CREATED 30


#define  IDX_BTLIB_VTUNE_CODE_DELETED 31


#define  IDX_BTLIB_VTUNE_ENTERING_DYNAMIC_CODE 32


#define  IDX_BTLIB_VTUNE_EXITING_DYNAMIC_CODE 33


#define  IDX_BTLIB_VTUNE_CODE_TO_TIA_DMP_FILE 34

#define  IDX_BTLIB_PSEUDO_OPEN_FILE 35

#define  IDX_BTLIB_PSEUDO_CLOSE_FILE 36

#define  IDX_BTLIB_PSEUDO_WRITE_FILE 37

/*#endif          // - to be uncommented */

#define  NO_OF_BTLIB_APIS 38

extern PLABEL_PTR_TYPE btl_place_holder_table[NO_OF_BTGENERIC_APIS];

#ifdef __cplusplus
}
#endif

#endif /* IA32EXEC_OS_API_H */
