/*  
 * ia32x_utils_tls.c
 *  
 * allocate and init thread local storage needed by gcc
 * 
*/
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
/* tls stuff */
/* based on __libc_setup_tls in sysdeps/generic/libc-tls.c */

#include "../ia32x_utils.h"
#include "../include/ia32x_utils_internal.h"

/* dtv and tcbhead are copied from nptl sources sysdeps/ia64/tls.h */
/* Type of the dtv.  */
typedef union dtv {
	unsigned long counter;
	void *pointer;
} dtv_t;
/* Type of the TCB.  */
typedef struct tcbhead {
	dtv_t *dtv;                   /* Vector of pointers to TLS data.  */
	void * private;
} tcbhead_t;

#define INITIAL_TLS_ALIGN __alignof__ (tcbhead_t)
#define TCB_SIZE sizeof (tcbhead_t)


static unsigned long m_filesz = 0;
static void * m_initimage = NULL;
static unsigned long m_tls_size = 0;
static unsigned long m_tls_align = 0;
/*
 * layout of the tls is
 *
 *                     r13
 *              PRE-   |
 *               TCB   |   TCB       TSD
 *            |--------|----|----|------------------|
 *                       |       ^
 *                       |       |
 *                       |       --------|
 *                       |               |
 *            	         ------>|------|------|
 *               	              DTV
 *
 */
/*****************************************************************************/
/*****************************************************************************/
void get_tls_static_info (unsigned long *sizep, unsigned long *alignp)
{
	if (sizep != NULL) {
		*sizep = m_tls_size + TCB_SIZE;
	}
	if (alignp != NULL) {
		*alignp = m_tls_align;
	}
}
/*****************************************************************************/
/* allocate and init the dtv */
/*****************************************************************************/
long allocate_and_init_dtv (tcbhead_t * tcb)
{
	tcb->private = NULL;
	if (m_tls_size > 0) {
		dtv_t * dtv;

		/* prepare the dtv */
		dtv = calloc (sizeof (dtv_t), 2ul);
		if (dtv == NULL) {
			return -ENOMEM;
		}
		/* dtv[0].counter = 0; - no need because we used calloc */
		dtv[1].pointer = (void*) ROUNDUP ((unsigned long)tcb + TCB_SIZE,
				m_tls_align);

		/* copy the initialized data. all the rest will have 
		 * zero value because we allocated the mem with calloc */
		if (m_initimage != NULL) {
			memcpy (dtv[1].pointer, m_initimage, m_filesz);
		}

		/* install the dtv */
		tcb->dtv = dtv;
	} else {
		tcb->dtv = NULL;
	}

	return 0;
}
/*****************************************************************************/
/* free the dtv */
/*****************************************************************************/
long deallocate_dtv (tcbhead_t * tcb)
{
	if (tcb->dtv != NULL) {
		free (tcb->dtv);
	}
	return 0;
}
/*****************************************************************************/
/*****************************************************************************/
long allocate_initial_tls ()
{
	Elf64_Phdr * phdr;
	void * tls;
	tcbhead_t * tcb;
	long retval;

	/* find the size and alignment of tls section */
	if (g_at_phdr != NULL) {
		for (phdr = (Elf64_Phdr*)g_at_phdr;
				phdr < &(((Elf64_Phdr*)g_at_phdr)[g_at_phnum]);
				++phdr) {
			if (phdr->p_type == PT_TLS) {
				m_tls_size = phdr->p_memsz;
				m_tls_align = phdr->p_align;
				m_initimage = (void*)phdr->p_vaddr;
				m_filesz = phdr->p_filesz;
				break;
			}
		}
	}
	m_tls_size += TCB_SIZE + TLS_PRE_TCB_SIZE + m_tls_align + INITIAL_TLS_ALIGN;

	/* allocate a minimal tls - this memory will never get
	 * deallocated. it's not a big thing since it's only
	 * done for the first thread */

	tls = calloc (m_tls_size, 1ul);
	if (tls == NULL) {
		return -ENOMEM;
	}
	tcb = (void*)ROUNDUP ((unsigned long)(tls + TLS_PRE_TCB_SIZE),
		       __alignof__ (tcbhead_t));

	retval = allocate_and_init_dtv (tcb);
	if (IS_SYS_ERROR (retval)) {
		return retval;
	}
	
	/* set r13 to point at the tcb */
	__asm__ __volatile__ ("add r13=r0, %0;;\n"::"r"(tcb));

	return 0;
}





