/* 
 * mutex.c
 *
 * implement the mutex services as declared in mutex.h
 *
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "../ia32x_utils.h"
#include "../include/mutex_internal.h"
#include "../include/ia32x_utils_internal.h"


#define SYS_futex __NR_futex
#ifndef FUTEX_WAIT
#define FUTEX_WAIT (0)
#endif
#ifndef FUTEX_WAKE
#define FUTEX_WAKE (1)
#endif

static int m_has_futex = -1;

#define lll_futex_clobbers \
  "out5", "out6", "out7",                             \
  /* Non-stacked integer registers, minus r8, r10, r15.  */           \
  "r2", "r3", "r9", "r11", "r12", "r13", "r14", "r16", "r17", "r18",          \
  "r19", "r20", "r21", "r22", "r23", "r24", "r25", "r26", "r27",          \
  "r28", "r29", "r30", "r31",                             \
  /* Predicate registers.  */                             \
  "p6", "p7", "p8", "p9", "p10", "p11", "p12", "p13", "p14", "p15",       \
  /* Non-rotating fp registers.  */                       \
  "f6", "f7", "f8", "f9", "f10", "f11", "f12", "f13", "f14", "f15",       \
  /* Branch registers.  */                            \
  "b6", "b7",                                     \
  "memory"

#define lll_futex_wait(ftx, val) lll_futex_timed_wait (ftx, val, 0)

#define lll_futex_timed_wait(ftx, val, timespec) \
  ({                                          \
     register long int __o0 asm ("out0") = (long int) (ftx);            \
     register long int __o1 asm ("out1") = FUTEX_WAIT;                \
     register int __o2 asm ("out2") = (int) (val);                \
     register long int __o3 asm ("out3") = (long int) (timespec);         \
     register long int __r8 asm ("r8");                       \
     register long int __r10 asm ("r10");                     \
     register long int __r15 asm ("r15") = SYS_futex;                 \
                                          \
     __asm __volatile ("break %7;;"                       \
               : "=r" (__r8), "=r" (__r10), "=r" (__r15),         \
             "=r" (__o0), "=r" (__o1), "=r" (__o2), "=r" (__o3)   \
               : "i" (0x100000), "2" (__r15), "3" (__o0), "4" (__o1), \
             "5" (__o2), "6" (__o3)                   \
               : "out4", lll_futex_clobbers);                 \
     __r10 == -1 ? -__r8 : __r8;                          \
  })

#define lll_futex_wake(futex, nr) \
  ({                                          \
     register long int __o0 asm ("out0") = (long int) (futex);            \
     register long int __o1 asm ("out1") = FUTEX_WAKE;                \
     register int __o2 asm ("out2") = (int) (nr);                 \
     register long int __r8 asm ("r8");                       \
     register long int __r10 asm ("r10");                     \
     register long int __r15 asm ("r15") = SYS_futex;                 \
                                          \
     __asm __volatile ("break %6;;"                       \
               : "=r" (__r8), "=r" (__r10), "=r" (__r15),         \
             "=r" (__o0), "=r" (__o1), "=r" (__o2)            \
               : "i" (0x100000), "2" (__r15), "3" (__o0), "4" (__o1), \
             "5" (__o2)                       \
               : "out3", "out4", lll_futex_clobbers);             \
     __r10 == -1 ? -__r8 : __r8;                          \
  })


/*****************************************************************************/
/*****************************************************************************/
static void __lll_lock_wait (int * ftx)
{
	do {
		int	oldval = __sync_val_compare_and_swap_si (ftx, 1, 2);
		if (oldval != 0) {
			lll_futex_wait (ftx, 2);
		}
	} while (__sync_bool_compare_and_swap_si (ftx, 0, 2) != 1);
}




/*****************************************************************************/
/* return 0 on success */
/*****************************************************************************/
static inline int __attribute__ ((always_inline)) __lll_mutex_trylock (int *ftx)
{
	int is_free = __sync_bool_compare_and_swap_si (ftx, 0, 1);
	return ! is_free; /* is_free ? 0 : 1; */
}

/*****************************************************************************/
/*****************************************************************************/
static inline void __attribute__ ((always_inline)) __lll_mutex_lock (int *ftx)
{
	int is_free = __sync_bool_compare_and_swap_si (ftx, 0, 1);
	if ( ! is_free ) {
		__lll_lock_wait (ftx);
	}
}

/*****************************************************************************/
/*****************************************************************************/
static inline void __attribute__ ((always_inline)) __lll_mutex_unlock (int *ftx)
{
	int is_waiting = __sync_lock_test_and_set_si (ftx, 0);
	__sync_synchronize ();
	if (is_waiting > 1) {
		lll_futex_wake (ftx, 1);
	}
}

/*****************************************************************************/
/*****************************************************************************/
static int ftx_mutex_lock (mutex_t * mutex)
{
	register unsigned long _tp asm ("r13");
	/* Check whether we already hold the mutex.  */
	if (mutex->__data.__owner == _tp) {
		/* Just bump the counter.  */
		if (mutex->__data.__count + 1 == 0)
			/* Overflow of the counter.  */
			return -EAGAIN;

		++mutex->__data.__count;
	} else {
		/* We have to get the mutex.  */
		__lll_mutex_lock (&mutex->__data.__lock);

		/* Record the ownership.  */
		mutex->__data.__owner = _tp;
		mutex->__data.__count = 1;
	}
	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static int ftx_mutex_unlock (mutex_t * mutex)
{
	register unsigned long _tp asm ("r13");
	if (mutex->__data.__owner != _tp)
		return -EPERM;

	if (--mutex->__data.__count != 0)
		/* We still hold the mutex.  */
		return 0;
	/* Always reset the owner field.  */
	mutex->__data.__owner = 0ul;

	/* Unlock.  */
	__lll_mutex_unlock (&mutex->__data.__lock);

	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static int ftx_mutex_init (mutex_t * mutex)
{
	mutex->__data.__lock = 0;
	mutex->__data.__count = 0;
	mutex->__data.__owner = 0;
	return 0;
}

/*****************************************************************************/
/* return 0 on success */
/*****************************************************************************/
static int ftx_mutex_trylock (mutex_t * mutex)
{
	register unsigned long _tp asm ("r13");
	/* Check whether we already hold the mutex.  */
	if (mutex->__data.__owner == _tp) {
		/* Just bump the counter.  */
		if (mutex->__data.__count + 1 == 0)
			/* Overflow of the counter.  */
			return -EAGAIN;

		++mutex->__data.__count;
		return 0;
	}

	if (__lll_mutex_trylock (&mutex->__data.__lock) == 0) {
		/* Record the ownership.  */
		mutex->__data.__owner = _tp;
		mutex->__data.__count = 1;
		return 0;
	}
	return -EBUSY;
}

/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_lock (mutex_t * mutex)
{
	int cont = 1;
	unsigned long locked;
	register unsigned long _tp asm ("r13");

	do {
		locked = __sync_val_compare_and_swap_di (&mutex->__data.__owner, 0l, (long)_tp);
		__sync_synchronize ();
		if (locked == _tp) {
			/* recurese mutex */
			if (mutex->__data.__count + 1 == 0) {
				return -EAGAIN;
			}
			mutex->__data.__count ++;
			cont = 0;
		} else if (locked == 0) {
			mutex->__data.__count = 1;
			cont = 0;
		} else {
			/* mutex is busy */
			SYSCALL (sched_yield, ());
		}
	} while (cont);

	return 0;
}

/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_unlock (mutex_t * mutex)
{
	int count;
	register unsigned long _tp asm ("r13");

	if (mutex->__data.__owner != _tp) {
		return -EPERM;
	}
	count = -- mutex->__data.__count;
	__sync_synchronize ();
	if (count == 0) {
		mutex->__data.__owner = 0;
	}
	return 0;
}
/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_init (mutex_t * mutex)
{
	mutex->__data.__lock = 0;
	mutex->__data.__count = 0;
	mutex->__data.__owner = 0;
	return 0;
}
/*****************************************************************************/
/*****************************************************************************/
static int spl_mutex_trylock (mutex_t * mutex)
{
	unsigned long locked;
	int ret = 0;
	register unsigned long _tp asm ("r13");

	locked = __sync_val_compare_and_swap_di (&mutex->__data.__owner, 0l, (long)_tp);
	__sync_synchronize ();
	if (locked == _tp) {
		/* recurese mutex */
		if (mutex->__data.__count + 1 == 0) {
			ret = -EAGAIN;
		} else {
			mutex->__data.__count ++;
		}
	} else if (locked == 0) {
		mutex->__data.__count = 1;
	} else {
		/* mutex is busy */
		ret = -EBUSY;
	}

	return ret;
}

/*****************************************************************************/
/* export functions */
/*****************************************************************************/
int mutex_module_init ()
{
	int ret = SYSCALL (futex, ((int*)1, FUTEX_WAKE, 1, NULL, NULL));
	if (ret == -ENOSYS) {
		m_has_futex = 0;
	} else {
		m_has_futex = 1;
	}
	DBPRINT (0, "ret=%d m_has_futex=%d\n", ret, m_has_futex);
	return 0;
}
int mutex_lock (mutex_handle mh) {
	mutex_t * m = (mutex_t *)mh;
	DBPRINT (0, "enter: mh=%p lock=%d count=%d owner=%lx\n",
			mh, m->__data.__lock, m->__data.__count, m->__data.__owner);
	if (m_has_futex) {
		return ftx_mutex_lock (m);
	} else {
		return spl_mutex_lock (m);
	}
}
int mutex_unlock (mutex_handle mh) {
	mutex_t * m = (mutex_t *)mh;
	DBPRINT (0, "enter: mh=%p lock=%d count=%d owner=%lx\n",
			mh, m->__data.__lock, m->__data.__count, m->__data.__owner);
	if (m_has_futex) {
		return ftx_mutex_unlock (m);
	} else {
		return spl_mutex_unlock (m);
	}
}
int mutex_init (mutex_handle mh) {
	mutex_t * m = (mutex_t *)mh;
	if (m_has_futex) {
		return ftx_mutex_init (m);
	} else {
		return spl_mutex_init (m);
	}
}
int mutex_trylock (mutex_handle mh) {
	mutex_t * m = (mutex_t *)mh;
	DBPRINT (0, "enter: mh=%p lock=%d count=%d owner=%lx\n",
			mh, m->__data.__lock, m->__data.__count, m->__data.__owner);
	if (m_has_futex) {
		return ftx_mutex_trylock (m);
	} else {
		return spl_mutex_trylock (m);
	}
}
unsigned long get_mutex_owner (mutex_handle mh) {
	mutex_t * m = (mutex_t *)mh;
	return m->__data.__owner;
}
mutex_handle mutex_create ()
{
	mutex_t * m = malloc (sizeof (mutex_t));
	return (mutex_handle)m;
}
void mutex_destroy (mutex_handle mh)
{
	free (mh);
}
