/* 
 * sprintf.c
 * a basic and limited implementation of sprintf
 */
 /*
 *
 * This file has been derived from GNU libc. Please see COPYING.LIBC
 * for more information
 *
 */
 /*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * or view it at http://www.gnu.org/licenses/lgpl.html
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/

#include "../ia32x_utils.h"

static unsigned long ulong2string (char * dest, unsigned long src, long base);
static unsigned long long2string (char * dest, long src, long base);
static inline char * zero_pad (char * p, unsigned long len, unsigned long min_width);
static unsigned long string2ulong (const char * p);

int sprintf(char *str, const char *format, ...)
{
	char * p = str;
	const char * pf = format;
	va_list ap;
	int state = 0;
	unsigned long len;
	unsigned long min_width = 0;

	unsigned long ul;
	long l;
	char * s;

	va_start (ap, format);

	while (1) {

		if (*pf == '\0') {
			break;
		}

		switch (state) {
			case 0:
				if (*pf == '%') {
					state = 1;
				} else {
					*p = *pf;
					p++;
				}
				break;
			case 1:
				if (*pf == '%') {
					*p = *pf;
					p++;
					state = 0;
				} else if (*pf == 'l') {
					state = 2;
				} else if (*pf == '0') {
					/* zero padding */
					/* get the min width of the field */
					min_width = string2ulong (pf + 1);
					assert (min_width != 0); 
					/* skip the min-width specification */
					while ((*(pf + 1) >= '0') && (*(pf + 1) <= '9')) {
						pf ++;
					}
					/* continue with state 1 */
				} else {
					switch (*pf) {
						case 'd':
							l = va_arg (ap, int);
							len = long2string (p, l, 10l);
							p = zero_pad (p, len, min_width);
							break;
						case 'u':
							ul = va_arg (ap, unsigned int);
							len = ulong2string (p, ul, 10l);
							p = zero_pad (p, len, min_width);
							break;
						case 'x':
							ul = va_arg (ap, unsigned int);
							len = ulong2string (p, ul, 16l);
							p = zero_pad (p, len, min_width);
							break;
						case 'c':
							l = va_arg (ap, int);
							*p++ = (char)l;
							break;
						case 's':
							s = va_arg (ap, char *);
							strcpy (p, s);
							p += strlen (s);
							break;
						case 'o':
							ul = va_arg (ap, unsigned long);
							len = ulong2string(p, ul, 8l);
							p += len;
							break;
						case 'p':
							ul = va_arg (ap, unsigned long);
							len = ulong2string(p, ul, 16l);
							p += len;
							break;
						default:
							assert ( ! "bt_sprintf: unknown type 1");
					}
					state=0;
					min_width = 0;
				}
				break;
			case 2:
				switch (*pf) {
					case 'd':
						l = va_arg (ap, long);
						len = long2string (p, l, 10l);
						p = zero_pad (p, len, min_width);
						break;
					case 'u':
						ul = va_arg (ap, unsigned long);
						len = ulong2string(p, ul, 10l);
						p = zero_pad (p, len, min_width);
						break;
					case 'x':
						ul = va_arg (ap, unsigned long);
						len = ulong2string(p, ul, 16l);
						p = zero_pad (p, len, min_width);
						break;
					default:
						assert ( ! "bt_sprintf: unknown type 2");
				}
				state=0;
				min_width = 0;
				break;
			default:
				assert ( ! "should not get here");
		}
		pf++;
	}
	va_end (ap);
	*p = '\0';
	return (int)(p-str);
}

static char digits [] = {'0','1','2','3','4','5','6','7','8','9',
						'a','b','c','d','e','f'};


/*****************************************************************************/
/* convert unsigned long number to string
 * input: 	dest - the char buffer to be written
 * 			src - the number to be converted
 * 			base - the base of conversion
 * 	output: number of characters written
 * 	remarks: the dest parameter should point at a large enough
 * 			buffer. no validity checks are made.
 */
/*****************************************************************************/
static unsigned long ulong2string (char * dest, unsigned long src, long base)
{
	unsigned long src1 = src;
	unsigned long src2;
	long d = 0;
	char str[256];
	long j = 0;
	unsigned long len;

	assert ((base >= 2) && (base <= 16));
	j = sizeof(str)-1;
	do {
		src2 = src1/base;
		d = src1-(src2*base);
		assert ((d>=0) && (d<base));
		str[j--] = digits[d];
		src1 = src2;
	} while (src1>0);
	len = sizeof(str)-j-1;
	memcpy (dest, &str[j+1], len);
	return len;
}

/*****************************************************************************/
/* convert long number to string
 * input: 	dest - the char buffer to be written
 * 			src - the number to be converted
 * 			base - the base of conversion
 * 	output: number of characters written
 * 	remarks: the dest parameter should point at a large enough
 * 			buffer. no validity checks are made.
 */
/*****************************************************************************/
static unsigned long long2string (char * dest, long src, long base)
{
	char str[256];
	unsigned long len;
	unsigned long src1;

	if (src < 0) {
		src1 = 0-src;
		*dest++ = '-';
	} else {
		src1 = src;
	}
	len = ulong2string (str, src1, base);
	memcpy (dest, str, len);
	return (src < 0) ? len+1 : len;
}
/*****************************************************************************/
/* pad a numeric string with zero's
 * input: p - the strint
 *	  len - the lenght of the string
 *	  min_width - the minimal width of the string
 * output: pointer just after the string */
/*****************************************************************************/
static inline char * zero_pad (char * p, unsigned long len, unsigned long min_width)
{
	char * retval;
	unsigned long j;

	if (len >= min_width) {
		retval = p + len;
	} else {
		unsigned long pad_size = min_width - len;
		retval = p + min_width;
		while (len > 0) {
			p[min_width-- - 1] = p[len-- - 1];
		}
		for (j = 0; j < pad_size; j ++) {
			p[j] = '0';
		}
	}

	return retval;
}
/*****************************************************************************/
/* convert string to unsigned long
 * input: p - the string
 * output: the converted value */
/*****************************************************************************/
static unsigned long string2ulong (const char * p)
{
#define MAX_INPUT_LEN 20
	unsigned long j;
	unsigned long un = 0;

	j = 0;
	while ((p[j] >= '0') && (p[j] <= '9')) {
		un *= 10;
		un += p[j] - '0';
		j ++;
		assert (j < MAX_INPUT_LEN);
	}
	return un;
}

